
(****************************************************************************
                                                                             
          IEEE754  :  FroundMult                                                 
                                                                             
          Laurent Thery                                                      
                                                                             
*****************************************************************************
*)
Require Export FroundMult.
Require Export ClosestProp.
Section FRoundP.
Variable b:Fbound.
Variable radix:nat.
Variable precision:nat.

Local FtoRradix := (FtoR radix).
Coercion FtoRradix : float >-> R.
Hypothesis radixMoreThanOne:(lt (S O) radix).
Hypothesis precisionGreaterThanOne:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp radix precision) (1)).

Theorem closestLessMultPos:
  (p:float) (r:R) (Closest b radix r p) -> (Rle R0 r) ->(Rle p (Rmult (2) r)).
Intros p r H' H'0.
Case ClosestMinOrMax with 1 := H'; Intros H'3.
Apply Rle_trans with r; Auto with real.
Intuition.
Case (MinEx b radix precision) with r := r; Auto with arith.
Intros min Hmin.
Apply Rle_trans with (Rplus min p); Auto with real.
Apply Rle_anti_compatibility with r := (Ropp p).
Replace (Rplus (Ropp p) p) with R0; [Idtac | Ring].
Replace (Rplus (Ropp p) (Rplus min p)) with (FtoRradix min);
 [Apply (RleMinR0 b radix precision) with r := r; Auto | Ring].
Apply Rle_anti_compatibility with r := (Ropp r).
Apply Rle_anti_compatibility with r := (Ropp min).
Replace (Rplus (Ropp min) (Rplus (Ropp r) (Rplus min p)))
     with (Rabsolu (Rminus p r)).
Replace (Rplus (Ropp min) (Rplus (Ropp r) (Rmult (2) r)))
     with (Rabsolu (Rminus min r)).
Case H'.
Intros H'1 H'2; Apply H'2; Auto.
Intuition.
Rewrite Rabsolu_simpl2; Simpl.
Ring; Auto.
Apply Rle_minus; Intuition.
Rewrite Rabsolu_simpl1; Simpl.
Ring; Auto.
Apply Rle_anti_compatibility with r := r.
Replace (Rplus r R0) with r; [Idtac | Ring].
Replace (Rplus r (Rminus p r)) with (FtoRradix p); [Intuition | Ring].
Qed.

Theorem closestLessMultNeg:
  (p:float) (r:R) (Closest b radix r p) -> (Rle r R0) ->(Rle (Rmult (2) r) p).
Intros p r H' H'0.
Replace (Rmult (2) r) with (Ropp (Rmult (2) (Ropp r))); [Idtac | Ring].
Replace (FtoRradix p) with (Ropp (Ropp p));
 [Unfold FtoRradix; Rewrite <- Fopp_correct | Ring].
Apply Rle_Ropp.
Apply closestLessMultPos; Auto.
Apply ClosestOpp; Auto.
Replace R0 with (Ropp R0); [Auto with real | Ring].
Qed.

Theorem closestLessMultAbs:
  (p:float) (r:R) (Closest b radix r p) ->
  (Rle (Rabsolu p) (Rmult (2) (Rabsolu r))).
Intros p r H'; Case (Rle_or_lt R0 r); Intros H'1.
Repeat Rewrite Rabsolu_simpl1; Auto.
Apply closestLessMultPos; Auto.
Apply (RleRoundedR0 b radix precision) with P := (Closest b radix) r := r; Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Repeat Rewrite Rabsolu_simpl2; Auto.
Replace (Rmult (2) (Ropp r)) with (Ropp (Rmult (2) r)); [Apply Rle_Ropp | Ring].
Apply closestLessMultNeg; Auto.
Apply Rlt_le; Auto.
Apply Rlt_le; Auto.
Apply (RleRoundedLessR0 b radix precision) with P := (Closest b radix) r := r;
 Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply Rlt_le; Auto.
Qed.
End FRoundP.
