
(****************************************************************************
                                                                             
          IEEE754  :  ClosestProp                                                     
                                                                             
          Laurent Thery    & Sylvie Boldo                                           
                                                                             
*****************************************************************************
*)
Require Export FroundProp.
Require Export Closest.
Section Fclosestp.
Variable b:Fbound.
Variable radix:nat.
Variable precision:nat.

Local FtoRradix := (FtoR radix).
Coercion FtoRradix : float >-> R.
Hypothesis radixMoreThanOne:(lt (S O) radix).
Hypothesis precisionGreaterThanOne:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp radix precision) (1)).

Theorem Rltdouble: (r:R) (Rlt R0 r) ->(Rlt r (Rmult (2) r)).
Intros r H'.
Pattern 1 r; Replace r with (Rplus r R0); Try Ring.
Replace (Rmult (2) r) with (Rplus r r); Simpl; Try Ring; Auto with real.
Apply Rlt_compatibility; Auto.
Qed.

Theorem RleFexpFabs:
  (p:float) ~ <R> p==R0 ->(Rle (Float (1) (Fexp p)) (Fabs p)).
Intros p H'.
Unfold FtoRradix FtoR Fabs; Simpl.
Apply Rle_monotony_r; Auto with real arith.
Rewrite Zabs_absolu.
Replace R1 with (INR (1)); Auto with real.
Repeat Rewrite <- INR_IZR_INZ; Apply Rle_INR; Auto.
Cut ~ (absolu (Fnum p))=O; Auto with arith.
Contradict H'.
Unfold FtoRradix FtoR; Simpl.
Replace (Fnum p) with ZERO; Try Ring.
Generalize H'; Case (Fnum p); Simpl; Auto with zarith arith; Intros p0 H'3;
 Contradict H'3; Auto with zarith arith.
Qed.

Theorem ClosestOpp:
  (p:float) (r:R) (Closest b radix r p) ->(Closest b radix (Ropp r) (Fopp p)).
Intros p r H'; Split.
Apply oppBounded; Auto.
Intuition.
Intros f H'0.
Rewrite Fopp_correct.
Replace (Rminus (Ropp (FtoR radix p)) (Ropp r))
     with (Ropp (Rminus (FtoR radix p) r)); [Idtac | Ring].
Replace (Rminus (FtoR radix f) (Ropp r))
     with (Ropp (Rminus (Ropp (FtoR radix f)) r)); [Idtac | Ring].
Rewrite <- Fopp_correct.
Repeat Rewrite Rabsolu_Ropp.
Case H'; Auto with float.
Qed.

Theorem ClosestFabs:
  (p:float) (r:R) (Closest b radix r p) ->(Closest b radix (Rabsolu r) (Fabs p)).
Intros p r H'; Case (Rle_or_lt R0 r); Intros Rl0.
Rewrite Rabsolu_simpl1; Auto.
Replace (Fabs p) with p; Auto.
Unfold Fabs; Apply floatEq; Simpl; Auto.
Cut (Zle ZERO (Fnum p)).
Case (Fnum p); Simpl; Auto; Intros p H0; Contradict H0; Apply Zlt_not_le; Red;
 Simpl; Auto with zarith.
Apply LeR0Fnum with radix := radix; Auto.
Apply RleRoundedR0
     with b := b precision := precision P := (Closest b radix) r := r; Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Rewrite Rabsolu_simpl2; Auto.
Replace (Fabs p) with (Fopp p).
Apply ClosestOpp; Auto.
Unfold Fabs; Apply floatEq; Simpl; Auto.
Cut (Zle (Fnum p) ZERO).
Case (Fnum p); Simpl; Auto; Intros p H0; Contradict H0; Apply Zlt_not_le; Red;
 Simpl; Auto with zarith.
Apply R0LeFnum with radix := radix; Auto.
Apply RleRoundedLessR0
     with b := b precision := precision P := (Closest b radix) r := r; Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply Rlt_le; Auto.
Apply Rlt_le; Auto.
Qed.

Theorem ClosestUlp:
  (p:R) (q:float) (Closest b radix p q) ->
  (Rle (Rmult (2) (Rabsolu (Rminus p q))) (Fulp b radix precision q)).
Intros p q H'.
Case (Req_EM p q); Intros Eqpq.
Rewrite Eqpq.
Replace (Rabsolu (Rminus q q)) with R0;
 [AutoRewrite [Rsimpl] |
   Replace (Rminus q q) with R0; Try Ring; Rewrite Rabsolu_simpl1;
    Auto with real].
Unfold Fulp; Apply Rlt_le; Auto with real arith.
Replace (Rmult (2) (Rabsolu (Rminus p q)))
     with (Rplus (Rabsolu (Rminus p q)) (Rabsolu (Rminus p q)));
 [Idtac | Simpl; AutoRewrite [Rsimpl]; Ring].
Case ClosestMinOrMax with 1 := H'; Intros H'1.
Apply Rle_trans
     with (Rplus
             (Rabsolu (Rminus p q))
             (Rabsolu (Rminus (FNSucc b radix precision q) p))).
Apply Rle_compatibility.
Rewrite <- (Rabsolu_Ropp (Rminus p q)).
Rewrite Rminus_Ropp.
Elim H'; Auto.
Intros H'0 H'2; Apply H'2; Auto.
Apply FcanonicBound with radix := radix precision := precision;
 Auto with float arith.
Repeat Rewrite Rminus_Rabsolu.
Replace (Rplus (Rminus p q) (Rminus (FNSucc b radix precision q) p))
     with (Rminus (FNSucc b radix precision q) q); [Idtac | Ring].
Unfold FtoRradix; Apply FulpSuc; Auto.
Intuition.
Case MinMax with 3 := pGivesBound r := p p := q; Auto with arith.
Intros H'0 H'2; Elim H'2; Intros H'3 H'4; Apply H'3; Clear H'2; Auto.
Intuition.
Apply Rle_trans
     with (Rplus
             (Rabsolu (Rminus p q))
             (Rabsolu (Rminus p (FNPred b radix precision q)))).
Apply Rle_compatibility.
Rewrite <- (Rabsolu_Ropp (Rminus p q));
 Rewrite <- (Rabsolu_Ropp (Rminus p (FNPred b radix precision q))).
Repeat Rewrite Rminus_Ropp.
Elim H'; Auto.
Intros H'0 H'2; Apply H'2; Auto.
Apply FcanonicBound with radix := radix precision := precision;
 Auto with float arith.
Rewrite <- (Rabsolu_Ropp (Rminus p q)); Rewrite Rminus_Ropp.
Repeat Rewrite Rminus_Rabsolu.
Replace (Rplus (Rminus q p) (Rminus p (FNPred b radix precision q)))
     with (Rminus q (FNPred b radix precision q)); [Idtac | Ring].
Unfold FtoRradix; Apply FulpPred; Auto.
Intuition.
Case MaxMin with 3 := pGivesBound r := p p := q; Auto with arith.
Intros H'0 H'2; Elim H'2; Intros H'3 H'4; Apply H'3; Clear H'2; Auto.
Intuition.
Qed.

Theorem ClosestExp:
  (p:R) (q:float) (Closest b radix p q) ->
  (Rle (Rmult (2) (Rabsolu (Rminus p q))) (powerRZ radix (Fexp q))).
Intros p q H'.
Apply Rle_trans with (Fulp b radix precision q).
Apply (ClosestUlp p q); Auto.
Replace (powerRZ radix (Fexp q)) with (FtoRradix (Float (1) (Fexp q))).
Apply (FulpLe b radix); Auto.
Apply RoundedModeBounded with radix := radix P := (Closest b radix) r := p; Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Unfold FtoRradix FtoR; Simpl.
Ring.
Qed.

Theorem ClosestErrorExpStrict:
  (p, q:float)
  (x:R)
  (Fbounded b p) ->
  (Fbounded b q) ->
  (Closest b radix x p) -> <R> q==(Rminus x p) -> ~ <R> q==R0 ->
  (Zlt (Fexp q) (Fexp p)).
Intros.
Apply Zlt_powerRZ with e := (INR radix).
Replace R1 with (INR (1)); Auto with real arith.
Apply Rlt_le_trans with (Rmult (2) (powerRZ radix (Fexp q))).
Apply Rltdouble; Auto with real arith.
Apply Rle_trans with (Rmult (2) (Fabs q)).
Apply Rle_monotony; Auto with real arith.
Replace R0 with (INR O); Auto with real arith.
Replace (powerRZ radix (Fexp q)) with (FtoRradix (Float (1) (Fexp q))); Auto.
Apply RleFexpFabs; Auto with arith.
Unfold FtoRradix FtoR; Simpl; Ring.
Rewrite (Fabs_correct radix); Auto with arith.
Replace (FtoR radix q) with (Rminus x p); Auto.
Apply ClosestExp; Auto.
Qed.

Theorem ClosestIdem:
  (p, q:float) (Fbounded b p) -> (Closest b radix p q) -><R> p==q.
Intros p q H' H'0.
Apply r_Rplus_plus with r := (Ropp p).
Replace (Rplus (Ropp p) p) with R0; [Idtac | Ring].
Replace (Rplus (Ropp p) q) with (Rminus q p); [Idtac | Ring].
Apply sym_eqT.
Apply RabsoluR0Inv; Auto.
Replace R0 with (Rabsolu R0); [Idtac | AutoRewrite [Rsimpl]]; Auto.
Replace R0 with (Rminus p p); [Idtac | Ring].
Case H'0; Auto.
Qed.

Theorem ClosestM1:
  (r1, r2:R)
  (min, max, p, q:float)
  (isMin b radix r1 min) ->
  (isMax b radix r1 max) ->
  (Rlt (Rplus min max) (Rmult (2) r2)) ->
  (Closest b radix r1 p) -> (Closest b radix r2 q) ->(Rle p q).
Intros r1 r2 min max p q H' H'0 H'1 H'2 H'3.
Case (Rle_or_lt r2 max); Intros H'4.
2:Apply (ClosestMonotone b radix) with p := r1 q := r2; Auto.
2:Apply Rle_lt_trans with (FtoRradix max); Auto.
2:Intuition.
Case H'4; Clear H'4; Intros H'4.
2:Replace (FtoRradix q) with (FtoRradix max).
2:Case ClosestMinOrMax with 1 := H'2; Intros H'5.
2:Replace (FtoRradix p) with (FtoRradix min).
2:Apply Rle_trans with r1; Intuition.
2:Apply MinEq with 1 := H'; Auto.
2:Replace (FtoRradix p) with (FtoRradix max); Auto with real.
2:Apply MaxEq with 1 := H'0; Auto.
2:Apply ClosestIdem; Auto.
2:Intuition.
2:Rewrite <- H'4; Auto.
Cut (Rlt min r2).
2:Apply Rlt_mult_anticompatibility with z := (INR (2)); Auto with real.
2:Replace (Rmult (2) min) with (Rplus min min); [Idtac | Simpl; Ring].
2:Apply Rle_lt_trans with 2 := H'1.
2:Apply Rle_compatibility; Auto with real.
2:Apply Rle_trans with r1; Intuition.
Intros H'5.
Replace (FtoRradix q) with (FtoRradix max).
Case ClosestMinOrMax with 1 := H'2; Intros H'6.
Replace (FtoRradix p) with (FtoRradix min).
Apply Rle_trans with r1; Intuition.
Apply MinEq with 1 := H'; Auto.
Replace (FtoRradix p) with (FtoRradix max); Auto with real.
Apply MaxEq with 1 := H'0; Auto.
Apply sym_eqT.
Apply (ClosestMaxEq b radix) with r := r2 min := min; Auto.
Apply isMinComp with 2 := H'0; Auto.
Apply isMaxComp with 1 := H'; Auto.
Qed.

Theorem FmultRadixInv:
  (x, z:float)
  (y:R)
  (Fbounded b x) -> (Closest b radix y z) -> (Rlt (Rmult (Rinv (2)) x) y) ->
  (Rle (Rmult (Rinv (2)) x) z).
Intros x z y H' H'0 H'1.
Case MinEx with r := (Rmult (Rinv (2)) x) 3 := pGivesBound; Auto with arith.
Intros min isMin.
Case MaxEx with r := (Rmult (Rinv (2)) x) 3 := pGivesBound; Auto with arith.
Intros max isMax.
Case (Rle_or_lt y max); Intros Rl1.
Case Rl1; Clear Rl1; Intros Rl1.
Replace (FtoRradix z) with (FtoRradix max).
Intuition.
Apply sym_eqT.
Unfold FtoRradix; Apply ClosestMaxEq with b := b r := y min := min; Auto.
Apply isMinComp with r1 := (Rmult (Rinv (2)) x) max := max; Auto.
Apply Rle_lt_trans with 2 := H'1; Auto.
Intuition.
Apply isMaxComp with r1 := (Rmult (Rinv (2)) x) min := min; Auto.
Apply Rle_lt_trans with 2 := H'1; Auto.
Intuition.
Replace (Rplus (FtoR radix min) (FtoR radix max)) with (FtoRradix x).
Apply Rlt_mult_anticompatibility with z := (Rinv (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_l; AutoRewrite [Rsimpl]; Auto with real.
Unfold FtoRradix; Apply (div2IsBetween b radix precision); Auto.
Cut (Closest b radix max z); [Intros C0 | Idtac].
Replace (FtoRradix z) with (FtoRradix max); Auto.
Rewrite <- Rl1; Auto.
Apply Rlt_le; Auto.
Apply ClosestIdem; Auto.
Intuition.
Apply (ClosestCompatible b radix y max z z); Auto.
Intuition.
Apply Rle_trans with (FtoRradix max); Auto.
Intuition.
Apply (ClosestMonotone b radix (FtoRradix max) y); Auto.
Apply (RoundedModeProjectorIdem b radix (Closest b radix)); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Intuition.
Qed.

Theorem ClosestErrorBound:
  (p, q:float)
  (x:R) (Fbounded b p) -> (Closest b radix x p) -> <R> q==(Rminus x p) ->
  (Rle (Rabsolu q) (Rmult (Float (1) (Fexp p)) (Rinv (2)))).
Intros.
Apply Rle_trans with (Rmult (Fulp b radix precision p) (Rinv (2))).
Rewrite H1.
Replace (Rabsolu (Rminus x p))
     with (Rmult (Rmult (2) (Rabsolu (Rminus x p))) (Rinv (2))); [Idtac | Ring].
Apply Rle_monotony_r; Auto with real.
Apply Rlt_le.
Apply Rlt_Rinv; Auto with real.
Apply ClosestUlp; Auto.
Replace (Rmult (Rinv (2)) (Rmult (Rabsolu (Rminus x p)) (2)))
     with (Rmult (Rabsolu (Rminus x p)) (Rmult (2) (Rinv (2)))).
Replace (Rmult (2) (Rinv (2))) with R1.
Ring.
Apply sym_eqT; Apply Rinv_r; Auto with real.
Ring.
Apply Rle_monotony_r.
Apply Rlt_le.
Apply Rlt_Rinv; Auto with real.
Unfold FtoRradix; Apply FulpLe; Auto.
Qed.

Theorem ClosestErrorExp:
  (p, q:float)
  (x:R)
  (Fbounded b p) ->
  (Fbounded b q) -> (Closest b radix x p) -> <R> q==(Rminus x p) ->
  (Ex [error:float]
  (Fbounded b error) /\
  (<R> error==q /\
   (Zle (Fexp error) (Zmax (Zminus (Fexp p) precision) (Zopp (dExp b)))))).
Intros.
Exists (Fnormalize radix b precision q).
Cut (Fcanonic radix b precision (Fnormalize radix b precision q));
 [Intros | Apply FnormalizeCanonic; Auto with arith].
Split.
Apply FcanonicBound with radix := radix precision := precision; Auto.
Split.
Unfold FtoRradix; Apply FnormalizeCorrect; Auto.
Case H3; Intros.
Apply Zle_trans with (Zminus (Fexp p) precision); Auto with zarith.
Apply Zsimpl_le_plus_l with (inject_nat precision).
Replace (Zplus precision (Zminus (Fexp p) precision)) with (Fexp p).
Replace (Zplus precision (Fexp (Fnormalize radix b precision q)))
     with (Zs (Zplus (Zpred precision) (Fexp (Fnormalize radix b precision q)))).
Apply Zgt_le_S.
Apply Zlt_gt.
Apply Zlt_powerRZ with (INR radix); Auto with real arith.
Apply Rlt_le; Auto with real arith.
Rewrite (powerRZ_add
           radix (Zpred precision) (Fexp (Fnormalize radix b precision q)));
 Auto with real arith.
Elim H4; Unfold Fdigit; Intros.
Apply Rle_lt_trans
     with (Rmult
             (absolu (Fnum (Fnormalize radix b precision q)))
             (powerRZ radix (Fexp (Fnormalize radix b precision q)))).
Apply Rle_monotony_r.
Apply powerRZ_le; Auto with real arith.
Cut <Z> (Zpred precision)=(pred precision); Intros.
Rewrite H7.
Pattern 1 precision; Rewrite <- H6.
Replace (powerRZ
           radix
           (pred (digit radix (absolu (Fnum (Fnormalize radix b precision q))))))
     with (INR
             (exp
                radix
                (pred
                   (digit
                      radix (absolu (Fnum (Fnormalize radix b precision q))))))).
Apply Rle_INR.
Apply digitLess.
Apply absolu_lt_nz.
Cut ~ (is_Fzero (Fnormalize radix b precision q)) ->
    ~ <Z> (Fnum (Fnormalize radix b precision q))=ZERO.
Intros H'; Apply H';
 Apply FnormalNotZero with radix := radix b := b precision := precision;
 Auto with arith.
Intros.
Contradict H8.
Apply is_Fzero_rep2 with radix; Auto.
Unfold FtoRradix; Unfold FtoR; Simpl.
Apply without_div_Oi; Auto.
Left; Apply IZR_zero; Auto with zarith.
Rewrite H6.
Case precision; Simpl; Auto with real arith.
Intros n; Apply sym_eqT; Apply powerRZ_is_exp; Auto with real arith.
Generalize precisionGreaterThanOne; Case precision.
Intros.
Absurd (lt (1) O); Auto with arith.
Intros; Apply sym_eq.
Apply inj_pred; Auto with zarith arith.
Replace (Rmult
           (absolu (Fnum (Fnormalize radix b precision q)))
           (powerRZ radix (Fexp (Fnormalize radix b precision q))))
     with (Rabsolu (Fnormalize radix b precision q)).
Replace (FtoRradix (Fnormalize radix b precision q)) with (FtoRradix q);
 [Idtac | Unfold FtoRradix; Apply sym_eqT; Apply FnormalizeCorrect; Auto].
Apply Rle_lt_trans with (Rmult (Float (1) (Fexp p)) (Rinv (2))).
Apply ClosestErrorBound with x := x; Auto.
Unfold FtoRradix; Unfold FtoR; Simpl.
Pattern 2 (powerRZ radix (Fexp p));
 Replace (powerRZ radix (Fexp p)) with (Rmult (powerRZ radix (Fexp p)) R1);
 [Idtac | Ring].
Replace (Rmult R1 (powerRZ radix (Fexp p))) with (powerRZ radix (Fexp p));
 [Apply Rlt_monotony | Ring].
Apply powerRZ_lt; Auto with arith real.
Pattern 3 R1; Replace R1 with (Rinv R1).
Apply Rlt_Rinv_R1; Auto with real.
Replace (Rplus R1 R1) with (INR (2)); Auto with real arith.
Apply R1_Rinv.
Replace (Rabsolu (Fnormalize radix b precision q))
     with (FtoRradix (Fabs (Fnormalize radix b precision q))).
Unfold Fabs; Unfold FtoRradix; Unfold FtoR; Simpl; Auto with real.
Rewrite (Zabs_absolu (Fnum (Fnormalize radix b precision q))); Auto with real.
Rewrite <- INR_IZR_INZ; Auto with real.
Unfold FtoRradix; Apply Fabs_correct; Auto with arith.
Unfold Zs; Unfold Zpred; Ring.
Ring.
Elim H4; Intros.
Elim H6; Intros; Clear H6.
Apply Zle_trans with (Zopp (dExp b)); Auto with zarith.
Qed.
End Fclosestp.
Hints Resolve ClosestOpp ClosestFabs ClosestUlp :float.
