
(****************************************************************************
                                                                             
          IEEE754  :  EFast2Sum                                                     
                                                                             
          Laurent Thery                                                      
                                                                             
*****************************************************************************
*)
Require Export Fast2Sum.
Section EFast.
Variable b:Fbound.
Variable precision:nat.

Local FtoRradix := (FtoR (2)).
Coercion FtoRradix : float >-> R.
Hypothesis precisionGreaterThanOne:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp (2) precision) (1)).
Variable Iplus:float -> float ->float.
Hypothesis IplusCorrect:
           (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
           (Closest b (2) (Rplus p q) (Iplus p q)).
Hypothesis IplusComp:
           (p, q, r, s:float)
           (Fbounded b p) ->
           (Fbounded b q) ->
           (Fbounded b r) -> (Fbounded b s) -> <R> p==r -> <R> q==s ->
           <R> (Iplus p q)==(Iplus r s).
Hypothesis IplusSym:(p, q:float)(Iplus p q)=(Iplus q p).
Hypothesis IplusOp:(p, q:float)(Fopp (Iplus p q))=(Iplus (Fopp p) (Fopp q)).
Variable Iminus:float -> float ->float.
Hypothesis IminusPlus:(p, q:float)(Iminus p q)=(Iplus p (Fopp q)).

Theorem IminusComp:
  (p, q, r, s:float)
  (Fbounded b p) ->
  (Fbounded b q) -> (Fbounded b r) -> (Fbounded b s) -> <R> p==r -> <R> q==s ->
  <R> (Iminus p q)==(Iminus r s).
Intros p q r s H' H'0 H'1 H'2 H'3 H'4.
Repeat Rewrite IminusPlus.
Apply IplusComp; Auto; Try Apply oppBounded; Auto.
Unfold FtoRradix; Repeat Rewrite Fopp_correct; Auto.
Unfold FtoRradix in H'4; Rewrite H'4; Auto.
Qed.

Theorem EvenBound:
  (p:float)
  (m:Z)
  (Even (absolu m)) ->
  (Zle (minus (exp (2) precision) (1)) m) ->
  (Zle m (minus (exp (2) (S precision)) (2))) -> (Fbounded b p) ->
  (Ex [q:float] (Fbounded b q) /\ <R> q==(Float m (Fexp p))).
Intros p m H' H'0 H'1 H'2.
Case (EvenD ? H').
Intros am2 H'3.
Exists (Float am2 (Zs (Fexp p))); Split.
Repeat Split; Simpl.
Apply Zle_trans with (Zopp O); Auto with zarith.
Apply Zle_Zmult_right2 with z := (inject_nat (2)); Auto with zarith.
Apply Zlt_gt; Replace ZERO with (inject_nat O); Auto with zarith arith.
Rewrite Zmult_sym.
Rewrite <- inj_mult.
Rewrite <- H'3.
Rewrite inj_abs.
Rewrite pGivesBound.
Rewrite <- inj_mult.
Rewrite mult_minus_distr.
Simpl; Rewrite mult_sym; Auto.
Apply Zle_trans with 2 := H'0.
Replace ZERO with (inject_nat O); Auto with zarith arith.
Apply Zle_trans with (Fexp p); Auto with zarith arith.
Intuition.
Rewrite <- (inj_abs m).
Rewrite H'3.
Rewrite mult_sym; Simpl.
Unfold FtoRradix FtoR; Simpl.
Rewrite powerRZ_Zs.
AutoRewrite [Rsimpl].
Replace (INR (2)) with (Rplus R1 R1).
Ring.
Simpl; Auto.
Replace (Rplus R1 R1) with (INR (2)).
Auto with real arith.
Simpl; Auto.
Apply Zle_trans with 2 := H'0.
Replace ZERO with (inject_nat O); Auto with zarith arith.
Qed.

Theorem ExtMDekker1:
  (p, q:float)
  (Fbounded b p) -> (Fbounded b q) -> (Zle (Fexp q) (Fexp p)) -> (Rle R0 p) ->
  <R> (Iminus (Iplus p q) p)==(Rminus (Iplus p q) p).
Intros p q H' H'0 H'1 H'2.
Cut (Zle ZERO (Fnum p)); [Intros Z1 | Apply LeR0Fnum with radix := (2); Auto].
Case (Rle_or_lt (Rabsolu q) (Rabsolu p)); Intros Rl0.
Apply MDekker with b := b precision := precision; Auto.
Rewrite Rabsolu_simpl1 in Rl0; Auto.
Case (LessExpBound ? TwoMoreThanOne b p (Fabs q)); Auto.
Apply absFBounded; Auto.
Unfold FtoRradix; Rewrite Fabs_correct; Auto; Apply Rlt_le; Auto.
Replace (Fexp (Fabs q)) with (Fexp q); [Idtac | Case q; Simpl; Auto].
Intros m E; Elim E; Intros H'4 H'5; Clear E.
Cut (Zle ZERO (Fnum (Float m (Fexp q))));
 [Simpl; Intros Z1 |
   Apply LeR0Fnum with radix := (2); Auto; Unfold FtoRradix in H'4; Rewrite H'4];
 Auto.
Unfold FtoRradix; Rewrite <- H'4; Auto.
Apply trans_eqT
     with y :=
          (FtoRradix (Iminus (Iplus (Float m (Fexp q)) q) (Float m (Fexp q)))).
Apply IminusComp; Auto.
Apply IplusBounded; Auto.
Apply IplusBounded; Auto.
Fold FtoRradix.
Rewrite (IplusComp p q (Float m (Fexp q)) q); Auto.
Case (Rle_or_lt R0 q); Intros Rl1.
Cut (Zle ZERO (Fnum q)); [Intros Z2 | Apply LeR0Fnum with radix := (2); Auto].
Rewrite Rabsolu_simpl1 in Rl0; Auto.
Case (Zle_or_lt (Zplus m (Fnum q)) (vNum b)); Intros Rl1.
Apply (MDekkerAux3 b); Auto.
Unfold Fplus; Simpl.
Repeat Rewrite Zmin_n_n; Repeat Rewrite <- Zminus_n_n; Simpl.
Rewrite Zpower_nat_O; Repeat Rewrite Zmult_n_1.
Repeat Split; Simpl; Auto.
Apply Zle_trans with (Zplus ZERO ZERO); Auto with zarith.
Intuition.
Cut (Zle (exp (2) precision) (Zplus m (Fnum q))); [Intros Rl3 | Idtac].
2:Replace (inject_nat (exp (2) precision)) with (Zs (vNum b)); Auto with zarith.
2:Rewrite pGivesBound; Auto with zarith arith.
2:Rewrite <- inj_S.
2:Replace (S (minus (exp (2) precision) (1)))
       with (plus (1) (minus (exp (2) precision) (1))); Auto with arith.
2:Rewrite le_plus_minus_r; Auto.
2:Replace (1) with (exp (2) O); Auto with arith.
Cut (Zle (Zplus m (Fnum q)) (minus (exp (2) (S precision)) (2)));
 [Intros Rl4 | Idtac].
2:Replace (minus (exp (2) (S precision)) (2))
       with (plus
               (minus (exp (2) precision) (1)) (minus (exp (2) precision) (1)));
   Auto with arith.
2:Rewrite inj_plus.
2:Apply Zle_plus_plus; Auto; Rewrite <- pGivesBound; Intuition.
2:Replace (plus (minus (exp (2) precision) (1)) (minus (exp (2) precision) (1)))
       with (mult (2) (minus (exp (2) precision) (1))); Auto with arith.
2:Rewrite mult_sym; Rewrite mult_minus_distr; Rewrite mult_sym; Simpl; Auto.
2:Simpl; Ring.
Case (OddEvenDec (absolu (Zplus m (Fnum q)))); Intros E0.
Case (RoundedModeRep b (2) precision) with P := (Closest b (2))
                                           p := (Fplus (2) (Float m (Fexp q)) q)
                                           q := (Iplus (Float m (Fexp q)) q);
 Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Rewrite Fplus_correct; Auto.
Replace (Fexp (Fplus (2) (Float m (Fexp q)) q)) with (Fexp q).
Simpl; Intros m' Hm'.
2:Unfold Fplus; Simpl.
2:Repeat Rewrite Zmin_n_n; Repeat Rewrite <- Zminus_n_n; Simpl; Auto.
Cut (Zle (Zminus m' m) (Zs (Fnum q))); [Intros Rl5 | Idtac].
Case (Zle_lt_or_eq ? ? Rl5); Clear Rl5; Intros Rl5.
Cut (Zle (Zpred (Fnum q)) (Zminus m' m)); [Intros Rl6 | Idtac].
Unfold FtoRradix; Rewrite Hm'.
Unfold FtoRradix; Rewrite <- Fminus_correct; Auto.
Apply sym_eqT; Apply (ClosestIdem b (2)); Auto.
Unfold Fminus Fplus; Simpl.
Repeat Rewrite Zmin_n_n; Repeat Rewrite <- Zminus_n_n; Simpl.
Rewrite Zpower_nat_O; Repeat Rewrite Zmult_n_1.
Repeat Split; Simpl; Auto.
Apply Zle_trans with (Zpred ZERO); Auto with zarith.
Replace (Zpred ZERO) with (Zopp (1)); Auto with zarith.
Apply Zle_Zopp; Auto.
Apply inj_le.
Apply lt_le_S; Auto.
Apply (lt_vNum (2)) with precision := precision; Auto with arith.
Apply Zle_trans with (Zpred (Fnum q)); Auto with zarith.
Unfold Zpred; Auto with zarith.
Apply Zle_trans with (Fnum q); Auto with zarith.
Intuition.
Intuition.
Rewrite Fminus_correct; Auto.
Rewrite <- Hm'; Auto.
Apply (IminusCorrect b Iplus); Auto.
Apply IplusBounded; Auto.
Apply Zsimpl_le_plus_l with p := m; Auto.
Replace (Zplus m (Zminus m' m)) with m'; Auto with zarith.
Apply le_IZR.
Apply Rle_mult_anticompatibility with z := (powerRZ (2) (Fexp q));
 Auto with real arith.
Repeat Rewrite (Rmult_sym (powerRZ (2) (Fexp q))).
Change (Rle (Float (Zplus m (Zpred (Fnum q))) (Fexp q)) (Float m' (Fexp q))).
Replace (Zplus m (Zpred (Fnum q))) with (Zpred (Zplus m (Fnum q))).
2:Unfold Zpred; Ring.
Case (EvenBound (Float m (Fexp q)) (Zpred (Zplus m (Fnum q)))); Auto.
Apply OddSEvenInv; Auto.
Rewrite <- absolu_Zs; Auto.
Replace (Zs (Zpred (Zplus m (Fnum q)))) with (Zplus m (Fnum q));
 [Idtac | Unfold Zs Zpred; Ring]; Auto.
Case (Zle_lt_or_eq ZERO (Zplus m (Fnum q))).
Replace ZERO with (Zplus ZERO ZERO); Auto with zarith.
Intros H'3; Apply Zlt_n_Sm_le.
Cut (r:Z)(Zs (Zpred r))=r;
 [Intros Er; Rewrite Er; Auto | Intros r; Unfold Zs Zpred; Ring].
Intros H'3; Contradict E0; Rewrite <- H'3; Simpl.
Apply EvenNOdd; Apply EvenO.
Apply Zlt_n_Sm_le.
Cut (r:Z)(Zs (Zpred r))=r;
 [Intros Er; Rewrite Er; Auto | Intros r; Unfold Zs Zpred; Ring].
Elim pGivesBound; Auto.
Apply Zle_trans with 2 := Rl4; Auto with zarith.
Apply Zle_pred_n.
Simpl; Intros x H'3; Elim H'3; Intros H'6 H'7; Rewrite <- H'7; Clear H'3.
Unfold FtoRradix; Rewrite <- Hm'.
Apply (ClosestMonotone b (2) x (Rplus (Float m (Fexp q)) q)); Auto.
Unfold FtoRradix; Rewrite <- Fplus_correct; Auto.
Unfold FtoRradix in H'7; Rewrite H'7.
Unfold Fplus.
Repeat Rewrite Zmin_n_n; Repeat Rewrite <- Zminus_n_n; Simpl; Auto.
Rewrite Zpower_nat_O; Repeat Rewrite Zmult_n_1.
Unfold FtoR; Simpl.
Repeat Rewrite [x:Z](Rmult_sym x (powerRZ (Rplus R1 R1) (Fexp q))).
Apply Rlt_monotony; Auto with real arith.
Apply Rlt_IZR; Auto.
Apply Zlt_pred_n_n; Auto.
Apply (RoundedModeProjectorIdem b (2) (Closest b (2))); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Case (Zle_lt_or_eq (Fnum q) (vNum b)); Intros H'3 Orelse Intuition.
Cut (Closest
       b (2)
       (Rminus
          (FtoR (2) (Iplus (Float m (Fexp q)) q)) (FtoR (2) (Float m (Fexp q))))
       (Iminus (Iplus (Float m (Fexp q)) q) (Float m (Fexp q)))).
Unfold FtoRradix; Rewrite Hm'.
Rewrite <- Fminus_correct; Auto.
Unfold Fminus Fplus; Simpl; Auto.
Repeat Rewrite Zmin_n_n; Repeat Rewrite <- Zminus_n_n; Simpl; Auto.
Rewrite Zpower_nat_O; Repeat Rewrite Zmult_n_1.
Replace (Zplus m' (Zopp m)) with (Zminus m' m); [Idtac | Ring].
Intros H'4; Apply sym_eqT; Apply (ClosestIdem b (2)); Auto.
Repeat Split; Simpl; Auto.
Rewrite Rl5; Auto.
Apply Zle_trans with (Fnum q); Auto with zarith.
Rewrite Rl5; Auto with zarith.
Intuition.
Apply (IminusCorrect b Iplus); Auto.
Apply IplusBounded; Auto.
Cut (Closest
       b (2)
       (Rminus
          (FtoR (2) (Iplus (Float m (Fexp q)) q)) (FtoR (2) (Float m (Fexp q))))
       (Iminus (Iplus (Float m (Fexp q)) q) (Float m (Fexp q)))).
Unfold FtoRradix; Rewrite Hm'.
Rewrite <- Fminus_correct; Auto.
Unfold Fminus Fplus; Simpl; Auto.
Repeat Rewrite Zmin_n_n; Repeat Rewrite <- Zminus_n_n; Simpl; Auto.
Rewrite Zpower_nat_O; Repeat Rewrite Zmult_n_1.
Replace (Zplus m' (Zopp m)) with (Zminus m' m); [Idtac | Ring].
Rewrite Rl5.
Rewrite H'3.
Replace (FtoR (2) (Float (Zs (vNum b)) (Fexp q)))
     with (FtoR (2) (Float (nNormMin (2) precision) (Zs (Fexp q)))).
Intros H'6; Apply sym_eqT; Apply (ClosestIdem b (2)); Auto.
Repeat Split; Simpl; Auto.
Apply Zle_trans with (inject_nat O); Auto with zarith arith.
Apply inj_le; Apply nNrMMimLevNum; Auto with arith.
Apply Zle_trans with (Fexp q); Auto with zarith.
Intuition.
Unfold nNormMin.
Rewrite pGivesBound.
Rewrite <- inj_S.
Replace (S (minus (exp (2) precision) (1)))
     with (plus (1) (minus (exp (2) precision) (1))).
Rewrite le_plus_minus_r; Auto.
Unfold FtoR; Simpl.
Rewrite powerRZ_Zs.
Pattern 2 precision; Replace precision with (S (pred precision)).
Simpl; AutoRewrite [Rsimpl].
Ring.
Generalize precisionGreaterThanOne; Case precision; Simpl; Auto; Intros H'6;
 Inversion H'6.
Replace (Rplus R1 R1) with (INR (2)).
Auto with real arith.
Simpl; Auto.
Replace (1) with (exp (2) O); Auto with arith.
Simpl; Auto.
Apply (IminusCorrect b Iplus); Auto.
Apply IplusBounded; Auto.
Apply Zsimpl_le_plus_l with p := m; Auto.
Replace (Zplus m (Zminus m' m)) with m'; Auto with zarith.
Apply le_IZR.
Apply Rle_mult_anticompatibility with z := (powerRZ (2) (Fexp q));
 Auto with real arith.
Repeat Rewrite (Rmult_sym (powerRZ (2) (Fexp q))).
Change (Rle (Float m' (Fexp q)) (Float (Zplus m (Zs (Fnum q))) (Fexp q))).
Replace (Zplus m (Zs (Fnum q))) with (Zs (Zplus m (Fnum q)));
 [Idtac | Unfold Zs; Ring].
Case (EvenBound (Float m (Fexp q)) (Zs (Zplus m (Fnum q)))); Auto.
Rewrite absolu_Zs; Auto.
Apply OddSEven; Auto.
Replace ZERO with (Zplus ZERO ZERO); Auto with zarith.
Apply Zle_trans with (Zplus m (Fnum q)); Auto with zarith.
Apply Zle_trans with 2 := Rl3; Auto with zarith arith.
Case (Zle_next ? ? Rl4); Auto.
Intros H'3; Contradict E0.
Rewrite <- H'3; Auto.
Rewrite absolu_INR.
Apply EvenNOdd; Auto.
Apply EvenPlusInv1 with b := b radix := (2) precision := precision n := (2);
 Auto.
Rewrite le_plus_minus_r; Auto with arith.
Apply EvenExp with b := b radix := (2) precision := precision; Auto.
Apply EvenSS; Auto.
Apply EvenO; Auto.
Replace (2) with (exp (2) (1)); Auto with arith.
Apply EvenSS; Auto.
Apply EvenO; Auto.
Simpl; Intros x H'3; Elim H'3; Intros H'6 H'7; Rewrite <- H'7; Clear H'3.
Unfold FtoRradix; Rewrite <- Hm'.
Apply (ClosestMonotone b (2) (Rplus (Float m (Fexp q)) q) x); Auto.
Rewrite H'7.
Unfold FtoRradix; Rewrite <- Fplus_correct; Auto.
Unfold Fplus.
Repeat Rewrite Zmin_n_n; Repeat Rewrite <- Zminus_n_n; Simpl; Auto.
Rewrite Zpower_nat_O; Repeat Rewrite Zmult_n_1.
Unfold FtoR; Simpl.
Repeat Rewrite [x:Z](Rmult_sym x (powerRZ (Rplus R1 R1) (Fexp q))).
Apply Rlt_monotony; Auto with real arith.
Apply Rlt_IZR; Auto with zarith.
Apply (RoundedModeProjectorIdem b (2) (Closest b (2))); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply (MDekkerAux2 b); Auto.
Cut (Closest b (2) (Rplus (Float m (Fexp q)) q) (Iplus (Float m (Fexp q)) q));
 Auto.
Unfold FtoRradix; Rewrite <- Fplus_correct; Auto with arith.
Unfold Fplus.
Repeat Rewrite Zmin_n_n; Repeat Rewrite <- Zminus_n_n; Simpl; Auto.
Rewrite Zpower_nat_O; Repeat Rewrite Zmult_n_1.
Case (EvenBound (Float m (Fexp q)) (Zplus m (Fnum q))); Auto.
Apply Zle_trans with 2 := Rl3; Auto with zarith arith.
Simpl; Intros x H'3; Elim H'3; Intros H'6 H'7; Unfold FtoRradix in H'7;
 Rewrite <- H'7; Clear H'3.
Intros H'3.
Apply sym_eqT; Apply (ClosestIdem b (2)); Auto with arith.
Apply (MDekkerAux3 b); Auto.
Replace (Fplus (2) (Float m (Fexp q)) q)
     with (Fminus (2) (Float m (Fexp q)) (Fopp q)).
Apply BminusSameExp; Auto.
Apply oppBounded; Auto.
Unfold FtoRradix in H'4; Rewrite H'4; Auto.
Replace R0 with (Ropp R0); Auto with real.
Rewrite Fopp_correct; Apply Rlt_le; Auto with real.
Unfold Fminus; Rewrite Fopp_Fopp; Auto.
Qed.

Theorem ExtMDekker:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) -> (Zle (Fexp q) (Fexp p)) ->
  <R> (Iminus (Iplus p q) p)==(Rminus (Iplus p q) p).
Intros p q H' H'0 H'1.
Case (Rle_or_lt R0 p); Intros H1.
Apply ExtMDekker1; Auto.
Apply (MDekkerAux5 b); Auto.
Apply ExtMDekker1; Auto.
Apply oppBounded; Auto.
Apply oppBounded; Auto.
Apply Rlt_le; Unfold FtoRradix; Rewrite Fopp_correct; Replace R0 with (Ropp R0);
 Auto with real.
Qed.

Theorem ExtDekker:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) -> (Zle (Fexp q) (Fexp p)) ->
  <R> (Iminus q (Iminus (Iplus p q) p))==(Rminus (Rplus p q) (Iplus p q)).
Intros p q H' H'0 H'1.
Apply (MDekkerAux1 b precision); Auto.
Apply ExtMDekker; Auto.
Qed.
End EFast.
