
(****************************************************************************
                                                                             
          IEEE754  :  Fast2Diff                                                     
                                                                             
          Laurent Thery                                                      
                                                                             
*****************************************************************************
*)
Require Export EFast2Sum.
Section EDiff.
Variable b:Fbound.
Variable precision:nat.

Local FtoRradix := (FtoR (2)).
Coercion FtoRradix : float >-> R.
Hypothesis precisionGreaterThanOne:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp (2) precision) (1)).
Variable Iplus:float -> float ->float.
Hypothesis IplusCorrect:
           (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
           (Closest b (2) (Rplus p q) (Iplus p q)).
Hypothesis IplusComp:
           (p, q, r, s:float)
           (Fbounded b p) ->
           (Fbounded b q) ->
           (Fbounded b r) -> (Fbounded b s) -> <R> p==r -> <R> q==s ->
           <R> (Iplus p q)==(Iplus r s).
Hypothesis IplusSym:(p, q:float)(Iplus p q)=(Iplus q p).
Hypothesis IplusOp:(p, q:float)(Fopp (Iplus p q))=(Iplus (Fopp p) (Fopp q)).
Variable Iminus:float -> float ->float.
Hypothesis IminusPlus:(p, q:float)(Iminus p q)=(Iplus p (Fopp q)).

Theorem MDekkerDiffAux1:
  (p, q:float)
  <R> (Iminus p (Iminus p q))==(Rminus p (Iminus p q)) ->
  (Fbounded b p) -> (Fbounded b q) ->
  <R> (Iminus (Iminus p (Iminus p q)) q)==(Rminus (Rminus p q) (Iminus p q)).
Intros p q H' H'0 H'1.
Elim (ErrorBoundedIplus b precision) with Iplus := Iplus p := p q := (Fopp q);
 Auto.
Intros error H'2; Elim H'2; Intros H'3 H'4; Clear H'2.
Cut (Closest
       b (2) (Rminus (Iminus p (Iminus p q)) q)
       (Iminus (Iminus p (Iminus p q)) q)); Auto.
Rewrite H'.
Replace (Rminus (Rminus p (Iminus p q)) q)
     with (Rminus (Rminus p q) (Iminus p q)); [Idtac | Ring].
Replace (Rminus p q) with (Rplus p (Ropp q)); [Idtac | Ring].
Rewrite (IminusPlus p q).
Unfold FtoRradix; Rewrite <- Fopp_correct.
Rewrite <- H'3.
Intros H'2.
Apply sym_eqT; Apply (ClosestIdem b (2)); Auto.
Apply (IminusCorrect b Iplus); Auto.
Apply (IminusBounded b Iplus); Auto.
Apply (IminusBounded b Iplus); Auto.
Apply oppBounded; Auto.
Qed.

Theorem MDekkerDiff:
  (p, q:float)
  (Fbounded b p) -> (Fbounded b q) -> (Rle (Rabsolu q) (Rabsolu p)) ->
  <R> (Iminus p (Iminus p q))==(Rminus p (Iminus p q)).
Intros p q H' H'0 H'1.
Pattern 2 (Iminus p q); Rewrite IminusPlus.
Replace (Rminus p (Iplus p (Fopp q))) with (Ropp (Rminus (Iplus p (Fopp q)) p));
 [Idtac | Ring].
Unfold FtoRradix; Rewrite <- (MDekker b precision) with Iminus := Iminus; Auto.
Rewrite <- Fopp_correct.
Repeat (Rewrite IminusPlus Orelse Rewrite IplusOp Orelse Rewrite Fopp_Fopp).
Rewrite IplusSym; Auto.
Apply oppBounded; Auto.
Rewrite Fopp_correct; Auto.
Rewrite Rabsolu_Ropp; Auto.
Qed.

Theorem DekkerDiff:
  (p, q:float)
  (Fbounded b p) -> (Fbounded b q) -> (Rle (Rabsolu q) (Rabsolu p)) ->
  <R> (Iminus (Iminus p (Iminus p q)) q)==(Rminus (Rminus p q) (Iminus p q)).
Intros p q H' H'0 H'1.
Apply MDekkerDiffAux1; Auto.
Apply MDekkerDiff; Auto.
Qed.

Theorem ExtMDekkerDiff:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) -> (Zle (Fexp q) (Fexp p)) ->
  <R> (Iminus p (Iminus p q))==(Rminus p (Iminus p q)).
Intros p q H' H'0 H'1.
Pattern 2 (Iminus p q); Rewrite IminusPlus.
Replace (Rminus p (Iplus p (Fopp q))) with (Ropp (Rminus (Iplus p (Fopp q)) p));
 [Idtac | Ring].
Unfold FtoRradix; Rewrite <- (ExtMDekker b precision) with Iminus := Iminus;
 Auto.
Rewrite <- Fopp_correct.
Repeat (Rewrite IminusPlus Orelse Rewrite IplusOp Orelse Rewrite Fopp_Fopp).
Rewrite IplusSym; Auto.
Apply oppBounded; Auto.
Qed.

Theorem ExtDekkerDiff:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) -> (Zle (Fexp q) (Fexp p)) ->
  <R> (Iminus (Iminus p (Iminus p q)) q)==(Rminus (Rminus p q) (Iminus p q)).
Intros p q H' H'0 H'1.
Apply MDekkerDiffAux1; Auto.
Apply ExtMDekkerDiff; Auto.
Qed.
End EDiff.

