Require Export FroundPlus.
Require Export FroundMult.
Require Export PolyList.
Section mf.
Variable b:Fbound.
Variable radix:nat.
Variable precision:nat.

Local FtoRradix := (FtoR radix).
Coercion FtoRradix : float >-> R.
Hypothesis radixMoreThanOne:(lt (S O) radix).
Hypothesis precisionGreaterThanOne:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp radix precision) (1)).
(* because the flow of algorithm goes from the smallest element of the
    expansion to the biggest, expansion are defined as list of the smallest 
    to the biggest *)

Inductive IsExpansion: (list float) ->Prop :=
     IsExpansionNil: (IsExpansion (nil ?))
    | IsExpansionSingle:
        (x:float) (Fbounded b x) ->(IsExpansion (cons x (nil ?)))
    | IsExpansionTop1:
        (x:float)
        (L:(list float)) (is_Fzero x) -> (Fbounded b x) -> (IsExpansion L) ->
        (IsExpansion (cons x L))
    | IsExpansionTop2:
        (x, y:float)
        (L:(list float))
        (Fbounded b x) ->
        (is_Fzero y) -> (Fbounded b y) -> (IsExpansion (cons x L)) ->
        (IsExpansion (cons x (cons y L)))
    | IsExpansionTop:
        (x, y:float)
        (L:(list float))
        ~ (is_Fzero x) ->
        ~ (is_Fzero y) ->
        (Fbounded b x) ->
        (Fbounded b y) ->
        (Zlt (MSB radix x) (LSB radix y)) -> (IsExpansion (cons y L)) ->
        (IsExpansion (cons x (cons y L))) .

Theorem Zlt_MSB_LSB:
  (x, y:float) ~ (is_Fzero y) -> (Zlt (MSB radix x) (LSB radix y)) ->
  (Rlt (Fabs x) (Fabs y)).
Intros x y H' H'0; Case (is_FzeroP x); Intros F0.
Replace (FtoRradix (Fabs x)) with R0.
Case (RabsoluR0 y).
Unfold FtoRradix; Rewrite Fabs_correct; Auto with arith.
Intros H'1; Contradict H'.
Apply is_Fzero_rep2 with radix := radix; Auto.
Apply RabsoluR0Inv; Auto.
Rewrite H'1; Auto with real.
Unfold Fabs; Rewrite F0; Unfold FtoRradix FtoR; Simpl; AutoRewrite [Rsimpl];
 Auto.
Case (Rle_or_lt (Fabs y) (Fabs x)); Auto.
Intros H'1; Contradict H'0.
Apply Zle_not_lt; Auto.
Apply Zle_trans with (MSB radix y).
Apply LSB_le_MSB; Auto.
Apply MSB_monotone; Auto.
Qed.

Theorem IsExpansion_comp1:
  (L:(list float))
  (x, y:float) (Fbounded b y) -> <R> x==y -> (IsExpansion (cons x L)) ->
  (IsExpansion (cons y L)).
Induction L.
Intros x y H' H'0 H'1; Apply IsExpansionSingle; Auto.
Intros a l H' x y H'0 H'1 H'2; Inversion H'2.
Apply IsExpansionTop1; Auto.
Apply (is_Fzero_rep2 radix); Auto with arith; Unfold FtoRradix in H'1;
 Rewrite <- H'1; Apply is_Fzero_rep1; Auto.
Apply IsExpansionTop2; Auto.
Apply (H' x); Auto.
Apply IsExpansionTop; Auto.
Contradict H2; Apply (is_Fzero_rep2 radix); Auto with arith;
 Unfold FtoRradix in H'1; Rewrite H'1; Apply is_Fzero_rep1; Auto.
Rewrite (MSB_comp radix) with y := x; Auto.
Contradict H2; Apply (is_Fzero_rep2 radix); Auto with arith;
 Unfold FtoRradix in H'1; Rewrite H'1; Apply is_Fzero_rep1; Auto.
Qed.

Theorem IsExpansion_comp2:
  (L:(list float))
  (x, y, z:float)
  (Fbounded b z) -> <R> y==z -> (IsExpansion (cons x (cons y L))) ->
  (IsExpansion (cons x (cons z L))).
Intros L x y z H' H'0 H'1; Inversion H'1.
Apply IsExpansionTop1; Auto; Apply IsExpansion_comp1 with y; Auto.
Apply IsExpansionTop2; Auto.
Apply (is_Fzero_rep2 radix); Auto with arith; Unfold FtoRradix in H'0;
 Rewrite <- H'0; Apply is_Fzero_rep1; Auto.
Apply IsExpansionTop; Auto.
Contradict H3; Apply (is_Fzero_rep2 radix); Auto with arith;
 Unfold FtoRradix in H'0; Rewrite H'0; Apply is_Fzero_rep1; Auto.
Rewrite (LSB_comp radix) with y := y; Auto.
Contradict H3; Apply (is_Fzero_rep2 radix); Auto with arith;
 Unfold FtoRradix in H'0; Rewrite H'0; Apply is_Fzero_rep1; Auto.
Apply IsExpansion_comp1 with x := y; Auto.
Qed.

Fixpoint expValue[L:(list float)]: float :=
   Cases L of
       nil => (Fzero ZERO)
      | (cons x L1) => (Fplus radix x (expValue L1))
   end.

Fixpoint expNormalize[L:(list float)]: (list float) :=
   Cases L of
       nil => (nil ?)
      | (cons x L1) => Cases (Fnum x) of
                           ZERO => (expNormalize L1)
                          | (POS p) => (cons x (expNormalize L1))
                          | (NEG p) => (cons x (expNormalize L1))
                       end
   end.

Theorem expInv:
  (a:float) (L:(list float)) (IsExpansion (cons a L)) ->(IsExpansion L).
Intros a L; Generalize a; Clear a; Elim L; Auto.
Intros a H'; Apply IsExpansionNil.
Intros a l H' a0 H'0; Inversion H'0; Auto.
Apply IsExpansionTop1; Auto.
Apply (H' a0); Auto.
Qed.

Theorem expBoundedInv:
  (a:float) (L:(list float)) (IsExpansion (cons a L)) ->(Fbounded b a).
Intros a L H'; Inversion H'; Auto.
Qed.

Theorem expNormalizeCorrect:
  (L:(list float))<R> (expValue L)==(expValue (expNormalize L)).
Intros L; Elim L; Simpl; Auto.
Intros a l H'; CaseEq (Fnum a); Simpl; Unfold FtoRradix;
 Repeat Rewrite Fplus_correct; Auto with arith; Unfold FtoRradix in H';
 Rewrite H'; Auto.
Intros H'0; Rewrite is_Fzero_rep1; Auto; AutoRewrite [Rsimpl]; Auto.
Qed.

Theorem expNormalizeExp:
  (L:(list float)) (IsExpansion L) ->(IsExpansion (expNormalize L)).
Intros L H'; Elim H'; Auto.
Simpl; Apply IsExpansionNil; Auto.
Simpl; Intros a Ha; Case (Fnum a); Try Apply IsExpansionNil; Intros;
 Apply IsExpansionSingle; Auto.
Simpl; Intros x L0 H'0 H'1 H'2 H'3; Rewrite H'0; Auto.
Intros x y L0 H'0 H'1 H'2 H'3; Case (is_FzeroP x); Intros Z0.
Simpl; Rewrite Z0; Rewrite H'1; Auto.
Generalize Z0; Unfold is_Fzero; Simpl; Case (Fnum x);
 Try (Intros tmp; Case tmp; Auto; Fail); Rewrite H'1; Auto.
Intros x y L0 H'0 H'1 H'2 H'3 H'4 H'5; Simpl; Generalize H'0 H'1;
 Unfold is_Fzero; Case (Fnum x); Try (Intros tmp; Case tmp; Auto; Fail);
 Intros p H; Case (Fnum y); Try (Intros tmp; Case tmp; Auto; Fail); Intros;
 Apply IsExpansionTop; Auto.
Qed.
(*A notion of expansion without zero*)

Inductive IsNzExpansion: (list float) ->Prop :=
     IsNzExpansionNil: (IsNzExpansion (nil ?))
    | IsNzExpansionSingle:
        (x:float) ~ (is_Fzero x) -> (Fbounded b x) ->
        (IsNzExpansion (cons x (nil ?)))
    | IsNzExpansionTop:
        (x, y:float)
        (L:(list float))
        ~ (is_Fzero x) ->
        ~ (is_Fzero y) ->
        (Fbounded b x) ->
        (Fbounded b y) ->
        (Zlt (MSB radix x) (LSB radix y)) -> (IsNzExpansion (cons y L)) ->
        (IsNzExpansion (cons x (cons y L))) .

Theorem isNzExpansionIsExpansion:
  (l:(list float)) (IsNzExpansion l) ->(IsExpansion l).
Intros l H'; Elim H'; Simpl; Auto.
Apply IsExpansionNil.
Intros x H'0 H'1; Apply IsExpansionSingle; Auto.
Intros x y L H'0 H'1 H'2 H'3 H'4 H'5 H'6.
Apply IsExpansionTop; Auto.
Qed.

Theorem IsNzExpansionInv:
  (a:float) (l:(list float)) (IsNzExpansion (cons a l)) ->(IsNzExpansion l).
Intros a l H'; Inversion H'; Auto.
Apply IsNzExpansionNil; Auto.
Qed.
(*A notion of expansion without zero. in reverse order*)

Inductive IsNzExpansionR: (list float) ->Prop :=
     IsNzExpansionNilR: (IsNzExpansionR (nil ?))
    | IsNzExpansionSingleR:
        (x:float) ~ (is_Fzero x) -> (Fbounded b x) ->
        (IsNzExpansionR (cons x (nil ?)))
    | IsNzExpansionTopR:
        (x, y:float)
        (L:(list float))
        ~ (is_Fzero x) ->
        ~ (is_Fzero y) ->
        (Fbounded b x) ->
        (Fbounded b y) ->
        (Zlt (MSB radix y) (LSB radix x)) -> (IsNzExpansionR (cons y L)) ->
        (IsNzExpansionR (cons x (cons y L))) .

Theorem isNzExpansionRIsNzExpansion:
  (l:(list float)) (IsNzExpansionR l) ->(IsNzExpansion (rev l)).
Intros l H'; Elim H'; Simpl; Auto.
Apply IsNzExpansionNil.
Exact IsNzExpansionSingle.
Intros x y L H'0 H'1 H'2 H'3 H'4 H'5; Elim (rev L); Simpl.
Intros H'6; Apply IsNzExpansionTop; Auto.
Apply IsNzExpansionSingle; Auto.
Intros a l0; Case l0; Simpl.
Intros H'6 H'7; Apply IsNzExpansionTop; Auto.
Inversion H'7; Auto.
Inversion H'7; Auto.
Inversion H'7; Auto.
Apply IsNzExpansionTop; Auto.
Apply IsNzExpansionSingle; Auto.
Intros f l1 H'6 H'7; Apply IsNzExpansionTop; Auto.
Inversion H'7; Auto.
Inversion H'7; Auto.
Inversion H'7; Auto.
Inversion H'7; Auto.
Inversion H'7; Auto.
Apply H'6.
Inversion H'7; Auto.
Qed.

Theorem IsNzExpansionRInv:
  (a:float) (l:(list float)) (IsNzExpansionR (cons a l)) ->(IsNzExpansionR l).
Intros a l H'; Inversion H'; Auto.
Apply IsNzExpansionNilR; Auto.
Qed.

Theorem mkIsNzAux:
  (n:nat)
  (p:float)
  (Zle (Zopp (dExp b)) (LSB radix p)) ->
  (Zlt (MSB radix p) (Zplus (Zopp (dExp b)) n)) ->
  (Ex [l:(list float)] (IsNzExpansionR l) /\ (<R> p==(expValue l) /\ Cases l of
                                                                         (cons
                                                                            a _)
                                                                           =>
                                                                           (ToZeroP
                                                                              b
                                                                              radix
                                                                              p
                                                                              a)
                                                                        | nil =>
                                                                            True
                                                                     end)).
Intros n; Elim n.
Intros p H' H'0; Case (is_FzeroP p); Intros H'1.
Exists (nil float); Repeat Split; Simpl; Auto.
Apply IsNzExpansionNilR.
Replace (FtoRradix (Fzero ZERO)) with R0.
Unfold FtoRradix; Apply is_Fzero_rep1; Auto.
Unfold FtoRradix FtoR Fzero; Simpl; AutoRewrite [Rsimpl]; Auto.
Absurd (Zle (LSB radix p) (MSB radix p)).
Apply Zlt_not_le.
Apply Zlt_le_trans with 1 := H'0.
Replace (Zplus (Zopp (dExp b)) O) with (Zopp (dExp b)); [Auto | Ring].
Apply LSB_le_MSB; Auto.
Intros n0 H' p H'0 H'1.
Case (is_FzeroP p); Intros H'2.
Exists (nil float); Repeat Split; Simpl; Auto.
Apply IsNzExpansionNilR.
Replace (FtoRradix (Fzero ZERO)) with R0.
Unfold FtoRradix; Apply is_Fzero_rep1; Auto.
Unfold FtoRradix FtoR Fzero; Simpl; AutoRewrite [Rsimpl]; Auto.
Cut (TotalP (ToZeroP b radix));
 [Intros tmp1; Case (tmp1 (FtoRradix p)); Clear tmp1; Intros a Ha |
   Apply ToZeroTotal with precision := precision]; Auto.
Cut ~ (is_Fzero a); [Intros Na | Idtac].
2:Red; Intros H'6; Absurd <R> a==R0.
2:Apply (MSBBoundNotZero b radix precision) with P := (ToZeroP b radix) f1 := p;
   Auto.
2:Apply ToZeroRoundedModeP with precision := precision; Auto.
2:Contradict H'2.
2:Apply is_Fzero_rep2 with radix := radix; Auto.
2:Apply Zle_trans with 1 := H'0.
2:Apply LSB_le_MSB; Auto.
2:Unfold FtoRradix; Apply is_Fzero_rep1; Auto.
Case (is_FzeroP (Fminus radix p a)); Intros H'3.
Exists (cons a (nil ?)); Repeat Split; Auto.
Apply IsNzExpansionSingleR; Auto.
Apply RoundedModeBounded
     with radix := radix P := (ToZeroP b radix) r := (FtoRradix p); Auto.
Apply ToZeroRoundedModeP with precision := precision; Auto.
Simpl.
Replace (FtoRradix p) with (FtoRradix (Fplus radix (Fminus radix p a) a)).
Unfold FtoRradix; Repeat Rewrite Fplus_correct; Auto with real arith.
Replace (FtoR radix (Fminus radix p a)) with R0.
Replace (FtoR radix (Fzero ZERO)) with R0.
Apply Rplus_sym.
Unfold FtoRradix FtoR Fzero; Simpl; AutoRewrite [Rsimpl]; Auto.
Apply sym_eqT; Apply is_Fzero_rep1; Auto.
Unfold FtoRradix; Repeat Rewrite Fplus_correct; Auto with real arith;
 Repeat Rewrite Fminus_correct; Auto with real arith; Ring.
Elim (H' (Fminus radix p a));
 [Intros l E; Elim E; Intros H'7 H'8; Elim H'8; Intros H'9 H'10; Clear H'8 E |
   Idtac | Idtac]; Auto.
Exists (cons a l); Repeat Split; Auto.
Generalize H'7 H'10; Case l.
Intros H'4 H'5; Apply IsNzExpansionSingleR; Auto.
Apply RoundedModeBounded
     with radix := radix P := (ToZeroP b radix) r := (FtoRradix p); Auto.
Apply ToZeroRoundedModeP with precision := precision; Auto.
Intros f l0 H'4 H'5.
Apply IsNzExpansionTopR; Auto.
Inversion H'4; Auto.
Apply RoundedModeBounded
     with radix := radix P := (ToZeroP b radix) r := (FtoRradix p); Auto.
Apply ToZeroRoundedModeP with precision := precision; Auto.
Inversion H'4; Auto.
Replace (MSB radix f) with (MSB radix (Fminus radix p a)).
Apply (MSBroundLSB b radix precision) with P := (ToZeroP b radix); Auto.
Apply ToZeroRoundedModeP with precision := precision; Auto.
Apply (MSBtoZero b radix precision); Auto.
Inversion H'4; Auto.
Simpl; Unfold FtoRradix; Rewrite Fplus_correct; Auto with arith.
Unfold FtoRradix in H'9; Rewrite <- H'9.
Rewrite Fminus_correct; Auto with arith; Ring.
Apply Zle_trans with (Zmin (LSB radix p) (LSB radix a)).
Apply Zmin_Zle; Auto.
Apply Zle_trans with (Fexp a).
Cut (Fbounded b a); [Intros Fb0; Intuition | Idtac].
Apply RoundedModeBounded
     with radix := radix P := (ToZeroP b radix) r := (FtoRradix p); Auto.
Apply ToZeroRoundedModeP with precision := precision; Auto.
Apply Fexp_le_LSB; Auto.
Apply LSBMinus; Auto.
Apply Zlt_le_trans with (LSB radix a).
Apply (MSBroundLSB b radix precision) with P := (ToZeroP b radix); Auto.
Apply ToZeroRoundedModeP with precision := precision; Auto.
Apply Zle_trans with (MSB radix a); Auto.
Apply LSB_le_MSB; Auto.
Rewrite <- (MSBtoZero b radix precision) with f1 := p f2 := a; Auto.
Apply Zlt_n_Sm_le.
Replace (Zs (Zplus (Zopp (dExp b)) n0)) with (Zplus (Zopp (dExp b)) (S n0));
 Auto.
Rewrite inj_S; Unfold Zs; Ring.
Qed.

Theorem expValueApp:
  (l1, l2:(list float))
  <R> (expValue (app l1 l2))==(Rplus (expValue l1) (expValue l2)).
Intros l1; Elim l1; Simpl; Auto.
Intros l2; Unfold FtoRradix FtoR Fzero; Simpl; AutoRewrite [Rsimpl]; Auto.
Intros a l H' l2.
Unfold FtoRradix; Repeat Rewrite Fplus_correct; Auto with zarith.
Rewrite H'; Unfold FtoRradix; Ring.
Qed.

Theorem expValueRev: (l:(list float))<R> (expValue (rev l))==(expValue l).
Intros l; Elim l; Simpl; Auto.
Intros a l0 H'; Rewrite expValueApp; Simpl.
Rewrite H'.
Unfold FtoRradix; Repeat Rewrite Fplus_correct; Auto with zarith.
Replace (FtoR radix (Fzero ZERO)) with R0; [Ring | Idtac].
Unfold FtoRradix FtoR Fzero; Simpl; AutoRewrite [Rsimpl]; Auto.
Qed.

Theorem existExp:
  (p:float) (Zle (Zopp (dExp b)) (LSB radix p)) ->
  (Ex [l:(list float)] (IsExpansion l) /\ <R> p==(expValue l)).
Intros p H'.
Elim (mkIsNzAux (plus (S (dExp b)) (absolu (MSB radix p))) p);
 [Intros l E; Elim E; Intros H'3 H'4; Elim H'4; Intros H'5 H'6; Clear H'4 E |
   Idtac | Idtac]; Auto.
Exists (rev l); Split; Auto.
Apply isNzExpansionIsExpansion.
Apply isNzExpansionRIsNzExpansion; Auto.
Rewrite expValueRev; Auto.
Apply Zle_lt_trans with (inject_nat (absolu (MSB radix p))).
Rewrite <- Zabs_absolu.
Case (MSB radix p); Simpl; Auto with zarith; Intros; Unfold Zle; Simpl; Red;
 Intros; Discriminate.
Rewrite inj_plus.
Rewrite Zplus_assoc.
Replace (Zplus (Zopp (dExp b)) (S (dExp b))) with (Zs ZERO).
Auto with zarith.
Rewrite inj_S; Unfold Zs; Ring.
Qed.
End mf.

