
(****************************************************************************
                                                                             
          IEEE754  :  Fexp2                                                     
                                                                             
                 Sylvie Boldo                                           
                                                                             
*****************************************************************************
*)
Require Export ThreeSumProps.
Require Export PolyList.
Section Fexp2.
Variable b:Fbound.
Variable precision:nat.

Local FtoRradix := (FtoR (2)).
Coercion FtoRradix : float >-> R.
Hypothesis precisionGreaterThanOne:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp (2) precision) (1)).
Hypothesis Ngd:(Rle R1 (Rmult (vNum b) (Rminus R1 (Rinv (2))))).
Hypothesis Ngd2:
           (Rle (6) (Rmult (vNum b) (Rminus R1 (Rmult (Rinv (2)) (Rinv (2)))))).

Inductive IsExp: (list float) ->Prop :=
     IsExpNil: (IsExp (nil ?))
    | IsExpSingle: (x:float) (Fbounded b x) ->(IsExp (cons x (nil ?)))
    | IsExpTop:
        (x, y:float)
        (L:(list float))
        (Fbounded b x) ->
        (Fbounded b y) -> (Zle (Fexp y) (Fexp x)) -> (IsExp (cons y L)) ->
        (IsExp (cons x (cons y L))) .

Inductive NearEqual: (list float) -> (list float) ->Prop :=
     IsEqual: (x:(list float))(NearEqual x x)
    | OneMoreR:
        (x:(list float)) (e:float) (Fbounded b e) ->(NearEqual x (cons e x)) .

Fixpoint sum[L:(list float)]: R :=
   Cases L of
       nil => R0
      | (cons x L1) => (Rplus (FtoRradix x) (sum L1))
   end.

Definition hdexp :=
   [L:(list float)]
      Cases L of
          nil => (Zopp (dExp b))
         | (cons x L1) => (Fexp x)
      end.

Fixpoint lastexp[L:(list float)]: Z :=
   Cases L of
       nil => (Zopp (dExp b))
      | (cons x nil) => (Fexp x)
      | (cons x L1) => (lastexp L1)
   end.

Definition hd :=
   [L:(list float)]
      Cases L of
          nil => (Fzero (Zopp (dExp b)))
         | (cons x L1) => x
      end.

Theorem IsExpZle:
  (i:float) (L:(list float)) (IsExp (cons i L)) ->(Zle (hdexp L) (Fexp i)).
Induction L; Simpl; Intros.
Cut (Fbounded b i).
Intros H'; Elim H'; Intros; Auto.
Inversion H; Auto.
Inversion H0; Auto.
Qed.

Theorem sum_IsExp:
  (L:(list float))
  (x:float) (m:R) (IsExp (cons x L)) -> (Rle (Float (vNum b) (Fexp x)) m) ->
  (Rle (Rabsolu (sum (cons x L))) (Rmult (length (cons x L)) m)).
Intros.
Replace (sum (cons x L)) with (Rplus x (sum L)); [Idtac | Simpl; Auto with real].
Apply Rle_trans with (Rplus (Rabsolu x) (Rabsolu (sum L)));
 [Apply Rabsolu_triang |
   Replace (INR (length (cons x L))) with (Rplus R1 (length L))].
Replace (Rmult (Rplus R1 (length L)) m) with (Rplus m (Rmult (length L) m));
 [Idtac | Ring].
Apply Rle_trans with (Rplus m (Rabsolu (sum L))).
Apply Rle_compatibility_r.
Unfold FtoRradix; Apply Rle_trans with (FtoR (2) (Float (vNum b) (Fexp x)));
 [Apply MaxFloat with precision | Idtac]; Auto.
Inversion H; Auto.
Apply Rle_compatibility.
Induction L; Intros.
Simpl; Right; Ring; Apply Rabsolu_R0.
Replace (sum (cons a L)) with (Rplus a (sum L)); [Idtac | Simpl; Auto with real].
Apply Rle_trans with (Rplus (Rabsolu a) (Rabsolu (sum L)));
 [Apply Rabsolu_triang |
   Replace (INR (length (cons a L))) with (Rplus R1 (length L))].
Replace (Rmult (Rplus R1 (length L)) m) with (Rplus m (Rmult (length L) m));
 [Idtac | Ring].
Apply Rle_trans with (Rplus m (Rabsolu (sum L))).
Apply Rle_compatibility_r.
Unfold FtoRradix; Apply Rle_trans with (FtoR (2) (Float (vNum b) (Fexp a)));
 [Apply MaxFloat with precision | Idtac]; Auto.
Inversion H; Auto.
Apply Rle_trans with (FtoRradix (Float (vNum b) (Fexp x))); Auto.
Unfold FtoRradix; Unfold FtoR; Simpl.
Rewrite <- INR_IZR_INZ.
Apply Rle_monotony;
 [Apply vNumPos with precision; Auto | Apply Rle_powerRZ; Auto with real].
Replace (Rplus R1 R1) with (INR (2)); [Replace R1 with (INR (1)) | Idtac];
 Auto with real arith.
Inversion H; Auto.
Apply Rle_compatibility.
Apply HrecL.
Generalize H; Case L; Intros.
Apply IsExpSingle; Inversion H1; Auto.
Apply IsExpTop; Inversion H1; Auto; Inversion H8; Auto.
Apply Zle_trans with (Fexp a); Auto.
Rewrite (Rplus_sym R1 (length L)); Simpl; Auto with real arith.
Case (length L); Intros; Auto with real arith.
Rewrite (Rplus_sym R1 (length L)); Simpl; Auto with real arith.
Case (length L); Intros; Auto with real arith.
Qed.

Inductive IsRleExp: (list float) ->Prop :=
     IsRleExpNil: (IsRleExp (nil ?))
    | IsRleExpSingle: (x:float) (Fbounded b x) ->(IsRleExp (cons x (nil ?)))
    | IsRleExpTop:
        (x, y:float)
        (L:(list float))
        (Fbounded b x) ->
        (Fbounded b y) ->
        (Rle (Rabsolu x) (Rabsolu y)) -> (IsRleExp (cons y L)) ->
        (IsRleExp (cons x (cons y L))) .

Inductive EqListFloat: (list float) -> (list float) ->Prop :=
     EqListFloatnil: (EqListFloat (nil float) (nil float))
    | EqListFloatTop:
        (x, y:float)
        (L, L':(list float))
        (Fbounded b x) -> (Fbounded b y) -> <R> x==y -> (EqListFloat L L') ->
        (EqListFloat (cons x L) (cons y L')) .

Theorem sum_app:
  (L:(list float)) (x:float)
  <R> (Rplus x (sum L))==((sum (app L (cons x (nil float))))).
Intros.
Induction L; Intros; Auto.
Simpl.
Replace (Rplus x (Rplus a (sum L))) with (Rplus a (Rplus x (sum L)));
 [Rewrite HrecL | Ring]; Auto.
Qed.

Theorem cons_neq: (x:float) (L:(list float))~ <(list float)> (cons x L)=L.
Cut (L1, L2:(list float)) (lt (length L1) (length L2)) ->~ L1=L2.
Intros Z.
Intros; Apply sym_not_eq; Apply Z.
Induction L; Intros; Simpl; Auto.
Intros.
Contradict H.
Replace L1 with L2; Auto with arith.
Qed.

Definition endof :=
   [all, part:(list float)](Ex [rest:(list float)] all=(app rest part)).

Theorem app_length:
  (l1, l2:(list float))(length (app l1 l2))=(plus (length l1) (length l2)).
Intros l1; Induction l1.
Intros; Simpl; Auto.
Intros; Simpl.
Rewrite (Hrecl1 l2); Auto.
Qed.

Theorem endof_length:
  (L, l:(list float)) (endof L l) ->(le (length l) (length L)).
Intros.
Inversion H.
Rewrite H0.
Rewrite (app_length x l); Auto with arith.
Qed.

Inductive IsCanExp: (list float) ->Prop :=
     IsCanExpNil: (IsCanExp (nil ?))
    | IsCanExpTop:
        (x:float)
        (L:(list float)) (Fcanonic (2) b precision x) -> (IsCanExp L) ->
        (IsCanExp (cons x L)) .

Theorem IsCanExpBounded:
  (i:float) (L:(list float)) (IsCanExp (cons i L)) ->(Fbounded b i).
Intros; Apply FcanonicBound with (2) precision.
Inversion H; Auto.
Qed.

Inductive IsRleExpRev: (list float) ->Prop :=
     IsRleExpRevNil: (IsRleExpRev (nil ?))
    | IsRleExpRevSingle:
        (x:float) (Fbounded b x) ->(IsRleExpRev (cons x (nil ?)))
    | IsRleRevExpTop:
        (x, y:float)
        (L:(list float))
        (Fbounded b x) ->
        (Fbounded b y) ->
        (Rle (Rabsolu y) (Rabsolu x)) -> (IsRleExpRev (cons y L)) ->
        (IsRleExpRev (cons x (cons y L))) .

Theorem IsRleExpRevIsExp:
  (L:(list float)) (IsRleExpRev L) ->
  (Ex [L':(list float)]
  (IsCanExp L') /\ ((IsRleExpRev L') /\ ((EqListFloat L L') /\ (IsExp L')))).
Intros L; Induction L; Intros.
Exists (nil float); Repeat Split;
 [Apply IsCanExpNil | Apply IsRleExpRevNil | Apply EqListFloatnil |
   Apply IsExpNil].
Elim HrecL; [Intros L' E; Clear HrecL | Inversion H; Auto; Apply IsRleExpRevNil].
Elim E; Intros; Clear E.
Elim H1; Intros; Clear H1.
Elim H3; Intros; Clear H3.
Cut (Fbounded b a); [Intros | Inversion H; Auto].
Cut (Fcanonic (2) b precision (Fnormalize (2) b precision a));
 [Intros | Apply FnormalizeCanonic; Auto with arith].
Cut (Fbounded b (Fnormalize (2) b precision a));
 [Intros | Apply FnormalizeBounded; Auto with arith].
Cut <R> (Fnormalize (2) b precision a)==a;
 [Intros | Unfold FtoRradix; Apply FnormalizeCorrect; Auto with arith].
Exists (cons (Fnormalize (2) b precision a) L'); Split.
Apply IsCanExpTop; Auto.
Split; [Generalize H1 H2; Case L'; Auto; Intros | Idtac].
Apply IsRleExpRevSingle; Auto.
Apply IsRleRevExpTop; Auto.
Inversion H9; Auto.
Rewrite H7; Inversion H8; Inversion H; Intros.
Generalize H8; Rewrite <- H19; Intros.
Inversion H20; Auto.
Replace (FtoRradix f) with (FtoRradix y0); Auto; Generalize H18; Rewrite <- H14;
 Intros; Auto.
Rewrite <- H15; Inversion H23; Auto.
Split; [Apply EqListFloatTop; Auto | Idtac].
Generalize H0 H4 H2 H1; Case L'; Auto; Intros.
Apply IsExpSingle; Auto.
Apply IsExpTop; Auto.
Inversion H10; Auto.
Apply Fcanonic_Rle_Zle with (2) b precision; Auto with arith.
Inversion H8; Auto.
Generalize H11 H1 H; Case L; Intros.
Inversion H12; Auto.
Inversion H12; Inversion H14; Intros.
Fold FtoRradix; Rewrite H7.
Inversion H12; Intros.
Rewrite <- H36; Auto.
Qed.

Fixpoint last[L:(list float)]: float :=
   Cases L of
       nil => (Fzero (Zopp (dExp b)))
      | (cons x nil) => x
      | (cons x L1) => (last L1)
   end.

Theorem ExpRev_aux:
  (l:(list float))
  (x:float)
  (Fbounded b x) -> (IsRleExpRev l) -> (Rle (Rabsolu x) (Rabsolu (last l))) ->
  (IsRleExpRev (app l (cons x (nil float)))).
Intros l; Induction l.
Simpl; Intros; Apply IsRleExpRevSingle; Auto.
Intros.
Generalize Hrecl H1 H0; Case l.
Simpl; Intros.
Apply IsRleRevExpTop; Auto.
Inversion H3; Auto.
Apply IsRleExpRevSingle; Auto.
Intros.
LApply (Hrecl0 x);
 [Intros H'0; LApply H'0;
   [Intros H'1; LApply H'1;
     [Intros H'2; Try Exact H'2; Clear H'1 H'0 | Clear H'1 H'0] | Clear H'0] |
   Idtac]; Auto.
Replace (app (cons a (cons f l0)) (cons x (nil float)))
     with (cons a (cons f (app l0 (cons x (nil float)))));
 [Apply IsRleRevExpTop | Idtac]; Auto; Try (Inversion H3; Auto).
Inversion H3; Auto.
Qed.

Theorem Exp_aux:
  (l:(list float))
  (x:float)
  (Fbounded b x) -> (IsRleExp l) -> (Rle (Rabsolu (last l)) (Rabsolu x)) ->
  (IsRleExp (app l (cons x (nil float)))).
Intros l; Induction l.
Simpl; Intros; Apply IsRleExpSingle; Auto.
Intros.
Generalize Hrecl H1 H0; Case l.
Simpl; Intros.
Apply IsRleExpTop; Auto.
Inversion H3; Auto.
Apply IsRleExpSingle; Auto.
Intros.
LApply (Hrecl0 x);
 [Intros H'0; LApply H'0;
   [Intros H'1; LApply H'1; [Intros H'2; Clear H'1 H'0 | Clear H'1 H'0] |
     Clear H'0] | Idtac]; Auto.
Replace (app (cons a (cons f l0)) (cons x (nil float)))
     with (cons a (cons f (app l0 (cons x (nil float)))));
 [Apply IsRleExpTop | Idtac]; Auto; Try (Inversion H3; Auto).
Inversion H3; Auto.
Qed.

Theorem last_hd: (l:(list float))(last l)=(hd (rev l)).
Intros l; Induction l.
Simpl; Auto.
Generalize Hrecl; Case l; Intros; Auto.
Replace (last (cons a (cons f l0))) with (last (cons f l0));
 [Idtac | Simpl; Auto].
Replace (hd (rev (cons a (cons f l0)))) with (hd (rev (cons f l0)));
 [Auto | Simpl; Auto].
Simpl.
Cut (l, l':(list float)) ~ l=(nil float) ->(hd (app l l'))=(hd l).
Intros K.
LApply (K (app (rev l0) (cons f (nil float))) (cons a (nil float)));
 [Intros H'1 | Idtac]; Auto.
Apply sym_not_eq; Apply app_cons_not_nil.
Intros l1; Case l1; Auto; Intros.
Absurd <(list float)> (nil float)=(nil float); Auto.
Qed.

Theorem IsRleExpRev_IsRleExp:
  (l:(list float)) (IsRleExpRev l) ->(IsRleExp (rev l)).
Intros l; Induction l; Intros; Auto.
Simpl; Apply IsRleExpNil.
LApply Hrecl;
 [Intros H'; Clear Hrecl | Clear Hrecl; Inversion H; Auto; Apply IsRleExpRevNil].
Simpl.
Generalize H H'; Induction l; Intros.
Simpl; Apply IsRleExpSingle; Inversion H; Auto.
Apply Exp_aux; Auto.
Inversion H; Auto.
Replace (last (rev (cons a0 l))) with a0; [Inversion H0; Auto | Auto].
Rewrite (last_hd (rev (cons a0 l))).
Rewrite (idempot_rev (cons a0 l)); Auto.
Qed.

Theorem IsRleExp_IsRleExpRev:
  (l:(list float)) (IsRleExp l) ->(IsRleExpRev (rev l)).
Intros l; Induction l; Intros; Auto.
Simpl; Apply IsRleExpRevNil.
LApply Hrecl;
 [Intros H'; Clear Hrecl | Clear Hrecl; Inversion H; Auto; Apply IsRleExpNil].
Simpl.
Generalize H H'; Induction l; Intros.
Simpl; Apply IsRleExpRevSingle; Inversion H; Auto.
Apply ExpRev_aux; Auto.
Inversion H; Auto.
Replace (last (rev (cons a0 l))) with a0; [Inversion H0; Auto | Auto].
Rewrite (last_hd (rev (cons a0 l))).
Rewrite (idempot_rev (cons a0 l)); Auto.
Qed.

Theorem EqListFloat_length:
  (l, l':(list float)) (EqListFloat l l') ->(length l)=(length l').
Intros l; Induction l; Intros.
Inversion H; Auto.
Inversion H; Intros.
LApply (Hrecl L'); [Intros H'0 | Idtac]; Auto.
Replace (length (cons a l)) with (S (length l));
 Replace (length (cons y L')) with (S (length L')); Auto with arith.
Qed.

Theorem EqListFloat_sum:
  (l, l':(list float)) (EqListFloat l l') -><R> ((sum l))==((sum l')).
Intros l; Induction l; Intros.
Inversion H; Auto.
Inversion H; Intros.
LApply (Hrecl L'); [Intros H'0 | Idtac]; Auto.
Replace (sum (cons a l)) with (Rplus a (sum l));
 Replace (sum (cons y L')) with (Rplus y (sum L')); Auto with arith.
Rewrite H4; Auto with real.
Qed.

Theorem rev_sum: (l:(list float))<R> ((sum l))==((sum (rev l))).
Intros l; Induction l.
Simpl; Auto with real.
Simpl.
Rewrite <- (sum_app (rev l) a); Auto with real.
Qed.

Theorem rev_length: (l:(list float))(length l)=(length (rev l)).
Intros l; Induction l; Simpl; Auto.
Rewrite (app_length (rev l) (cons a (nil float))).
Rewrite Hrecl; Simpl; Ring.
Qed.
End Fexp2.
