
(****************************************************************************
                                                                             
          IEEE754  :  FexpAdd                                                     
                                                                             
                 Sylvie Boldo                                           
                                                                             
*****************************************************************************
*)
Require Export ThreeSumProps.
Require Export PolyList.
Require Export Fexp2.
Section FexpAdd.
Variable b:Fbound.
Variable precision:nat.

Local FtoRradix := (FtoR (2)).
Coercion FtoRradix : float >-> R.
Hypothesis precisionGreaterThanOne:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp (2) precision) (1)).
Hypothesis Ngd:(Rle R1 (Rmult (vNum b) (Rminus R1 (Rinv (2))))).
Hypothesis Ngd2:
           (Rle (6) (Rmult (vNum b) (Rminus R1 (Rmult (Rinv (2)) (Rinv (2)))))).

Inductive NearEqual: (list float) -> (list float) ->Prop :=
     IsEqual: (x:(list float))(NearEqual x x)
    | OneMoreR:
        (x:(list float)) (e:float) (Fbounded b e) ->(NearEqual x (cons e x)) .

Definition Step :=
   [x, y, i, x', y':float] [input, output, output':(list float)]
   (Fbounded b x') /\
   ((Fbounded b y') /\
    ((NearEqual output output') /\
     (<R>
        (Rplus x (Rplus y (Rplus (sum (cons i input)) (sum output))))==
        (Rplus x' (Rplus y' (Rplus (sum input) (sum output')))) /\
      ((Zle (Fexp i) (Fexp y')) /\
       ((Zle (Fexp y') (Fexp x')) /\
        ((IsExp b (cons i input)) /\
         ((<R> x'==R0 \/
           (<R> y'==R0 \/
            (Rle
               (Rabsolu y')
               (Rmult
                  (vNum b)
                  (Rminus (Float (1) (Fexp x')) (Float (1) (hdexp b input)))))
            /\
            (Rle
               (Rabsolu y')
               (Rmult
                  (vNum b) (Rminus (Float (1) (Fexp x')) (Float (1) (Fexp i)))))))
          /\
          ((output'=output \/
            (Rle
               (Rabsolu (Rplus x' y'))
               (Rmult
                  (Rmult (Rmult (3) (2)) (Rinv (vNum b)))
                  (Rabsolu (hd b output'))))) /\
           (output'=(nil float) \/
            (Rle
               (Rabsolu (hd b input))
               (Rmult
                  (3)
                  (Rmult
                     (2)
                     (Rmult
                        (Rabsolu (hd b output')) (Rinv (Rminus (vNum b) (1)))))))
            /\
            (input=(nil float) \/
             (Rle
                (Float (vNum b) (Fexp (hd b input)))
                (Rmult
                   (3)
                   (Rmult
                      (2)
                      (Rmult
                         (Rabsolu (hd b output')) (Rinv (Rminus (vNum b) (1)))))))
             /\
             (Rle
                (Rabsolu (sum input))
                (Rmult
                   (length input)
                   (Rmult
                      (3)
                      (Rmult
                         (2)
                         (Rmult
                            (Rabsolu (hd b output'))
                            (Rinv (Rminus (vNum b) (1)))))))))))))))))).

Theorem AddStep:
  (x, y, i:float)
  (input, output:(list float))
  (Fbounded b x) ->
  (Fbounded b y) ->
  (IsExp b (cons i input)) ->
  (Zle (Fexp i) (Fexp y)) ->
  (Zle (Fexp y) (Fexp x)) ->
  (<R> (FtoR (2) x)==R0 \/ (<R> (FtoR (2) y)==R0 \/ <R> (FtoR (2) i)==R0)) \/
  (Rle
     (Rabsolu (FtoR (2) y))
     (Rmult
        (vNum b)
        (Rminus (FtoR (2) (Float (1) (Fexp x))) (FtoR (2) (Float (1) (Fexp i)))))) ->
  output=(nil float) \/
  (Rle
     (Rabsolu i)
     (Rmult
        (3)
        (Rmult (2) (Rmult (Rabsolu (hd b output)) (Rinv (Rminus (vNum b) (1)))))))
  /\
  ((Rle
      (Float (vNum b) (Fexp i))
      (Rmult
         (3)
         (Rmult
            (2) (Rmult (Rabsolu (hd b output)) (Rinv (Rminus (vNum b) (1)))))))
   /\
   (Rle
      (Rabsolu (sum (cons i input)))
      (Rmult
         (length (cons i input))
         (Rmult
            (3)
            (Rmult
               (2) (Rmult (Rabsolu (hd b output)) (Rinv (Rminus (vNum b) (1))))))))) ->
  (Ex [x':float]
  (Ex [y':float]
  (Ex [output':(list float)] (Step x y i x' y' input output output')))).
Cut (Rlt R0 (Rminus (vNum b) R1)); [Intros Z1 | Idtac].
2:Apply Rlt_anti_compatibility with (INR (1)).
2:Replace (Rplus (1) R0) with (INR (1));
   Replace (Rplus (1) (Rminus (vNum b) R1)) with (INR (vNum b)); Try Ring.
2:Rewrite pGivesBound.
2:Apply Rlt_anti_compatibility with (INR (1)).
2:Replace (Rplus (1) (1)) with (INR (2));
   Replace (Rplus (1) (minus (exp (2) precision) (1)))
        with (INR (exp (2) precision)); Try Ring.
2:Apply Rlt_INR.
2:Pattern 1 (2); Replace (2) with (exp (2) (1));
   [Apply exp_monotone_lt | Idtac]; Auto with arith.
2:Rewrite (minus_INR (exp (2) precision) (1)); [Ring | Idtac].
2:Replace (1) with (exp (2) (0)); [Apply exp_monotone_le | Idtac];
   Auto with arith.
2:Rewrite (minus_INR (exp (2) precision) (1)); [Ring | Idtac].
2:Replace (1) with (exp (2) (0)); [Apply exp_monotone_le | Idtac];
   Auto with arith.
Cut (Rlt R0 (vNum b)); [Intros Z2 | Idtac].
2:Replace R0 with (INR (0)); Auto with real.
2:Apply Rlt_INR.
2:Apply lt_vNum with precision := precision radix := (2); Auto with real arith.
Intros x y i input output H H0 H1 H2 H3 H4 G.
Cut (Fbounded b i); [Intros K | Inversion H1; Auto].
Generalize ThreeSumLoop; Intros L.
Generalize (ClosestTotal b (2)); Intros ExC.
LApply (ExC precision);
 [Intros H'0; LApply H'0;
   [Intros H'1; LApply H'1; [Intros H'2; Clear H'1 H'0 | Clear H'1 H'0] |
     Clear H'0] | Idtac]; Clear ExC; Auto.
Generalize H'2; Unfold TotalP; Intros ExC; Clear H'2.
Generalize (ExC (Rplus y i)); Intros Z; Case Z; Intros u Hu; Clear Z.
Generalize (ExC (Rplus x u)); Intros Z; Case Z; Intros p' Hp'; Clear Z.
Generalize (ExC (Rplus (Rminus (Rplus x u) p') (Rminus (Rplus y i) u)));
 Intros Z; Case Z; Intros q' Hq'; Clear Z.
LApply (L b precision);
 [Intros H'1; LApply H'1; [Intros H'2; Clear H'1 | Clear H'1] | Idtac]; Auto.
Generalize errorBoundedPlus; Intros.
LApply (H5 b (2) precision);
 [Intros H'3; LApply H'3;
   [Intros H'4; LApply H'4; [Intros H'5; Clear H'4 H'3 | Clear H'4 H'3] |
     Clear H'3] | Idtac]; Auto.
Generalize (H'5 y i u); Intros; Auto.
LApply H6;
 [Intros H'; LApply H';
   [Intros H'0; LApply H'0; [Intros H'1; Clear H'0 H' H6 | Clear H'0 H' H6] |
     Clear H' H6] | Clear H6]; Auto.
Elim H'1; Intros v E; Clear H'1.
Case E; Intros H'1 H'2; Clear E.
Case H'0; Intros; Clear H'0.
Generalize (H'5 x u p'); Intros; Auto.
LApply H8;
 [Intros H'; LApply H';
   [Intros H'0; LApply H'0; [Intros H'3; Clear H'0 H' H8 | Clear H'0 H' H8] |
     Clear H' H8] | Clear H8]; Auto.
Elim H'3; Intros w E; Clear H'3.
Case E; Intros H'3 H'2; Clear E.
Case H'0; Intros; Clear H'0.
Generalize (H'5 w v q'); Intros; Auto.
LApply H10;
 [Intros H'; LApply H';
   [Intros H'0; LApply H'0; [Intros H'4; Clear H'0 H' H10 | Clear H'0 H' H10] |
     Clear H' H10] | Clear H10]; Auto.
Elim H'4; Intros r' E; Clear H'4.
Case E; Intros H'4 H'2; Clear E.
Case H'0; Intros; Clear H'0.
2:Replace (FtoR (2) v)
       with (Rminus (Rplus (FtoR (2) y) (FtoR (2) i)) (FtoR (2) u));
   Replace (FtoR (2) w)
        with (Rminus (Rplus (FtoR (2) x) (FtoR (2) u)) (FtoR (2) p'));
   Auto with real.
LApply (H'2 x y i u v w p' q' r');
 [Intros H'13; LApply H'13;
   [Intros H'14; LApply H'14;
     [Intros H'15; LApply H'15;
       [Intros H'16; LApply H'16;
         [Intros H'17; LApply H'17;
           [Intros H'18; LApply H'18;
             [Intros H'19; LApply H'19;
               [Intros H'20; LApply H'20;
                 [Intros H'21; LApply H'21;
                   [Intros H'22; LApply H'22;
                     [Intros H'23; LApply H'23;
                       [Intros H'24; LApply H'24;
                         [Intros H'25; LApply H'25;
                           [Intros H'26; LApply H'26;
                             [Intros H'27; LApply H'27;
                               [Intros H'28; LApply H'28;
                                 [Intros H'29; LApply H'29;
                                   [Intros H'30; LApply H'30;
                                     [Intros H'31; Try Exact H'31;
                                       Clear H'30 H'29 H'28 H'27 H'26 H'25 H'24
                                             H'23 H'22 H'21 H'20 H'19 H'18 H'17
                                             H'16 H'15 H'14 H'13 |
                                       Clear H'30 H'29 H'28 H'27 H'26 H'25 H'24
                                             H'23 H'22 H'21 H'20 H'19 H'18 H'17
                                             H'16 H'15 H'14 H'13] |
                                     Clear H'29 H'28 H'27 H'26 H'25 H'24 H'23
                                           H'22 H'21 H'20 H'19 H'18 H'17 H'16
                                           H'15 H'14 H'13] |
                                   Clear H'28 H'27 H'26 H'25 H'24 H'23 H'22 H'21
                                         H'20 H'19 H'18 H'17 H'16 H'15 H'14 H'13] |
                                 Clear H'27 H'26 H'25 H'24 H'23 H'22 H'21 H'20
                                       H'19 H'18 H'17 H'16 H'15 H'14 H'13] |
                               Clear H'26 H'25 H'24 H'23 H'22 H'21 H'20 H'19
                                     H'18 H'17 H'16 H'15 H'14 H'13] |
                             Clear H'25 H'24 H'23 H'22 H'21 H'20 H'19 H'18 H'17
                                   H'16 H'15 H'14 H'13] |
                           Clear H'24 H'23 H'22 H'21 H'20 H'19 H'18 H'17 H'16
                                 H'15 H'14 H'13] |
                         Clear H'23 H'22 H'21 H'20 H'19 H'18 H'17 H'16 H'15 H'14
                               H'13] |
                       Clear H'22 H'21 H'20 H'19 H'18 H'17 H'16 H'15 H'14 H'13] |
                     Clear H'21 H'20 H'19 H'18 H'17 H'16 H'15 H'14 H'13] |
                   Clear H'20 H'19 H'18 H'17 H'16 H'15 H'14 H'13] |
                 Clear H'19 H'18 H'17 H'16 H'15 H'14 H'13] |
               Clear H'18 H'17 H'16 H'15 H'14 H'13] |
             Clear H'17 H'16 H'15 H'14 H'13] | Clear H'16 H'15 H'14 H'13] |
         Clear H'15 H'14 H'13] | Clear H'14 H'13] | Clear H'13] | Idtac];
 Auto with arith zarith float real.
2:Replace (FtoR (2) w)
       with (Rminus (Rplus (FtoR (2) x) (FtoR (2) u)) (FtoR (2) p'));
   Auto with real.
2:Replace (FtoR (2) v)
       with (Rminus (Rplus (FtoR (2) y) (FtoR (2) i)) (FtoR (2) u));
   Auto with real.
2:Apply RoundedModeBounded
       with b := b radix := (2) P := (Closest b (2))
            r := (Rplus (Rminus (Rplus x u) p') (Rminus (Rplus y i) u)); Auto.
2:Apply ClosestRoundedModeP with precision := precision; Auto.
2:Apply RoundedModeBounded
       with b := b radix := (2) P := (Closest b (2)) r := (Rplus x u); Auto.
2:Apply ClosestRoundedModeP with precision := precision; Auto.
2:Apply RoundedModeBounded
       with b := b radix := (2) P := (Closest b (2)) r := (Rplus y i); Auto.
2:Apply ClosestRoundedModeP with precision := precision; Auto.
2:Apply RoundedModeBounded
       with b := b radix := (2) P := (Closest b (2)) r := (Rplus y i); Auto.
2:Apply ClosestRoundedModeP with precision := precision; Auto.
Elim H'31; Intros p'' E; Elim E; Intros q'' E0; Elim E0; Intros r'' E1;
 Clear E0 E H'31.
Elim E1; Intros H' H'0; Elim H'0; Intros H'6 H'7; Elim H'6; Intros H'8 H'9;
 Elim H'9; Intros H'10 H'11; Clear H'9 H'6 H'0 E1.
Elim H'11; Intros H'0 H'6; Clear H'11.
Elim H'0; Intros; Clear H'0.
Elim H'8; Intros H'0 H'9; Clear H'8.
Elim H'9; Intros; Clear H'9.
Elim H'; Intros H'8 H'9; Clear H'.
Elim H'9; Intros; Clear H'9.
Clear H'5 H5 H'2 L.
Case (Req_EM r' R0); Intros Hr'.
Exists p''; Exists q''; Exists output.
Unfold Step.
Repeat (Split; Auto).
Apply IsEqual.
Replace (Rplus p'' (Rplus q'' (Rplus (sum input) (sum output))))
     with (Rplus (Rplus p'' (Rplus q'' r'')) (Rplus (sum input) (sum output)));
 [Idtac | Ring].
Replace (Rplus p'' (Rplus q'' r'')) with (Rplus x (Rplus y i));
 Replace (sum (cons i input)) with (Rplus i (sum input)).
Ring.
Simpl; Auto with real.
Replace (FtoRradix r'') with (FtoRradix r');
 Replace (FtoRradix r') with (Rminus (Rplus w v) q'); Auto with real.
Replace (FtoRradix q') with (FtoRradix q''); Ring.
Replace (FtoRradix p'') with (FtoRradix p'); Auto with real.
Replace (FtoRradix w)
     with (Rminus (Rplus (FtoR (2) x) (FtoR (2) u)) (FtoR (2) p')).
Replace (FtoRradix v) with (Rminus (Rplus y i) u); Auto with real.
Unfold FtoRradix; Ring; Ring.
Simpl; Auto with real.
Replace (FtoRradix r'') with (FtoRradix r'); Replace (FtoRradix r') with R0;
 Auto with real; Ring.
Apply Zle_trans with (Fexp r''); Auto.
Case H'7; Auto; Intros.
Elim H5; Intros; Auto.
Elim H5; Intros; Auto.
Case (Req_EM p'' R0); Intros; Auto.
Right; Right.
Case H21; Auto; Intros.
Absurd <R> (FtoR (2) p'')==R0; Auto with real.
Split; Auto.
2:Elim H5; Intros.
2:Absurd <R> (FtoR (2) r')==R0; Auto with real.
Apply Rle_trans
     with (Rmult
             (vNum b)
             (Rminus
                (FtoR (2) (Float (1) (Fexp p'')))
                (FtoR (2) (Float (1) (Fexp i))))); Auto.
Apply Rle_monotony; Auto.
Apply vNumPos with precision := precision; Auto.
Fold FtoRradix;
 Replace (Rminus (Float (1) (Fexp p'')) (Float (1) (hdexp b input)))
      with (Rplus (Float (1) (Fexp p'')) (Ropp (Float (1) (hdexp b input))));
 Replace (Rminus (Float (1) (Fexp p'')) (Float (1) (Fexp i)))
      with (Rplus (Float (1) (Fexp p'')) (Ropp (Float (1) (Fexp i)))); Try Ring.
Apply Rle_compatibility.
Apply Rle_Ropp.
Unfold FtoRradix; Unfold FtoR; Simpl.
Replace (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp i)))
     with (powerRZ (Rplus R1 R1) (Fexp i));
 Replace (Rmult R1 (powerRZ (Rplus R1 R1) (hdexp b input)))
      with (powerRZ (Rplus R1 R1) (hdexp b input)); Try Ring.
Apply Rle_powerRZ.
Replace (Rplus R1 R1) with (INR (2)); Auto with real arith; Apply Rlt_le;
 Auto with real arith.
Apply IsExpZle; Auto.
Case G; Intros; Auto.
Elim H5; Intros H' H'2; Clear H5.
Elim H'2; Intros; Clear H'2.
Right.
Generalize H1 H18 H5 H'; Case input; Intros; Auto.
Repeat Split.
Replace (FtoRradix (hd b (nil float))) with R0;
 [Rewrite Rabsolu_R0 | Simpl; Auto with real].
Apply Rmult_Rle_O; [Apply Rlt_le; Auto with real arith | Idtac].
Apply Rmult_Rle_O; [Apply Rlt_le; Auto with real arith | Idtac].
Apply Rmult_Rle_O; Auto with real arith.
Apply Rlt_le; Apply Rlt_Rinv; Auto with real arith.
Unfold FtoRradix; Apply sym_eqT; Apply FzeroisReallyZero.
Left; Auto.
Inversion H19; Intros; Repeat Split.
Apply Rle_trans with (FtoRradix (Float (vNum b) (Fexp i))); Auto with real.
Replace (hd b (cons f l)) with f;
 [Apply Rle_trans with (FtoRradix (Float (vNum b) (Fexp f))) |
   Simpl; Auto with real].
Unfold FtoRradix; Apply MaxFloat with precision; Auto.
Unfold FtoRradix; Unfold FtoR; Simpl.
Rewrite <- INR_IZR_INZ.
Apply Rle_monotony;
 [Apply vNumPos with precision; Auto | Apply Rle_powerRZ; Auto with real].
Replace (Rplus R1 R1) with (INR (2)); [Replace R1 with (INR (1)) | Idtac];
 Auto with real arith.
Right; Split.
Replace (hd b (cons f l)) with f; [Idtac | Simpl; Auto with real].
Apply Rle_trans with (FtoRradix (Float (vNum b) (Fexp i))); Auto with real.
Unfold FtoRradix; Unfold FtoR; Simpl.
Rewrite <- INR_IZR_INZ; Auto with real.
Apply Rle_monotony;
 [Apply vNumPos with precision; Auto | Apply Rle_powerRZ; Auto with real].
Replace (Rplus R1 R1) with (INR (2)); [Replace R1 with (INR (1)) | Idtac];
 Auto with real arith.
Apply sum_IsExp with b precision; Auto.
Apply Rle_trans with (FtoRradix (Float (vNum b) (Fexp i))); Auto with real.
Unfold FtoRradix; Unfold FtoR; Simpl.
Rewrite <- INR_IZR_INZ; Auto with real.
Apply Rle_monotony;
 [Apply vNumPos with precision; Auto | Apply Rle_powerRZ; Auto with real].
Replace (Rplus R1 R1) with (INR (2)); [Replace R1 with (INR (1)) | Idtac];
 Auto with real arith.
Exists q''; Exists r''; Exists (cons p'' output).
Unfold Step; Repeat (Split; Auto).
Apply OneMoreR; Auto.
Replace (sum (cons i input)) with (Rplus i (sum input));
 [Idtac | Simpl; Auto with real].
Replace (sum (cons p'' output)) with (Rplus p'' (sum output));
 [Idtac | Simpl; Auto with real].
Replace (Rplus q'' (Rplus r'' (Rplus (sum input) (Rplus p'' (sum output)))))
     with (Rplus (Rplus p'' (Rplus q'' r'')) (Rplus (sum input) (sum output)));
 [Idtac | Ring].
Replace (Rplus p'' (Rplus q'' r'')) with (Rplus x (Rplus y i));
 Replace (sum (cons i input)) with (Rplus i (sum input)).
Ring.
Simpl; Auto with real.
Replace (FtoRradix r'') with (FtoRradix r');
 Replace (FtoRradix r') with (Rminus (Rplus w v) q'); Auto with real.
Replace (FtoRradix q') with (FtoRradix q''); Ring.
Replace (FtoRradix p'') with (FtoRradix p'); Auto with real.
Replace (FtoRradix w)
     with (Rminus (Rplus (FtoR (2) x) (FtoR (2) u)) (FtoR (2) p')).
Replace (FtoRradix v) with (Rminus (Rplus y i) u); Auto with real.
Unfold FtoRradix; Ring; Ring.
Simpl; Auto with real.
Case H'7; Auto; Intros.
Elim H5; Intros; Auto.
Absurd <R> (FtoR (2) r')==R0; Auto with real.
Elim H5; Intros; Auto.
Right; Right.
Split; Auto.
Apply Rle_trans
     with (Rmult
             (vNum b)
             (Rminus
                (FtoR (2) (Float (1) (Fexp q'')))
                (FtoR (2) (Float (1) (Fexp i))))); Auto.
Apply Rle_monotony; Auto.
Apply vNumPos with precision := precision; Auto.
Fold FtoRradix;
 Replace (Rminus (Float (1) (Fexp q'')) (Float (1) (hdexp b input)))
      with (Rplus (Float (1) (Fexp q'')) (Ropp (Float (1) (hdexp b input))));
 Replace (Rminus (Float (1) (Fexp q'')) (Float (1) (Fexp i)))
      with (Rplus (Float (1) (Fexp q'')) (Ropp (Float (1) (Fexp i)))); Try Ring.
Apply Rle_compatibility.
Apply Rle_Ropp.
Unfold FtoRradix; Unfold FtoR; Simpl.
Replace (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp i)))
     with (powerRZ (Rplus R1 R1) (Fexp i));
 Replace (Rmult R1 (powerRZ (Rplus R1 R1) (hdexp b input)))
      with (powerRZ (Rplus R1 R1) (hdexp b input)); Try Ring.
Apply Rle_powerRZ.
Replace (Rplus R1 R1) with (INR (2)); Auto with real arith; Apply Rlt_le;
 Auto with real arith.
Apply IsExpZle; Auto.
Right.
Replace (hd b (cons p'' output)) with p''; [Idtac | Simpl; Auto with real].
Apply Rlt_le; Unfold FtoRradix;
 Apply bound3Sum
      with precision := precision w := w v := v p := x q := y r := i u := u;
 Auto.
Apply RoundedModeBounded with (2) (Closest b (2)) (Rplus y i); Auto;
 Apply ClosestRoundedModeP with precision; Auto.
Apply (ClosestCompatible b (2) (Rplus x u) (Rplus x u) p'); Auto with real.
Replace (FtoR (2) p'') with (FtoR (2) p'); Auto with real.
Apply (ClosestCompatible b (2) (Rplus w v) (Rplus w v) q'); Auto with real.
Replace (FtoRradix w) with (Rminus (Rplus x u) p');
 Replace (FtoRradix v) with (Rminus (Rplus y i) u); Auto with real.
Replace (FtoR (2) r'') with (FtoR (2) r');
 Replace (FtoR (2) q'') with (FtoR (2) q'); Auto with real.
Replace (FtoR (2) r'') with (FtoR (2) r'); Auto with real.
Right.
Replace (hd b (cons p'' output)) with p''; [Idtac | Simpl; Auto with real].
Cut (z:float) (Zle (Fexp z) (Fexp i)) -> (Fbounded b z) ->
    (Rle
       (Rabsolu z)
       (Rmult
          (3) (Rmult (2) (Rmult (Rabsolu p'') (Rinv (Rminus (vNum b) (1)))))));
 [Intros | Idtac].
Split; [Apply H5; Auto with zarith | Idtac].
Replace (Fexp (hd b input)) with (hdexp b input).
Apply IsExpZle; Auto.
Case input; Simpl; Auto with zarith.
Generalize H1; Case input; Intros; Simpl.
Apply FboundedFzero.
Inversion H18; Auto.
Generalize H1; Induction input; Intros; Auto.
Right.
Cut (Rle
       (Float (vNum b) (Fexp a))
       (Rmult
          (3) (Rmult (2) (Rmult (Rabsolu p'') (Rinv (Rminus (vNum b) (1)))))));
 [Intros | Idtac].
Replace (hd b (cons a input)) with a; [Split; Auto | Simpl; Auto with real].
Apply sum_IsExp with b precision; Auto.
Inversion H18; Auto.
Replace (FtoRradix (Float (vNum b) (Fexp a)))
     with (Rabsolu (Float (vNum b) (Fexp a))).
Apply H5; Auto.
Simpl; Inversion H18; Auto.
Apply maxFbounded; Auto.
Inversion H18; Auto; Elim H23; Auto.
Replace (Rabsolu (Float (vNum b) (Fexp a)))
     with (FtoRradix (Fabs (Float (vNum b) (Fexp a))));
 [Unfold Fabs; Simpl | Idtac]; Auto with real zarith.
Replace (Zabs (vNum b)) with (inject_nat (vNum b)); Auto with zarith float real.
Apply sym_eq; Apply Zabs_eq; Auto with zarith arith.
Unfold FtoRradix; Apply Fabs_correct; Auto.
Intros.
Apply Rle_trans with (FtoRradix (Float (vNum b) (Fexp z))).
Unfold FtoRradix; Apply MaxFloat with precision; Auto with arith.
Replace (FtoRradix (Float (vNum b) (Fexp z)))
     with (Rmult (vNum b) (powerRZ (2) (Fexp z)));
 [Idtac | Unfold FtoRradix; Unfold FtoR; Simpl].
Replace (Rmult
           (3) (Rmult (2) (Rmult (Rabsolu p'') (Rinv (Rminus (vNum b) (1))))))
     with (Rmult
             (vNum b)
             (Rmult
                (3)
                (Rmult
                   (2)
                   (Rmult
                      (2)
                      (Rmult
                         (Rabsolu p'')
                         (Rinv
                            (Rmult (vNum b) (Rminus (Rmult (2) (vNum b)) (2))))))))).
Apply Rle_monotony.
Apply vNumPos with precision; Auto.
Apply Rle_trans with (FtoRradix (Float (1) (Fexp i))).
Unfold FtoRradix; Unfold FtoR; Simpl.
Replace (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp i)))
     with (powerRZ (Rplus R1 R1) (Fexp i)); [Idtac | Ring].
Apply Rle_powerRZ; Auto with real.
Replace (Rplus R1 R1) with (INR (2)); [Replace R1 with (INR (1)) | Idtac];
 Auto with real arith.
Apply Rlt_le;
 Apply OutSum3
      with precision := precision p := x q := y r := i u := u v := v w := w
           p' := p'' q' := q'' r' := r''; Auto with real arith.
Apply RoundedModeBounded with (2) (Closest b (2)) (Rplus y i); Auto;
 Apply ClosestRoundedModeP with precision; Auto.
Apply (ClosestCompatible b (2) (Rplus x u) (Rplus x u) p'); Auto with real.
Replace (FtoR (2) p'') with (FtoR (2) p'); Auto with real.
Apply (ClosestCompatible b (2) (Rplus w v) (Rplus w v) q'); Auto with real.
Replace (FtoRradix w) with (Rminus (Rplus x u) p');
 Replace (FtoRradix v) with (Rminus (Rplus y i) u); Auto with real.
Replace (FtoR (2) r'') with (FtoR (2) r');
 Replace (FtoR (2) q'') with (FtoR (2) q'); Auto with real.
Replace (FtoR (2) r'') with (FtoR (2) r'); Auto with real.
Ring; Ring; Apply Rmult_mult_r.
Ring; Ring; Apply Rmult_mult_r.
Rewrite (Rinv_Rmult (vNum b) (Rminus (Rmult (2) (vNum b)) (2))); Auto.
Replace (Rmult
           (2)
           (Rmult
              (vNum b)
              (Rmult
                 (Rabsolu p'')
                 (Rmult
                    (Rinv (vNum b)) (Rinv (Rminus (Rmult (2) (vNum b)) (2)))))))
     with (Rmult
             (2)
             (Rmult
                (Rinv (Rminus (Rmult (2) (vNum b)) (2)))
                (Rmult (Rabsolu p'') (Rmult (Rinv (vNum b)) (vNum b)))));
 [Idtac | Ring].
Replace (Rmult (Rinv (vNum b)) (vNum b)) with R1; [Ring | Idtac].
Replace (Rmult (2) (Rinv (Rminus (Rmult (2) (vNum b)) (2))))
     with (Rmult (Rinv (Rminus (vNum b) (1))) (Rmult (2) (Rinv (2)))); Ring.
Replace (Rmult (2) (Rinv (2))) with R1; Ring.
Apply sym_eqT; Ring; Apply Rinv_r; Auto with real arith.
Rewrite <- (Rinv_Rmult (Rminus (vNum b) (1)) (2)); Auto with real arith.
Replace (Rmult (Rminus (vNum b) (1)) (2)) with (Rminus (Rmult (2) (vNum b)) (2));
 Ring; Auto with real.
Apply sym_not_eqT; Apply imp_not_Req; Left; Auto.
Apply sym_eqT; Ring; Apply Rinv_r; Auto with real arith.
Apply sym_not_eqT; Apply imp_not_Req; Left; Auto.
Apply sym_not_eqT; Apply imp_not_Req; Left; Auto.
Replace (Rminus (Rmult (2) (vNum b)) (2)) with (Rmult (Rminus (vNum b) (1)) (2));
 [Auto with real | Ring].
Apply sym_not_eqT; Apply imp_not_Req; Left; Auto.
Replace R0 with (Rmult (Rminus (vNum b) (1)) R0); [Apply Rlt_monotony | Ring];
 Auto with real arith.
Rewrite <- INR_IZR_INZ; Auto with real.
Qed.

Definition endof :=
   [all, part:(list float)](Ex [rest:(list float)] all=(app rest part)).

Theorem endof_length:
  (L, l:(list float)) (endof L l) ->(le (length l) (length L)).
Intros.
Inversion H.
Rewrite H0.
Rewrite (app_length x l); Auto with arith.
Qed.
Variable input:(list float).

Inductive IsRleEpsExp: (list float) ->Prop :=
     IsRleEpsExpNil: (IsRleEpsExp (nil ?))
    | IsRleEpsExpSingle:
        (x:float) (Fbounded b x) ->(IsRleEpsExp (cons x (nil ?)))
    | IsRleEpsExpTop:
        (x, y:float)
        (L:(list float))
        (Fbounded b x) ->
        (Fbounded b y) ->
        (Rle
           (Rabsolu x)
           (Rmult
              (Rmult
                 (Rplus (Rmult (6) (length input)) (6))
                 (Rinv
                    (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input)))))
              (Rabsolu y))) -> (IsRleEpsExp (cons y L)) ->
        (IsRleEpsExp (cons x (cons y L))) .

Theorem endof_Rle_length:
  (P, Q:(list float)) (k:float) (endof input (app P (cons k Q))) ->
  (Rle (length P) (Rminus (length input) R1)).
Intros P Q k H; Inversion H.
Rewrite H0.
Repeat Rewrite app_length; Repeat Rewrite plus_INR.
Replace (INR (length (cons k Q))) with (Rplus (length Q) R1);
 Auto with real arith zarith.
Apply Rle_trans with (Rplus (length x) (Rplus (length P) (length Q)));
 [Auto with real arith zarith | Right; Ring].
Apply Rle_trans with (Rplus (length x) (length P));
 [Idtac | Apply Rle_compatibility]; Auto with real arith zarith.
Rewrite <- plus_INR; Auto with real arith zarith.
Rewrite <- plus_INR; Auto with real arith zarith.
Case Q; Simpl; Auto with real arith zarith.
Qed.

Theorem FexpAdd_aux:
  (L, output:(list float))
  (x, y:float)
  (endof input L) ->
  (Rlt (Rmult (Rmult (6) (length input)) (Rinv (Rminus (vNum b) (1)))) R1) ->
  (IsExp b L) ->
  (IsRleEpsExp output) ->
  output=(nil float) \/
  (Rle
     (Rabsolu (Rplus x y))
     (Rmult
        (Rmult
           (Rmult (Rminus (vNum b) (1)) (Rinv (vNum b)))
           (Rmult
              (Rplus (Rmult (6) (length input)) (6))
              (Rinv (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input))))))
        (Rabsolu (hd b output)))) ->
  output=(nil float) \/
  (Ex [L1:(list float)]
  (Ex [x1:float]
  (Ex [y1:float]
  (IsExp b (app L1 L)) /\
  ((endof input (app L1 L)) /\
   (<R> (Rplus x1 (Rplus y1 (sum L1)))==(Rplus x y) /\
    ((Rle
        (Rabsolu (sum L1))
        (Rmult
           (length L1)
           (Rmult
              (3)
              (Rmult
                 (2)
                 (Rmult (Rabsolu (hd b output)) (Rinv (Rminus (vNum b) (1))))))))
     /\
     (Rle
        (Rabsolu (Rplus x1 y1))
        (Rmult (Rmult (Rmult (3) (2)) (Rinv (vNum b))) (Rabsolu (hd b output)))))))))) ->
  (Fbounded b x) ->
  (Fbounded b y) ->
  (Zle (hdexp b L) (Fexp y)) ->
  (Zle (Fexp y) (Fexp x)) ->
  (<R> (FtoR (2) x)==R0 \/ <R> (FtoR (2) y)==R0) \/
  (Rle
     (Rabsolu (FtoR (2) y))
     (Rmult
        (vNum b)
        (Rminus
           (FtoR (2) (Float (1) (Fexp x))) (FtoR (2) (Float (1) (hdexp b L)))))) ->
  output=(nil float) \/
  (L=(nil float) \/
   (Rle
      (Rabsolu (hd b L))
      (Rmult
         (3)
         (Rmult
            (2) (Rmult (Rabsolu (hd b output)) (Rinv (Rminus (vNum b) (1)))))))
   /\
   ((Rle
       (Float (vNum b) (Fexp (hd b L)))
       (Rmult
          (3)
          (Rmult
             (2) (Rmult (Rabsolu (hd b output)) (Rinv (Rminus (vNum b) (1)))))))
    /\
    (Rle
       (Rabsolu (sum L))
       (Rmult
          (length L)
          (Rmult
             (3)
             (Rmult
                (2) (Rmult (Rabsolu (hd b output)) (Rinv (Rminus (vNum b) (1)))))))))) ->
  (Ex [x':float]
  (Ex [y':float]
  (Ex [output':(list float)]
  (Fbounded b x') /\
  ((Fbounded b y') /\
   (<R>
      (Rplus x (Rplus y (Rplus (sum output) (sum L))))==
      (Rplus x' (Rplus y' (sum output'))) /\
    ((le (length output') (plus (length L) (length output))) /\
     ((Zle (Fexp y') (Fexp x')) /\
      ((IsRleEpsExp output') /\
       ((output=(nil float) \/
         (Ex [L1:(list float)]
         (Ex [x1:float]
         (Ex [y1:float]
         (IsExp b (app L1 L)) /\
         ((endof input (app L1 L)) /\
          (<R> (Rplus x1 (Rplus y1 (sum L1)))==(Rplus x y) /\
           ((Rle
               (Rabsolu (sum L1))
               (Rmult
                  (length L1)
                  (Rmult
                     (3)
                     (Rmult
                        (2)
                        (Rmult
                           (Rabsolu (hd b output)) (Rinv (Rminus (vNum b) (1))))))))
            /\
            (Rle
               (Rabsolu (Rplus x1 y1))
               (Rmult
                  (Rmult (Rmult (3) (2)) (Rinv (vNum b)))
                  (Rabsolu (hd b output))))))))))) /\
        ((endof output' output) /\
         (output'=(nil float) \/
          (Rle
             (Rabsolu (Rplus x' y'))
             (Rmult
                (Rmult
                   (Rmult (Rminus (vNum b) (1)) (Rinv (vNum b)))
                   (Rmult
                      (Rplus (Rmult (6) (length input)) (6))
                      (Rinv
                         (Rminus
                            (Rminus (vNum b) (1)) (Rmult (6) (length input))))))
                (Rabsolu (hd b output'))))))))))))))).
Cut (Rlt R0 (Rminus (vNum b) R1)); [Intros Z1 | Idtac].
2:Apply Rlt_anti_compatibility with (INR (1)).
2:Replace (Rplus (1) R0) with (INR (1));
   Replace (Rplus (1) (Rminus (vNum b) R1)) with (INR (vNum b)); Try Ring.
2:Rewrite pGivesBound.
2:Apply Rlt_anti_compatibility with (INR (1)).
2:Replace (Rplus (1) (1)) with (INR (2));
   Replace (Rplus (1) (minus (exp (2) precision) (1)))
        with (INR (exp (2) precision)); Try Ring.
2:Apply Rlt_INR.
2:Pattern 1 (2); Replace (2) with (exp (2) (1));
   [Apply exp_monotone_lt | Idtac]; Auto with arith.
2:Rewrite (minus_INR (exp (2) precision) (1)); [Ring | Idtac].
2:Replace (1) with (exp (2) (0)); [Apply exp_monotone_le | Idtac];
   Auto with arith.
2:Rewrite (minus_INR (exp (2) precision) (1)); [Ring | Idtac].
2:Replace (1) with (exp (2) (0)); [Apply exp_monotone_le | Idtac];
   Auto with arith.
Cut (Rlt R0 (vNum b)); [Intros Z2 | Idtac].
2:Replace R0 with (INR (0)); Auto with real.
2:Apply Rlt_INR.
2:Apply lt_vNum with precision := precision radix := (2); Auto with real arith.
Cut (Rlt (Rmult (Rmult (6) (length input)) (Rinv (Rminus (vNum b) (1)))) R1) ->
    (Rlt R0 (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input))));
 [Intros H665 | Intros].
2:Apply Rlt_anti_compatibility with (Rmult (6) (length input)).
2:Replace (Rplus (Rmult (6) (length input)) R0) with (Rmult (6) (length input));
   [Idtac | Ring].
2:Replace (Rplus
             (Rmult (6) (length input))
             (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input))))
       with (Rminus (vNum b) (1)); [Idtac | Ring].
2:Apply Rlt_mult_anticompatibility with (Rinv (Rminus (vNum b) (1))).
2:Rewrite Rmult_sym; Apply Rlt_le_trans with R1; Auto; Right.
2:Rewrite Rmult_sym; Rewrite Rinv_r; Auto with real.
2:Apply sym_not_eqT; Apply imp_not_Req; Left.
2:Replace (INR (1)) with R1; Auto with real.
2:Apply Rlt_Rinv; Replace (INR (1)) with R1; Auto with real.
Cut (Rlt (Rmult (Rmult (6) (length input)) (Rinv (Rminus (vNum b) (1)))) R1) ->
    (Rle
       (Rmult (Rmult (3) (2)) (Rinv (vNum b)))
       (Rmult
          (Rmult (Rminus (vNum b) (1)) (Rinv (vNum b)))
          (Rmult
             (Rplus (Rmult (6) (length input)) (6))
             (Rinv (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input)))))));
 [Intros H666 | Intros].
2:Rewrite <- Rmult_INR; Replace (mult (3) (2)) with (6); Auto with arith.
2:Apply Rle_trans
       with (Rmult
               (Rmult
                  (Rplus (Rmult (6) (length input)) (6))
                  (Rmult
                     (Rminus (vNum b) (1))
                     (Rinv
                        (Rminus
                           (Rminus (vNum b) (1)) (Rmult (6) (length input))))))
               (Rinv (vNum b))); [Idtac | Right; Ring; Ring].
2:Apply Rle_monotony_r; Auto with real.
2:Apply Rlt_le; Apply Rlt_Rinv; Auto.
2:Apply Rle_trans with (Rmult (Rplus (Rmult (6) (length input)) (6)) R1).
2:Apply Rle_trans with (Rmult (Rplus (Rmult (6) R0) (6)) R1);
   [Right; Ring | Idtac].
2:Apply Rle_monotony_r; [Apply Rlt_le | Idtac]; Auto with real zarith.
2:Apply Rle_compatibility_r; Apply Rle_monotony; Replace R0 with (INR (0));
   Auto with real arith.
2:Apply Rle_monotony.
2:Apply Rle_trans with (INR (6));
   [Apply Rlt_le; Auto with real arith |
     Pattern 1 (INR (6)); Replace (INR (6)) with (Rplus (Rmult (6) R0) (6))].
2:Apply Rle_compatibility_r; Apply Rle_monotony; Replace R0 with (INR (0));
   Auto with real arith.
2:Ring; Auto with real arith.
2:LApply H665; [Intros H'; Clear H665 | Clear H665]; Auto.
2:Apply Rle_mult_anticompatibility
       with (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input))); Auto.
2:Apply Rle_trans
       with (Rmult
               (Rminus (vNum b) (1))
               (Rmult
                  (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input)))
                  (Rinv
                     (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input))))));
   [Rewrite Rinv_r | Right; Ring; Ring].
2:Apply Rle_monotony_r; [Apply Rlt_le | Idtac]; Auto with real.
2:Pattern 2 (Rminus (vNum b) (1));
   Replace (Rminus (vNum b) (1)) with (Rminus (Rminus (vNum b) (1)) R0);
   Unfold Rminus; [Apply Rle_compatibility | Ring].
2:Apply Rle_Ropp; Apply Rmult_Rle_O; Replace R0 with (INR (0));
   Auto with real arith.
2:Apply imp_not_Req; Auto.
Intros L; Induction L.
Intros.
LApply H665; Clear H665; Auto; Intros H665.
LApply H666; Clear H666; Auto; Intros H666.
Exists x; Exists y; Exists output.
Repeat (Split; Auto).
Simpl; Ring.
Unfold endof; Exists (nil float); Auto.
Intros output; Induction output.
Intros x y V H H0 H1 M H2 H3 H4 H5 H6 H7 M.
LApply H665; Clear H665; Auto; Intros H665.
LApply H666; Clear H666; Auto; Intros H666.
Generalize AddStep; Intros AS.
LApply (AS x y a L (nil float));
 [Intros H'4; LApply H'4;
   [Intros H'5; LApply H'5;
     [Intros H'6; LApply H'6;
       [Intros H'7; LApply H'7;
         [Intros H'8; LApply H'8;
           [Intros H'9; Try Exact H'9; Clear H'8 H'7 H'6 H'5 H'4 |
             Clear H'8 H'7 H'6 H'5 H'4] | Clear H'7 H'6 H'5 H'4] |
         Clear H'6 H'5 H'4] | Clear H'5 H'4] | Clear H'4] | Idtac]; Auto.
2:Case (Req_EM (FtoR (2) x) R0); Intros; Auto.
2:Case (Req_EM (FtoR (2) y) R0); Intros; Auto.
2:Case H7; Intros; Auto with real.
2:Case H10; Intros; Auto with real.
LApply H'9; [Intros N; Clear H'9 | Clear H'9]; Auto.
Elim N; Intros x' E; Elim E; Intros y' E0; Elim E0; Intros output' E1;
 Clear N E E0.
Elim E1; Intros.
Elim H9; Intros H' H'0; Elim H'0; Intros H'1 H'2; Elim H'2; Intros H'3 H'4;
 Elim H'4; Intros H'5 H'6; Elim H'6; Intros H'7 H'8; Elim H'8; Intros H'9 H'10;
 Clear H'8 H'6 H'4 H'2 H'0 H9.
Elim H'10; Intros; Clear H'10.
LApply (HrecL output' x' y');
 [Intros H'8; LApply H'8;
   [Intros H'10; LApply H'10;
     [Intros H'11; LApply H'11;
       [Intros H'12; LApply H'12;
         [Intros H'13; LApply H'13;
           [Intros H'14; LApply H'14;
             [Intros H'15; LApply H'15;
               [Intros H'16; LApply H'16;
                 [Intros H'17; LApply H'17;
                   [Intros H'18; LApply H'18;
                     [Intros H'19; Try Exact H'19;
                       Clear H'18 H'17 H'16 H'15 H'14 H'13 H'12 H'11 H'10 H'8 |
                       Clear H'18 H'17 H'16 H'15 H'14 H'13 H'12 H'11 H'10 H'8] |
                     Clear H'17 H'16 H'15 H'14 H'13 H'12 H'11 H'10 H'8] |
                   Clear H'16 H'15 H'14 H'13 H'12 H'11 H'10 H'8] |
                 Clear H'15 H'14 H'13 H'12 H'11 H'10 H'8] |
               Clear H'14 H'13 H'12 H'11 H'10 H'8] |
             Clear H'13 H'12 H'11 H'10 H'8] | Clear H'12 H'11 H'10 H'8] |
         Clear H'11 H'10 H'8] | Clear H'10 H'8] | Clear H'8] | Idtac]; Auto.
2:Case (Req_EM (FtoR (2) x') R0); Intros; Auto.
2:Case (Req_EM (FtoR (2) y') R0); Intros; Auto.
2:Case H9; Intros; Auto with real.
2:Case H13; Intros; Auto with real.
2:Elim H14; Auto with real.
2:Apply Zle_trans with (Fexp a); Auto with zarith.
2:Apply IsExpZle; Auto.
2:Elim H10; Intros; Clear H10.
2:Generalize H12 H11; Case output'; Intros; Auto.
2:Right; Exists (nil float); Exists x'; Exists y'.
2:Repeat (Split; Auto with real).
2:Simpl; Auto; Inversion H'9; Auto; Apply IsExpNil.
2:Simpl; Auto.
2:Unfold endof; Inversion V.
2:Exists (app x0 (cons a (nil float))).
2:Replace (app (app x0 (cons a (nil float))) L) with (app x0 (cons a L)); Auto.
2:Rewrite (app_ass x0 (cons a (nil float)) L).
2:Replace (app (cons a (nil float)) L) with (cons a L); Auto.
2:Simpl; Auto with real.
2:Replace (Rabsolu R0) with R0; [Right; Ring | Apply sym_eqT; Apply Rabsolu_R0].
2:Elim H13; Auto.
2:Intros; Absurd <(list float)> (cons f l)=(nil float); Auto.
2:Apply sym_not_eq; Apply nil_cons.
3:Inversion H'1; Auto.
3:Apply IsRleEpsExpSingle; Auto.
3:Inversion H'9; Auto; Apply IsExpNil.
3:Unfold endof; Inversion V.
3:Exists (app x0 (cons a (nil float))).
3:Replace (app (app x0 (cons a (nil float))) L) with (app x0 (cons a L)); Auto.
3:Rewrite (app_ass x0 (cons a (nil float)) L).
3:Replace (app (cons a (nil float)) L) with (cons a L); Auto.
2:Generalize H10; Case output'; Intros; Auto.
2:Right; Elim H11; Intros; Clear H11; Auto.
2:Case H12; Intros; Clear H12; Auto.
2:Absurd <(list float)> (cons f l)=(nil float); Auto.
2:Apply sym_not_eq; Apply nil_cons.
2:Apply Rle_trans
       with (Rmult
               (Rmult (Rmult (3) (2)) (Rinv (vNum b)))
               (Rabsolu (hd b (cons f l)))); Auto.
2:Apply Rle_monotony_r; Auto with real.
LApply H'19; [Clear H'19; Intros H'19 | Clear H'19].
2:Generalize H10; Case output'; Case L; Auto; Intros.
2:Elim H11; Intros H12 H13; Case H12; Clear H11; Intros; Auto.
2:Case H13; Intros; Auto.
2:Elim H14; Intros H15 H16; Case H16; Auto.
Elim H'19; Intros x'0 E; Elim E; Clear H'19; Intros y'0 E0; Elim E0;
 Intros output'0 E2; Elim E2; Intros H'0 H'2; Elim H'2; Intros H'4 H'6;
 Elim H'6; Intros H'8 H'10; Elim H'10; Intros H'11 H'12;
 Clear H'10 H'6 H'2 E2 E0 E.
Elim H'12; Intros H11 K; Clear H'12.
Elim K; Intros H12 K1; Clear K.
Exists x'0; Exists y'0; Exists output'0.
Repeat (Split; Auto).
Rewrite (Rplus_sym (sum (nil float)) (sum (cons a L))); Rewrite H'3.
Rewrite <- H'8; Ring.
Generalize H'11; Case H'1; Intros; Simpl; Auto with arith.
Replace (S (plus (length L) (length x0)))
     with (plus (length L) (length (cons e x0)));
 [Auto | Simpl; Auto with arith].
Unfold endof; Exists output'0; Apply app_nil_end.
Elim K1; Intros H13 H14; Elim H14; Auto.
Intros x y V' H H0 H1 M H2 H3 H4 H5 H6 H7 H8.
LApply H665; Clear H665; Auto; Intros H665.
LApply H666; Clear H666; Auto; Intros H666.
Generalize AddStep; Intros AS.
LApply (AS x y a L (cons a0 output));
 [Intros H'6; LApply H'6;
   [Intros H'7; LApply H'7;
     [Intros H'8; LApply H'8;
       [Intros H'9; LApply H'9;
         [Intros H'10; LApply H'10;
           [Intros H'11; Try Exact H'11; Clear H'10 H'9 H'8 H'7 H'6 |
             Clear H'10 H'9 H'8 H'7 H'6] | Clear H'9 H'8 H'7 H'6] |
         Clear H'8 H'7 H'6] | Clear H'7 H'6] | Clear H'6] | Idtac]; Auto.
LApply H'11; [Intros N; Clear H'11 | Clear H'11].
2:Case H8; Intros; Auto.
2:Case H9; Intros.
2:Absurd <(list float)> (cons a L)=(nil float); Auto.
2:Apply sym_not_eq; Apply nil_cons.
2:Elim H10; Simpl; Auto with real.
2:Case H7; Auto; Intros tmp; Case tmp; Auto.
Elim N; Intros x' E; Elim E; Intros y' E0; Elim E0; Intros output' E1;
 Clear E0 E N.
Elim E1; Intros.
Elim H10; Intros H'0 H'2; Elim H'2; Intros H'3 H'4; Elim H'4; Intros H'5 H'6;
 Elim H'6; Intros H'7 H'8; Elim H'8; Intros H'9 H'10; Elim H'10;
 Intros H'11 H'12; Clear H'10 H'8 H'6 H'4 H'2 H10.
Elim H'12; Intros; Clear E1 H'12.
Elim H11; Intros; Clear H11.
Cut (IsRleEpsExp output'); [Intros RR | Idtac].
LApply (HrecL output' x' y');
 [Intros H'6; LApply H'6;
   [Intros H'8; LApply H'8;
     [Intros H'10; LApply H'10;
       [Intros H'12; LApply H'12;
         [Intros H'13; LApply H'13;
           [Intros H'14; LApply H'14;
             [Intros H'15; LApply H'15;
               [Intros H'16; LApply H'16;
                 [Intros H'17; LApply H'17;
                   [Intros H'18; LApply H'18;
                     [Intros H'19; Try Exact H'19;
                       Clear H'18 H'17 H'16 H'15 H'14 H'13 H'12 H'10 H'8 H'6 |
                       Clear H'18 H'17 H'16 H'15 H'14 H'13 H'12 H'10 H'8 H'6] |
                     Clear H'17 H'16 H'15 H'14 H'13 H'12 H'10 H'8 H'6] |
                   Clear H'16 H'15 H'14 H'13 H'12 H'10 H'8 H'6] |
                 Clear H'15 H'14 H'13 H'12 H'10 H'8 H'6] |
               Clear H'14 H'13 H'12 H'10 H'8 H'6] | Clear H'13 H'12 H'10 H'8 H'6] |
           Clear H'12 H'10 H'8 H'6] | Clear H'10 H'8 H'6] | Clear H'8 H'6] |
     Clear H'6] | Idtac]; Auto.
2:Case (Req_EM (FtoR (2) x') R0); Intros; Auto.
2:Case (Req_EM (FtoR (2) y') R0); Intros; Auto.
2:Case H10; Intros; Auto with real.
2:Case H15; Intros; Auto with real.
2:Elim H16; Auto with real.
2:Apply Zle_trans with (Fexp a); Auto with zarith.
2:Apply IsExpZle; Auto.
4:Inversion H'11; Auto; Apply IsExpNil.
4:Unfold endof; Inversion V'.
4:Exists (app x0 (cons a (nil float))).
4:Replace (app (app x0 (cons a (nil float))) L) with (app x0 (cons a L)); Auto.
4:Rewrite (app_ass x0 (cons a (nil float)) L); Auto.
LApply H'19; [Clear H'19; Intros H'19 | Idtac].
Elim H'19; Intros x'0 E; Elim E; Intros y'0 E0; Elim E0; Intros output'0 E1;
 Elim E1; Intros H'2 H'4; Elim H'4; Intros H'6 H'8; Elim H'8; Intros H'10 H'12;
 Elim H'12; Intros H'13 H'14; Clear H'12 H'8 H'4 E1 E0 E H'19.
Elim H'14; Intros; Clear H'14.
Elim H14; Intros; Clear H14.
Exists x'0; Exists y'0; Exists output'0.
Repeat (Split; Auto).
Rewrite <- H'10; Auto with real.
Rewrite (Rplus_sym (sum output') (sum L)); Rewrite <- H'5; Ring; Ring.
Apply le_trans with (plus (length L) (length output')); Auto.
Case H'3; Intros; Auto with arith.
Apply le_refl_eq; Simpl; Auto with arith.
Elim H16; Intros; Clear H16.
Elim H17; Intros; Clear H17.
Inversion H'3; Auto.
Rewrite H19; Auto.
Inversion H16; Unfold endof; Exists (app x1 (cons e (nil float))).
Rewrite app_ass;
 Replace (app (cons e (nil float)) (cons a0 output))
      with (cons e (cons a0 output)); Auto.
Rewrite H20; Auto.
Elim H16; Intros X1 X2; Elim X2; Auto.
Generalize H'3 H13 H12 H10 H'11 H'5 H8 H7 H5 H2 H0 RR; Case output'; Case L;
 Auto; Clear H'3 H13 H12 H10 H'11 H'5 H8 H7 H5 H2 H0 RR.
Intros f l f0 l0 H20; Case H20; Intros.
Generalize H2 H8; Case x0; Intros; Auto.
Case H14; Intros; Auto.
Case H15; Intros; Auto.
Absurd <(list float)> (cons a (cons f l))=(nil float); Auto.
Apply sym_not_eq; Apply nil_cons.
Right; Right.
Cut (Rle
       (Float (vNum b) (Fexp f))
       (Rmult
          (3)
          (Rmult
             (2)
             (Rmult (Rabsolu (hd b (cons f1 l1))) (Rinv (Rminus (vNum b) (1)))))));
 [Intros | Idtac].
Replace (hd b (cons f l)) with f; [Repeat (Split; Auto) | Simpl; Auto with real].
Apply Rle_trans with (FtoRradix (Float (vNum b) (Fexp f))); Auto.
Unfold FtoRradix; Apply MaxFloat with precision; Auto.
Inversion H0; Auto.
Apply sum_IsExp with b precision; Auto.
Inversion H0; Auto.
Elim H16; Intros; Clear H16.
Apply Rle_trans
     with (FtoRradix (Float (vNum b) (Fexp (hd b (cons a (cons f l))))));
 Auto with real.
Unfold FtoRradix; Unfold FtoR; Simpl.
Rewrite <- INR_IZR_INZ.
Apply Rle_monotony; [Apply vNumPos with precision; Auto | Apply Rle_powerRZ].
Replace (Rplus R1 R1) with (INR (2)); [Replace R1 with (INR (1)) | Idtac];
 Auto with real arith.
Inversion H0; Auto.
Elim H18; Auto with real.
Right; Right.
Case H13; Intros.
Absurd <(list float)> (cons e x0)=(nil float); Auto.
Apply sym_not_eq; Apply nil_cons.
Elim H14; Intros; Auto.
Case H16; Intros.
Absurd <(list float)> (cons f l)=(nil float); Auto.
Apply sym_not_eq; Apply nil_cons.
Elim H17; Intros; Clear H17.
Split; Auto.
Generalize RR H13 H12 H'5 H'3; Case output'; Intros; Auto;
 Clear RR H13 H12 H'5 H'3.
Case H2; Intros.
Absurd <(list float)> (cons a0 output)=(nil float); Auto.
Apply sym_not_eq; Apply nil_cons.
Elim H12; Intros L1 E; Elim E; Intros x1 E0; Elim E0; Intros y1 E1; Elim E1;
 Intros H' H'4; Elim H'4; Intros H'6 H'8; Clear H'4 E1 E0 E H12.
Elim H'8; Intros; Clear H'8.
Elim H13; Intros; Clear H13.
Cut ~ (cons a0 output)=(nil float);
 [Intros D | Apply sym_not_eq; Apply nil_cons].
Generalize D H16 H15 H12 H'6 H' H'1 H14 H11 RR0 H10 H'11 H8 H7; Case H'2;
 Clear D H16 H15 H12 H'6 H' H'1 H14 H11 RR0 H10 H'11 H8 H7; Intros.
Right; Exists (app L1 (cons a (nil float))); Exists x1; Exists y1.
Repeat (Split; Auto).
Replace (app (app L1 (cons a (nil float))) L) with (app L1 (cons a L)); Auto.
Rewrite (app_ass L1 (cons a (nil float)) L).
Replace (app (cons a (nil float)) L) with (cons a L); Auto.
Replace (app (app L1 (cons a (nil float))) L) with (app L1 (cons a L)); Auto.
Rewrite (app_ass L1 (cons a (nil float)) L).
Replace (app (cons a (nil float)) L) with (cons a L); Auto.
Replace (Rplus x1 (Rplus y1 (sum (app L1 (cons a (nil float))))))
     with (Rplus a (Rplus x1 (Rplus y1 (sum L1)))); Rewrite H12.
Apply r_Rplus_plus with (sum L); Apply r_Rplus_plus with (sum x0).
Replace (Rplus (sum x0) (Rplus (sum L) (Rplus a (Rplus x y))))
     with (Rplus x (Rplus y (Rplus (sum (cons a L)) (sum x0)))).
Rewrite H'1; Ring.
Simpl; Ring.
Replace (sum (app L1 (cons a (nil float)))) with (Rplus a (sum L1));
 [Rewrite <- H12; Ring | Idtac].
Apply sum_app.
Rewrite <- (sum_app L1 a).
Apply Rle_trans with (Rplus (Rabsolu a) (Rabsolu (sum L1)));
 [Apply Rabsolu_triang | Idtac].
Replace (INR (length (app L1 (cons a (nil float)))))
     with (Rplus R1 (INR (length L1))).
Replace (Rmult
           (Rplus R1 (length L1))
           (Rmult
              (3)
              (Rmult
                 (2) (Rmult (Rabsolu (hd b x0)) (Rinv (Rminus (vNum b) (1)))))))
     with (Rplus
             (Rmult
                (3)
                (Rmult
                   (2) (Rmult (Rabsolu (hd b x0)) (Rinv (Rminus (vNum b) (1))))))
             (Rmult
                (length L1)
                (Rmult
                   (3)
                   (Rmult
                      (2)
                      (Rmult (Rabsolu (hd b x0)) (Rinv (Rminus (vNum b) (1))))))));
 [Idtac | Ring; Ring].
Apply Rle_trans
     with (Rplus
             (Rabsolu a)
             (Rmult
                (length L1)
                (Rmult
                   (3)
                   (Rmult
                      (2)
                      (Rmult (Rabsolu (hd b x0)) (Rinv (Rminus (vNum b) (1))))))));
 [Apply Rle_compatibility | Apply Rle_compatibility_r]; Auto.
Elim H8; Intros; Clear H8; Auto.
Absurd <(list float)> x0=(nil float); Auto.
Elim H13; Intros; Clear H13.
Absurd <(list float)> (cons a L)=(nil float); Auto.
Apply sym_not_eq; Apply nil_cons.
Elim H8; Intros; Clear H8; Auto.
Rewrite (app_length L1 (cons a (nil float))); Simpl; Auto with arith real.
Replace R1 with (INR (1)); [Rewrite plus_INR | Auto with real arith];
 Apply Rplus_sym.
Right; Exists (nil float); Exists x'; Exists y'.
Repeat (Split; Auto).
Simpl; Inversion H'11; Auto; Apply IsExpNil.
Simpl; Auto.
Inversion H'6; Intros.
Unfold endof; Exists (app x2 (app L1 (cons a (nil float)))).
Replace (app (app x2 (app L1 (cons a (nil float)))) L)
     with (app x2 (app L1 (cons a L))); Auto.
Rewrite (app_ass x2 (app L1 (cons a (nil float))) L); Auto.
Rewrite (app_ass L1 (cons a (nil float)) L); Auto.
Simpl; Ring.
Simpl; Right; Ring.
Apply Rabsolu_R0.
Case H14; Intros; Auto.
Absurd <(list float)> (cons e x0)=x0; Auto.
Apply cons_neq.
Cut ~ (cons a0 output)=(nil float);
 [Intros D | Apply sym_not_eq; Apply nil_cons].
Cut (Ex [l:(list float)] (cons a0 output)=(cons a0 l));
 [Intros E | Exists output; Auto].
Generalize RR E D H'5 H13 H12 H10 H8 H2 H1; Case H'3; Intros;
 Clear RR H'5 H13 H12 D E H10 H8 H2 H1 HrecL Hrecoutput AS; Auto.
Case H17; Intros; Auto; Clear H17.
Elim H1; Intros L1 E; Elim E; Intros x1 E1; Elim E1; Intros y1 E2; Elim E2;
 Intros H' H'2; Elim H'2; Intros H'4 H'5; Elim H'5; Intros H'6 H'8; Elim H'8;
 Intros H'10 H'12; Clear H'8 H'5 H'2 E2 E1 E H1.
Right; Replace (Rplus x' y') with (Rplus a (Rplus (Rplus x1 y1) (sum L1))).
Apply Rle_trans
     with (Rplus (Rabsolu a) (Rabsolu (Rplus (Rplus x1 y1) (sum L1))));
 [Apply Rabsolu_triang | Idtac].
Apply Rle_trans
     with (Rplus (Rabsolu a) (Rplus (Rabsolu (Rplus x1 y1)) (Rabsolu (sum L1)))).
Apply Rle_compatibility; Apply Rabsolu_triang.
Apply Rle_trans
     with (Rplus
             (Rmult
                (3)
                (Rmult
                   (2) (Rmult (Rabsolu (hd b x0)) (Rinv (Rminus (vNum b) (1))))))
             (Rplus (Rabsolu (Rplus x1 y1)) (Rabsolu (sum L1))));
 [Apply Rle_compatibility_r | Idtac].
Case H16; Intros; Clear H16.
Absurd <(list float)> x0=(nil float); Auto.
Case H1; Intros; Clear H1.
Absurd <(list float)> (cons a L)=(nil float); Auto; Apply sym_not_eq;
 Apply nil_cons.
Elim H2; Simpl; Auto.
Apply Rle_trans
     with (Rplus
             (Rmult
                (3)
                (Rmult
                   (2) (Rmult (Rabsolu (hd b x0)) (Rinv (Rminus (vNum b) (1))))))
             (Rplus
                (Rmult
                   (Rmult (Rmult (3) (2)) (Rinv (vNum b))) (Rabsolu (hd b x0)))
                (Rabsolu (sum L1))));
 [Apply Rle_compatibility; Apply Rle_compatibility_r; Auto | Idtac].
Apply Rle_trans
     with (Rplus
             (Rmult
                (3)
                (Rmult
                   (2) (Rmult (Rabsolu (hd b x0)) (Rinv (Rminus (vNum b) (1))))))
             (Rplus
                (Rmult
                   (Rmult (Rmult (3) (2)) (Rinv (vNum b))) (Rabsolu (hd b x0)))
                (Rmult
                   (length L1)
                   (Rmult
                      (3)
                      (Rmult
                         (2)
                         (Rmult
                            (Rabsolu (hd b x0)) (Rinv (Rminus (vNum b) (1)))))))));
 [Apply Rle_compatibility; Apply Rle_compatibility; Auto | Idtac].
Apply Rle_trans
     with (Rplus
             (Rmult
                (3)
                (Rmult
                   (2) (Rmult (Rabsolu (hd b x0)) (Rinv (Rminus (vNum b) (1))))))
             (Rplus
                (Rmult
                   (Rmult (Rmult (3) (2)) (Rinv (Rminus (vNum b) (1))))
                   (Rabsolu (hd b x0)))
                (Rmult
                   (length L1)
                   (Rmult
                      (3)
                      (Rmult
                         (2)
                         (Rmult
                            (Rabsolu (hd b x0)) (Rinv (Rminus (vNum b) (1)))))))));
 [Apply Rle_compatibility; Apply Rle_compatibility_r; Auto | Idtac].
Apply Rle_monotony_r; Auto with real.
Apply Rle_monotony; Auto with real.
Apply Rmult_Rle_O; Replace R0 with (INR (0)); Auto with real arith.
Apply Rle_Rinv; Auto with real arith.
Pattern 2 (INR (vNum b)); Replace (INR (vNum b)) with (Rminus (vNum b) R0);
 Unfold Rminus;
 [Apply Rle_compatibility; Apply Rle_Ropp; Auto with real arith | Ring].
Replace R0 with (INR (0)); Auto with real arith.
Apply Rle_trans
     with (Rmult
             (Rmult
                (Rplus
                   (Rmult (3) (2))
                   (Rplus (Rmult (3) (2)) (Rmult (length L1) (Rmult (3) (2)))))
                (Rinv (Rminus (vNum b) (1)))) (Rabsolu (hd b x0)));
 [Right; Ring; Ring | Idtac].
Apply Rle_monotony_r; Auto with real.
Replace (Rmult (3) (2)) with (INR (6)).
2:Rewrite <- Rmult_INR; Auto with arith zarith real.
Apply Rle_trans
     with (Rmult
             (Rplus (6) (Rplus (6) (Rmult (Rminus (length input) R1) (6))))
             (Rinv (Rminus (vNum b) (1)))).
Apply Rle_monotony_r.
Apply Rlt_le; Apply Rlt_Rinv; Auto.
Apply Rle_compatibility; Apply Rle_compatibility; Apply Rle_monotony_r;
 Auto with arith zarith real.
Replace R0 with (INR (0)); Auto with arith zarith real.
Apply endof_Rle_length with L a; Auto.
Replace (Rplus (6) (Rmult (Rminus (length input) R1) (6)))
     with (Rmult (6) (length input)); [Idtac | Ring; Ring].
Apply Rle_trans
     with (Rmult
             (Rplus (6) (Rmult (6) (length input)))
             (Rmult
                (Rmult (Rminus (vNum b) (1)) (Rinv (vNum b)))
                (Rinv (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input))))));
 [Idtac | Right; Ring; Ring].
Apply Rle_monotony; Auto with arith zarith real.
Rewrite <- Rmult_INR; Rewrite <- plus_INR; Replace R0 with (INR (0));
 Auto with arith zarith real.
Apply Rle_mult_anticompatibility with (Rminus (vNum b) (1)); Auto.
Apply Rle_mult_anticompatibility
     with (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input))); Auto.
Apply Rle_mult_anticompatibility with (INR (vNum b)); Auto.
Rewrite Rinv_r; [Idtac | Apply imp_not_Req; Auto].
Apply Rle_trans
     with (Rmult
             (Rminus (vNum b) (1))
             (Rmult
                (Rminus (vNum b) (1))
                (Rmult
                   (Rmult (vNum b) (Rinv (vNum b)))
                   (Rmult
                      (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input)))
                      (Rinv
                         (Rminus
                            (Rminus (vNum b) (1)) (Rmult (6) (length input))))))));
 [Idtac | Right; Ring; Ring].
Rewrite Rinv_r; [Idtac | Apply imp_not_Req; Auto].
Rewrite Rinv_r; [Idtac | Apply imp_not_Req; Auto].
Replace (INR (1)) with R1; Auto with real arith.
Apply Rle_trans
     with (Rplus
             (Rmult (vNum b) (vNum b))
             (Ropp (Rmult (vNum b) (Rplus (Rmult (6) (length input)) R1))));
 [Right; Ring; Ring | Idtac].
Apply Rle_trans
     with (Rplus
             (Rmult (vNum b) (vNum b))
             (Rplus (Ropp (Rmult (vNum b) (Rplus R1 R1))) R1));
 [Idtac | Right; Ring; Ring].
Apply Rle_compatibility.
Apply Rle_trans with (Ropp (Rmult (vNum b) (Rplus R1 R1))).
Apply Rle_Ropp; Apply Rle_monotony.
Apply Rlt_le; Auto.
Apply Rle_compatibility_r; Inversion H'4.
Apply Rle_trans with (Rmult (INR (6)) (INR (1))).
Rewrite <- Rmult_INR; Replace R1 with (INR (1)); Auto with real arith;
 Replace (INR (mult (6) (1))) with (INR (6)); Auto with real arith.
Apply Rle_monotony; [Replace R0 with (INR (0)) | Idtac]; Auto with real arith.
Rewrite H1; Repeat Rewrite app_length.
Apply Rle_INR; Simpl; Auto with arith.
Replace (plus (length x2) (plus (length L1) (S (length L))))
     with (S (plus (length x2) (plus (length L1) (length L))));
 [Auto with arith | Ring; Ring].
Cut (n, m:nat)(plus n (S m))=(S (plus n m)); [Intros tmp | Intros].
Rewrite (tmp (length L1) (length L)).
Rewrite (tmp (length x2) (plus (length L1) (length L))); Auto with arith.
Case n; Auto with arith zarith.
Apply Rle_trans with (Rplus (Ropp (Rmult (vNum b) (Rplus R1 R1))) R0);
 [Right; Ring | Apply Rle_compatibility; Apply Rlt_le; Auto with real].
Rewrite Rplus_assoc; Rewrite H'6.
Apply r_Rplus_plus with (sum x0); Apply r_Rplus_plus with (sum L).
Replace (Rplus (sum L) (Rplus (sum x0) (Rplus x' y')))
     with (Rplus x' (Rplus y' (Rplus (sum L) (sum x0))));
 [Rewrite <- H'1 | Ring; Ring].
Replace (sum (cons a L)) with (Rplus a (sum L));
 [Ring; Ring | Simpl; Auto with real].
Right; Case H15; Intros.
Absurd <(list float)> (cons e x0)=x0; Auto; Apply cons_neq.
Apply Rle_trans
     with (Rmult
             (Rmult (Rmult (3) (2)) (Rinv (vNum b)))
             (Rabsolu (hd b (cons e x0)))); Auto.
Apply Rle_monotony_r; Auto with real.
Cut ~ (cons a0 output)=(nil float);
 [Intros D | Apply sym_not_eq; Apply nil_cons].
Cut (Ex [f:float] (Ex [l:(list float)] (cons a0 output)=(cons f l)));
 [Intros E | Exists a0; Exists output; Auto].
Generalize E D H'5 H13 H12 H10 H8 H2 H1; Case H'3; Intros;
 Clear H'5 H13 H12 D E H10 H8 H2 H1 HrecL Hrecoutput AS; Auto.
Elim E0; Intros f E; Elim E; Intros l E1; Clear E E0.
Rewrite E1.
Apply IsRleEpsExpTop; Auto.
Generalize H19; Rewrite E1; Clear H19; Intros H19; Inversion H19; Auto.
2:Rewrite <- E1; Auto.
Apply Rle_mult_anticompatibility
     with (Rminus
             R1
             (Rmult (Rmult (6) (length (cons a L))) (Rinv (Rminus (vNum b) R1)))).
Apply Rle_trans
     with (Rmult
             (Rminus
                R1
                (Rplus
                   (Rmult (6) (Rinv (vNum b)))
                   (Rmult (Rmult (6) (length L)) (Rinv (Rminus (vNum b) R1)))))
             (Rabsolu e)).
Apply Rle_monotony_r; Auto with real.
Unfold Rminus; Apply Rle_compatibility; Apply Rle_Ropp.
Replace (Rmult
           (Rmult (6) (length (cons a L))) (Rinv (Rplus (vNum b) (Ropp R1))))
     with (Rplus
             (Rmult (6) (Rinv (Rplus (vNum b) (Ropp R1))))
             (Rmult (Rmult (6) (length L)) (Rinv (Rplus (vNum b) (Ropp R1)))));
 [Apply Rle_compatibility_r | Idtac].
Apply Rle_monotony; Apply Rlt_le; Auto with real arith.
Apply Rlt_Rinv_R1; Auto with real arith.
Apply Rle_anti_compatibility with (INR (1)).
Replace (Rplus (1) R1) with (INR (2));
 Replace (Rplus (1) (Rplus (vNum b) (Ropp R1))) with (INR (vNum b)); Try Ring.
Rewrite pGivesBound.
Apply Rle_anti_compatibility with (INR (1)).
Replace (Rplus (1) (2)) with (INR (3));
 Replace (Rplus (1) (minus (exp (2) precision) (1)))
      with (INR (exp (2) precision)); Try Ring.
Apply Rle_INR.
Apply le_trans with (4); Auto with arith.
Replace (4) with (exp (2) (2)); [Apply exp_monotone_le | Idtac]; Auto with arith.
Rewrite (minus_INR (exp (2) precision) (1)); [Ring | Idtac].
Replace (1) with (exp (2) (0)); [Apply exp_monotone_le | Idtac]; Auto with arith.
Replace R1 with (INR (1)); Auto with real arith.
Rewrite <- (plus_INR (1) (2)); Auto with arith.
Rewrite (minus_INR (exp (2) precision) (1)); [Ring | Idtac].
Replace (1) with (exp (2) (0)); [Apply exp_monotone_le | Idtac]; Auto with arith.
Pattern 1 (vNum b); Replace (INR (vNum b)) with (Rplus (vNum b) (Ropp R0));
 [Apply Rlt_compatibility | Ring]; Auto with real.
Replace (INR (length (cons a L))) with (Rplus (length L) R1);
 [Ring; Ring | Simpl; Auto with real].
Case L; Intros; Auto.
Simpl; Auto; Ring.
Case H15; Intros.
Absurd <(list float)> (cons e x0)=x0; Auto; Apply cons_neq.
Replace (Rmult
           (Rminus
              R1
              (Rplus
                 (Rmult (6) (Rinv (vNum b)))
                 (Rmult (Rmult (6) (length L)) (Rinv (Rminus (vNum b) R1)))))
           (Rabsolu e))
     with (Rminus
             (Rabsolu e)
             (Rplus
                (Rmult (Rmult (6) (Rinv (vNum b))) (Rabsolu e))
                (Rmult
                   (Rmult (Rmult (6) (length L)) (Rinv (Rminus (vNum b) R1)))
                   (Rabsolu e)))); [Idtac | Ring; Ring].
Apply Rle_trans
     with (Rminus
             (Rabsolu e)
             (Rplus
                (Rabsolu (Rplus x' y'))
                (Rmult
                   (Rmult (Rmult (6) (length L)) (Rinv (Rminus (vNum b) R1)))
                   (Rabsolu e)))).
Unfold Rminus; Apply Rle_compatibility.
Apply Rle_Ropp; Apply Rle_compatibility_r.
Replace (INR (6)) with (Rmult (3) (2)); Replace e with (hd b (cons e x0));
 Auto with real.
Rewrite <- (Rmult_INR (3) (2)); Auto with arith.
Case H14; Intros.
Absurd <(list float)> (cons e x0)=(nil float); Auto; Apply sym_not_eq;
 Apply nil_cons.
Elim H2; Intros; Clear H2.
Cut (Rle
       (Rabsolu (sum L))
       (Rmult
          (length L)
          (Rmult
             (3)
             (Rmult
                (2)
                (Rmult
                   (Rabsolu (hd b (cons e x0))) (Rinv (Rminus (vNum b) (1))))))));
 [Intros | Idtac].
Apply Rle_trans
     with (Rminus (Rabsolu e) (Rplus (Rabsolu (Rplus x' y')) (Rabsolu (sum L)))).
Unfold Rminus; Apply Rle_compatibility; Apply Rle_Ropp; Apply Rle_compatibility.
Apply Rle_trans
     with (Rmult
             (length L)
             (Rmult
                (3)
                (Rmult
                   (2)
                   (Rmult
                      (Rabsolu (hd b (cons e x0))) (Rinv (Rminus (vNum b) (1)))))));
 Auto.
Right; Simpl; Ring.
Apply Rle_trans with (Rabsolu (Rplus e (Rplus (Rplus x' y') (sum L)))).
Rewrite <- (Rabsolu_Ropp (Rplus x' y')); Rewrite <- (Rabsolu_Ropp (sum L)).
Replace (Rminus
           (Rabsolu e)
           (Rplus (Rabsolu (Ropp (Rplus x' y'))) (Rabsolu (Ropp (sum L)))))
     with (Rminus
             (Rminus (Rabsolu e) (Rabsolu (Ropp (Rplus x' y'))))
             (Rabsolu (Ropp (sum L)))); [Idtac | Ring; Ring].
Apply Rle_trans
     with (Rminus
             (Rabsolu (Rminus e (Ropp (Rplus x' y')))) (Rabsolu (Ropp (sum L)))).
Unfold Rminus; Apply Rle_compatibility_r.
Replace (Rplus (Rabsolu e) (Ropp (Rabsolu (Ropp (Rplus x' y')))))
     with (Rminus (Rabsolu e) (Rabsolu (Ropp (Rplus x' y'))));
 Replace (Rplus e (Ropp (Ropp (Rplus x' y'))))
      with (Rminus e (Ropp (Rplus x' y'))); Try Ring.
Apply Rabsolu_triang_inv.
Replace (Rplus e (Rplus (Rplus x' y') (sum L)))
     with (Rminus (Rminus e (Ropp (Rplus x' y'))) (Ropp (sum L))); Try Ring.
Apply Rabsolu_triang_inv.
2:Case H10; Auto; Intros.
2:Replace L with (nil float); Auto; Simpl; Right; Ring; Apply Rabsolu_R0.
2:Elim H2; Auto.
Replace (Rplus e (Rplus (Rplus x' y') (sum L)))
     with (Rplus x (Rplus y (sum (cons a L)))).
2:Apply r_Rplus_plus with (sum x0).
2:Replace (Rplus (sum x0) (Rplus x (Rplus y (sum (cons a L)))))
       with (Rplus x (Rplus y (Rplus (sum (cons a L)) (sum x0))));
   [Rewrite H'1; Simpl; Ring | Ring].
Case H18; Intros.
Absurd <(list float)> x0=(nil float); Auto.
Elim H12; Intros L1 E; Elim E; Intros x1 E0; Elim E0; Intros y1 E2; Elim E2;
 Intros H' H'2; Elim H'2; Intros H'4 H'5; Clear H'2 E2 E0 E H12.
Elim H'5; Intros; Clear H'5.
Replace (Rplus x (Rplus y (sum (cons a L))))
     with (Rplus (Rplus x y) (sum (cons a L))); [Rewrite <- H12 | Ring].
Replace (Rplus (Rplus x1 (Rplus y1 (sum L1))) (sum (cons a L)))
     with (Rplus (Rplus x1 y1) (Rplus (sum L1) (sum (cons a L))));
 [Idtac | Ring].
Apply Rle_trans
     with (Rplus
             (Rabsolu (Rplus x1 y1)) (Rabsolu (Rplus (sum L1) (sum (cons a L)))));
 [Apply Rabsolu_triang | Idtac].
Apply Rle_trans
     with (Rplus
             (Rmult (Rmult (6) (Rinv (vNum b))) (Rabsolu f))
             (Rabsolu (Rplus (sum L1) (sum (cons a L))))).
Elim H13; Intros; Clear H13.
Apply Rle_compatibility_r; Apply Rle_trans with 1 := H21.
Replace (Rmult (3) (2)) with (INR (6)); Replace (hd b x0) with f; Auto with real.
Rewrite E1; Auto.
Rewrite <- (Rmult_INR (3) (2)); Auto with arith.
Rewrite E1; Auto.
Apply Rle_trans
     with (Rplus
             (Rmult (Rmult (6) (Rinv (vNum b))) (Rabsolu f))
             (Rplus (Rabsolu (sum L1)) (Rabsolu (sum (cons a L)))));
 [Apply Rle_compatibility; Apply Rabsolu_triang | Idtac].
Apply Rle_trans
     with (Rplus
             (Rmult (Rmult (6) (Rinv (vNum b))) (Rabsolu f))
             (Rplus
                (Rmult
                   (length L1)
                   (Rmult
                      (3)
                      (Rmult
                         (2)
                         (Rmult
                            (Rabsolu (hd b x0)) (Rinv (Rminus (vNum b) (1)))))))
                (Rabsolu (sum (cons a L))))).
Apply Rle_compatibility; Apply Rle_compatibility_r; Auto.
Case H17; Intros.
Absurd <(list float)> x0=(nil float); Auto.
Case H20; Intros.
Absurd <(list float)> (cons a L)=(nil float); Auto; Apply sym_not_eq;
 Apply nil_cons.
Elim H21; Intros; Clear H21.
Elim H13; Intros; Auto.
Apply Rle_trans
     with (Rplus
             (Rmult (Rmult (6) (Rinv (vNum b))) (Rabsolu f))
             (Rplus
                (Rmult
                   (length L1)
                   (Rmult
                      (3)
                      (Rmult
                         (2)
                         (Rmult
                            (Rabsolu (hd b x0)) (Rinv (Rminus (vNum b) (1)))))))
                (Rmult
                   (length (cons a L))
                   (Rmult
                      (3)
                      (Rmult
                         (2)
                         (Rmult
                            (Rabsolu (hd b x0)) (Rinv (Rminus (vNum b) (1))))))))).
Apply Rle_compatibility; Apply Rle_compatibility; Auto.
Case H17; Intros.
Absurd <(list float)> x0=(nil float); Auto.
Case H20; Intros.
Absurd <(list float)> (cons a L)=(nil float); Auto.
Apply sym_not_eq; Apply nil_cons.
Elim H21; Intros.
Elim H23; Auto.
Replace (hd b x0) with f; [Idtac | Rewrite E1; Auto].
Replace (Rplus
           (Rmult
              (length L1)
              (Rmult
                 (3)
                 (Rmult (2) (Rmult (Rabsolu f) (Rinv (Rminus (vNum b) (1)))))))
           (Rmult
              (length (cons a L))
              (Rmult
                 (3)
                 (Rmult (2) (Rmult (Rabsolu f) (Rinv (Rminus (vNum b) (1))))))))
     with (Rmult
             (Rplus (length L1) (length (cons a L)))
             (Rmult
                (Rmult (3) (2)) (Rmult (Rabsolu f) (Rinv (Rminus (vNum b) (1))))));
 [Idtac | Ring; Ring].
Replace (Rmult (3) (2)) with (INR (6)).
Replace (Rplus
           (Rmult (Rmult (6) (Rinv (vNum b))) (Rabsolu f))
           (Rmult
              (Rplus (length L1) (length (cons a L)))
              (Rmult (6) (Rmult (Rabsolu f) (Rinv (Rminus (vNum b) (1)))))))
     with (Rmult
             (Rplus
                (Rmult (6) (Rinv (vNum b)))
                (Rmult
                   (Rplus (length L1) (length (cons a L)))
                   (Rmult (6) (Rinv (Rminus (vNum b) (1)))))) (Rabsolu f));
 [Idtac | Ring].
Apply Rle_trans
     with (Rmult
             (Rmult
                (Rminus
                   R1
                   (Rmult
                      (Rmult (6) (length (cons a L)))
                      (Rinv (Rminus (vNum b) R1))))
                (Rmult
                   (Rplus (Rmult (6) (length input)) (6))
                   (Rinv
                      (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input))))))
             (Rabsolu f)); [Idtac | Right; Ring].
Apply Rle_monotony_r; Auto with real.
Apply Rle_trans
     with (Rplus
             (Rmult (6) (Rinv (Rminus (vNum b) (1))))
             (Rmult
                (Rplus (length L1) (length (cons a L)))
                (Rmult (6) (Rinv (Rminus (vNum b) (1)))))).
Apply Rle_compatibility_r; Apply Rle_monotony; Auto with real arith.
Apply Rlt_le; Replace R0 with (INR (0)); Auto with real arith.
Apply Rlt_le.
Apply Rlt_Rinv_R1; Auto with real arith.
Apply Rle_anti_compatibility with (INR (1)).
Replace (Rplus (1) R1) with (INR (2));
 Replace (Rplus (1) (Rminus (vNum b) (1))) with (INR (vNum b)); Try Ring.
Rewrite pGivesBound.
Apply Rle_anti_compatibility with (INR (1)).
Replace (Rplus (1) (2)) with (INR (3));
 Replace (Rplus (1) (minus (exp (2) precision) (1)))
      with (INR (exp (2) precision)); Try Ring.
Apply Rle_INR.
Apply le_trans with (4); Auto with arith.
Replace (4) with (exp (2) (2)); [Apply exp_monotone_le | Idtac]; Auto with arith.
Rewrite (minus_INR (exp (2) precision) (1)); [Ring | Idtac].
Replace (1) with (exp (2) (0)); [Apply exp_monotone_le | Idtac]; Auto with arith.
Replace R1 with (INR (1)); Auto with real arith.
Rewrite <- (plus_INR (1) (2)); Auto with arith.
Rewrite (minus_INR (exp (2) precision) (1)); [Ring | Idtac].
Replace (1) with (exp (2) (0)); [Apply exp_monotone_le | Idtac]; Auto with arith.
Unfold Rminus; Pattern 1 (vNum b);
 Replace (INR (vNum b)) with (Rplus (vNum b) (Ropp R0));
 [Apply Rlt_compatibility | Ring]; Auto with real.
Replace (Rplus
           (Rmult (6) (Rinv (Rminus (vNum b) (1))))
           (Rmult
              (Rplus (length L1) (length (cons a L)))
              (Rmult (6) (Rinv (Rminus (vNum b) (1))))))
     with (Rmult
             (Rplus R1 (Rplus (length L1) (length (cons a L))))
             (Rmult (6) (Rinv (Rminus (vNum b) (1))))); [Idtac | Ring].
Replace (Rplus (length L1) (length (cons a L)))
     with (INR (length (app L1 (cons a L)))).
Cut (Rle R0 (Rmult (6) (Rinv (Rminus (vNum b) (1))))); [Intros | Idtac].
Apply Rle_trans
     with (Rmult
             (Rplus R1 (length input)) (Rmult (6) (Rinv (Rminus (vNum b) (1)))));
 [Apply Rle_monotony_r; Auto with real | Idtac].
Apply Rle_compatibility; Apply Rle_INR.
Apply endof_length; Auto.
Apply Rle_trans
     with (Rmult
             (Rmult R1 (Rinv (Rminus (vNum b) (1))))
             (Rplus (Rmult (6) (length input)) (6)));
 [Right; Ring; Ring | Idtac].
Apply Rle_trans
     with (Rmult
             (Rmult
                (Rinv (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input))))
                (Rminus
                   R1
                   (Rmult
                      (Rmult (6) (length (cons a L)))
                      (Rinv (Rminus (vNum b) R1)))))
             (Rplus (Rmult (6) (length input)) (6))); [Idtac | Right; Ring].
Apply Rle_monotony_r; Auto with real.
Apply Rle_trans with (Rplus R0 (6)); [Apply Rlt_le | Apply Rle_compatibility_r];
 Auto with real arith.
Replace (Rplus R0 (6)) with (INR (6)); Auto with arith real.
Apply Rmult_Rle_O; Replace R0 with (INR (0)); Auto with real arith.
Replace (Rminus
           R1
           (Rmult (Rmult (6) (length (cons a L))) (Rinv (Rminus (vNum b) R1))))
     with (Rmult
             (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length (cons a L))))
             (Rinv (Rminus (vNum b) R1))).
Rewrite <- Rmult_assoc; Apply Rle_monotony_r.
Apply Rlt_le; Apply Rlt_Rinv; Auto.
Apply Rle_trans
     with (Rmult
             (Rinv (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input))))
             (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input)))).
Rewrite Rmult_sym; Rewrite Rinv_r; Auto with real.
Apply sym_not_eqT; Apply imp_not_Req; Left; Auto.
Apply Rle_monotony.
Apply Rlt_le; Apply Rlt_Rinv; Auto.
Unfold Rminus; Apply Rle_compatibility; Apply Rle_Ropp.
Apply Rle_monotony; Auto with real arith.
Replace R0 with (INR (0)); Auto with real arith.
Apply Rle_INR; Apply endof_length; Auto.
Pattern 2 R1;
 Replace R1 with (Rmult (Rminus (vNum b) R1) (Rinv (Rminus (vNum b) R1)));
 [Ring | Idtac].
Apply Rinv_r; Auto with real arith.
Apply sym_not_eqT; Apply imp_not_Req; Left; Auto.
Apply Rmult_Rle_O; Auto with real arith.
Replace R0 with (INR (0)); Auto with real arith.
Apply Rlt_le; Apply Rlt_Rinv; Replace (INR (1)) with R1; Auto with real.
Rewrite <- (plus_INR (length L1) (length (cons a L))).
Rewrite (app_length L1 (cons a L)); Auto with real.
Rewrite <- Rmult_INR; Auto with real arith.
Replace (FtoR (2) e) with (FtoRradix e); Auto with real; Ring.
Apply Rlt_anti_compatibility
     with (Rmult (Rmult (6) (length (cons a L))) (Rinv (Rminus (vNum b) R1))).
Replace (Rplus
           (Rmult (Rmult (6) (length (cons a L))) (Rinv (Rminus (vNum b) R1)))
           R0)
     with (Rmult (Rmult (6) (length (cons a L))) (Rinv (Rminus (vNum b) R1)));
 [Idtac | Ring].
Replace (Rplus
           (Rmult (Rmult (6) (length (cons a L))) (Rinv (Rminus (vNum b) R1)))
           (Rminus
              R1
              (Rmult
                 (Rmult (6) (length (cons a L))) (Rinv (Rminus (vNum b) R1)))))
     with R1; [Auto | Ring].
Apply Rle_lt_trans
     with (Rmult (Rmult (6) (length input)) (Rinv (Rminus (vNum b) R1)));
 [Apply Rle_monotony_r | Auto].
Apply Rlt_le; Apply Rlt_Rinv; Auto with real.
Apply Rle_monotony; [Replace R0 with (INR (0)); Auto with real arith | Idtac].
Apply Rle_INR; Apply endof_length; Auto.
Qed.

Theorem FexpAdd_aux2:
  (L:(list float))
  L=input ->
  (IsExp b input) ->
  (Rlt (Rmult (Rmult (6) (length input)) (Rinv (Rminus (vNum b) (1)))) R1) ->
  (Ex [output:(list float)]
  (~ input==(nil float) ->(le (length output) (S (length input)))) /\
  (<R> ((sum input))==((sum output)) /\ (IsRleEpsExp output))).
Intros L; Case L; Intros.
Rewrite <- H; Exists (nil float); Repeat (Split; Auto); Apply IsRleEpsExpNil.
Generalize H0; Rewrite <- H; Intros.
Cut (Fbounded b f); [Intros Z0 | Inversion H2; Auto].
Cut (Fbounded b (Fzero (Fexp f))); [Intros Z1 | Apply FboundedZeroSameExp; Auto].
Cut <R> (Fzero (Fexp f))==R0;
 [Intros Z2 | Unfold FtoRradix; Apply FzeroisReallyZero].
Cut (Rle R0 (vNum b)); [Intros Z3 | Idtac].
2:Replace R0 with (INR (0)); Auto with real arith.
Cut (Rlt R0 (Rminus (vNum b) (1))); [Intros Z4 | Idtac].
2:Apply Rlt_anti_compatibility with (INR (1)).
2:Replace (Rplus (1) R0) with (INR (1));
   Replace (Rplus (1) (Rminus (vNum b) (1))) with (INR (vNum b)); Try Ring.
2:Rewrite pGivesBound.
2:Apply Rlt_anti_compatibility with (INR (1)).
2:Replace (Rplus (1) (1)) with (INR (2));
   Replace (Rplus (1) (minus (exp (2) precision) (1)))
        with (INR (exp (2) precision)); Try Ring.
2:Apply Rlt_INR.
2:Pattern 1 (2); Replace (2) with (exp (2) (1));
   [Apply exp_monotone_lt | Idtac]; Auto with arith.
2:Rewrite (minus_INR (exp (2) precision) (1)); [Ring | Idtac].
2:Replace (1) with (exp (2) (0)); [Apply exp_monotone_le | Idtac];
   Auto with arith.
2:Rewrite (minus_INR (exp (2) precision) (1)); [Ring | Idtac].
2:Replace (1) with (exp (2) (0)); [Apply exp_monotone_le | Idtac];
   Auto with arith.
LApply (FexpAdd_aux l (nil float) f (Fzero (Fexp f)));
 [Intros H'3; LApply H'3;
   [Intros H'4; LApply H'4;
     [Intros H'5; LApply H'5;
       [Intros H'6; LApply H'6;
         [Intros H'7; LApply H'7;
           [Intros H'8; LApply H'8;
             [Intros H'9; LApply H'9;
               [Intros H'10; LApply H'10;
                 [Intros H'11; LApply H'11;
                   [Intros H'12; LApply H'12;
                     [Intros H'13; LApply H'13;
                       [Intros H'14; Try Exact H'14;
                         Clear H'13 H'12 H'11 H'10 H'9 H'8 H'7 H'6 H'5 H'4 H'3 |
                         Clear H'13 H'12 H'11 H'10 H'9 H'8 H'7 H'6 H'5 H'4 H'3] |
                       Clear H'12 H'11 H'10 H'9 H'8 H'7 H'6 H'5 H'4 H'3] |
                     Clear H'11 H'10 H'9 H'8 H'7 H'6 H'5 H'4 H'3] |
                   Clear H'10 H'9 H'8 H'7 H'6 H'5 H'4 H'3] |
                 Clear H'9 H'8 H'7 H'6 H'5 H'4 H'3] |
               Clear H'8 H'7 H'6 H'5 H'4 H'3] | Clear H'7 H'6 H'5 H'4 H'3] |
           Clear H'6 H'5 H'4 H'3] | Clear H'5 H'4 H'3] | Clear H'4 H'3] |
     Clear H'3] | Idtac]; Auto with arith zarith real float.
2:Simpl; Apply IsExpZle; Auto.
2:Apply IsRleEpsExpNil.
2:Inversion H2; Auto; Apply IsExpNil.
2:Rewrite <- H; Unfold endof; Exists (cons f (nil float)); Auto.
Elim H'14; Intros x' E; Elim E; Intros y' E0; Elim E0; Intros output' E1;
 Elim E1; Intros H' H'0; Elim H'0; Intros H'1 H'2; Elim H'2; Intros H'3 H'4;
 Elim H'4; Intros H'5 H'6; Elim H'6; Intros H'7 H'8; Elim H'8; Intros H'9 H'10;
 Elim H'10; Intros H'11 H'12; Elim H'12; Intros H'13 H'15;
 Clear H'12 H'10 H'8 H'6 H'4 H'2 H'0 E1 E0 E H'14 H'11.
LApply (ClosestTotal b (2) precision);
 [Intros H'6; LApply H'6;
   [Intros H'8; LApply H'8; [Intros H'10; Clear H'8 H'6 | Clear H'8 H'6] |
     Clear H'6] | Idtac]; Auto.
Generalize H'10; Unfold TotalP; Intros ExC.
Elim (ExC (Rplus x' y')); Intros p E0.
LApply (errorBoundedPlus b (2) precision);
 [Intros H'6; LApply H'6;
   [Intros H'8; LApply H'8;
     [Intros H'11; Elim (H'11 x' y' p);
       [Intros q E; Elim E; Intros H'20 H'21; Elim H'21; Intros H'22 H'23;
         Try Exact H'23; Clear H'21 E H'8 H'6 | Clear H'8 H'6 | Clear H'8 H'6 |
         Clear H'8 H'6] | Clear H'8 H'6] | Clear H'6] | Idtac]; Auto.
Clear H'11 ExC H'10.
Exists (cons q (cons p output')); Repeat Split.
Intros; Simpl.
Apply le_trans with (S (S (plus (length l) (length (nil float)))));
 Auto with arith.
Simpl; Apply le_refl_eq; Auto with arith zarith.
Generalize H'3; Rewrite Z2; Simpl; Fold FtoRradix.
Replace (Rplus f (Rplus R0 (Rplus R0 (sum l)))) with (Rplus f (sum l));
 [Intros H'25; Rewrite H'25 | Ring].
Unfold FtoRradix; Rewrite H'20; Fold FtoRradix; Ring.
Cut (Fbounded b p); [Intros | Idtac].
2:Apply RoundedModeBounded with (2) (Closest b (2)) (Rplus x' y'); Auto;
   Apply ClosestRoundedModeP with precision; Auto.
Apply IsRleEpsExpTop; Auto.
Case (Req_EM p R0); Intros.
Cut (FtoRradix q)==R0; Intros.
Rewrite H5; Rewrite H4; Rewrite Rabsolu_R0; Right; Ring.
Unfold FtoRradix; Rewrite H'20; Apply TwoSumNul with b precision; Auto.
Apply Rle_trans
     with (Rmult
             (Rmult (Rabsolu p) (Rinv (2)))
             (Rmult (2) (Rinv (plus (vNum b) (1))))).
Unfold FtoRradix; Rewrite H'20; Rewrite <- Rabsolu_Ropp.
Replace (Ropp (Rminus (Rplus (FtoR (2) x') (FtoR (2) y')) (FtoR (2) p)))
     with (Rminus (FtoR (2) p) (Rplus (FtoR (2) x') (FtoR (2) y')));
 [Idtac | Ring].
Apply plusErrorBound1bis with precision; Auto.
Contradict H4; Unfold FtoRradix; Apply is_Fzero_rep1; Auto.
Rewrite Rmult_assoc; Rewrite Rmult_sym; Apply Rle_monotony_r; Auto with real.
Rewrite <- Rmult_assoc.
Replace (Rmult (Rinv (2)) (2)) with R1;
 [Idtac | Ring; Rewrite Rinv_r; Auto with real arith].
Apply Rle_trans
     with (Rmult
             (Rplus (Rmult (6) (length input)) (6)) (Rinv (plus (vNum b) (1))));
 [Apply Rle_monotony_r | Apply Rle_monotony]; Auto with real arith zarith.
Apply Rlt_le; Apply Rlt_Rinv; Auto with real arith.
Replace R1 with (INR (1)); Auto with real arith zarith.
Rewrite <- Rmult_INR; Rewrite <- plus_INR; Apply Rle_INR; Auto with arith zarith.
Replace R0 with (INR (0)); Auto with real arith zarith.
Rewrite <- Rmult_INR; Rewrite <- plus_INR; Apply Rle_INR; Auto with arith zarith.
Apply Rle_Rinv.
Apply Rlt_anti_compatibility with (Rmult (6) (length input)).
Replace (Rplus (Rmult (6) (length input)) R0) with (Rmult (6) (length input));
 [Idtac | Ring].
Replace (Rplus
           (Rmult (6) (length input))
           (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input))))
     with (Rminus (vNum b) (1)); [Idtac | Ring].
Apply Rlt_mult_anticompatibility with (Rinv (Rminus (vNum b) (1))).
Rewrite Rmult_sym;
 Replace (Rmult (Rinv (Rminus (vNum b) (1))) (Rminus (vNum b) (1))) with R1;
 Auto with real.
Rewrite Rmult_sym; Rewrite Rinv_r; Auto with real.
Apply imp_not_Req; Auto.
Apply Rlt_Rinv; Auto.
Apply Rle_trans with (Rminus (plus (vNum b) (1)) (Rmult (6) (length input)));
 Auto with real arith zarith.
Unfold Rminus; Apply Rle_compatibility_r; Auto with real arith zarith.
Rewrite plus_INR; Apply Rle_compatibility; Replace (INR (1)) with R1;
 Auto with real arith.
Apply Rlt_le; Apply Rlt_trans with R0; Auto with real arith zarith.
Rewrite <- Ropp_O; Apply Rlt_Ropp; Auto with real arith zarith.
Pattern 2 (INR (plus (vNum b) (1)));
 Replace (INR (plus (vNum b) (1))) with (Rminus (plus (vNum b) (1)) R0);
 Unfold Rminus; Try Ring; Apply Rle_compatibility; Apply Rle_Ropp;
 Auto with real arith zarith.
Apply Rmult_Rle_O; Replace R0 with (INR (0)); Auto with real arith zarith.
Generalize H'15 H'9; Case output'; Clear H'15 H'9; Intros.
Apply IsRleEpsExpSingle; Auto.
Apply IsRleEpsExpTop; Auto.
Inversion H'9; Auto.
Case H'15; Intros; Clear H'15.
Absurd <(list float)> (cons f0 l0)=(nil float); Auto; Apply sym_not_eq;
 Apply nil_cons.
Apply Rle_trans
     with (Rmult
             (Rmult (Rinv (Rminus (vNum b) (1))) (vNum b))
             (Rabsolu (Rplus x' y'))).
Apply Rle_trans
     with (Rmult
             (Rabsolu (Rplus (FtoR (2) x') (FtoR (2) y')))
             (Rmult
                (Rmult (2) (vNum b)) (Rinv (Rminus (Rmult (2) (vNum b)) (2))))).
Unfold FtoRradix; Apply RoundBound with precision; Auto.
Pattern 1 (2); Replace (2) with (mult (2) (1)); Auto with arith.
Apply lt_comp_mult_l; Auto with arith.
Rewrite pGivesBound.
Apply lt_Rlt; Apply Rlt_anti_compatibility with (INR (1)).
Replace (Rplus (1) (1)) with (INR (2));
 Replace (Rplus (1) (minus (exp (2) precision) (1)))
      with (INR (exp (2) precision)); Try Ring.
Apply Rlt_INR.
Pattern 1 (2); Replace (2) with (exp (2) (1)); [Apply exp_monotone_lt | Idtac];
 Auto with arith.
Rewrite (minus_INR (exp (2) precision) (1)); [Ring | Idtac].
Replace (1) with (exp (2) (0)); [Apply exp_monotone_le | Idtac]; Auto with arith.
Rewrite (minus_INR (exp (2) precision) (1)); [Ring | Idtac].
Replace (1) with (exp (2) (0)); [Apply exp_monotone_le | Idtac]; Auto with arith.
Rewrite Rmult_sym; Fold FtoRradix; Apply Rle_monotony_r; Auto with real; Right.
Replace (Rminus (Rmult (2) (vNum b)) (2)) with (Rmult (2) (Rminus (vNum b) (1)));
 [Idtac | Ring].
Rewrite Rinv_Rmult; Auto with arith real.
Apply trans_eqT
     with (Rmult
             (Rmult (2) (Rinv (2)))
             (Rmult (vNum b) (Rinv (Rminus (vNum b) (1)))));
 [Ring; Ring | Idtac].
Rewrite Rinv_r; [Ring | Auto with real arith].
Apply imp_not_Req; Auto.
Apply Rle_mult_anticompatibility with (Rinv (vNum b)); Auto.
2:Apply Rlt_Rinv; Replace R0 with (INR (0)); Auto with real arith.
2:Apply Rlt_INR; Apply lt_vNum with (2) precision; Auto with arith.
Apply Rle_mult_anticompatibility with (Rminus (vNum b) (1)); Auto.
Apply Rle_trans
     with (Rmult
             (Rabsolu (Rplus x' y'))
             (Rmult
                (Rmult (Rminus (vNum b) (1)) (Rinv (Rminus (vNum b) (1))))
                (Rmult (vNum b) (Rinv (vNum b))))); [Right; Ring; Ring | Idtac].
Rewrite Rinv_r; [Idtac | Apply imp_not_Req; Auto].
Rewrite Rinv_r; [Idtac | Apply sym_not_eqT; Apply imp_not_Req; Auto].
2:Left; Replace R0 with (INR (0)); Auto with real arith; Apply Rlt_INR;
   Apply lt_vNum with (2) precision; Auto with arith.
Apply Rle_trans with (Rabsolu (Rplus x' y')); [Right; Ring | Idtac].
Apply Rle_trans
     with (Rmult
             (Rmult
                (Rmult (Rminus (vNum b) (1)) (Rinv (vNum b)))
                (Rmult
                   (Rplus (Rmult (6) (length input)) (6))
                   (Rinv
                      (Rminus (Rminus (vNum b) (1)) (Rmult (6) (length input))))))
             (Rabsolu (hd b (cons f0 l0)))); [Auto | Simpl; Right; Ring; Ring].
Qed.

Theorem FexpAdd_main:
  (IsExp b input) ->
  (Rlt (Rmult (Rmult (6) (length input)) (Rinv (Rminus (vNum b) (1)))) R1) ->
  (Ex [output:(list float)]
  (~ input==(nil float) ->(le (length output) (S (length input)))) /\
  (<R> ((sum input))==((sum output)) /\ (IsRleEpsExp output))).
LApply (FexpAdd_aux2 input); Auto.
Qed.
End FexpAdd.
