
(****************************************************************************
                                                                             
          IEEE754  :  FexpDiv                                                    
                                                                             
                 Sylvie Boldo                                           
                                                                             
*****************************************************************************
*)
Require Import Arith.
Require Import Reals.
Require Export Faux.
Require Export sTactic.
Section Div.
Variables wi, wi1, qi, D, N, qi1, wi2, ai, bi, eps1, eps2, eps3, ep:R.
Hypothesis Hw:<R> wi==(Rminus wi1 (Rmult qi D)).
Hypothesis Ha:(Rle (Rabsolu (Rmult (Rinv wi1) (Rminus ai wi1))) eps1).
Hypothesis Hb:(Rle (Rabsolu (Rmult (Rinv D) (Rminus bi D))) eps2).
Hypothesis Hq:
           (Rle
              (Rabsolu
                 (Rmult
                    (Rinv (Rmult (Rinv bi) ai)) (Rminus qi (Rmult (Rinv bi) ai))))
              eps3).
Hypothesis NZwi:~ wi1==R0.
Hypothesis NZD:~ D==R0.
Hypothesis NZbi:~ bi==R0.
Hypothesis NZai:~ ai==R0.
Hypothesis PosEps1:(Rle R0 eps1).
Hypothesis PosEps2:(Rle R0 eps2).
Hypothesis PosEps3:(Rle R0 eps3).
Hypothesis LeEps1:(Rlt eps1 R1).
Hypothesis LeEps2:(Rlt eps2 R1).
Hypothesis LeEps3:(Rlt eps3 R1).
Hypothesis Hep:<R> ep==(Rmax (Rmax eps1 eps2) eps3).

Theorem InegAbsInf:
  (x, y, eps:R) ~ x==R0 -> (Rle (Rabsolu (Rmult (Rinv x) (Rminus y x))) eps) ->
  (Rle (Rabsolu y) (Rmult (Rplus R1 eps) (Rabsolu x))).
Intros.
Replace (Rmult (Rplus R1 eps) (Rabsolu x))
     with (Rplus (Rabsolu x) (Rmult eps (Rabsolu x))).
2:Ring.
Replace (Rabsolu y) with (Rplus (Rabsolu x) (Rminus (Rabsolu y) (Rabsolu x))).
2:Ring.
Apply Rle_compatibility.
Apply Rle_trans with (Rabsolu (Rminus y x)).
Apply Rabsolu_triang_inv.
Apply Rle_mult_anticompatibility with (Rinv (Rabsolu x)).
Replace (Rmult (Rinv (Rabsolu x)) (Rmult eps (Rabsolu x))) with eps.
Replace (Rmult (Rinv (Rabsolu x)) (Rabsolu (Rminus y x)))
     with (Rabsolu (Rmult (Rinv x) (Rminus y x))); Auto.
Replace (Rinv (Rabsolu x)) with (Rabsolu (Rinv x)).
2:Apply Rabsolu_Rinv; Auto.
Apply Rabsolu_mult.
Replace (Rmult (Rinv (Rabsolu x)) (Rmult eps (Rabsolu x)))
     with (Rmult eps (Rmult (Rabsolu x) (Rinv (Rabsolu x)))).
Replace (Rmult (Rabsolu x) (Rinv (Rabsolu x))) with R1.
Ring.
Apply sym_eqT.
Apply Rinv_r.
2:Ring.
Apply Rabsolu_no_R0; Auto.
Replace (Rinv (Rabsolu x)) with (Rabsolu (Rinv x)).
Apply Rabsolu_pos_lt.
2:Apply Rabsolu_Rinv; Auto.
Apply Rinv_RN; Auto.
Qed.

Theorem InegAbsSup:
  (x, y, eps:R) ~ x==R0 -> (Rle (Rabsolu (Rmult (Rinv x) (Rminus y x))) eps) ->
  (Rle (Rmult (Rminus R1 eps) (Rabsolu x)) (Rabsolu y)).
Intros.
Replace (Rmult (Rminus R1 eps) (Rabsolu x))
     with (Rplus (Rabsolu x) (Ropp (Rmult eps (Rabsolu x)))).
2:Ring.
Replace (Rabsolu y)
     with (Rplus (Rabsolu x) (Ropp (Rminus (Rabsolu x) (Rabsolu y)))).
Apply Rle_compatibility.
Apply Rle_Ropp.
Apply Rle_trans with (Rabsolu (Rminus y x)).
Replace (Rabsolu (Rminus y x)) with (Rabsolu (Rminus x y)).
Apply Rabsolu_triang_inv.
Replace (Rminus y x) with (Ropp (Rminus x y)); Auto with real.
Apply sym_eqT.
Apply Rabsolu_Ropp.
Ring.
Apply Rle_mult_anticompatibility with (Rinv (Rabsolu x)).
Replace (Rmult (Rinv (Rabsolu x)) (Rmult eps (Rabsolu x))) with eps.
Replace (Rmult (Rinv (Rabsolu x)) (Rabsolu (Rminus y x)))
     with (Rabsolu (Rmult (Rinv x) (Rminus y x))); Auto.
Replace (Rinv (Rabsolu x)) with (Rabsolu (Rinv x)).
2:Apply Rabsolu_Rinv; Auto.
Apply Rabsolu_mult.
Replace (Rmult (Rinv (Rabsolu x)) (Rmult eps (Rabsolu x)))
     with (Rmult eps (Rmult (Rabsolu x) (Rinv (Rabsolu x)))).
Replace (Rmult (Rabsolu x) (Rinv (Rabsolu x))) with R1.
Ring.
Apply sym_eqT.
Apply Rinv_r.
2:Ring.
Apply Rabsolu_no_R0; Auto.
Replace (Rinv (Rabsolu x)) with (Rabsolu (Rinv x)).
Apply Rabsolu_pos_lt.
2:Apply Rabsolu_Rinv; Auto.
Apply Rinv_RN; Auto.
Ring.
Qed.

Theorem DivFirstCase:
  (Rle
     (Rmult (Rminus (Rabsolu wi1) (Rabsolu (Rmult qi D))) (Rinv (Rabsolu wi1)))
     (Rminus
        R1
        (Rmult (Rmult (Rminus R1 eps3) (Rminus R1 eps1)) (Rinv (Rplus R1 eps2))))).
Apply Rle_trans
     with (Rminus R1 (Rmult (Rabsolu (Rmult qi D)) (Rinv (Rabsolu wi1)))).
Right.
Replace R1 with (Rmult (Rabsolu wi1) (Rinv (Rabsolu wi1))); [Ring | Idtac].
Apply Rinv_r; Apply Rabsolu_no_R0; Auto with real.
Unfold Rminus; Apply Rle_compatibility; Apply Rle_Ropp.
Fold (Rminus R1 eps1); Fold (Rminus R1 eps3).
Apply Rle_trans
     with (Rmult
             (Rmult (Rmult (Rminus R1 eps3) (Rminus R1 eps1)) (Rabsolu D))
             (Rinv (Rmult (Rplus R1 eps2) (Rabsolu D)))); [Right | Idtac].
Replace (Rmult
           (Rmult (Rmult (Rminus R1 eps3) (Rminus R1 eps1)) (Rabsolu D))
           (Rinv (Rmult (Rplus R1 eps2) (Rabsolu D))))
     with (Rmult
             (Rmult
                (Rmult (Rminus R1 eps3) (Rminus R1 eps1))
                (Rmult (Rabsolu D) (Rinv (Rabsolu D)))) (Rinv (Rplus R1 eps2)));
 [Replace (Rmult (Rabsolu D) (Rinv (Rabsolu D))) with R1 | Idtac].
Ring.
Apply sym_eqT; Apply Rinv_r; Apply Rabsolu_no_R0; Auto with real.
Replace (Rinv (Rmult (Rplus R1 eps2) (Rabsolu D)))
     with (Rmult (Rinv (Rplus R1 eps2)) (Rinv (Rabsolu D))); Ring.
Ring; Apply sym_eqT; Apply Rinv_Rmult;
 [Auto with real | Apply Rabsolu_no_R0; Auto].
Apply sym_not_eqT; Apply imp_not_Req; Left; Auto with real.
Apply Rlt_le_trans with R1; Auto with real.
Pattern 1 R1; Replace R1 with (Rplus R1 R0); Auto with real.
Apply Rle_trans
     with (Rmult
             (Rmult (Rmult (Rminus R1 eps3) (Rminus R1 eps1)) (Rabsolu D))
             (Rinv (Rabsolu bi))).
Apply Rle_monotony.
Apply Rmult_Rle_O; Auto with real.
Apply Rmult_Rle_O; Auto with real.
Apply Rle_anti_compatibility with eps3; Auto with real.
Replace (Rplus eps3 R0) with eps3;
 [Replace (Rplus eps3 (Rminus R1 eps3)) with R1;
   [Apply Rlt_le; Auto with real | Ring] | Ring].
Apply Rle_anti_compatibility with eps1; Auto with real.
Replace (Rplus eps1 R0) with eps1;
 [Replace (Rplus eps1 (Rminus R1 eps1)) with R1;
   [Apply Rlt_le; Auto with real | Ring] | Ring].
Apply Rle_Rinv; Auto with real.
Apply Rle_lt; Auto with real; Apply sym_not_eqT; Apply Rabsolu_no_R0; Auto.
Apply InegAbsInf; Auto.
Apply Rle_trans
     with (Rmult
             (Rmult
                (Rmult (Rminus R1 eps3) (Rmult (Rminus R1 eps1) (Rabsolu wi1)))
                (Rabsolu D)) (Rmult (Rinv (Rabsolu wi1)) (Rinv (Rabsolu bi))));
 [Right | Idtac].
Replace (Rmult
           (Rmult
              (Rmult (Rminus R1 eps3) (Rmult (Rminus R1 eps1) (Rabsolu wi1)))
              (Rabsolu D)) (Rmult (Rinv (Rabsolu wi1)) (Rinv (Rabsolu bi))))
     with (Rmult
             (Rmult
                (Rmult
                   (Rminus R1 eps3) (Rmult (Rminus R1 eps1) (Rinv (Rabsolu bi))))
                (Rabsolu D)) (Rmult (Rabsolu wi1) (Rinv (Rabsolu wi1))));
 [Replace (Rmult (Rabsolu wi1) (Rinv (Rabsolu wi1))) with R1; [Ring | Idtac] |
   Ring].
Apply sym_eqT; Apply Rinv_r; Apply Rabsolu_no_R0; Auto with real.
Apply Rle_trans
     with (Rmult
             (Rmult (Rmult (Rminus R1 eps3) (Rabsolu ai)) (Rabsolu D))
             (Rmult (Rinv (Rabsolu wi1)) (Rinv (Rabsolu bi)))).
Apply Rle_monotony_r.
Apply Rmult_Rle_O.
Apply Rlt_le; Apply Rlt_Rinv; Apply Rle_lt; Auto with real; Apply sym_not_eqT;
 Apply Rabsolu_no_R0; Auto with real.
Apply Rlt_le; Apply Rlt_Rinv; Apply Rle_lt; Auto with real; Apply sym_not_eqT;
 Apply Rabsolu_no_R0; Auto with real.
Apply Rle_monotony_r; Auto with real.
Apply Rle_monotony; Auto with real.
Apply Rle_anti_compatibility with eps3; Auto with real.
Replace (Rplus eps3 R0) with eps3;
 [Replace (Rplus eps3 (Rminus R1 eps3)) with R1;
   [Apply Rlt_le; Auto with real | Ring] | Ring].
Apply InegAbsSup; Auto.
Replace (Rmult
           (Rmult (Rmult (Rminus R1 eps3) (Rabsolu ai)) (Rabsolu D))
           (Rmult (Rinv (Rabsolu wi1)) (Rinv (Rabsolu bi))))
     with (Rmult
             (Rmult
                (Rmult
                   (Rinv (Rabsolu bi)) (Rmult (Rminus R1 eps3) (Rabsolu ai)))
                (Rabsolu D)) (Rinv (Rabsolu wi1)));
 [Apply Rle_monotony_r | Ring].
Apply Rlt_le; Apply Rlt_Rinv; Apply Rle_lt; Auto with real; Apply sym_not_eqT;
 Apply Rabsolu_no_R0; Auto with real.
Rewrite (Rabsolu_mult qi D).
Apply Rle_monotony_r; Auto with real.
Replace (Rmult (Rinv (Rabsolu bi)) (Rmult (Rminus R1 eps3) (Rabsolu ai)))
     with (Rmult (Rminus R1 eps3) (Rmult (Rinv (Rabsolu bi)) (Rabsolu ai)));
 [Idtac | Ring; Ring].
Replace (Rmult (Rinv (Rabsolu bi)) (Rabsolu ai))
     with (Rabsolu (Rmult (Rinv bi) ai)).
Apply InegAbsSup; Auto.
Contradict NZai.
Apply Rmult_anticompatibility with (Rinv bi); Auto.
Apply Rinv_RN; Auto.
Ring; Auto with real.
Rewrite (Rabsolu_mult (Rinv bi) ai).
Rewrite (Rabsolu_Rinv bi); Auto with real.
Qed.

Theorem DivSecondCase:
  (Rle
     (Rmult (Rminus (Rabsolu (Rmult qi D)) (Rabsolu wi1)) (Rinv (Rabsolu wi1)))
     (Rminus
        (Rmult (Rmult (Rplus R1 eps3) (Rplus R1 eps1)) (Rinv (Rminus R1 eps2)))
        R1)).
Apply Rle_trans
     with (Rminus (Rmult (Rabsolu (Rmult qi D)) (Rinv (Rabsolu wi1))) R1);
 [Right | Idtac].
Ring; Ring.
Replace (Rmult (Rinv (Rabsolu wi1)) (Rabsolu wi1)) with R1; Ring.
Apply sym_eqT; Apply Rinv_r; Apply Rabsolu_no_R0; Auto with real.
Replace (Rminus (Rmult (Rabsolu (Rmult qi D)) (Rinv (Rabsolu wi1))) R1)
     with (Rplus (Ropp R1) (Rmult (Rabsolu (Rmult qi D)) (Rinv (Rabsolu wi1))));
 Replace (Rminus
            (Rmult
               (Rmult (Rplus R1 eps3) (Rplus R1 eps1)) (Rinv (Rminus R1 eps2)))
            R1)
      with (Rplus
              (Ropp R1)
              (Rmult
                 (Rmult (Rplus R1 eps3) (Rplus R1 eps1)) (Rinv (Rminus R1 eps2))));
 Try Ring.
Apply Rle_compatibility.
Rewrite (Rabsolu_mult qi D).
Apply Rle_trans
     with (Rmult
             (Rmult
                (Rmult (Rplus R1 eps3) (Rabsolu (Rmult (Rinv bi) ai)))
                (Rabsolu D)) (Rinv (Rabsolu wi1))).
Apply Rle_monotony_r; Auto with real.
Apply Rlt_le; Apply Rlt_Rinv; Apply Rle_lt; Auto with real; Apply sym_not_eqT;
 Apply Rabsolu_no_R0; Auto with real.
Apply Rle_monotony_r; Auto with real.
Apply InegAbsInf; Auto.
Contradict NZai.
Apply Rmult_anticompatibility with (Rinv bi); Auto.
Apply Rinv_RN; Auto.
Ring; Auto with real.
Rewrite (Rabsolu_mult (Rinv bi) ai).
Apply Rle_trans
     with (Rmult
             (Rmult
                (Rmult
                   (Rplus R1 eps3)
                   (Rmult
                      (Rabsolu (Rinv bi)) (Rmult (Rplus R1 eps1) (Rabsolu wi1))))
                (Rabsolu D)) (Rinv (Rabsolu wi1))).
Apply Rle_monotony_r; Auto with real.
Apply Rlt_le; Apply Rlt_Rinv; Apply Rle_lt; Auto with real; Apply sym_not_eqT;
 Apply Rabsolu_no_R0; Auto with real.
Apply Rle_monotony_r; Auto with real.
Apply Rle_monotony; Auto with real.
Apply Rle_trans with (Rplus R1 R0); Auto with real.
Apply Rlt_le; Auto with real.
Replace (Rplus R1 R0) with R1; Auto with real; Ring.
Apply Rle_monotony; Auto with real.
Apply InegAbsInf; Auto.
Replace (Rmult
           (Rmult
              (Rmult
                 (Rplus R1 eps3)
                 (Rmult
                    (Rabsolu (Rinv bi)) (Rmult (Rplus R1 eps1) (Rabsolu wi1))))
              (Rabsolu D)) (Rinv (Rabsolu wi1)))
     with (Rmult
             (Rmult (Rplus R1 eps3) (Rplus R1 eps1))
             (Rmult
                (Rmult (Rabsolu (Rinv bi)) (Rabsolu D))
                (Rmult (Rabsolu wi1) (Rinv (Rabsolu wi1)))));
 [Idtac | Ring; Ring].
Replace (Rmult (Rabsolu wi1) (Rinv (Rabsolu wi1))) with R1;
 [Idtac | Apply sym_eqT; Apply Rinv_r; Apply Rabsolu_no_R0; Auto with real].
Replace (Rmult (Rmult (Rabsolu (Rinv bi)) (Rabsolu D)) R1)
     with (Rmult (Rabsolu (Rinv bi)) (Rabsolu D));
 [Apply Rle_monotony; Auto with real | Ring].
Apply Rmult_Rle_O.
Apply Rle_trans with (Rplus R1 R0); Auto with real.
Replace (Rplus R1 R0) with R1; [Apply Rlt_le; Auto with real | Ring].
Apply Rle_trans with (Rplus R1 R0); Auto with real.
Replace (Rplus R1 R0) with R1; [Apply Rlt_le; Auto with real | Ring].
Cut (Rlt R0 (Rminus R1 eps2)); [Intros | Idtac].
Apply Rle_mult_anticompatibility with (Rminus R1 eps2); Auto with real.
Rewrite (Rinv_r (Rminus R1 eps2)); Auto with real.
Rewrite (Rabsolu_Rinv bi); Auto.
Cut (Rlt R0 (Rabsolu bi)); [Intros | Auto with real].
Apply Rle_mult_anticompatibility with (Rabsolu bi); Auto with real.
Apply Rle_trans
     with (Rmult
             (Rminus R1 eps2)
             (Rmult (Rabsolu D) (Rmult (Rabsolu bi) (Rinv (Rabsolu bi)))));
 [Right; Ring | Rewrite (Rinv_r (Rabsolu bi))]; Auto with real.
Replace (Rmult (Rabsolu D) R1) with (Rabsolu D);
 Replace (Rmult (Rabsolu bi) R1) with (Rabsolu bi); Try Ring.
Apply InegAbsSup; Auto.
Apply Rabsolu_no_R0; Auto.
Apply Rle_lt; Auto with real; Apply sym_not_eqT; Apply Rabsolu_no_R0; Auto.
Apply sym_not_eqT; Apply imp_not_Req; Auto.
Apply Rlt_anti_compatibility with eps2; Auto with real.
Replace (Rplus eps2 R0) with eps2;
 [Replace (Rplus eps2 (Rminus R1 eps2)) with R1; [Auto with real | Ring] | Ring].
Qed.

Definition dsd := [x, y:R](Rle R0 x) /\ (Rle R0 y) \/ (Rle x R0) /\ (Rle y R0).

Theorem dsdAbs:
  (x, y:R) (dsd x y) ->
  (Rabsolu (Rminus x y))==(Rabsolu (Rminus (Rabsolu x) (Rabsolu y))).
Unfold dsd; Intros.
Case H; Intros H1; Elim H1; Intros; Clear H1 H.
Rewrite (Rabsolu_simpl1 x); Auto with real.
Rewrite (Rabsolu_simpl1 y); Auto with real.
Rewrite (Rabsolu_simpl2 x); Auto with real.
Rewrite (Rabsolu_simpl2 y); Auto with real.
Replace (Rminus (Ropp x) (Ropp y)) with (Ropp (Rminus x y));
 [Rewrite (Rabsolu_Ropp (Rminus x y)) | Ring]; Auto with real.
Qed.

Theorem dsdsym: (x, y:R) (dsd x y) ->(dsd y x).
Unfold dsd; Intros; Auto.
Case H; Intros H1; Case H1; Intros; Clear H H1; Auto.
Qed.

Theorem Inegdsd:
  (x, y, eps:R)
  ~ x==R0 ->
  (Rlt eps R1) -> (Rle (Rabsolu (Rmult (Rinv x) (Rminus y x))) eps) ->(dsd x y).
Intros; Unfold dsd.
Cut (Rlt (Rabsolu (Rmult (Rinv x) (Rminus y x))) R1);
 [Intros | Apply Rle_lt_trans with eps; Auto with real].
Case (Rle_or_lt x R0); Intros; Case (Rle_or_lt y R0); Intros; Auto.
3:Left; Split; Apply Rlt_le; Auto with real.
Left; Split; [Idtac | Apply Rlt_le; Auto with real].
Contradict H2.
Apply Rlt_not.
Rewrite (Rabsolu_mult (Rinv x) (Rminus y x)).
Rewrite (Rabsolu_Rinv x); Auto.
Apply Rle_mult_anticompatibility with (Rabsolu x).
Replace (Rmult (Rabsolu x) (Rmult (Rinv (Rabsolu x)) (Rabsolu (Rminus y x))))
     with (Rmult (Rabsolu (Rminus y x)) (Rmult (Rabsolu x) (Rinv (Rabsolu x))));
 Try Ring.
Rewrite (Rinv_r (Rabsolu x)).
Apply Rle_monotony_r; Auto with real.
Apply Rlt_le; Auto with real.
Rewrite (Rabsolu_simpl2 x); Auto.
Rewrite (Rabsolu_simpl1 (Rminus y x)); Auto with real.
Replace (Ropp x) with (Rplus (Ropp x) R0);
 [Replace (Rminus y x) with (Rplus (Ropp x) y); [Auto with real | Ring] | Ring].
Apply Rle_compatibility; Apply Rlt_le; Auto.
Apply Rle_trans with y; [Apply Rlt_le; Auto | Unfold Rminus].
Pattern 1 y; Replace y with (Rplus y (Ropp R0));
 [Apply Rle_compatibility; Apply Rle_Ropp; Auto with real | Ring].
Apply Rabsolu_no_R0; Auto.
Apply Rle_lt; Auto with real; Apply sym_not_eqT; Apply Rabsolu_no_R0; Auto.
Right; Split; Auto.
Contradict H2.
Apply Rlt_not.
Rewrite (Rabsolu_mult (Rinv x) (Rminus y x)).
Rewrite (Rabsolu_Rinv x); Auto.
Apply Rle_mult_anticompatibility with (Rabsolu x).
Replace (Rmult (Rabsolu x) (Rmult (Rinv (Rabsolu x)) (Rabsolu (Rminus y x))))
     with (Rmult (Rabsolu (Rminus y x)) (Rmult (Rabsolu x) (Rinv (Rabsolu x))));
 Try Ring.
Rewrite (Rinv_r (Rabsolu x)).
Apply Rle_monotony_r; Auto with real.
Apply Rlt_le; Auto with real.
Rewrite (Rabsolu_simpl1 x); Auto.
Rewrite (Rabsolu_simpl2 (Rminus y x)); Auto with real.
Replace (Ropp (Rminus y x)) with (Rplus x (Ropp y)); [Idtac | Ring].
Pattern 1 x; Replace x with (Rplus x (Ropp R0));
 [Apply Rle_compatibility; Apply Rle_Ropp; Auto with real | Ring].
Apply Rle_trans with y; Auto with real.
Unfold Rminus; Pattern 2 y; Replace y with (Rplus y (Ropp R0));
 [Apply Rle_compatibility; Apply Rle_Ropp; Auto with real | Ring].
Apply Rlt_le; Auto.
Apply Rlt_le; Auto.
Apply Rabsolu_no_R0; Auto.
Apply Rle_lt; Auto with real; Apply sym_not_eqT; Apply Rabsolu_no_R0; Auto.
Qed.

Theorem dsdtrans: (x, y, z:R) (dsd x y) -> (dsd y z) -> ~ y==R0 ->(dsd x z).
Unfold dsd; Intros.
Case H0; Intros H'; Elim H'; Intros; Clear H0 H'.
Case H; Intros H'; Elim H'; Intros; Clear H H'; Auto.
Absurd (Rle y R0); Auto with real.
Apply Rle_not; Apply Rle_lt; Auto.
Case H; Intros H'; Elim H'; Intros; Clear H H'; Auto.
Absurd (Rle y R0); Auto with real.
Apply Rle_not; Apply Rle_lt; Auto.
Qed.

Theorem dsdinv: (x, y:R) (dsd x y) -> ~ y==R0 ->(dsd x (Rinv y)).
Unfold dsd; Intros.
Case H; Intros H1; Elim H1; Intros; Clear H H1.
Left; Split; Auto.
Apply Rlt_le; Apply Rlt_Rinv; Apply Rle_lt; Auto.
Right; Split; Auto.
Apply Rlt_le; Apply Rlt_Rinv2; Apply Rle_lt; Auto.
Qed.

Theorem dsdmult: (x, y, r:R) (dsd x y) ->(dsd (Rmult r x) (Rmult r y)).
Unfold dsd; Intros.
Case (Rle_or_lt r R0); Intros.
Case H; Intros H1; Elim H1; Intros; Clear H H1.
Right; Split; Apply Ropp_Rle; Rewrite Ropp_O; Rewrite <- Ropp_mul1;
 Apply Rmult_Rle_O; Auto; Rewrite <- Ropp_O; Apply Rle_Ropp; Auto.
Left; Split.
Replace (Rmult r x) with (Rmult (Ropp r) (Ropp x)); [Idtac | Ring].
Apply Rmult_Rle_O; Rewrite <- Ropp_O; Apply Rle_Ropp; Auto.
Replace (Rmult r y) with (Rmult (Ropp r) (Ropp y)); [Idtac | Ring].
Apply Rmult_Rle_O; Rewrite <- Ropp_O; Apply Rle_Ropp; Auto.
Cut (Rle R0 r); [Intros | Apply Rlt_le; Auto].
Case H; Intros H'; Elim H'; Intros; Clear H H'.
Left; Split; Apply Rmult_Rle_O; Auto.
Right; Split; Apply Ropp_Rle; Rewrite Ropp_O.
Replace (Ropp (Rmult r x)) with (Rmult r (Ropp x)); [Idtac | Ring].
Apply Rmult_Rle_O; Auto; Rewrite <- Ropp_O; Apply Rle_Ropp; Auto.
Replace (Ropp (Rmult r y)) with (Rmult r (Ropp y)); [Idtac | Ring].
Apply Rmult_Rle_O; Auto; Rewrite <- Ropp_O; Apply Rle_Ropp; Auto.
Qed.

Theorem dsdwi1qiD: (dsd wi1 (Rmult qi D)).
Case (Req_EM qi R0); Intros.
Rewrite H; Unfold dsd.
Replace (Rmult R0 D) with R0; [Auto with real | Ring].
Case (Rle_or_lt wi1 R0); Intros; Auto with real.
Left; Split; Auto with real; Apply Rlt_le; Auto.
Cut (dsd wi1 ai); [Intros | Apply Inegdsd with eps1; Auto].
Cut (dsd D bi); [Intros | Apply Inegdsd with eps2; Auto].
Cut (dsd (Rmult (Rinv bi) ai) qi); [Intros | Apply Inegdsd with eps3; Auto].
Apply dsdtrans with ai; Auto.
Apply dsdtrans with (Rmult qi bi).
Replace ai with (Rmult bi (Rmult (Rinv bi) ai)).
Rewrite (Rmult_sym qi bi); Apply dsdmult; Auto.
Replace (Rmult bi (Rmult (Rinv bi) ai)) with (Rmult ai (Rmult bi (Rinv bi)));
 [Rewrite (Rinv_r bi); Auto; Ring | Ring].
Apply dsdmult; Apply dsdsym; Auto.
Contradict H.
Apply r_Rmult_mult with bi; Auto.
Ring; Ring; Auto with real.
Contradict NZai.
Apply r_Rmult_mult with (Rinv bi); Auto.
Ring; Auto with real.
Apply Rinv_RN; Auto.
Qed.

Theorem Maxwiwi1:
  (Rle
     (Rmult (Rabsolu wi) (Rinv (Rabsolu wi1)))
     (Rmax
        (Rminus
           R1
           (Rmult
              (Rmult (Rminus R1 eps3) (Rminus R1 eps1)) (Rinv (Rplus R1 eps2))))
        (Rminus
           (Rmult
              (Rmult (Rplus R1 eps3) (Rplus R1 eps1)) (Rinv (Rminus R1 eps2)))
           R1))).
Rewrite Hw.
Rewrite (dsdAbs wi1 (Rmult qi D)); [Idtac | Apply dsdwi1qiD].
Case (Rle_or_lt (Rminus (Rabsolu wi1) (Rabsolu (Rmult qi D))) R0); Intros.
Rewrite Rabsolu_simpl2; Auto.
Replace (Ropp (Rminus (Rabsolu wi1) (Rabsolu (Rmult qi D))))
     with (Rminus (Rabsolu (Rmult qi D)) (Rabsolu wi1)); [Idtac | Ring].
Apply Rle_trans
     with (Rminus
             (Rmult
                (Rmult (Rplus R1 eps3) (Rplus R1 eps1)) (Rinv (Rminus R1 eps2)))
             R1); [Apply DivSecondCase; Auto | Auto with real].
Apply RmaxLess2.
Rewrite Rabsolu_simpl1; [Idtac | Apply Rlt_le; Auto].
Apply Rle_trans
     with (Rminus
             R1
             (Rmult
                (Rmult (Rminus R1 eps3) (Rminus R1 eps1)) (Rinv (Rplus R1 eps2))));
 [Apply DivFirstCase; Auto | Auto with real].
Apply RmaxLess1.
Qed.

Axiom confiance:(P:Prop)P.

Theorem Rmax_simpl1: (p, q:R) (Rle p q) ->(Rmax p q)==q.
Intros p q H; Unfold Rmax.
Case (total_order_Rle p q); Auto.
Intros H1; Apply Rle_antisym; Auto with real.
Case (Rle_or_lt q p); Auto; Intros H2.
Case H1; Auto with real.
Qed.

Theorem RmaxProp: (P:R ->Prop) (x, y:R) (P x) -> (P y) ->(P (Rmax x y)).
Intros.
Unfold Rmax; Case (total_order_Rle x y); Simpl; Auto.
Qed.

Theorem ep_aux:
  ((Rle R0 ep) /\ (Rlt ep R1)) /\
  ((Rle eps1 ep) /\ ((Rle eps2 ep) /\ (Rle eps3 ep))).
Rewrite Hep.
Split.
Split; Apply RmaxProp; Auto; Apply RmaxProp; Auto.
Repeat Split.
Apply Rle_trans with (Rmax eps1 eps2); Apply RmaxLess1.
Apply Rle_trans with (Rmax eps1 eps2); [Apply RmaxLess2 | Apply RmaxLess1].
Apply RmaxLess2.
Qed.

Theorem ConvDiv:
  (Rle
     (Rmult (Rabsolu wi) (Rinv (Rabsolu wi1)))
     (Rmult
        (Rplus ep (Rplus ep (Rplus ep (Rmult ep ep)))) (Rinv (Rminus R1 ep)))).
Generalize ep_aux; Intros H; Elim H; Intros H1 H2; Elim H1; Intros; Elim H2;
 Intros H4 H5; Elim H5; Intros; Clear H H1 H2 H5.
Cut (Rlt R0 (Rminus R1 ep)); [Intros | Idtac].
Apply Rle_trans
     with (Rmax
             (Rminus
                R1
                (Rmult
                   (Rmult (Rminus R1 eps3) (Rminus R1 eps1))
                   (Rinv (Rplus R1 eps2))))
             (Rminus
                (Rmult
                   (Rmult (Rplus R1 eps3) (Rplus R1 eps1))
                   (Rinv (Rminus R1 eps2))) R1)); [Apply Maxwiwi1 | Idtac].
Apply RmaxProp.
Cut (Rlt R0 (Rplus R1 eps2)); [Intros | Idtac].
Pattern 1 R1; Replace R1 with (Rmult (Rplus R1 eps2) (Rinv (Rplus R1 eps2)));
 [Idtac | Apply Rinv_r; Apply imp_not_Req; Auto with real].
Apply Rle_trans
     with (Rmult
             (Rminus (Rplus R1 eps2) (Rmult (Rminus R1 eps3) (Rminus R1 eps1)))
             (Rinv (Rplus R1 eps2))); [Right; Ring; Ring | Idtac].
Replace (Rmult (Rminus R1 eps3) (Rminus R1 eps1))
     with (Rplus R1 (Rplus (Ropp eps3) (Rplus (Ropp eps1) (Rmult eps3 eps1))));
 [Idtac | Ring; Ring].
Replace (Rminus
           (Rplus R1 eps2)
           (Rplus R1 (Rplus (Ropp eps3) (Rplus (Ropp eps1) (Rmult eps3 eps1)))))
     with (Rplus eps2 (Rplus eps3 (Rplus eps1 (Ropp (Rmult eps3 eps1)))));
 [Idtac | Ring; Ring].
Apply Rle_trans with (Rmult (Rplus ep (Rplus ep ep)) (Rinv (Rplus R1 eps2))).
Apply Rle_monotony_r.
Apply Rlt_le; Apply Rlt_Rinv; Auto.
Apply Rplus_le; Auto.
Apply Rplus_le; Auto.
Apply Rle_trans with (Rplus ep (Ropp (Rmult eps3 eps1)));
 [Apply Rplus_le; Auto with real | Idtac].
Pattern 2 ep; Replace ep with (Rplus ep (Ropp R0));
 [Apply Rplus_le; Auto with real; Apply Rle_Ropp; Auto with real | Ring].
Apply Rmult_Rle_O; Auto.
Apply Rle_trans with (Rmult (Rplus ep (Rplus ep ep)) (Rinv (Rminus R1 ep)));
 [Apply Rle_monotony | Idtac].
Replace R0 with (Rplus R0 (Rplus R0 R0));
 [Apply Rplus_le; Auto; Apply Rplus_le; Auto | Ring].
Apply Rle_Rinv; Auto.
Unfold Rminus; Apply Rle_compatibility; Apply Rle_trans with R0; Auto.
Rewrite <- Ropp_O; Apply Rle_Ropp; Auto.
Apply Rle_monotony_r;
 [Apply Rlt_le; Apply Rlt_Rinv; Auto |
   Apply Rle_compatibility; Apply Rle_compatibility].
Pattern 1 ep; Replace ep with (Rplus ep R0);
 [Apply Rle_compatibility; Apply Rmult_Rle_O; Auto | Ring].
Apply Rlt_le_trans with R1; [Auto with real | Idtac].
Pattern 1 R1; Replace R1 with (Rplus R1 R0);
 [Apply Rle_compatibility; Auto | Ring].
Pattern 4 R1; Replace R1 with (Rmult (Rminus R1 eps2) (Rinv (Rminus R1 eps2)));
 [Idtac | Apply Rinv_r; Auto with real].
Apply Rle_trans
     with (Rmult
             (Rplus eps3 (Rplus eps1 (Rplus eps2 (Rmult eps3 eps1))))
             (Rinv (Rminus R1 eps2))); [Right; Ring; Ring | Idtac].
Apply Rle_trans
     with (Rmult
             (Rplus ep (Rplus ep (Rplus ep (Rmult ep ep))))
             (Rinv (Rminus R1 eps2)));
 [Apply Rle_monotony_r | Apply Rle_monotony].
Apply Rlt_le; Apply Rlt_Rinv.
Apply Rlt_anti_compatibility with eps2.
Replace (Rplus eps2 R0) with eps2;
 [Replace (Rplus eps2 (Rminus R1 eps2)) with R1; [Idtac | Ring] | Ring]; Auto.
Repeat (Apply Rplus_le; Auto).
Apply Rle_Rmult_comp; Auto.
Repeat (Replace R0 with (Rplus R0 R0); [Apply Rplus_le; Auto | Ring]).
Apply Rmult_Rle_O; Auto.
Apply Rle_Rinv;
 [Auto | Unfold Rminus; Apply Rle_compatibility; Apply Rle_Ropp; Auto].
Apply sym_not_eqT; Apply imp_not_Req; Left.
Apply Rlt_anti_compatibility with eps2.
Replace (Rplus eps2 R0) with eps2;
 [Replace (Rplus eps2 (Rminus R1 eps2)) with R1; [Idtac | Ring] | Ring]; Auto.
Apply Rlt_anti_compatibility with ep.
Replace (Rplus ep R0) with ep;
 [Replace (Rplus ep (Rminus R1 ep)) with R1; [Idtac | Ring] | Ring]; Auto.
Qed.
End Div.
