Require Export Fexp.
Section mf.
Variable radix:nat.
Hypothesis radixMoreThanOne:(lt (1) radix).

Local FtoRradix := (FtoR radix).
Coercion FtoRradix : float >-> R.
Require Import PolyList.
Require Import ClosestPlus.
Variable b:Fbound.
Variable precision:nat.
Hypothesis precisionGreaterThanOne:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp radix precision) (1)).
Variable TwoSum:float -> float ->float * float.
Hypothesis TwoSum1:
           (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
           (Closest b radix (Rplus p q) (Fst (TwoSum p q))).
Hypothesis TwoSum2:
           (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
           <R> (Snd (TwoSum p q))==(Rminus (Rplus p q) (Fst (TwoSum p q))).
Hypothesis TwoSum3:
           (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
           (Fbounded b (Fst (TwoSum p q))).
Hypothesis TwoSum4:
           (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
           (Fbounded b (Snd (TwoSum p q))).
Hypothesis TwoSumEq1:
           (p, q, r, s:float)
           (Fbounded b p) ->
           (Fbounded b q) ->
           (Fbounded b r) -> (Fbounded b s) -> <R> p==q -> <R> r==s ->
           <R> (Fst (TwoSum p r))==(Fst (TwoSum q s)).
Hypothesis TwoSumEq2:
           (p, q, r, s:float)
           (Fbounded b p) ->
           (Fbounded b q) ->
           (Fbounded b r) -> (Fbounded b s) -> <R> p==q -> <R> r==s ->
           <R> (Snd (TwoSum p r))==(Snd (TwoSum q s)).

Fixpoint growExp[p:float; L:(list float)]: (list float) :=
   Cases L of
       nil => (cons p (nil ?))
      | (cons x L1) => Cases (TwoSum p x) of
                           (pair h c) => (cons c (growExp h L1))
                       end
   end.

Theorem TwoSumExp:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
  (IsExpansion
     b radix (cons (Snd (TwoSum p q)) (cons (Fst (TwoSum p q)) (nil ?)))).
Intros p q H' H'0.
Case (Z_zerop (Fnum (Snd (TwoSum p q)))); Intros Z1.
Apply IsExpansionTop1; Auto.
Apply IsExpansionSingle; Auto.
Case (Z_zerop (Fnum (Fst (TwoSum p q)))); Intros Z2.
Apply IsExpansionTop2; Auto.
Apply IsExpansionSingle; Auto.
Apply IsExpansionTop; Auto.
Cut <R> (Snd (TwoSum p q))==(Fminus radix (Fplus radix p q) (Fst (TwoSum p q)));
 [Intros Eq1 | Idtac].
Rewrite (MSB_comp radix) with 4 := Eq1; Auto.
Apply (MSBroundLSB b radix precision) with P := (Closest b radix); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply (ClosestCompatible b radix (Rplus p q)) with p0 := (Fst (TwoSum p q));
 Auto.
Apply sym_eqT; Apply (Fplus_correct radix); Auto with arith.
Apply NisFzeroComp with radix := radix x := (Snd (TwoSum p q)); Auto.
Rewrite (Fminus_correct radix); Auto with arith; Rewrite (Fplus_correct radix);
 Auto with arith.
Apply IsExpansionSingle; Auto.
Qed.

Theorem TwoSumOl1:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) -> (is_Fzero q) ->
  <R> (Fst (TwoSum p q))==p.
Intros p q H' H'0 H'1; Generalize (TwoSum1 p q H' H'0); Case (TwoSum p q);
 Simpl; Auto.
Intros f H'2 H'3.
(Apply sym_eqT;
  Apply (RoundedModeProjectorIdemEq b radix precision)
       with P := (Closest b radix)); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply (ClosestCompatible b radix) with 1 := H'3; Auto.
Replace (FtoRradix q) with R0; [Unfold FtoRradix; Ring | Idtac].
Apply sym_eqT; Apply (is_Fzero_rep1 radix); Auto.
Apply RoundedModeBounded
     with radix := radix P := (Closest b radix) r := (Rplus p q); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Qed.

Theorem TwoSumOl2:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) -> (is_Fzero q) ->
  <R> (Snd (TwoSum p q))==R0.
Intros p q H' H'0 H'1; Generalize (TwoSumOl1 p q H' H'0 H'1);
 Generalize (TwoSum2 p q H' H'0); Case (TwoSum p q); Simpl; Auto.
Intros f f0 H'2 H'3.
Rewrite H'2.
Rewrite H'3.
Replace (FtoRradix q) with R0; [Unfold FtoRradix; Ring | Idtac].
Apply sym_eqT; Apply (is_Fzero_rep1 radix); Auto.
Qed.

Theorem TwoSumOr1:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) -> (is_Fzero p) ->
  <R> (Fst (TwoSum p q))==q.
Intros p q H' H'0 H'1; Generalize (TwoSum1 p q H' H'0); Case (TwoSum p q);
 Simpl; Auto.
Intros f H'2 H'3.
(Apply sym_eqT;
  Apply (RoundedModeProjectorIdemEq b radix precision)
       with P := (Closest b radix)); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply (ClosestCompatible b radix) with 1 := H'3; Auto.
Replace (FtoRradix p) with R0; [Unfold FtoRradix; Ring | Idtac].
Apply sym_eqT; Apply (is_Fzero_rep1 radix); Auto.
Apply RoundedModeBounded
     with radix := radix P := (Closest b radix) r := (Rplus p q); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Qed.

Theorem TwoSumOr2:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) -> (is_Fzero p) ->
  <R> (Snd (TwoSum p q))==R0.
Intros p q H' H'0 H'1; Generalize (TwoSumOr1 p q H' H'0 H'1);
 Generalize (TwoSum2 p q H' H'0); Case (TwoSum p q); Simpl; Auto.
Intros f f0 H'2 H'3.
Rewrite H'2.
Rewrite H'3.
Replace (FtoRradix p) with R0; [Unfold FtoRradix; Ring | Idtac].
Apply sym_eqT; Apply (is_Fzero_rep1 radix); Auto.
Qed.

Theorem growExpIsVal:
  (L:(list float)) (IsExpansion b radix L) ->
  (p:float) (Fbounded b p) ->
  <R> (expValue radix (growExp p L))==(Rplus p (expValue radix L)).
Intros L; Elim L.
Simpl; Auto.
Intros H' p H'0; Rewrite (Fplus_correct radix); Auto with arith.
Intros a l H' H'0 p H'1; Simpl.
CaseEq (TwoSum p a).
Intros f f0 H'2; Simpl.
Repeat Rewrite (Fplus_correct radix); Auto with arith.
Rewrite H'.
Repeat Rewrite <- Rplus_assoc.
Cut (Fbounded b a); [Intros Fb1 | Apply (expBoundedInv b radix) with L := l];
 Auto.
Generalize (TwoSum2 ? ? H'1 Fb1); Rewrite H'2; Simpl.
Unfold FtoRradix; Intros H'3; Rewrite H'3; Ring; Ring.
Apply expInv with a := a; Auto.
Cut (Fbounded b a); [Intros Fb1 | Apply (expBoundedInv b radix) with L := l];
 Auto.
Generalize (TwoSum3 ? ? H'1 Fb1); Rewrite H'2; Simpl; Auto.
Qed.

Theorem IsExpansionCons:
  (L:(list float)) (IsExpansion b radix L) ->
  (p:float)
  ~ (is_Fzero p) ->
  (Fbounded b p) ->
  ((q:float) (In q L) -> ~ (is_Fzero q) ->(Zlt (MSB radix p) (LSB radix q))) ->
  (IsExpansion b radix (cons p L)).
Intros L; Elim L.
Intros H' p H'0 H'1 H'2; Apply IsExpansionSingle; Auto.
Intros a l H' H'0 p H'1 H'2 H'3.
Case (is_FzeroP a); Intros Z1.
Apply IsExpansionTop2; Auto.
Apply expBoundedInv with radix := radix L := l; Auto.
Apply H'; Auto.
Apply expInv with a := a; Auto.
Intros q H'4 H'5; Apply H'3; Auto.
Simpl; Auto.
Apply IsExpansionTop; Auto.
Apply expBoundedInv with radix := radix L := l; Auto.
Apply H'3; Auto with datatypes.
Qed.

Theorem IsExpansionConsInvAux:
  (L:(list float)) (IsExpansion b radix L) ->
  (L':(list float))
  (p, q:float) ~ (is_Fzero p) -> L=(cons p L') -> (In q L') -> ~ (is_Fzero q) ->
  (Zlt (MSB radix p) (LSB radix q)).
Intros L H'; Elim H'; Auto.
Intros L' p q H'0 H'1; Discriminate.
Intros x H'0 L' p q H'1 H'2; Inversion H'2; Simpl.
Intros H'3; Elim H'3.
Intros x L0 H'0 H'1 H'2 H'3 L' p q H'4 H'5; Inversion H'5; Auto.
Case H'4; Rewrite <- H0; Auto.
Intros x y L0 H'0 H'1 H'2 H'3 H'4 L' p q H'5 H'6; Inversion H'6.
Simpl; Auto.
Intros H'7 H'8; Case H'7; Intros H'9.
Case H'8; Rewrite <- H'9; Auto.
Apply H'4 with L' := L0; Auto.
Rewrite H0; Auto.
Intros x y L0 H'0 H'1 H'2 H'3 H'4 H'5 H'6 L' p q H'7 H'8; Inversion H'8; Auto.
Simpl; Auto.
Intros H'9 H'10; Case H'9; Intros H'11.
Rewrite <- H'11; Rewrite <- H0; Auto.
Apply Zlt_trans with (LSB radix y); Auto.
Rewrite <- H0; Auto.
Apply Zle_lt_trans with (MSB radix y); Auto.
Apply LSB_le_MSB; Auto.
Apply H'6 with L' := L0; Auto.
Qed.

Theorem IsExpansionConsInv:
  (L:(list float))
  (p, q:float)
  ~ (is_Fzero p) ->
  (IsExpansion b radix (cons p L)) -> (In q L) -> ~ (is_Fzero q) ->
  (Zlt (MSB radix p) (LSB radix q)).
Intros L p q H' H'0 H'1 H'2.
Apply IsExpansionConsInvAux with L := (cons p L) L' := L; Auto.
Qed.

Theorem IsExpansionSkip:
  (L:(list float)) (p, q:float) (IsExpansion b radix (cons p (cons q L))) ->
  (IsExpansion b radix (cons p L)).
Intros L p q H'.
Case (is_FzeroP p); Intros Z1.
Apply IsExpansionTop1; Auto.
Apply expBoundedInv with radix := radix L := (cons q L); Auto.
Apply expInv with a := q; Auto.
Apply expInv with a := p; Auto.
Apply IsExpansionCons; Auto.
Apply expInv with a := q; Auto.
Apply expInv with a := p; Auto.
Apply expBoundedInv with radix := radix L := (cons q L); Auto.
Intros q0 H'0 H'1.
Apply IsExpansionConsInv with L := (cons q L); Auto with datatypes.
Qed.

Theorem TwoSumLt1:
  (p, q:float)
  (Fbounded b p) ->
  (Fbounded b q) ->
  ~ (is_Fzero p) -> ~ (is_Fzero q) -> ~ (is_Fzero (Fst (TwoSum p q))) ->
  (Zle (Zmin (LSB radix p) (LSB radix q)) (LSB radix (Fst (TwoSum p q)))).
Intros p q H'0 H'1 H'2 H'3 H'4.
Apply Zle_trans with (LSB radix (Fplus radix p q)); Auto.
Case (LSB_rep_min radix radixMoreThanOne p); Auto; Intros p' Hp'.
Case (LSB_rep_min radix radixMoreThanOne q); Auto; Intros q' Hp'.
Apply LSBPlus; Auto.
Contradict H'4; Auto.
Apply (is_Fzero_rep2 radix); Auto.
Rewrite <- (FzeroisZero radix b); Auto.
Apply sym_eqT;
 Apply RoundedModeProjectorIdemEq
      with b := b precision := precision P := (Closest b radix); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply FboundedFzero; Auto.
Apply (ClosestCompatible b radix) with 1 := (TwoSum1 ? ? H'0 H'1); Auto.
Rewrite (FzeroisZero radix b); Auto.
Replace R0 with (FtoRradix (Fplus radix p q)); Auto.
Rewrite (Fplus_correct radix); Auto with arith.
Apply (is_Fzero_rep1 radix); Auto.
Repeat Rewrite (Fplus_correct radix); Auto with arith.
Apply RoundLSBMax with b := b precision := precision P := (Closest b radix);
 Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply (ClosestCompatible b radix) with 1 := (TwoSum1 ? ? H'0 H'1); Auto.
Rewrite Fplus_correct; Auto with arith.
Qed.

Theorem TwoSumLt2:
  (p, q:float)
  (Fbounded b p) ->
  (Fbounded b q) ->
  ~ (is_Fzero p) -> ~ (is_Fzero q) -> ~ (is_Fzero (Snd (TwoSum p q))) ->
  (Zle (Zmin (LSB radix p) (LSB radix q)) (LSB radix (Snd (TwoSum p q)))).
Intros p q H' H'0 H'1 H'2 H'3.
Case (is_FzeroP (Fst (TwoSum p q))); Intros Z1.
Cut <R> (Snd (TwoSum p q))==(Fplus radix p q); [Intros Eq1 | Idtac].
Replace (LSB radix (Snd (TwoSum p q))) with (LSB radix (Fplus radix p q)); Auto.
Apply LSBPlus; Auto.
Apply (NisFzeroComp radix) with 3 := Eq1; Auto.
Apply LSB_comp; Auto.
Apply (NisFzeroComp radix) with 3 := Eq1; Auto.
Rewrite TwoSum2; Auto.
Unfold FtoRradix; Rewrite (is_Fzero_rep1 radix) with 1 := Z1; Auto.
Rewrite Fplus_correct; Auto with arith; Ring.
Cut <R> (Snd (TwoSum p q))==(Fminus radix (Fplus radix p q) (Fst (TwoSum p q)));
 [Intros Eq1 | Idtac].
Replace (LSB radix (Snd (TwoSum p q)))
     with (LSB radix (Fminus radix (Fplus radix p q) (Fst (TwoSum p q)))).
Apply Zle_trans
     with (Zmin (LSB radix (Fplus radix p q)) (LSB radix (Fst (TwoSum p q))));
 Auto.
Apply Zmin_Zle; Auto.
Apply LSBPlus; Auto.
Contradict Z1; Auto.
Apply (is_Fzero_rep2 radix); Auto.
Rewrite <- (FzeroisZero radix b); Auto.
Apply sym_eqT;
 Apply RoundedModeProjectorIdemEq
      with b := b precision := precision P := (Closest b radix); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply FboundedFzero; Auto.
Apply (ClosestCompatible b radix) with 1 := (TwoSum1 ? ? H' H'0); Auto.
Rewrite (FzeroisZero radix b); Auto.
Replace R0 with (FtoRradix (Fplus radix p q)); Auto.
Rewrite (Fplus_correct radix); Auto with arith.
Apply (is_Fzero_rep1 radix); Auto.
Apply TwoSumLt1; Auto.
Apply LSBMinus; Auto.
Apply (NisFzeroComp radix) with 3 := Eq1; Auto.
Apply LSB_comp; Auto.
Apply (NisFzeroComp radix) with 3 := Eq1; Auto.
Rewrite TwoSum2; Auto.
Rewrite (Fminus_correct radix); Auto with arith; Rewrite (Fplus_correct radix);
 Auto with arith.
Qed.

Theorem IsExpansionGrowConsInvAux:
  (L:(list float)) (IsExpansion b radix L) ->
  (L':(list float))
  (p, q, r:float)
  (Fbounded b r) ->
  ~ (is_Fzero p) -> L=(cons p L') -> (In q (growExp r L')) -> ~ (is_Fzero q) ->
  ((is_Fzero r) ->(Zlt (MSB radix p) (LSB radix q))) /\
  (~ (is_Fzero r) ->
   (Zlt (MSB radix p) (LSB radix q)) \/ (Zle (LSB radix r) (LSB radix q))).
Intros L H'; Elim H'; Simpl; Auto.
Intros L' p q r H'0 H'1 H'2; Discriminate.
Intros x H'0 L' p q r H'1 H'2 H'3; Inversion H'3; Simpl; Auto.
Intros H'4 H'5; Split; Case H'4.
Intros H'6 H'7; Case H'5; Rewrite <- H'6; Auto.
Intros H'6; Elim H'6; Auto.
Intros H'6; Rewrite <- H'6; Auto with zarith.
Intros H'6; Elim H'6.
Intros x L0 H'0 H'1 H'2 H'3 L' p q r H'4 H'5 H'6; Inversion H'6; Auto.
Case H'5; Rewrite <- H0; Auto.
Intros x y L0 H'0 H'1 H'2 H'3 H'4 L' p q r H'5 H'6 H'7; Inversion H'7.
Simpl.
CaseEq (TwoSum r y); Simpl; Auto.
Intros f f0 H'8 H'9; Elim H'9; Intros H'10.
Intros H'11; Case H'11; Rewrite <- H'10.
Apply (is_Fzero_rep2 radix); Auto.
Replace (FtoR radix f0) with (FtoRradix (Snd (TwoSum r y))); Auto.
Apply TwoSumOl2; Auto.
Rewrite H'8; Auto.
Split.
Intros H'11.
Case (H'4 L0 p q f); Auto.
Replace f with (Fst (TwoSum r y)); Auto.
Rewrite H'8; Auto.
Rewrite <- H0; Auto.
Intros H'12 H'13; Apply H'12; Auto.
Replace f with (Fst (TwoSum r y)); Auto.
Apply (is_Fzero_rep2 radix); Auto.
Replace R0 with (FtoRradix r).
Apply TwoSumOl1; Auto.
Apply (is_Fzero_rep1 radix); Auto.
Rewrite H'8; Auto.
Intros H'11.
Case (H'4 L0 p q f); Auto.
Replace f with (Fst (TwoSum r y)); Auto.
Rewrite H'8; Auto.
Rewrite <- H0; Auto.
Intros H'13 H'14.
Case (is_FzeroP f); Intros Z1; Auto.
Replace (LSB radix r) with (LSB radix f); Auto.
Apply LSB_comp; Auto.
Replace f with (Fst (TwoSum r y)); Auto.
Apply TwoSumOl1; Auto.
Rewrite H'8; Auto.
Intros x y L0 H'0 H'1 H'2 H'3 H'4 H'5 H'6 L' p q r H'7 H'8 H'9; Inversion H'9.
Simpl; Auto.
CaseEq (TwoSum r y); Simpl; Auto.
Intros f f0 H'10 H'11; Case H'11; Intros H'12.
Rewrite <- H'12; Auto.
Intros H'13; Split; Auto.
Intros H'14; Case H'13.
Replace f0 with (Snd (TwoSum r y)); Auto.
Apply (is_Fzero_rep2 radix); Auto.
Apply TwoSumOr2; Auto.
Rewrite H'10; Auto.
Intros H'14.
Cut (Zle
       (Zmin (LSB radix r) (LSB radix y))
       (LSB radix <float, float> Snd ((TwoSum r y)))).
Rewrite H'10; Simpl; Auto.
Unfold Zmin; Case (Zcompare (LSB radix r) (LSB radix y)); Auto.
Intros H'15; Left; Apply Zlt_le_trans with (LSB radix y); Auto.
Rewrite <- H0; Auto.
Apply TwoSumLt2; Auto.
Rewrite H'10; Auto.
Intros H'13.
Case (H'6 L0 y q f); Auto.
Replace f with (Fst (TwoSum r y)); Auto.
Rewrite H'10; Auto.
Intros H'14 H'15; Split.
Intros H'16.
Case H'15; Auto.
Apply (NisFzeroComp radix) with 2 := H'1; Auto.
Replace f with (Fst (TwoSum r y)); Auto.
Apply sym_eqT; Apply TwoSumOr1; Auto.
Rewrite H'10; Auto.
Intros H'17; Apply Zlt_trans with (LSB radix y).
Rewrite <- H0; Auto.
Apply Zle_lt_trans with (MSB radix y); Auto.
Apply LSB_le_MSB; Auto.
Intros H'17; Apply Zlt_le_trans with (LSB radix y).
Rewrite <- H0; Auto.
Replace (LSB radix y) with (LSB radix f); Auto.
Apply sym_equal; Apply LSB_comp; Auto.
Replace f with (Fst (TwoSum r y)); Auto.
Apply sym_eqT; Apply TwoSumOr1; Auto.
Rewrite H'10; Auto.
Intros H'16; Case (is_FzeroP f); Intros Z1.
Left; Apply Zlt_trans with (LSB radix y); Auto.
Rewrite <- H0; Auto.
Apply Zle_lt_trans with (MSB radix y); Auto.
Apply LSB_le_MSB; Auto.
Case H'15; Auto; Intros H'17.
Left; Apply Zlt_trans with (LSB radix y); Auto.
Rewrite <- H0; Auto.
Apply Zle_lt_trans with (MSB radix y); Auto.
Apply LSB_le_MSB; Auto.
Cut (Zle (Zmin (LSB radix r) (LSB radix y)) (LSB radix (Fst (TwoSum r y)))).
Rewrite H'10; Simpl; Auto.
Unfold Zmin; Case (Zcompare (LSB radix r) (LSB radix y)); Auto.
Intros H'18; Right; Apply Zle_trans with 1 := H'18; Auto.
Intros H'18; Right; Apply Zle_trans with 1 := H'18; Auto.
Intros H'18; Left; Apply Zlt_le_trans with (LSB radix y).
Rewrite <- H0; Auto.
Apply Zle_trans with 1 := H'18; Auto.
Apply TwoSumLt1; Auto.
Rewrite H'10; Auto.
Qed.

Theorem growExpIsExp:
  (L:(list float)) (IsExpansion b radix L) ->
  (p:float) (Fbounded b p) ->(IsExpansion b radix (growExp p L)).
Intros L; Elim L; Simpl; Auto.
Intros H' p H'0.
Apply IsExpansionSingle; Auto.
Intros a l H' H'0 p H'1; CaseEq (TwoSum p a); Auto.
Intros f f0 H'2.
Cut (Fbounded b a); [Intros Fba | Apply (expBoundedInv b radix) with L := l];
 Auto.
Case (is_FzeroP f0); Intros Z1.
Apply IsExpansionTop1; Auto.
Replace f0 with (Snd (TwoSum p a)); Auto.
Rewrite H'2; Auto.
Apply H'; Auto.
Apply expInv with a := a; Auto.
Replace f with (Fst (TwoSum p a)); Auto.
Rewrite H'2; Auto.
Apply IsExpansionCons; Auto.
Apply H'; Auto.
Apply expInv with a := a; Auto.
Replace f with (Fst (TwoSum p a)); Auto.
Rewrite H'2; Auto.
Replace f0 with (Snd (TwoSum p a)); Auto.
Rewrite H'2; Auto.
Intros q H'3 H'4.
Cut (IsExpansion b radix (cons f0 l)); [Intros IsE1 | Idtac].
Case (IsExpansionGrowConsInvAux (cons f0 l) IsE1 l f0 q f); Auto.
Replace f with (Fst (TwoSum p a)); Auto.
Rewrite H'2; Auto.
Intros H'5 H'6.
Case (is_FzeroP f); Intros Z2; Auto.
Case H'6; Auto.
Intros H'7; Apply Zlt_le_trans with 2 := H'7.
Generalize (TwoSumExp p a H'1 Fba); Rewrite H'2; Simpl; Intros IsE2;
 Inversion IsE2; Auto.
Case Z1; Auto.
Case Z2; Auto.
Apply IsExpansionCons; Auto.
Apply expInv with a := a; Auto.
Replace f0 with (Snd (TwoSum p a)); Auto.
Rewrite H'2; Auto.
Intros q0 H'5 H'6.
Apply Zle_lt_trans with (MSB radix a).
Apply MSB_monotone; Auto.
Contradict Z1.
Apply (is_Fzero_rep2 radix); Auto.
Replace f0 with (Snd (TwoSum p a)); Auto.
Apply TwoSumOl2; Auto.
Rewrite H'2; Auto.
Repeat Rewrite Fabs_correct; Auto with arith.
Replace f0 with (Snd (TwoSum p a)); Auto.
Rewrite TwoSum2; Auto.
Replace (FtoR radix a) with (Rminus (Rplus p a) p);
 [Idtac | Unfold FtoRradix; Ring].
Cut (x, y:R)(Rabsolu (Rminus x y))==(Rabsolu (Rminus y x));
 [Intros Re1; Repeat Rewrite (Re1 (Rplus p a)) |
   Intros x y; Rewrite <- (Rminus_Ropp x); Rewrite Rabsolu_Ropp; Auto].
Case (TwoSum1 ? ? H'1 Fba); Auto.
Rewrite H'2; Auto.
Apply IsExpansionConsInv with L := l; Auto.
Contradict Z1.
Apply (is_Fzero_rep2 radix); Auto.
Replace f0 with (Snd (TwoSum p a)); Auto.
Apply TwoSumOl2; Auto.
Rewrite H'2; Auto.
Qed.

Fixpoint addExp[L1:(list float)]: (list float) ->(list float) :=
   [L2:(list float)]
      Cases L1 of
          nil => L2
         | (cons x L'1) => Cases (growExp x L2) of
                               nil => L'1
                              | (cons y L'2) => (cons y (addExp L'1 L'2))
                           end
      end.

Theorem addExpIsVal:
  (L1, L2:(list float)) (IsExpansion b radix L1) -> (IsExpansion b radix L2) ->
  <R>
    (expValue radix (addExp L1 L2))==
    (Rplus (expValue radix L1) (expValue radix L2)).
Intros L1; Elim L1; Simpl; Auto.
Intros L2 HL1 HL2; Replace (FtoRradix (Fzero ZERO)) with R0;
 [Ring | Apply sym_eqT; Apply (is_Fzero_rep1 radix); Red; Simpl]; Auto.
Intros a l Rec L2 HL1 HL2.
Cut (Fbounded b a); [Intros Fba | Apply (expBoundedInv b radix) with L := l];
 Auto.
Generalize (growExpIsVal L2 HL2 a Fba); Generalize (growExpIsExp L2 HL2 a Fba);
 Case (growExp a L2); Simpl; Auto.
Intros H' H'0; Rewrite (Fplus_correct radix); Auto with zarith.
Replace (Rplus
           (Rplus (FtoR radix a) (FtoR radix (expValue radix l)))
           (expValue radix L2))
     with (Rplus (Rplus a (expValue radix L2)) (expValue radix l));
 [Idtac | Unfold FtoRradix; Ring].
Rewrite <- H'0; Auto.
Replace (FtoRradix (Fzero ZERO)) with R0;
 [Ring | Apply sym_eqT; Apply (is_Fzero_rep1 radix); Red; Simpl]; Auto.
Intros f l0 H' H'0; Repeat Rewrite (Fplus_correct radix); Auto with zarith.
Rewrite Rec; Auto.
Replace (Rplus
           (Rplus (FtoR radix a) (FtoR radix (expValue radix l)))
           (expValue radix L2))
     with (Rplus (Rplus a (expValue radix L2)) (expValue radix l));
 [Idtac | Unfold FtoRradix; Ring].
Rewrite <- H'0; Repeat Rewrite (Fplus_correct radix); Auto with zarith.
Unfold FtoRradix; Ring.
Apply expInv with a := a; Auto.
Apply expInv with a := f; Auto.
Qed.

Theorem IsExpansionAddInv:
  (L1, L2:(list float))
  (p, q:float)
  ~ (is_Fzero p) ->
  ~ (is_Fzero q) ->
  (IsExpansion b radix (cons p L1)) ->
  (IsExpansion b radix (cons p L2)) -> (In q (addExp L1 L2)) ->
  (Zlt (MSB radix p) (LSB radix q)).
Intros L1; Elim L1; Simpl; Auto.
Intros L2 p q H' H'0 H'1 H'2 H'3.
Apply IsExpansionConsInv with L := L2; Auto.
Intros a l H' L2 p q H'0 H'1 H'2 H'3; CaseEq (growExp a L2); Simpl.
Intros H'4 H'5.
Apply IsExpansionConsInv with L := (cons a l); Auto with datatypes.
Intros f l0 H'4 H'5; Elim H'5; Clear H'5; Intros H'5; [Rewrite <- H'5 | Idtac];
 Auto.
Generalize H'3 H'4; Case L2; Simpl; Auto.
Intros H'6 H'7; Inversion H'7.
Rewrite <- H0; Auto.
Apply IsExpansionConsInv with L := (cons a l); Auto with datatypes.
Rewrite H0; Auto.
Rewrite H'5; Auto.
Intros f0 l1 H'6; CaseEq (TwoSum a f0).
Intros f1 f2 H'7 H'8; Inversion H'8.
Cut ~ (is_Fzero a); [Intros Za | Idtac].
Cut ~ (is_Fzero f0); [Intros Za | Idtac].
Apply Zlt_le_trans with (Zmin (LSB radix a) (LSB radix f0)).
Unfold Zmin; Case (Zcompare (LSB radix a) (LSB radix f0)).
Apply IsExpansionConsInv with L := (cons a l); Auto with datatypes.
Apply IsExpansionConsInv with L := (cons a l); Auto with datatypes.
Apply IsExpansionConsInv with L := (cons f0 l1); Auto with datatypes.
(Replace f with (Snd (TwoSum a f0)); [Idtac | Rewrite H'7; Auto]);
 Apply TwoSumLt2; Auto.
Apply expBoundedInv with radix := radix L := l; Auto.
Apply expInv with a := p; Auto.
Apply expBoundedInv with radix := radix L := l1; Auto.
Apply expInv with a := p; Auto.
Rewrite H'7; Simpl.
Rewrite H0.
Rewrite H'5; Auto.
Contradict H'1.
Rewrite <- H'5.
Rewrite <- H0.
(Replace f2 with (Snd (TwoSum a f0)); [Idtac | Rewrite H'7; Auto]);
 Apply (is_Fzero_rep2 radix); Auto; Apply TwoSumOl2; Auto.
Apply expBoundedInv with radix := radix L := l; Auto.
Apply expInv with a := p; Auto.
Apply expBoundedInv with radix := radix L := l1; Auto.
Apply expInv with a := p; Auto.
Contradict H'1.
Rewrite <- H'5.
Rewrite <- H0.
(Replace f2 with (Snd (TwoSum a f0)); [Idtac | Rewrite H'7; Auto]);
 Apply (is_Fzero_rep2 radix); Auto; Apply TwoSumOr2; Auto.
Apply expBoundedInv with radix := radix L := l; Auto.
Apply expInv with a := p; Auto.
Apply expBoundedInv with radix := radix L := l1; Auto.
Apply expInv with a := p; Auto.
Apply (H' l0); Auto.
Apply IsExpansionSkip with q := a; Auto.
Apply IsExpansionCons; Auto.
Apply expInv with a := f; Auto.
Rewrite <- H'4.
Apply growExpIsExp; Auto.
Apply expInv with a := p; Auto.
Apply expBoundedInv with radix := radix L := l; Auto.
Apply expInv with a := p; Auto.
Apply expBoundedInv with radix := radix L := L2; Auto.
Intros q0 H'6 H'7.
Case (IsExpansionGrowConsInvAux (cons p L2) H'3 L2 p q0 a); Auto.
Apply expBoundedInv with radix := radix L := l; Auto.
Apply expInv with a := p; Auto.
Rewrite H'4; Auto with datatypes.
Intros H'8 H'9; Case (is_FzeroP a); Auto.
Intros H'10; Case H'9; Auto.
Intros H'11; Apply Zlt_le_trans with 2 := H'11; Auto.
Apply IsExpansionConsInv with L := (cons a l); Auto with datatypes.
Qed.

Theorem addExpIsExp:
  (L1, L2:(list float)) (IsExpansion b radix L1) -> (IsExpansion b radix L2) ->
  (IsExpansion b radix (addExp L1 L2)).
Intros L1; Elim L1; Simpl; Auto.
Intros a l Rec L2 HL1 HL2.
Cut (Fbounded b a); [Intros Fba | Apply (expBoundedInv b radix) with L := l];
 Auto.
Generalize (growExpIsVal L2 HL2 a Fba); Generalize (growExpIsExp L2 HL2 a Fba);
 CaseEq (growExp a L2); Simpl; Auto.
Intros H' H'0 H'1.
Apply expInv with a := a; Auto.
Intros f l0 H' H'0 H'1.
Case (is_FzeroP f); Intros Z1; Auto.
Apply IsExpansionTop1; Auto.
Apply expBoundedInv with radix := radix L := l0; Auto.
Apply Rec; Auto.
Apply expInv with a := a; Auto.
Apply expInv with a := f; Auto.
Apply IsExpansionCons; Auto.
Apply Rec; Auto.
Apply expInv with a := a; Auto.
Apply expInv with a := f; Auto.
Apply expBoundedInv with radix := radix L := l0; Auto.
Intros q H'2 H'3.
Apply IsExpansionAddInv with L1 := l L2 := l0; Auto.
Generalize H' HL2; Case L2; Simpl; Auto.
Intros H'4; Inversion H'4; Auto.
Rewrite <- H0; Auto.
Intros f0 l1; CaseEq (TwoSum a f0).
Intros f1 f2 H'4 H'5; Inversion H'5.
Intros H'6.
Cut ~ (is_Fzero a); [Intros Z2 | Idtac].
Apply IsExpansionCons; Auto.
Apply expInv with a := a; Auto.
Apply expBoundedInv with radix := radix L := l0; Auto.
Intros q0 H'7 H'8.
Apply Zle_lt_trans with (MSB radix a).
Apply MSB_monotone; Auto.
Rewrite <- H0; Auto.
Repeat Rewrite (Fabs_correct radix); Auto with arith.
Replace (FtoR radix f2) with (FtoRradix (Snd (TwoSum a f0)));
 [Idtac | Rewrite H'4; Auto].
Rewrite TwoSum2; Auto.
Replace (FtoR radix a) with (Rminus (Rplus a f0) f0);
 [Idtac | Unfold FtoRradix; Ring].
Cut (x, y, z:R)
    (Rabsolu (Rminus (Rplus x y) z))==(Rabsolu (Rminus z (Rplus x y)));
 [Intros Ra1; Repeat Rewrite Ra1 |
   Intros; Rewrite <- Rminus_Ropp; Rewrite Rabsolu_Ropp]; Auto.
Case (TwoSum1 a f0); Auto.
Apply expBoundedInv with radix := radix L := l1; Auto.
Intros H'9 H'10; Apply H'10; Auto.
Apply expBoundedInv with radix := radix L := l1; Auto.
Apply expBoundedInv with radix := radix L := l1; Auto.
Apply IsExpansionConsInv with L := l; Auto.
Contradict Z1.
Rewrite <- H0.
Replace f2 with (Snd (TwoSum a f0)).
Apply (is_Fzero_rep2 radix); Auto.
Apply TwoSumOr2; Auto.
Apply expBoundedInv with radix := radix L := l1; Auto.
Rewrite H'4; Auto.
Qed.
End mf.
