
(****************************************************************************
                                                                             
          IEEE754  :  Finduct                                                     
                                                                             
          Laurent Thery                                                      
                                                                             
*****************************************************************************
Define an induction principle on float*)
Require Import PolyList.
Require Import Float.
Require Import Fnorm.
Require Import Fop.
Require Import Fcomp.
Require Import FSucc.
Require Import FPred.
Section finduct.
Variable b:Fbound.
Variable radix:nat.
Variable precision:nat.

Local FtoRradix := (FtoR radix).
Coercion FtoRradix : float >-> R.
Hypothesis radixMoreThanOne:(lt (S O) radix).
Hypothesis precisionNotZero:~ precision=O.
Hypothesis pGivesBound:(vNum b)=(minus (exp radix precision) (1)).

Definition Fweight :=
   [p:float](Zplus (Fnum p) (Zmult (Fexp p) (exp radix precision))).

Theorem FweightLt:
  (p, q:float)
  (Fcanonic radix b precision p) ->
  (Fcanonic radix b precision q) -> (Rle R0 p) -> (Rlt p q) ->
  (Zlt (Fweight p) (Fweight q)).
Intros p q H' H'0 H'1 H'2.
Case (FcanonicLtPos ? radixMoreThanOne b precision) with p := p q := q;
 Auto with arith; Intros Zl1.
Unfold Fweight; Simpl.
Replace (Fexp q) with (Zplus (Zminus (Fexp q) (Fexp p)) (Fexp p));
 [Idtac | Ring].
Rewrite Zmult_plus_distr_l.
Rewrite Zplus_assoc.
Repeat Rewrite [x, y, z:Z](Zplus_sym x (Zmult y z)).
Apply Zlt_reg_l.
Apply Zlt_le_trans with (inject_nat (exp radix precision)).
Apply Zle_lt_trans with (Zpred (exp radix precision)).
Rewrite <- inj_pred; Auto with arith.
Replace (pred (exp radix precision)) with (vNum b); [Intuition | Idtac].
Rewrite pGivesBound.
Case (exp radix precision); Simpl; Auto; Intros n; Case n; Simpl; Auto.
Apply Zlt_pred_n_n.
Unfold Fweight; Simpl.
Apply Zle_trans with (Zmult (Zminus (Fexp q) (Fexp p)) (exp radix precision)).
Pattern 1 (inject_nat (exp radix precision));
 Replace (inject_nat (exp radix precision))
      with (Zmult (Zs ZERO) (exp radix precision)); Auto.
Apply Zle_Zmult_comp_r; Auto with zarith.
Simpl; Case (exp radix precision); Auto.
Cut (Zle ZERO (Fnum q)); Auto with zarith.
Apply (LeR0Fnum radix); Auto.
Apply Rle_trans with (FtoRradix p); Auto; Apply Rlt_le; Auto.
Elim Zl1; Intros H'3 H'4; Clear Zl1.
Unfold Fweight; Simpl.
Rewrite <- H'3.
Repeat Rewrite [x, y, z:Z](Zplus_sym x (Zmult y z)).
Apply Zlt_reg_l; Auto.
Qed.

Theorem FweightEq:
  (p, q:float)
  (Fcanonic radix b precision p) ->
  (Fcanonic radix b precision q) -> <R> p==q ->(Fweight p)=(Fweight q).
Intros p q H' H'0 H'1.
Rewrite (FcanonicUnique ? radixMoreThanOne b precision) with p := p q := q;
 Auto with arith.
Qed.

Theorem FweightZle:
  (p, q:float)
  (Fcanonic radix b precision p) ->
  (Fcanonic radix b precision q) -> (Rle R0 p) -> (Rle p q) ->
  (Zle (Fweight p) (Fweight q)).
Intros p q H' H'0 H'1 H'2; Casec H'2; Intros H'2.
Apply Zlt_le_weak.
Apply FweightLt; Auto.
Rewrite (FweightEq p q); Auto with zarith.
Qed.

Theorem FinductPosAux:
  (P:float ->Prop)
  (p:float)
  (Rle R0 p) ->
  (Fcanonic radix b precision p) ->
  (P p) ->
  ((q:float) (Fcanonic radix b precision q) -> (Rle p q) -> (P q) ->
   (P (FSucc b radix precision q))) ->
  (x:Z) (Zle ZERO x) ->
  (q:float)
  x=(Zminus (Fweight q) (Fweight p)) ->
  (Fcanonic radix b precision q) -> (Rle p q) ->(P q).
Intros P p H' H'0 H'1 H'2 x H'3; Pattern x.
Apply Z_lt_induction; Auto.
Intros x0 H'4 q H'5 H'6 H'7.
Casec H'7; Intros H'7.
Cut (Rle p (FPred b radix precision q)); [Intros Rl1 | Idtac].
Cut (P (FPred b radix precision q)); [Intros P1 | Idtac].
Rewrite <- (FSucPred b radix precision) with x := q; Auto with arith.
Apply H'2; Auto with float arith.
Apply H'4 with y := (Zminus (Fweight (FPred b radix precision q)) (Fweight p));
 Auto.
Split.
Cut (Zle (Fweight p) (Fweight (FPred b radix precision q))); Auto with zarith.
Apply FweightZle; Auto.
Apply FPredCanonic; Auto with arith.
Rewrite H'5.
Cut (Zlt (Fweight (FPred b radix precision q)) (Fweight q));
 [Auto with zarith | Idtac].
Apply FweightLt; Auto with float.
Apply (R0RltRlePred b radix precision); Auto.
Apply Rle_lt_trans with (FtoRradix p); Auto.
Apply (FPredLt b radix precision); Auto.
Apply (FPredCanonic b radix precision); Auto with arith.
Apply (FPredProp b radix precision); Auto with arith.
Rewrite <- (FcanonicUnique ? radixMoreThanOne b precision) with p := p q := q;
 Auto with arith.
Qed.

Theorem FinductPos:
  (P:float ->Prop)
  (p:float)
  (Rle R0 p) ->
  (Fcanonic radix b precision p) ->
  (P p) ->
  ((q:float) (Fcanonic radix b precision q) -> (Rle p q) -> (P q) ->
   (P (FSucc b radix precision q))) ->
  (q:float) (Fcanonic radix b precision q) -> (Rle p q) ->(P q).
Intros P p H' H'0 H'1 H'2 q H'3 H'4.
Apply FinductPosAux with p := p x := (Zminus (Fweight q) (Fweight p)); Auto.
Cut (Zle (Fweight p) (Fweight q)); [Auto with zarith | Idtac].
Apply FweightZle; Auto with float.
Qed.

Theorem FinductNegAux:
  (P:float ->Prop)
  (p:float)
  (Rle R0 p) ->
  (Fcanonic radix b precision p) ->
  (P p) ->
  ((q:float)
   (Fcanonic radix b precision q) -> (Rlt R0 q) -> (Rle q p) -> (P q) ->
   (P (FPred b radix precision q))) ->
  (x:Z) (Zle ZERO x) ->
  (q:float)
  x=(Zminus (Fweight p) (Fweight q)) ->
  (Fcanonic radix b precision q) -> (Rle R0 q) -> (Rle q p) ->(P q).
Intros P p H' H'0 H'1 H'2 x H'3; Pattern x.
Apply Z_lt_induction; Auto.
Intros x0 H'4 q H'5 H'6 H'7 H'8.
Casec H'8; Intros H'8.
Cut (Rle (FSucc b radix precision q) p); [Intros Rle1 | Idtac].
Cut (P (FSucc b radix precision q)); [Intros P1 | Idtac].
Rewrite <- (FPredSuc b radix precision) with x := q; Auto with arith.
Apply H'2; Auto with float arith.
Apply Rle_lt_trans with (FtoRradix q); Auto.
Apply (FSuccLt b radix); Auto with arith.
Apply H'4 with y := (Zminus (Fweight p) (Fweight (FSucc b radix precision q)));
 Auto.
Split.
Cut (Zle (Fweight (FSucc b radix precision q)) (Fweight p)); Auto with zarith.
Apply FweightZle; Auto.
Apply FSuccCanonic; Auto with arith.
Apply Rle_trans with (FtoRradix q); Auto; Apply Rlt_le.
Apply (FSuccLt b radix); Auto with arith.
Rewrite H'5.
Cut (Zlt (Fweight q) (Fweight (FSucc b radix precision q)));
 [Auto with zarith | Idtac].
Apply FweightLt; Auto with float.
Apply (FSuccLt b radix); Auto with arith.
Apply FSuccCanonic; Auto with arith.
Apply Rle_trans with (FtoRradix q); Auto; Apply Rlt_le.
Apply (FSuccLt b radix); Auto with arith.
Apply (FSuccProp b radix); Auto with arith.
Rewrite <- (FcanonicUnique ? radixMoreThanOne b precision) with p := p q := q;
 Auto with arith.
Qed.

Theorem FinductNeg:
  (P:float ->Prop)
  (p:float)
  (Rle R0 p) ->
  (Fcanonic radix b precision p) ->
  (P p) ->
  ((q:float)
   (Fcanonic radix b precision q) -> (Rlt R0 q) -> (Rle q p) -> (P q) ->
   (P (FPred b radix precision q))) ->
  (q:float) (Fcanonic radix b precision q) -> (Rle R0 q) -> (Rle q p) ->(P q).
Intros P p H' H'0 H'1 H'2 q H'3 H'4 H'5.
Apply FinductNegAux with p := p x := (Zminus (Fweight p) (Fweight q)); Auto.
Cut (Zle (Fweight q) (Fweight p)); [Auto with zarith | Idtac].
Apply FweightZle; Auto with float.
Qed.

Theorem radixRangeBoundExp:
  (p, q:float)
  (Fcanonic radix b precision p) ->
  (Fcanonic radix b precision q) ->
  (Rle R0 p) -> (Rlt p q) -> (Rlt q (Rmult radix p)) ->
  (Fexp p)=(Fexp q) \/ (Zs (Fexp p))=(Fexp q).
Intros p q H' H'0 H'1 H'2 H'3.
Case (FcanonicLtPos ? radixMoreThanOne b precision) with p := p q := q;
 Auto with arith.
2:Intros H'4; Elim H'4; Intros H'5 H'6; Clear H'4; Auto.
Intros H'4; Right.
Casec H'; Intros H'.
Case (FcanonicLtPos ? radixMoreThanOne b precision) with p := q
                                                         q :=
                                                         (Float
                                                            (Fnum p)
                                                            (Zs (Fexp p)));
 Auto with arith.
Left.
Repeat Split; Simpl.
Intuition.
Intuition.
Apply Zle_trans with (Fexp p); Auto with zarith; Intuition.
Unfold Fdigit; Simpl.
Intuition.
Apply Rle_trans with (FtoRradix p); Auto; Apply Rlt_le; Auto.
Unfold FtoR; Simpl.
Rewrite powerRZ_Zs; Auto with real arith; Auto.
Rewrite <- Rmult_assoc; Rewrite [x:R] [y:nat](Rmult_sym x y);
 Rewrite Rmult_assoc; Auto.
Simpl; Intros; Apply Zle_antisym; Auto with zarith.
Simpl; Auto.
Intros H'5; Elim H'5; Intros H'6 H'7; Auto.
Case (FcanonicLtPos ? radixMoreThanOne b precision) with p := q
                                                         q :=
                                                         (Float
                                                            (nNormMin
                                                               radix precision)
                                                            (Zs (Fexp p)));
 Auto with arith.
Left; Repeat Split; Simpl.
Apply Zle_trans with ZERO; Auto with zarith.
Auto with zarith float.
Case H'.
Intros H'5 H'6; Elim H'6; Intros H'7 H'8; Rewrite H'7; Clear H'6;
 Auto with zarith.
Unfold Fdigit; Simpl; Auto with float.
Rewrite absolu_INR; Auto with float.
Apply digitnNormMin with b := b; Auto with arith.
Apply Rle_trans with 1 := H'1; Apply Rlt_le; Auto.
Apply Rlt_trans with 1 := H'3.
Unfold FtoR; Simpl.
Rewrite powerRZ_Zs; Auto with real arith; Auto.
Rewrite <- Rmult_assoc; Rewrite [x:R] [y:nat](Rmult_sym x y);
 Rewrite Rmult_assoc; Auto.
Apply Rlt_monotony; Auto with real arith.
Case H'.
Intros H'5 H'6; Elim H'6; Intros H'7 H'8; Rewrite H'7; Clear H'6.
Change (Rlt p (firstNormalPos radix b precision)).
Apply (FsubnormalLtFirstNormalPos radix); Auto with arith.
Simpl.
Auto with zarith.
Simpl.
Intros H'5; Elim H'5; Intros H'6 H'7; Rewrite H'6; Clear H'5; Auto.
Qed.
End finduct.

