
(****************************************************************************
                                                                             
          IEEE754  :  Fprop                                                     
                                                                             
          Laurent Thery                                                      
                                                                             
*****************************************************************************
*)
Require Import Float.
Require Import Fcomp.
Require Import Fop.
Require Import Fnorm.
Section Fprop.
Variable radix:nat.
Hypothesis radixMoreThanOne:(lt (1) radix).

Local FtoRradix := (FtoR radix).
Coercion FtoRradix : float >-> R.
Variable b:Fbound.

Theorem SterbenzAux:
  (x, y:float)
  (Fbounded b x) -> (Fbounded b y) -> (Rle y x) -> (Rle x (Rmult (2) y)) ->
  (Fbounded b (Fminus radix x y)).
Intros x y H' H'0 H'1 H'2.
Cut (Rle R0 (Fminus radix x y)); [Intros Rle1 | Idtac].
Cut (Rle (Fminus radix x y) y); [Intros Rle2 | Idtac].
Case (Zle_or_lt (Fexp x) (Fexp y)); Intros Zle1.
Repeat Split.
Apply Zle_trans with ZERO; Auto with zarith.
Apply (LeR0Fnum radix); Auto.
Apply Zle_trans with (Fnum x).
Apply Rle_Fexp_eq_Zle with radix := radix; Auto with arith.
Apply Rle_trans with 2 := H'1; Auto.
Unfold Fminus; Simpl; Apply Zmin_le1; Auto.
Intuition.
Unfold Fminus; Simpl; Rewrite Zmin_le1; Auto; Intuition.
Repeat Split.
Apply Zle_trans with ZERO; Auto with zarith.
Apply (LeR0Fnum radix); Auto.
Apply Zle_trans with (Fnum y).
Apply Rle_Fexp_eq_Zle with radix := radix; Auto with arith.
Unfold Fminus; Simpl; Apply Zmin_le2; Auto with zarith.
Intuition.
Unfold Fminus; Simpl; Rewrite Zmin_le2; Auto; Intuition.
Rewrite (Fminus_correct radix); Auto with arith; Fold FtoRradix.
Apply Rle_anti_compatibility with r := (FtoRradix y); Auto.
Replace (Rplus y (Rminus x y)) with (FtoRradix x); [Idtac | Ring].
Replace (Rplus y y) with (Rmult (2) y); [Auto | Simpl; Ring].
Rewrite (Fminus_correct radix); Auto with arith; Fold FtoRradix.
Apply Rle_anti_compatibility with r := (FtoRradix y); Auto.
Replace (Rplus y (Rminus x y)) with (FtoRradix x); [Idtac | Ring].
Replace (Rplus y R0) with (FtoRradix y); [Auto | Simpl; Ring].
Qed.

Theorem Sterbenz:
  (x, y:float)
  (Fbounded b x) ->
  (Fbounded b y) -> (Rle (Rmult (Rinv (2)) y) x) -> (Rle x (Rmult (2) y)) ->
  (Fbounded b (Fminus radix x y)).
Intros x y H' H'0 H'1 H'2.
Cut (Rle y (Rmult (2) x)); [Intros Le1 | Idtac].
Case (Rle_or_lt x y); Intros Le2; Auto.
Apply oppBoundedInv; Auto.
Rewrite Fopp_Fminus.
Apply SterbenzAux; Auto with real.
Apply SterbenzAux; Auto with real.
Apply Rlt_le; Auto.
Apply Rle_mult_anticompatibility with z := (Rinv (2)).
Rewrite <- Rmult_assoc; Rewrite Rinv_l; Auto with real.
Rewrite Rmult_ne_r; Auto.
Apply Rlt_Rinv; Auto with real.
Qed.

Theorem BminusSameExpAux:
  (x, y:float)
  (Fbounded b x) ->
  (Fbounded b y) -> (Rle R0 y) -> (Rle y x) -> (Fexp x)=(Fexp y) ->
  (Fbounded b (Fminus radix x y)).
Intros x y H' H'0 H'1 H'2 H'3.
Rewrite minusSameExp; Auto.
Repeat Split; Simpl; Auto.
Apply Zle_trans with ZERO; Auto with zarith.
Apply Zsimpl_le_plus_l with p := (Fnum y); Auto.
Rewrite Zero_right; Rewrite Zle_plus_minus.
Apply Rle_Fexp_eq_Zle with radix := radix; Auto with zarith.
Apply Zle_trans with (Fnum x); Auto with zarith.
Cut (Zle ZERO (Fnum y)); Auto with zarith.
Apply LeR0Fnum with radix := radix; Auto with zarith.
Intuition.
Intuition.
Qed.

Theorem BminusSameExp:
  (x, y:float)
  (Fbounded b x) ->
  (Fbounded b y) -> (Rle R0 x) -> (Rle R0 y) -> (Fexp x)=(Fexp y) ->
  (Fbounded b (Fminus radix x y)).
Intros x y H' H'0 H'1 H'2 H'3.
Case (Rle_or_lt x y); Intros Le2; Auto.
Apply oppBoundedInv; Auto.
Rewrite Fopp_Fminus.
Apply BminusSameExpAux; Auto.
Apply BminusSameExpAux; Auto with real.
Red; Auto.
Qed.

Theorem BminusSameExpNeg:
  (x, y:float)
  (Fbounded b x) ->
  (Fbounded b y) -> (Rle x R0) -> (Rle y R0) -> (Fexp x)=(Fexp y) ->
  (Fbounded b (Fminus radix x y)).
Intros x y H' H'0 H'1 H'2 H'3.
Apply oppBoundedInv; Auto.
Rewrite Fopp_Fminus_dist.
Apply BminusSameExp; Auto with float.
Unfold FtoRradix; Rewrite Fopp_correct.
Replace R0 with (Ropp R0); Auto with real.
Unfold FtoRradix; Rewrite Fopp_correct.
Replace R0 with (Ropp R0); Auto with real.
Qed.
End Fprop.
Hints Resolve Sterbenz BminusSameExp BminusSameExpNeg :float.
