
(****************************************************************************
                                                                             
          IEEE754  :  FroundMult                                                 
                                                                             
          Laurent Thery  &  Sylvie Boldo                                                      
                                                                             
*****************************************************************************
*)
Require Export FroundProp.
Section FRoundP.
Variable b:Fbound.
Variable radix:nat.
Variable precision:nat.

Local FtoRradix := (FtoR radix).
Coercion FtoRradix : float >-> R.
Hypothesis radixMoreThanOne:(lt (S O) radix).
Hypothesis precisionGreaterThanOne:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp radix precision) (1)).

Theorem minRepMult:
  (p, q, fmin:float)
  (Fbounded b p) ->
  (Fbounded b q) ->
  (Rle R0 p) ->
  (Rle R0 q) ->
  (Zle (Zopp (dExp b)) (Zplus (Fexp p) (Fexp q))) ->
  (isMin b radix (Rmult p q) fmin) ->
  (Ex [r:float]
  r==(Rminus (Rmult p q) fmin) /\
  ((Fbounded b r) /\ (Fexp r)=(Zplus (Fexp p) (Fexp q)))).
Intros p q fmin Fp Fq H' H'0 H'1 H'2.
Cut (Zle ZERO (Zmult (Fnum p) (Fnum q)));
 [Intros multPos | Apply Zle_mult_gen; Apply (LeR0Fnum radix); Auto with arith].
Cut (ex Z [m:Z]<R> fmin==(Float m (Fexp (Fmult p q)))).
2:Unfold FtoRradix;
   Apply RoundedModeRep with b := b precision := precision P := (isMin b radix);
   Auto.
2:Apply MinRoundedModeP with precision := precision; Auto.
2:Rewrite (Fmult_correct radix); Auto with arith.
Intros H'3; Elim H'3; Intros m E; Clear H'3.
Exists (Fminus radix (Fmult p q) (Float m (Fexp (Fmult p q)))).
Split.
Rewrite E; Unfold FtoRradix; Repeat Rewrite Fminus_correct;
 Repeat Rewrite Fmult_correct; Auto with arith.
Split.
Cut (Rle fmin (Fmult p q));
 [Idtac |
   Unfold FtoRradix; Rewrite Fmult_correct; Auto; Case H'2;
    Auto with real arith; Intuition].
Rewrite E; Unfold Fmult Fminus Fopp Fplus; Simpl; Auto.
Repeat Rewrite Zmin_n_n; Repeat Rewrite <- Zminus_n_n; Auto.
Simpl; Repeat Rewrite Zpower_nat_O; Repeat Rewrite Zmult_n_1.
Unfold FtoRradix FtoR; Simpl.
Intros H'3;
 (Cut (Zle m (Zmult (Fnum p) (Fnum q)));
   [Idtac |
     Apply le_IZR;
      Apply Rle_mult_anticompatibility
           with z := (powerRZ radix (Zplus (Fexp p) (Fexp q)));
      Auto with real arith;
      Repeat Rewrite (Rmult_sym (powerRZ radix (Zplus (Fexp p) (Fexp q)))); Auto]);
 Intros H'4.
Repeat Split; Simpl; Auto.
Apply Zle_trans with (Zopp O); Auto with zarith.
Case (NquotientProp (absolu (Zmult (Fnum p) (Fnum q))) (exp radix precision));
 Auto with arith.
Intros x H'5; Elim H'5; Intros H'6 H'7; Clear H'5.
Cut (Rle
       (Rmult
          (Nquotient (absolu (Zmult (Fnum p) (Fnum q))) (exp radix precision))
          (powerRZ radix (Zplus precision (Zplus (Fexp p) (Fexp q))))) fmin);
 [Rewrite E; Intros H'8 | Idtac].
Cut (Rle
       (Rmult
          (Nquotient (absolu (Zmult (Fnum p) (Fnum q))) (exp radix precision))
          (powerRZ radix precision)) m); [Intros H'9 | Idtac].
Apply Zle_trans with (inject_nat x); Auto.
Replace (inject_nat x)
     with (Zplus
             (Zmult (Fnum p) (Fnum q))
             (Zopp
                (mult
                   (Nquotient
                      (absolu (Zmult (Fnum p) (Fnum q))) (exp radix precision))
                   (exp radix precision)))).
Apply Zle_reg_l; Auto.
Apply Zle_Zopp.
Apply le_IZR; Auto.
AutoRewrite [Rsimpl].
Rewrite <- powerRZ_is_exp; Auto with arith.
Pattern 1 (Zmult (Fnum p) (Fnum q)); Rewrite <- (inj_abs ? multPos).
Pattern 1 (absolu (Zmult (Fnum p) (Fnum q))); Rewrite H'6; Auto.
Rewrite inj_plus; Ring.
Apply inj_le.
Rewrite pGivesBound.
Generalize H'7; Case (exp radix precision); Simpl; Auto with arith.
Intros n; Rewrite <- minus_n_O; Auto with arith.
Apply Rle_mult_anticompatibility
     with z := (powerRZ radix (Zplus (Fexp p) (Fexp q))); Auto with real arith.
Repeat Rewrite (Rmult_sym (powerRZ radix (Zplus (Fexp p) (Fexp q)))); Auto.
Rewrite Rmult_assoc; Rewrite <- powerRZ_add; Auto with real arith.
Case (FboundedMbound ? radixMoreThanOne b precision) with z :=
                                                          (Zplus
                                                             precision
                                                             (Zplus
                                                                (Fexp p)
                                                                (Fexp q)))
                                                          m :=
                                                          (Nquotient
                                                             (absolu
                                                                (Zmult
                                                                   (Fnum p)
                                                                   (Fnum q)))
                                                             (exp
                                                                radix precision));
 Auto with arith.
Apply le_mult_anti_compatibility with n := (exp radix precision);
 Auto with arith.
Rewrite (mult_sym (exp radix precision)).
Apply le_trans with (absolu (Zmult (Fnum p) (Fnum q))).
Pattern 2 (absolu (Zmult (Fnum p) (Fnum q))); Rewrite H'6; Auto with arith.
Rewrite absolu_comp_mult.
Apply lte_comp_mult; Auto; Apply le_trans with (vNum b);
 Try Apply pBounded_absolu; Auto; Rewrite pGivesBound;
 Pattern 2 (exp radix precision); Rewrite (minus_n_O (exp radix precision));
 Apply minus_le; Auto with arith.
Auto with zarith.
Intros x0 H'5; Elim H'5; Intros H'8 H'9; Rewrite <- H'9; Clear H'5.
Case H'2.
Intros H'5 H'10; Elim H'10; Intros H'11 H'12; Apply H'12; Clear H'10; Auto.
Unfold FtoRradix in H'9; Rewrite H'9; Auto.
Rewrite powerRZ_add; Auto with real arith.
Rewrite <- Rmult_assoc.
Unfold FtoRradix; Rewrite <- Fmult_correct; Auto with arith.
Unfold Fmult FtoR; Simpl.
Repeat Rewrite [x:?](Rmult_sym x (powerRZ radix (Zplus (Fexp p) (Fexp q)))).
Apply Rle_monotony; Auto with real arith.
Rewrite powerRZ_is_exp; Auto with arith.
Pattern 2 (Zmult (Fnum p) (Fnum q));
 Rewrite <- (inj_abs (Zmult (Fnum p) (Fnum q))); Auto with arith.
Pattern 2 (absolu (Zmult (Fnum p) (Fnum q))); Rewrite H'6; Auto.
Rewrite <- Rmult_INR.
Repeat Rewrite <- INR_IZR_INZ.
Apply Rle_INR; Auto with arith.
Simpl; Auto.
Apply Zmin_n_n; Auto.
Qed.

Theorem maxRepMult:
  (p, q, fmax:float)
  (Fbounded b p) ->
  (Fbounded b q) ->
  (Rle R0 p) ->
  (Rle R0 q) ->
  (Zle (Zopp (dExp b)) (Zplus (Fexp p) (Fexp q))) ->
  (isMax b radix (Rmult p q) fmax) ->
  (Ex [r:float]
  r==(Rminus (Rmult p q) fmax) /\
  ((Fbounded b r) /\ (Fexp r)=(Zplus (Fexp p) (Fexp q)))).
Intros p q fmax Fp Fq H' H'0 H'1 H'2.
Cut (Zle ZERO (Zmult (Fnum p) (Fnum q)));
 [Intros multPos | Apply Zle_mult_gen; Apply (LeR0Fnum radix); Auto with arith].
Case (NquotientProp (absolu (Zmult (Fnum p) (Fnum q))) (exp radix precision));
 Auto with arith.
Intros x; Casec x.
Intros H'3; Elim H'3; Intros H'4 H'5; Clear H'3.
Exists (Fzero (Zplus (Fexp p) (Fexp q))); Split.
Replace (FtoRradix (Fzero (Zplus (Fexp p) (Fexp q)))) with R0.
Apply r_Rplus_plus with r := (FtoRradix fmax).
Replace (Rplus fmax R0) with (FtoRradix fmax); [Idtac | Ring].
Replace (Rplus fmax (Rminus (Rmult p q) fmax)) with (Rmult p q); [Idtac | Ring].
Unfold FtoRradix; Rewrite <- (Fmult_correct radix); Auto with arith.
Case (FboundedMbound ? radixMoreThanOne b precision) with z :=
                                                          (Zplus
                                                             precision
                                                             (Zplus
                                                                (Fexp p)
                                                                (Fexp q)))
                                                          m :=
                                                          (Nquotient
                                                             (absolu
                                                                (Zmult
                                                                   (Fnum p)
                                                                   (Fnum q)))
                                                             (exp
                                                                radix precision));
 Auto with arith.
Apply le_mult_anti_compatibility with n := (exp radix precision);
 Auto with arith.
Rewrite (mult_sym (exp radix precision)).
Apply le_trans with (absolu (Zmult (Fnum p) (Fnum q))).
Pattern 2 (absolu (Zmult (Fnum p) (Fnum q))); Rewrite H'4; Auto with arith.
Rewrite absolu_comp_mult.
Apply lte_comp_mult; Auto; Apply le_trans with (vNum b);
 Try Apply pBounded_absolu; Auto; Rewrite pGivesBound;
 Pattern 2 (exp radix precision); Rewrite (minus_n_O (exp radix precision));
 Apply minus_le; Auto with arith.
Apply Zle_trans with 1 := H'1; Auto with zarith.
Intros x H'3; Elim H'3; Intros H'6 H'7; Clear H'3.
Cut <R> (FtoR radix (Fmult p q))==(FtoR radix x).
Intros H'3; Rewrite H'3.
Apply sym_eqT; Apply (ProjectMax b radix); Auto.
Rewrite <- H'3; Auto.
Rewrite Fmult_correct; Auto with arith.
Unfold FtoRradix in H'7; Rewrite H'7.
Unfold FtoR; Simpl.
Rewrite powerRZ_add with n := (inject_nat precision); Auto with real arith.
Rewrite <- Rmult_assoc.
Rewrite powerRZ_is_exp; Auto with real arith.
Rewrite <- Rmult_INR.
Rewrite <- plus_n_O in H'4; Rewrite <- H'4; Auto.
Repeat Rewrite INR_IZR_INZ; Rewrite inj_abs; Auto.
Unfold FtoRradix FtoR Fzero; Simpl; Ring.
Split; Repeat Split; Simpl; Auto with zarith.
Intros x H'3; Elim H'3; Intros Hx Hx'; Clear H'3.
Cut (ex Z [m:Z]<R> fmax==(Float m (Fexp (Fmult p q)))).
2:Unfold FtoRradix;
   Apply RoundedModeRep with b := b precision := precision P := (isMax b radix);
   Auto.
2:Apply MaxRoundedModeP with precision := precision; Auto.
2:Rewrite (Fmult_correct radix); Auto with arith.
Intros H'3; Elim H'3; Intros m E; Clear H'3.
Exists (Fopp (Fminus radix (Float m (Fexp (Fmult p q))) (Fmult p q))).
Split.
Rewrite E; Unfold FtoRradix; Repeat Rewrite Fopp_correct;
 Repeat Rewrite Fminus_correct; Repeat Rewrite Fmult_correct; Auto with arith;
 Ring.
Split.
Apply oppBounded; Auto.
Cut (Rle (Fmult p q) fmax);
 [Idtac |
   Unfold FtoRradix; Rewrite Fmult_correct; Auto; Case H'2;
    Auto with real arith; Intuition].
Rewrite E; Unfold Fmult Fminus Fopp Fplus; Simpl; Auto.
Repeat Rewrite Zmin_n_n; Repeat Rewrite <- Zminus_n_n; Auto.
Simpl; Repeat Rewrite Zpower_nat_O; Repeat Rewrite Zmult_n_1.
Unfold FtoRradix FtoR; Simpl.
Intros H'3;
 (Cut (Zle (Zmult (Fnum p) (Fnum q)) m);
   [Idtac |
     Apply le_IZR;
      Apply Rle_mult_anticompatibility
           with z := (powerRZ radix (Zplus (Fexp p) (Fexp q)));
      Auto with real arith;
      Repeat Rewrite (Rmult_sym (powerRZ radix (Zplus (Fexp p) (Fexp q)))); Auto]);
 Intros H'4.
Repeat Split; Simpl; Auto.
Apply Zle_trans with (Zopp O); Auto with zarith.
Cut (Rle
       fmax
       (Rmult
          (S
             (Nquotient
                (absolu (Zmult (Fnum p) (Fnum q))) (exp radix precision)))
          (powerRZ radix (Zplus precision (Zplus (Fexp p) (Fexp q))))));
 [Rewrite E; Intros H'8 | Idtac].
Cut (Rle
       m
       (Rmult
          (S
             (Nquotient
                (absolu (Zmult (Fnum p) (Fnum q))) (exp radix precision)))
          (powerRZ radix precision))); [Intros H'9 | Idtac].
Apply Zle_trans with (Zminus (exp radix precision) (S x)); Auto.
Replace (inject_nat (S x))
     with (Zplus
             (Zmult (Fnum p) (Fnum q))
             (Zopp
                (mult
                   (Nquotient
                      (absolu (Zmult (Fnum p) (Fnum q))) (exp radix precision))
                   (exp radix precision)))).
Replace (Zminus
           (exp radix precision)
           (Zplus
              (Zmult (Fnum p) (Fnum q))
              (Zopp
                 (mult
                    (Nquotient
                       (absolu (Zmult (Fnum p) (Fnum q))) (exp radix precision))
                    (exp radix precision)))))
     with (Zplus
             (Zplus
                (exp radix precision)
                (mult
                   (Nquotient
                      (absolu (Zmult (Fnum p) (Fnum q))) (exp radix precision))
                   (exp radix precision))) (Zopp (Zmult (Fnum p) (Fnum q))));
 [Idtac | Ring].
Apply Zle_reg_r; Auto.
Replace (Zplus
           (exp radix precision)
           (mult
              (Nquotient
                 (absolu (Zmult (Fnum p) (Fnum q))) (exp radix precision))
              (exp radix precision)))
     with (inject_nat
             (mult
                (S
                   (Nquotient
                      (absolu (Zmult (Fnum p) (Fnum q))) (exp radix precision)))
                (exp radix precision))).
Apply le_IZR.
AutoRewrite [Rsimpl].
Rewrite <- powerRZ_is_exp; Auto with arith.
Simpl; Rewrite inj_plus; Auto.
Pattern 1 (Zmult (Fnum p) (Fnum q)); Rewrite <- (inj_abs ? multPos).
Pattern 1 (absolu (Zmult (Fnum p) (Fnum q))); Rewrite Hx; Auto.
Rewrite inj_plus; Ring.
Rewrite pGivesBound; Auto with zarith.
Rewrite inj_minus1; Auto with zarith.
Unfold Zminus; Apply Zle_reg_l; Apply Zle_Zopp; Apply inj_le; Auto with arith.
Apply Rle_mult_anticompatibility
     with z := (powerRZ radix (Zplus (Fexp p) (Fexp q))); Auto with real arith.
Repeat Rewrite (Rmult_sym (powerRZ radix (Zplus (Fexp p) (Fexp q)))); Auto.
Rewrite Rmult_assoc; Rewrite <- powerRZ_add; Auto with real arith.
Case (FboundedMbound ? radixMoreThanOne b precision) with z :=
                                                          (Zplus
                                                             precision
                                                             (Zplus
                                                                (Fexp p)
                                                                (Fexp q)))
                                                          m :=
                                                          (S
                                                             (Nquotient
                                                                (absolu
                                                                   (Zmult
                                                                      (Fnum p)
                                                                      (Fnum q)))
                                                                (exp
                                                                   radix
                                                                   precision)));
 Auto with arith.
Apply lt_le_S.
Apply lt_mult_anti_compatibility with n := (exp radix precision);
 Auto with arith.
Rewrite (mult_sym (exp radix precision)).
Apply lt_le_trans with (absolu (Zmult (Fnum p) (Fnum q))).
Pattern 2 (absolu (Zmult (Fnum p) (Fnum q))); Rewrite Hx; Auto with zarith.
Rewrite absolu_comp_mult.
Apply lte_comp_mult; Auto; Apply le_trans with (vNum b);
 Try Apply pBounded_absolu; Auto; Rewrite pGivesBound;
 Pattern 2 (exp radix precision); Rewrite (minus_n_O (exp radix precision));
 Apply minus_le; Auto with arith.
Auto with zarith.
Intros x0 H'5; Elim H'5; Intros H'8 H'9; Rewrite <- H'9; Clear H'5.
Case H'2.
Intros H'5 H'10; Elim H'10; Intros H'11 H'12; Apply H'12; Clear H'10; Auto.
Unfold FtoRradix in H'9; Rewrite H'9; Auto.
Rewrite powerRZ_add; Auto with real arith.
Rewrite <- Rmult_assoc.
Unfold FtoRradix; Rewrite <- Fmult_correct; Auto with arith.
Unfold Fmult FtoR.
Replace (Fnum (Float (Zmult (Fnum p) (Fnum q)) (Zplus (Fexp p) (Fexp q))))
     with (Zmult (Fnum p) (Fnum q)); [Idtac | Simpl; Auto].
Replace (Fexp (Float (Zmult (Fnum p) (Fnum q)) (Zplus (Fexp p) (Fexp q))))
     with (Zplus (Fexp p) (Fexp q)); [Idtac | Simpl; Auto].
Repeat Rewrite [x:?](Rmult_sym x (powerRZ radix (Zplus (Fexp p) (Fexp q)))).
Apply Rle_monotony; Auto with real arith.
Rewrite powerRZ_is_exp; Auto with arith.
Pattern 1 (Zmult (Fnum p) (Fnum q));
 Rewrite <- (inj_abs (Zmult (Fnum p) (Fnum q))); Auto with arith.
Pattern 1 (absolu (Zmult (Fnum p) (Fnum q))); Rewrite Hx; Auto.
Rewrite <- Rmult_INR.
Repeat Rewrite <- INR_IZR_INZ.
Apply Rle_INR; Auto with arith.
Simpl; Auto with zarith.
Simpl; Apply Zmin_n_n; Auto.
Qed.

Theorem multExpMin:
  (P:?) (RoundedModeP b radix P) ->
  (p, q, pq:float) (P (Rmult p q) pq) ->
  (Ex [s:float]
  (Fbounded b s) /\ (<R> s==pq /\ (Zle (Zplus (Fexp p) (Fexp q)) (Fexp s)))).
Intros P H' p q pq H'0.
Case (RoundedModeRep b radix precision) with p := (Fmult p q) q := pq P := P;
 Auto.
Rewrite Fmult_correct; Auto with arith.
Simpl; Intros x H'1.
Case (eqExpLess ? radixMoreThanOne b) with p := pq
                                           q := (Float x (Fexp (Fmult p q)));
 Auto.
Apply RoundedModeBounded with radix := radix P := P r := (Rmult p q); Auto.
Simpl; Intros x0 H'2; Elim H'2; Intros H'3 H'4; Elim H'4; Intros H'5 H'6;
 Clear H'4 H'2.
Exists x0; Split; [Idtac | Split]; Auto.
Unfold FtoRradix; Rewrite H'5; Auto.
Apply le_IZR; Auto.
Qed.

Theorem multExpUpperBound:
  (P:?) (RoundedModeP b radix P) ->
  (p, q, pq:float)
  (P (Rmult p q) pq) ->
  (Fbounded b p) ->
  (Fbounded b q) -> (Zle (Zopp (dExp b)) (Zplus (Fexp p) (Fexp q))) ->
  (Ex [r:float]
  (Fbounded b r) /\
  (<R> r==pq /\ (Zle (Fexp r) (Zplus precision (Zplus (Fexp p) (Fexp q)))))).
Intros P H' p q pq H'0 H'1 H'2 H'3.
Replace (Zplus precision (Zplus (Fexp p) (Fexp q)))
     with (Fexp (Float (vNum b) (Zplus precision (Zplus (Fexp p) (Fexp q)))));
 [Idtac | Simpl; Auto].
Unfold FtoRradix; Apply eqExpMax; Auto.
Apply RoundedModeBounded with radix := radix P := P r := (Rmult p q); Auto; Auto.
Apply maxFbounded; Auto.
Apply Zle_trans with 1 := H'3; Auto with zarith.
Replace (FtoR
           radix (Float (vNum b) (Zplus precision (Zplus (Fexp p) (Fexp q)))))
     with (Rmult
             radix
             (Float (vNum b) (Zplus (pred precision) (Zplus (Fexp p) (Fexp q))))).
Rewrite Fabs_correct; Auto with zarith.
(Unfold FtoRradix;
  Apply RoundedModeMultAbs
       with b := b precision := precision P := P r := (Rmult p q)); Auto.
Apply maxFbounded; Auto with zarith.
Rewrite Rabsolu_mult; Auto.
Apply Rle_trans
     with (FtoR
             radix (Fmult (Float (vNum b) (Fexp p)) (Float (vNum b) (Fexp q)))).
Rewrite Fmult_correct; Auto with arith.
Apply Rle_Rmult_comp; Auto with real.
Rewrite <- (Fabs_correct radix); Auto with arith; Apply maxMax; Auto with zarith.
Rewrite <- (Fabs_correct radix); Auto with arith; Apply maxMax; Auto with zarith.
Unfold Fmult FtoR; Simpl; Auto.
Rewrite powerRZ_add with n := (inject_nat (pred precision));
 Auto with real arith.
Repeat Rewrite <- Rmult_assoc.
Repeat Rewrite [z:Z] [x:R](Rmult_sym x (powerRZ radix z)); Auto.
Apply Rle_monotony; Auto with real arith.
Rewrite <- Rmult_assoc.
Rewrite [x:R](Rmult_sym x radix).
Rewrite <- powerRZ_Zs; Auto with real arith.
Replace (Zs (pred precision)) with (inject_nat precision).
Rewrite Rmult_IZR; Auto.
Apply Rle_Rmult_comp; Auto with real arith.
Replace R0 with (INR O); Try Rewrite <- INR_IZR_INZ; Auto with real arith.
Replace R0 with (INR O); Try Rewrite <- INR_IZR_INZ; Auto with real arith.
Rewrite pGivesBound; Auto.
Rewrite powerRZ_is_exp; Auto with real arith.
Try Rewrite <- INR_IZR_INZ; Apply Rle_INR; Auto with arith.
Case (exp radix precision); Simpl; Auto with arith; Intros n;
 Rewrite <- minus_n_O; Auto.
Rewrite inj_pred; Auto with arith zarith.
Unfold FtoRradix FtoR; Simpl.
Repeat Rewrite (Rmult_sym (IZR (vNum b))).
Rewrite <- Rmult_assoc.
Rewrite <- powerRZ_Zs; Auto with real arith.
Rewrite inj_pred; Auto with arith zarith.
Replace (Zs (Zplus (Zpred precision) (Zplus (Fexp p) (Fexp q))))
     with (Zplus precision (Zplus (Fexp p) (Fexp q))); Auto; Unfold Zs Zpred;
 Ring.
Qed.

Theorem RepMult:
  (P:?) (RoundedModeP b radix P) ->
  (p, q, f:float)
  (Fbounded b p) ->
  (Fbounded b q) ->
  (Rle R0 p) ->
  (Rle R0 q) ->
  (Zle (Zopp (dExp b)) (Zplus (Fexp p) (Fexp q))) -> (P (Rmult p q) f) ->
  (Ex [r:float]
  r==(Rminus (Rmult p q) f) /\
  ((Fbounded b r) /\ (Fexp r)=(Zplus (Fexp p) (Fexp q)))).
Intros.
Generalize minRepMult maxRepMult; Intros.
Cut (MinOrMaxP b radix P); [Intros | Intuition].
Case (H8 (Rmult p q) f); Auto; Intros.
Qed.

Theorem multExactExp_aux:
  (n1, n2:Z)
  (Zle (Zabs n1) (vNum b)) ->
  (Zle (Zabs n2) (vNum b)) ->
  (Ex [ny:Z]
  (Ex [ey:Z]
  <R> (Rmult n1 n2)==(Rmult ny (powerRZ radix ey)) /\ (Zle (Zabs ny) (vNum b)))) ->
  (Ex [nx:Z]
  (Ex [ex:Z]
  <R> (Rmult n1 n2)==(Rmult nx (powerRZ radix ex)) /\
  ((Zle (Zabs nx) (vNum b)) /\ ((Zle ZERO ex) /\ (Zle ex precision))))).
Intros.
Elim H1; Intros ny E; Elim E; Intros ey E0; Clear E H1.
Elim E0; Intros; Clear E0.
Case (Zle_or_lt ZERO ey); Intros.
Case (Zle_or_lt ey precision); Intros.
Exists ny; Exists ey.
Repeat (Split; Auto).
Exists (Zmult ny (Zpower_nat radix (absolu (Zminus ey precision))));
 Exists (inject_nat precision).
Repeat (Split; Auto with zarith).
Replace (IZR (Zmult ny (Zpower_nat radix (absolu (Zminus ey precision)))))
     with (Rmult ny (powerRZ radix (Zminus ey precision))).
Unfold Zminus; Rewrite powerRZ_add; Auto with arith real.
Rewrite Rmult_assoc; Rewrite Rmult_assoc.
Rewrite <- powerRZ_add; Auto with arith real.
Replace (powerRZ radix (Zplus (Zopp precision) precision)) with R1; Ring;
 Auto with real.
Rewrite H1; Auto with real.
Replace (Zplus (Zopp precision) precision) with ZERO; Simpl; Ring.
Rewrite Rmult_IZR.
Rewrite Zpower_nat_powerRZ_absolu; Auto with real.
Rewrite <- INR_IZR_INZ; Auto with real.
Apply Zsimpl_le_plus_l with (inject_nat precision).
Rewrite Zle_plus_minus.
Replace (Zplus precision ZERO) with (inject_nat precision); Auto with zarith.
Apply le_IZR.
Rewrite <- Rabsolu_Zabs.
Replace (IZR (Zmult ny (Zpower_nat radix (absolu (Zminus ey precision)))))
     with (Rmult ny (powerRZ radix (Zminus ey precision))).
2:Rewrite Rmult_IZR; Rewrite Zpower_nat_powerRZ_absolu; Auto with real;
   [Rewrite <- INR_IZR_INZ; Auto with real | Idtac].
2:Apply Zsimpl_le_plus_l with (inject_nat precision); Rewrite Zle_plus_minus;
   Replace (Zplus precision ZERO) with (inject_nat precision); Auto with zarith.
Unfold Zminus; Rewrite powerRZ_add; Auto with arith real.
Rewrite <- Rmult_assoc; Rewrite Rabsolu_mult; Rewrite <- H1.
Rewrite Rabsolu_mult; Rewrite Rmult_assoc.
Rewrite <- INR_IZR_INZ.
Replace (INR (vNum b)) with (Rmult (vNum b) R1); [Idtac | Ring].
Apply Rle_trans
     with (Rmult
             (vNum b)
             (Rmult (Rabsolu n2) (Rabsolu (powerRZ radix (Zopp precision)))));
 [Apply Rle_monotony_r; Auto with real | Idtac].
Apply Rmult_Rle_O; Auto with real.
Rewrite Rabsolu_Zabs; Auto with real zarith arith.
Rewrite INR_IZR_INZ; Apply Rle_IZR; Auto.
Apply Rle_monotony; Auto with real arith.
Apply Rlt_le; Replace R0 with (INR (0)); Auto with real; Apply Rlt_INR;
 Apply lt_vNum with radix precision; Auto with arith.
Apply Rle_trans with (Rmult (vNum b) (Rabsolu (powerRZ radix (Zopp precision))));
 [Apply Rle_monotony_r; Auto with real | Idtac].
Rewrite Rabsolu_Zabs; Auto with real zarith arith.
Rewrite INR_IZR_INZ; Apply Rle_IZR; Auto.
Rewrite Rabsolu_simpl1; [Idtac | Apply powerRZ_le; Auto with arith real].
Replace R1
     with (Rmult (powerRZ radix precision) (powerRZ radix (Zopp precision)));
 [Apply Rle_monotony_r | Idtac].
Apply powerRZ_le; Auto with arith real.
2:Rewrite <- powerRZ_add; Auto with arith real.
2:Replace (Zplus precision (Zopp precision)) with ZERO; Simpl; Ring.
Rewrite pGivesBound; Auto with arith real.
Replace (powerRZ radix precision) with (INR (exp radix precision));
 Auto with real arith.
Apply Rle_INR; Auto with arith.
Case (exp radix precision); Intros; Simpl; Auto with arith.
Rewrite <- (minus_n_O n); Auto with arith.
Apply sym_eqT; Apply powerRZ_is_exp; Auto with arith.
Exists (Zmult n1 n2); Exists ZERO.
Repeat (Split; Auto with zarith).
Simpl; Rewrite Rmult_IZR; Ring.
Apply le_IZR.
Rewrite <- Rabsolu_Zabs; Rewrite Rmult_IZR; Rewrite H1.
Rewrite Rabsolu_mult.
Rewrite (Rabsolu_simpl1 (powerRZ radix ey));
 [Idtac | Apply powerRZ_le; Auto with arith real].
Apply Rle_trans with (Rabsolu ny).
Pattern 2 (Rabsolu ny); Replace (Rabsolu ny) with (Rmult (Rabsolu ny) R1);
 [Apply Rle_monotony | Ring]; Auto with arith real.
Replace R1 with (powerRZ radix ZERO); [Apply Rle_powerRZ | Simpl];
 Auto with real arith zarith.
Apply Rlt_le; Auto with real arith.
Rewrite Rabsolu_Zabs; Auto with real zarith.
Qed.

Theorem Zabs_Fbounded1:
  (a:Z) (Zle (Zabs a) (vNum b)) ->(Zle (Zopp (vNum b)) a) /\ (Zle a (vNum b)).
Intros.
Case (Rle_or_lt ZERO a); Intros.
Cut (Zle ZERO a); [Intros | Apply le_IZR]; Auto.
Rewrite <- (Zabs_eq a); Auto; Split; Auto.
Apply Zle_trans with ZERO; Auto with arith zarith.
Rewrite Zabs_absolu; Auto with arith zarith.
Replace a with (Zopp (Zabs a)); [Split; Auto | Idtac].
Apply Zle_Zopp; Auto.
Apply Zle_trans with ZERO; Auto with arith zarith.
Replace ZERO with (Zopp ZERO); [Apply Zle_Zopp | Idtac]; Auto with zarith.
Rewrite Zabs_absolu; Auto with arith zarith.
Replace (Zabs a) with (Zopp a); Auto with zarith.
Rewrite Zabs_absolu; Rewrite <- absolu_Zopp.
Apply sym_eq; Apply inj_abs.
Replace ZERO with (Zopp ZERO); [Apply Zle_Zopp | Idtac]; Auto with zarith.
Apply Zlt_le_weak; Apply lt_IZR; Auto.
Qed.

Theorem multExactExp:
  (P:?) (RoundedModeP b radix P) ->
  (p, q, pq:float)
  (Fbounded b p) ->
  (Fbounded b q) ->
  (Rle R0 p) ->
  (Rle R0 q) ->
  (P (Rmult p q) pq) -> (Zle (Zopp (dExp b)) (Zplus (Fexp p) (Fexp q))) ->
  (ex
     float
     [r:float]
     (ex
        float
        [s:float]
        (Fbounded b r) /\
        ((Fbounded b s) /\
         (<R> r==pq /\
          (<R> s==(Rminus (Rmult p q) r) /\
           (<Z> (Fexp s)=(Zplus (Fexp p) (Fexp q)) /\
            ((Zle (Fexp s) (Fexp r)) /\
             (Zle (Fexp r) (Zplus precision (Zplus (Fexp p) (Fexp q))))))))))).
Intros.
Generalize RepMult multExpUpperBound; Intros.
LApply (H7 P);
 [Intros H'0; LApply (H'0 p q pq);
   [Intros H'4; LApply H'4;
     [Intros H'5; LApply H'5;
       [Intros H'6; LApply H'6;
         [Intros H'7; Try Exact H'7; Clear H'6 H'5 H'4 | Clear H'6 H'5 H'4] |
         Clear H'5 H'4] | Clear H'4] | Idtac] | Idtac]; Auto.
Elim H'7; Intros r E; Elim E; Intros H' H'1; Clear E H'7.
Elim H'1; Intros; Clear H'1 H'0 H7.
Case (Req_EM (Rminus (Rmult p q) pq) R0); Intros.
Clear H9 H8 H' r H6.
Case (Req_EM pq R0); Intros.
Cut (Fbounded b (Fzero (Zplus (Fexp p) (Fexp q)))); [Intros | Idtac].
Exists (Fzero (Zplus (Fexp p) (Fexp q)));
 Exists (Fzero (Zplus (Fexp p) (Fexp q))).
Simpl; Repeat (Split; Auto with zarith).
Rewrite H6; Unfold FtoRradix; Apply FzeroisReallyZero.
Unfold FtoRradix; Rewrite FzeroisReallyZero.
Generalize H7; Rewrite H6; Auto with real.
Repeat (Split; Auto); Simpl; Auto with arith zarith.
Generalize multExactExp_aux; Intros.
LApply (H8 (Fnum p) (Fnum q));
 [Intros H'1; LApply H'1;
   [Intros H'2; LApply H'2;
     [Intros H'3; Try Exact H'3; Clear H'2 H'1 | Clear H'2 H'1] | Clear H'1] |
   Idtac]; Auto.
Elim H'3; Intros nx E; Elim E; Intros ex E0; Elim E0; Intros H' H'0; Elim H'0;
 Intros H'1 H'2; Clear H'0 E0 E H'3.
Elim H'2; Intros; Clear H'2.
Exists (Float nx (Zplus ex (Zplus (Fexp p) (Fexp q))));
 Exists (Fzero (Zplus (Fexp p) (Fexp q))).
Simpl; Repeat (Split; Simpl; Auto with zarith).
LApply (Zabs_Fbounded1 nx); [Intros H'2; Elim H'2 | Idtac]; Auto.
LApply (Zabs_Fbounded1 nx); [Intros H'2; Elim H'2 | Idtac]; Auto.
Replace (FtoRradix pq) with (Rmult p q).
Unfold FtoRradix; Unfold FtoR; Simpl.
Rewrite powerRZ_add; Auto with arith real.
Rewrite <- Rmult_assoc; Rewrite <- H'; Rewrite powerRZ_add;
 [Ring | Auto with arith real].
Apply r_Rplus_plus with (Ropp pq); Ring; Ring; Auto with real.
Unfold FtoRradix; Rewrite FzeroisReallyZero.
Replace (FtoR radix (Float nx (Zplus ex (Zplus (Fexp p) (Fexp q)))))
     with (FtoR radix pq); Auto with real.
Replace (FtoR radix pq) with (Rmult p q).
Unfold FtoRradix; Unfold FtoR; Simpl.
Rewrite powerRZ_add; Auto with arith real.
Apply sym_eqT; Rewrite <- Rmult_assoc; Rewrite <- H'; Rewrite powerRZ_add;
 [Ring | Auto with arith real].
Fold FtoRradix; Apply r_Rplus_plus with (Ropp pq); Ring; Ring; Auto with real.
Exists (Fnum pq); Exists (Zminus (Fexp pq) (Zplus (Fexp p) (Fexp q))); Split.
Apply r_Rmult_mult with (powerRZ radix (Zplus (Fexp p) (Fexp q))).
2:Apply powerRZ_NOR; Auto with real arith.
Apply trans_eqT with (FtoRradix pq).
Apply trans_eqT with (Rmult p q).
Replace (Rmult p q) with (FtoRradix (Fmult p q));
 [Unfold Fmult; Unfold FtoRradix; Unfold FtoR; Simpl |
   Unfold FtoRradix; Apply Fmult_correct; Auto with real arith].
Rewrite Rmult_IZR; Ring.
Apply r_Rplus_plus with (Ropp pq); Ring; Ring; Auto with real.
Rewrite Rmult_sym; Rewrite Rmult_assoc; Rewrite <- powerRZ_add;
 Auto with real arith.
Replace (Zplus
           (Zminus (Fexp pq) (Zplus (Fexp p) (Fexp q)))
           (Zplus (Fexp p) (Fexp q))) with (Fexp pq);
 [Unfold FtoRradix; Unfold FtoR; Simpl; Auto with real | Ring].
Apply Fbounded_Zabs; Auto.
Apply RoundedModeBounded with radix P (Rmult p q); Auto.
Apply Fbounded_Zabs; Auto.
Apply Fbounded_Zabs; Auto.
LApply (H6 P);
 [Intros H'1; LApply (H'1 p q pq);
   [Intros H'5; LApply H'5;
     [Intros H'6; LApply H'6;
       [Intros H'7; LApply H'7;
         [Intros H'8; LApply H'8;
           [Intros H'9; LApply H'9;
             [Intros H'10; Try Exact H'10; Clear H'9 H'8 H'7 H'6 H'5 |
               Clear H'9 H'8 H'7 H'6 H'5] | Clear H'8 H'7 H'6 H'5] |
           Clear H'7 H'6 H'5] | Clear H'6 H'5] | Clear H'5] | Idtac] | Idtac];
 Auto.
Elim H'10; Intros s E; Elim E; Intros H'0 H'2; Clear E H'10.
Elim H'2; Intros; Clear H'2 H'1 H6.
Exists r; Exists s.
Repeat (Split; Auto).
Rewrite H8; Auto with real.
Apply Zlt_le_weak;
 Apply RoundedModeErrorExpStrict with b radix precision P (Rmult p q); Auto.
Cut (CompatibleP b radix P); [Intros | Intuition].
Apply H6 with (Rmult p q) pq; Auto with real.
Fold FtoRradix; Rewrite H'0; Auto with real.
Fold FtoRradix; Rewrite H'0; Auto with real.
Qed.
End FRoundP.
