
(****************************************************************************
                                                                             
          IEEE754  :  FroundPlus                                                     
                                                                             
          Laurent Thery                                                      
                                                                             
*****************************************************************************
*)
Require Export FroundProp.
Require Import Finduct.
Section FRoundP.
Variable b:Fbound.
Variable radix:nat.
Variable precision:nat.

Local FtoRradix := (FtoR radix).
Coercion FtoRradix : float >-> R.
Hypothesis radixMoreThanOne:(lt (S O) radix).
Hypothesis precisionGreaterThanOne:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp radix precision) (1)).

Theorem plusExpMin:
  (P:?) (RoundedModeP b radix P) ->
  (p, q, pq:float) (P (Rplus p q) pq) ->
  (Ex [s:float]
  (Fbounded b s) /\ (<R> s==pq /\ (Zle (Zmin (Fexp p) (Fexp q)) (Fexp s)))).
Intros P H' p q pq H'0.
Case (RoundedModeRep b radix precision) with p := (Fplus radix p q) q := pq
                                             P := P; Auto with float arith.
Rewrite Fplus_correct; Auto with float arith.
Simpl; Intros x H'1.
Case (eqExpLess ? radixMoreThanOne b) with p := pq
                                           q :=
                                           (Float x (Fexp (Fplus radix p q)));
 Auto.
Apply (RoundedModeBounded b radix) with P := P r := (Rplus p q); Auto.
Simpl; Intros x0 H'2; Elim H'2; Intros H'3 H'4; Elim H'4; Intros H'5 H'6;
 Clear H'4 H'2.
Exists x0; Split; [Idtac | Split]; Auto.
Unfold FtoRradix; Rewrite H'5; Auto.
Apply le_IZR; Auto.
Qed.

Theorem plusExpUpperBound:
  (P:?) (RoundedModeP b radix P) ->
  (p, q, pq:float) (P (Rplus p q) pq) -> (Fbounded b p) -> (Fbounded b q) ->
  (Ex [r:float]
  (Fbounded b r) /\ (<R> r==pq /\ (Zle (Fexp r) (Zs (Zmax (Fexp p) (Fexp q)))))).
Intros P H' p q pq H'0 H'1 H'2.
Replace (Zs (Zmax (Fexp p) (Fexp q)))
     with (Fexp (Float (vNum b) (Zs (Zmax (Fexp p) (Fexp q)))));
 [Idtac | Simpl; Auto].
Unfold FtoRradix; Apply eqExpMax; Auto.
Apply RoundedModeBounded with radix := radix P := P r := (Rplus p q);
 Auto with float arith.
Apply maxFbounded; Auto.
Apply Zle_trans with (Fexp p); [Intuition | Idtac].
Apply Zle_trans with (Zs (Fexp p)); [Auto with zarith | Idtac].
Generalize (ZmaxLe1 (Fexp p) (Fexp q)); Auto with zarith.
Replace (FtoR radix (Float (vNum b) (Zs (Zmax (Fexp p) (Fexp q)))))
     with (Rmult radix (Float (vNum b) (Zmax (Fexp p) (Fexp q)))).
Rewrite Fabs_correct; Auto with zarith.
(Unfold FtoRradix;
  Apply RoundedModeMultAbs
       with b := b precision := precision P := P r := (Rplus p q)); Auto.
Apply maxFbounded; Auto.
Apply Zle_trans with (Fexp p); [Intuition | Idtac].
Generalize (ZmaxLe1 (Fexp p) (Fexp q)); Auto with zarith.
Apply Rle_trans with (Rplus (Rabsolu p) (Rabsolu q)).
Apply Rabsolu_triang; Auto.
Apply Rle_trans
     with (Rmult (2) (FtoR radix (Float (vNum b) (Zmax (Fexp p) (Fexp q)))));
 Auto.
Cut (r:R)(Rmult (2) r)==(Rplus r r);
 [Intros tmp; Rewrite tmp; Clear tmp | Intros; Simpl; Ring].
Apply Rplus_le; Auto.
Rewrite <- (Fabs_correct radix); Auto with arith; Apply maxMax; Auto;
 Apply ZmaxLe1.
Rewrite <- (Fabs_correct radix); Auto with arith; Apply maxMax; Auto;
 Apply ZmaxLe2.
Apply Rle_Rmult_comp; Auto with real arith.
Replace R0 with (INR O); Auto with real arith.
Apply LeFnumZERO; Simpl; Auto; Replace ZERO with (inject_nat O);
 Auto with zarith.
Unfold FtoRradix FtoR; Simpl.
Rewrite powerRZ_Zs; Auto with real arith; Ring.
Qed.

Theorem plusExpBound:
  (P:?) (RoundedModeP b radix P) ->
  (p, q, pq:float) (P (Rplus p q) pq) -> (Fbounded b p) -> (Fbounded b q) ->
  (Ex [r:float]
  (Fbounded b r) /\
  (<R> r==pq /\
   ((Zle (Zmin (Fexp p) (Fexp q)) (Fexp r)) /\
    (Zle (Fexp r) (Zs (Zmax (Fexp p) (Fexp q))))))).
Intros P H' p q pq H'0 H'1 H'2.
Case (plusExpMin P H' ? ? ? H'0).
Intros r' H'3; Elim H'3; Intros H'4 H'5; Elim H'5; Intros H'6 H'7; Clear H'5 H'3.
Case (Zle_or_lt (Fexp r') (Zs (Zmax (Fexp p) (Fexp q)))); Intros Zl1.
Exists r'; Repeat (Split; Auto).
Case (plusExpUpperBound P H' ? ? ? H'0); Auto.
Intros r'' H'3; Elim H'3; Intros H'5 H'8; Elim H'8; Intros H'9 H'10;
 Clear H'8 H'3.
Exists (Fshift
          radix (absolu (Zminus (Fexp r') (Zs (Zmax (Fexp p) (Fexp q))))) r');
 Split.
Apply FboundedShiftLess with n := (absolu (Zminus (Fexp r') (Fexp r''))); Auto.
Apply ZleLe; Auto.
Repeat Rewrite <- Zabs_absolu.
Repeat Rewrite Zabs_eq; Auto with zarith.
Rewrite FshiftCorrectInv; Auto.
Apply trans_eqT with (FtoRradix pq); Auto.
Apply Zle_trans with 1 := H'10; Auto with zarith.
Split.
Unfold FtoRradix; Rewrite FshiftCorrect; Auto.
Split.
Simpl.
Repeat Rewrite inj_abs; Auto with zarith arith.
Apply Zle_trans with (Zmax (Fexp p) (Fexp q)); Auto with zarith.
Apply Zmin_Zmax; Auto.
Simpl.
Repeat Rewrite inj_abs; Auto with zarith arith.
Qed.

Theorem minusRoundRep:
  (P:?) (RoundedModeP b radix P) ->
  (p, q, qmp, qmmp:float)
  (Rle R0 p) ->
  (Rle p q) -> (P (Rminus q p) qmp) -> (Fbounded b p) -> (Fbounded b q) ->
  (Ex [r:float] (Fbounded b r) /\ r==(Rminus q qmp)).
Intros P H' p q qmp H'0 H'1 H'2 H'3 H'4 H'5.
Case (Rle_or_lt (Rmult (Rinv (2)) q) p); Intros Rle1.
Exists p; Split; Auto.
Replace (FtoRradix qmp) with (FtoRradix (Fminus radix q p)).
Rewrite (Fminus_correct radix); Auto with arith; Unfold FtoRradix; Ring.
Apply (RoundedModeProjectorIdemEq b radix precision) with P := P; Auto.
Rewrite <- Fopp_Fminus.
Apply oppBounded; Auto.
Apply Sterbenz; Auto.
Apply Rle_trans with (FtoRradix q); Auto with real.
Apply Rledouble; Auto.
Apply Rle_trans with (FtoRradix p); Auto with real.
Cut (CompatibleP b radix P); [Intros Cp | Intuition].
Apply (Cp (Rminus q p) (Fminus radix q p) qmp); Auto.
Rewrite (Fminus_correct radix); Auto with arith.
Apply RoundedModeBounded with radix := radix P := P r := (Rminus q p); Auto;
 Auto.
Exists (Fminus radix q qmp); Split.
Rewrite <- Fopp_Fminus.
Apply oppBounded; Auto.
Apply Sterbenz; Auto.
Apply RoundedModeBounded with radix := radix P := P r := (Rminus q p); Auto;
 Auto.
Case MaxEx with r := (Rmult (Rinv (2)) (FtoR radix q)) 3 := pGivesBound;
 Auto with arith.
Intros max H'6.
Apply Rle_trans with (FtoRradix max); [Intuition | Idtac].
Apply (RleBoundRoundl b radix precision) with P := P r := (Rminus q p); Auto;
 Fold FtoRradix.
Intuition.
Case MinEx with r := (Rmult (Rinv (2)) (FtoR radix q)) 3 := pGivesBound;
 Auto with arith.
Intros min H'7.
Replace (FtoRradix max) with (Rminus q min).
Apply Rle_anti_compatibility with r := (Ropp q).
Cut (p, q:R)(Rplus (Ropp p) (Rminus p q))==(Ropp q);
 [Intros tmp; Repeat Rewrite tmp; Clear tmp | Intros; Ring].
Apply Rle_Ropp.
Case H'7.
Intros H'8 H'9; Elim H'9; Intros H'10 H'11; Apply H'11; Clear H'9; Auto.
Apply Rlt_le; Auto.
Unfold FtoRradix;
 Rewrite (div2IsBetween b radix precision) with 5 := H'7 6 := H'6; Auto.
Ring.
Apply Rle_trans with (FtoRradix q); Auto with real.
Apply (RleBoundRoundr b radix precision) with P := P r := (Rminus q p); Auto;
 Fold FtoRradix.
Apply Rle_anti_compatibility with r := (Ropp q).
Cut (p, q:R)(Rplus (Ropp p) (Rminus p q))==(Ropp q);
 [Intros tmp; Repeat Rewrite tmp; Clear tmp | Intros; Ring].
Replace (Rplus (Ropp q) q) with (Ropp R0); [Auto with real | Ring].
Apply Rle_trans with (FtoRradix q); Auto with real.
Apply Rledouble; Auto.
Apply Rle_trans with (FtoRradix p); Auto with real.
Apply (Fminus_correct radix); Auto with arith.
Qed.

Theorem radixRangeBoundExp:
  (p, q:float)
  (Fcanonic radix b precision p) ->
  (Fcanonic radix b precision q) ->
  (Rle R0 p) -> (Rlt p q) -> (Rlt q (Rmult radix p)) ->
  (Fexp p)=(Fexp q) \/ (Zs (Fexp p))=(Fexp q).
Intros p q H' H'0 H'1 H'2 H'3.
Case (FcanonicLtPos ? radixMoreThanOne b precision) with p := p q := q;
 Auto with arith.
2:Intros H'4; Elim H'4; Intros H'5 H'6; Clear H'4; Auto.
Intros H'4; Right.
Casec H'; Intros H'.
Case (FcanonicLtPos ? radixMoreThanOne b precision) with p := q
                                                         q :=
                                                         (Float
                                                            (Fnum p)
                                                            (Zs (Fexp p)));
 Auto with arith.
Left.
Repeat Split; Simpl.
Intuition.
Intuition.
Apply Zle_trans with (Fexp p); Auto with zarith; Intuition.
Unfold Fdigit; Simpl.
Intuition.
Apply Rle_trans with (FtoRradix p); Auto; Apply Rlt_le; Auto.
Unfold FtoR; Simpl.
Rewrite powerRZ_Zs; Auto with real arith; Auto.
Rewrite <- Rmult_assoc; Rewrite [x:R] [y:nat](Rmult_sym x y);
 Rewrite Rmult_assoc; Auto.
Simpl; Intros; Apply Zle_antisym; Auto with zarith.
Simpl; Auto.
Intros H'5; Elim H'5; Intros H'6 H'7; Auto.
Case (FcanonicLtPos ? radixMoreThanOne b precision) with p := q
                                                         q :=
                                                         (Float
                                                            (nNormMin
                                                               radix precision)
                                                            (Zs (Fexp p)));
 Auto with arith.
Left; Repeat Split; Simpl.
Apply Zle_trans with ZERO; Auto with zarith.
Auto with zarith float.
Case H'.
Intros H'5 H'6; Elim H'6; Intros H'7 H'8; Rewrite H'7; Clear H'6;
 Auto with zarith.
Unfold Fdigit; Simpl; Auto with float.
Rewrite absolu_INR; Auto with float.
Apply digitnNormMin with b := b; Auto with arith.
Apply Rle_trans with 1 := H'1; Apply Rlt_le; Auto.
Apply Rlt_trans with 1 := H'3.
Unfold FtoR; Simpl.
Rewrite powerRZ_Zs; Auto with real arith; Auto.
Rewrite <- Rmult_assoc; Rewrite [x:R] [y:nat](Rmult_sym x y);
 Rewrite Rmult_assoc; Auto.
Apply Rlt_monotony; Auto with real arith.
Case H'.
Intros H'5 H'6; Elim H'6; Intros H'7 H'8; Rewrite H'7; Clear H'6.
Change (Rlt p (firstNormalPos radix b precision)).
Apply (FsubnormalLtFirstNormalPos radix); Auto with arith.
Simpl.
Auto with zarith.
Simpl.
Intros H'5; Elim H'5; Intros H'6 H'7; Rewrite H'6; Clear H'5; Auto.
Qed.

Theorem ExactMinusIntervalAux:
  (P:?) (RoundedModeP b radix P) ->
  (p, q:float)
  (Rlt R0 p) ->
  (Rlt (Rmult (2) p) q) ->
  (Fcanonic radix b precision p) ->
  (Fcanonic radix b precision q) ->
  (Ex [r:float] (Fbounded b r) /\ r==(Rminus q p)) ->
  (r:float)
  (Fcanonic radix b precision r) -> (Rlt (Rmult (2) p) r) -> (Rle r q) ->
  (Ex [r':float] (Fbounded b r') /\ r'==(Rminus r p)).
Intros P H' p q H'0 H'1 H'2 H'3 H'4 r H'5 H'6 H'7.
Cut (Rle R0 p); [Intros Rle0 | Apply Rlt_le; Auto].
Cut (Rle R0 r); [Intros Rle1 | Apply Rle_trans with (Rmult (2) p); Auto].
2:Apply Rle_trans with (FtoRradix p); Auto with float arith.
2:Apply Rledouble; Auto.
2:Apply Rlt_le; Auto.
Generalize H'6; Clear H'6; Pattern r;
 Apply (FinductNeg b radix precision) with p := q; Auto with arith.
Apply Rle_trans with (FtoRradix r); Auto.
Intros q0 H'6 H'8 H'9 H'10 H'11.
Elim H'10; [Intros r' E; Elim E; Intros H'13 H'14; Clear E H'10 | Clear H'10];
 Auto.
2:Apply Rlt_trans with 1 := H'11; Auto; Apply (FPredLt b radix precision);
   Auto with arith.
Cut (Rle R0 (Fnormalize radix b precision r')); [Intros Rle2 | Idtac].
2:Rewrite (FnormalizeCorrect radix); Auto with arith.
2:Unfold FtoRradix in H'14; Rewrite H'14.
2:Apply Rle_anti_compatibility with r := (FtoR radix p).
2:Replace (Rplus (FtoR radix p) R0) with (FtoR radix p); [Idtac | Ring].
2:Replace (Rplus (FtoR radix p) (Rminus (FtoR radix q0) (FtoR radix p)))
       with (FtoR radix q0); [Idtac | Ring].
2:Apply Rle_trans with (Rmult (2) p); Auto.
2:Apply Rledouble; Auto with real arith.
2:Apply Rlt_le; Apply Rlt_trans with 1 := H'11; Auto with float.
2:Apply (FPredLt b radix precision); Auto with arith.
Cut (Rlt (Fnormalize radix b precision r') q0); [Intros Rle3 | Idtac].
2:Rewrite (FnormalizeCorrect radix); Auto with arith.
2:Unfold FtoRradix in H'14; Rewrite H'14.
2:Apply Rlt_anti_compatibility with r := (Ropp q0).
2:Replace (Rplus (Ropp q0) (Rminus (FtoR radix q0) (FtoR radix p)))
       with (Ropp p); [Idtac | Unfold FtoRradix; Ring; Ring].
2:Replace (Rplus (Ropp q0) q0) with (Ropp R0); [Auto with real | Ring].
Case radixRangeBoundExp with p := (Fnormalize radix b precision r') q := q0;
 Auto with float arith; Fold FtoRradix.
Rewrite (FnormalizeCorrect radix); Auto with arith.
Apply Rlt_le_trans with (Rmult (2) r'); Auto.
Rewrite H'14.
Rewrite Rmult_Rminus_distr.
Pattern 1 (FtoRradix q0);
 (Replace (FtoRradix q0) with (Rminus (Rmult (2) q0) q0); [Idtac | Simpl; Ring]).
Unfold Rminus; Apply Rlt_compatibility; Apply Rlt_Ropp.
Apply Rlt_trans with 1 := H'11.
Apply (FPredLt b radix precision); Auto with arith.
Apply Rle_monotony_r; Auto with real arith.
(Unfold FtoRradix in Rle2; Rewrite (FnormalizeCorrect radix) in Rle2);
 Auto with arith.
Intros H'10.
Case (FcanonicLtPos ? radixMoreThanOne b precision) with p :=
                                                         (Fnormalize
                                                            radix b precision r')
                                                         q := q0;
 Auto with arith.
Apply FnormalizeCanonic; Auto with arith.
Intros; Contradict H'10; Auto with zarith.
Intros H'12; Elim H'12; Intros H'15 H'16; Clear H'12.
Exists (Float
          (Zpred (Fnum (Fnormalize radix b precision r')))
          (Fexp (Fnormalize radix b precision r'))).
Split.
Cut (Fbounded b (Fnormalize radix b precision r')); [Intros Fb0 | Idtac].
Repeat Split; Simpl.
Apply Zle_trans with (Zpred ZERO).
Replace (Zpred ZERO) with (Zopp (1)); [Idtac | Simpl; Auto].
Apply Zle_Zopp; Apply inj_le.
Apply lt_le_S.
Apply lt_vNum with radix := radix precision := precision; Auto with arith.
Cut (Zle ZERO (Fnum (Fnormalize radix b precision r')));
 [Unfold Zpred; Auto with zarith | Idtac].
Apply (LeR0Fnum radix); Auto.
Apply Zle_trans with (Fnum (Fnormalize radix b precision r'));
 [Unfold Zpred; Auto with zarith | Intuition].
Intuition.
Apply FcanonicBound with radix := radix precision := precision; Auto.
Apply FnormalizeCanonic; Auto with arith.
Replace (Float
           (Zpred (Fnum (Fnormalize radix b precision r')))
           (Fexp (Fnormalize radix b precision r')))
     with (Fminus
             radix (Fnormalize radix b precision r')
             (Fminus radix q0 (FPred b radix precision q0))).
Repeat Rewrite (Fopp_correct radix); Repeat Rewrite (Fminus_correct radix);
 Auto with arith.
Rewrite (FnormalizeCorrect radix); Auto with arith.
Unfold FtoRradix in H'14; Rewrite H'14.
Unfold FtoRradix; Ring; Ring.
Replace (FPred b radix precision q0) with (Float (Zpred (Fnum q0)) (Fexp q0));
 Auto.
Unfold Fminus Fopp Fplus; Simpl.
Repeat Rewrite Zmin_n_n; Repeat Rewrite <- Zminus_n_n; Simpl; Auto.
Rewrite H'10.
Repeat Rewrite Zmin_n_n; Repeat Rewrite <- Zminus_n_n; Simpl; Auto.
Rewrite Zpower_nat_O; Repeat Rewrite Zmult_n_1.
Apply floatEq; Simpl; Auto; Unfold Zpred; Ring.
Case (Z_eq_dec (Fnum q0) (nNormMin radix precision)); Intros Zeq2.
Case (Z_eq_dec (Fexp q0) (Zopp (dExp b))); Intros Zeq1.
Rewrite Zeq1; Rewrite Zeq2; Rewrite <- (FPredSimpl3 b radix); Auto with arith;
 Rewrite <- Zeq1; Rewrite <- Zeq2; Auto.
Contradict H'16.
Apply Zle_not_lt.
Rewrite Zeq2.
Rewrite <- (inj_abs (Fnum (Fnormalize radix b precision r'))).
Apply inj_le.
Apply pNormal_absolu_min with b := b; Auto with arith.
Cut (Fcanonic radix b precision (Fnormalize radix b precision r'));
 [Intros Ca1; Case Ca1; Auto | Auto with float arith].
Intros H'12; Case Zeq1; Rewrite <- H'10; Intuition.
Apply (LeR0Fnum radix); Auto.
Rewrite FPredSimpl4; Auto.
Contradict H'16; Rewrite H'16.
Apply Zle_not_lt.
Cut (Fbounded b (Fnormalize radix b precision r')); [Intuition | Idtac].
Apply (FcanonicBound radix b precision); Auto with float arith.
Intros H'10.
Case (Z_eq_dec (Fnum q0) (nNormMin radix precision)); Intros Zeq2.
Exists (Float
          (Zpred (Fnum (Fnormalize radix b precision r')))
          (Fexp (Fnormalize radix b precision r'))).
Cut (Fbounded b (Fnormalize radix b precision r')); [Intros Fb1 | Idtac].
Repeat Split; Simpl.
Apply Zle_trans with (Zpred ZERO).
Replace (Zpred ZERO) with (Zopp (1)); [Idtac | Simpl; Auto].
Apply Zle_Zopp; Apply inj_le.
Apply lt_le_S.
Apply lt_vNum with radix := radix precision := precision; Auto with arith.
Cut (Zle ZERO (Fnum (Fnormalize radix b precision r')));
 [Unfold Zpred; Auto with zarith | Idtac].
Apply (LeR0Fnum radix); Auto.
Apply Zle_trans with (Fnum (Fnormalize radix b precision r'));
 [Unfold Zpred; Auto with zarith | Intuition].
Intuition.
Rewrite FPredSimpl2; Auto with arith.
Rewrite <- H'10.
Cut (z:Z)(Zpred (Zs z))=z;
 [Intros tmp; Rewrite tmp; Clear tmp | Intros; Unfold Zs Zpred; Ring].
Unfold FtoRradix FtoR; Simpl.
Cut (x:Z)(Zpred x)=(Zminus x (1));
 [Intros tmp; Rewrite tmp; Clear tmp | Intros; Unfold Zpred; Simpl; Ring].
Rewrite Rminus_IZR; Auto.
Rewrite [x, y:?](Rmult_sym (Rminus x y)); Rewrite Rmult_Rminus_distr;
 Repeat Rewrite [x:?] [y:?](Rmult_sym (powerRZ x y)).
Replace (Rmult
           (Fnum (Fnormalize radix b precision r'))
           (powerRZ radix (Fexp (Fnormalize radix b precision r'))))
     with (FtoRradix (Fnormalize radix b precision r')).
Rewrite (FnormalizeCorrect radix); Auto.
Unfold FtoRradix in H'14; Rewrite H'14.
Unfold FtoR; Simpl.
Pattern 1 (Fexp q0); Rewrite <- H'10.
Rewrite Zeq2; Rewrite powerRZ_Zs; Auto with real arith.
Unfold nNormMin.
Rewrite <- Rmult_assoc.
Repeat Rewrite <- INR_IZR_INZ; Rewrite <- Rmult_INR.
Replace (mult (exp radix (pred precision)) radix) with (exp radix precision).
Rewrite pGivesBound.
Rewrite minus_INR; Auto with arith.
Ring.
Generalize precisionGreaterThanOne; Case precision; Simpl; Auto with arith;
 Intros tmp; Inversion tmp.
Auto.
Red; Intros H'12;
 Absurd (Zle (Zopp (dExp b)) (Fexp (Fnormalize radix b precision r'))).
Apply Zlt_not_le.
Rewrite <- H'12; Rewrite <- H'10; Unfold Zs; Auto with zarith.
Intuition.
Apply (FcanonicBound radix b precision); Auto with arith.
Apply FnormalizeCanonic; Auto with arith.
Exists (Float
          (Zminus (Fnum (Fnormalize radix b precision r')) radix)
          (Fexp (Fnormalize radix b precision r'))).
Cut (Fbounded b (Fnormalize radix b precision r')); [Intros Fb1 | Idtac].
Repeat Split; Simpl.
Apply Zle_trans with (Zopp radix).
Apply Zle_Zopp; Apply inj_le.
Apply le_trans with (nNormMin radix precision); Auto with float arith.
(Pattern 1 radix; Replace radix with (exp radix (1)));
 [Idtac | Simpl; Rewrite mult_sym; Simpl; Rewrite plus_sym; Simpl; Auto];
 Unfold nNormMin; Generalize precisionGreaterThanOne; Case precision;
 [Intros tmp; Inversion tmp | Idtac]; Simpl; Intros n; Case n; Simpl;
 Auto with arith; Intros tmp; Inversion tmp; Inversion H0.
Cut (Zle ZERO (Fnum (Fnormalize radix b precision r')));
 [Auto with zarith | Apply (LeR0Fnum radix)]; Auto.
Apply Zle_trans with (Fnum (Fnormalize radix b precision r'));
 [Auto with zarith | Intuition].
Intuition.
Rewrite FPredSimpl4; Auto with arith.
Rewrite <- H'10.
Unfold FtoRradix FtoR; Simpl.
Cut (x:Z)(Zpred x)=(Zminus x (1));
 [Intros tmp; Rewrite tmp; Clear tmp | Intros; Unfold Zpred; Simpl; Ring].
Repeat Rewrite Rminus_IZR; Auto.
Repeat Rewrite [x, y:?](Rmult_sym (Rminus x y));
 Repeat Rewrite Rmult_Rminus_distr;
 Repeat Rewrite [x:?] [y:?](Rmult_sym (powerRZ x y)).
Replace (Rmult
           (Fnum (Fnormalize radix b precision r'))
           (powerRZ radix (Fexp (Fnormalize radix b precision r'))))
     with (FtoRradix (Fnormalize radix b precision r')).
Rewrite (FnormalizeCorrect radix); Auto.
Unfold FtoRradix in H'14; Rewrite H'14.
Unfold FtoR; Simpl.
Rewrite <- H'10.
Repeat Rewrite <- INR_IZR_INZ.
Repeat Rewrite powerRZ_Zs; Auto with real arith.
Ring.
Unfold FtoR; Simpl; Auto.
Red; Intros H'12; Absurd (Zle ZERO (Fnum q0)); Auto.
Apply Zlt_not_le.
Rewrite H'12.
Replace ZERO with (Zopp O); [Apply Zlt_Zopp; Apply inj_lt | Simpl; Auto].
Apply (lt_vNum ? radixMoreThanOne b precision); Auto with arith.
Apply (LeR0Fnum radix); Auto.
Apply Rlt_le; Auto.
Apply (FcanonicBound radix b precision); Auto with arith.
Apply FnormalizeCanonic; Auto with arith.
Qed.

Theorem ExactMinusIntervalAux1:
  (P:?) (RoundedModeP b radix P) ->
  (p, q:float)
  (Rle R0 p) ->
  (Rle p q) ->
  (Fcanonic radix b precision p) ->
  (Fcanonic radix b precision q) ->
  (Ex [r:float] (Fbounded b r) /\ r==(Rminus q p)) ->
  (r:float) (Fcanonic radix b precision r) -> (Rle p r) -> (Rle r q) ->
  (Ex [r':float] (Fbounded b r') /\ r'==(Rminus r p)).
Intros P H' p q H'0 H'1 H'2 H'3 H'4 r H'5 H'6 H'7.
Casec H'0; Intros H'0.
Case (Rle_or_lt q (Rmult (2) p)); Intros Rl1.
Exists (Fminus radix r p); Split; Auto.
Rewrite <- Fopp_Fminus.
Apply oppBounded.
Apply Sterbenz; Auto.
Apply (FcanonicBound radix b precision); Auto with arith.
Apply (FcanonicBound radix b precision); Auto with arith.
Apply Rle_mult_anticompatibility with z := (INR (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_r; Auto with real arith;
 Rewrite Rmult_ne_r; Auto.
Apply Rle_trans with 1 := H'7; Auto.
Apply Rle_trans with 1 := H'6; Auto.
Apply Rledouble; Auto.
Apply Rle_trans with 2 := H'6; Apply Rlt_le; Auto.
Rewrite (Fminus_correct radix); Auto with arith.
Case (Rle_or_lt r (Rmult (2) p)); Intros Rl2.
Exists (Fminus radix r p); Split; Auto.
Rewrite <- Fopp_Fminus.
Apply oppBounded.
Apply Sterbenz; Auto.
Apply (FcanonicBound radix b precision); Auto with arith.
Apply (FcanonicBound radix b precision); Auto with arith.
Apply Rle_mult_anticompatibility with z := (INR (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_r; Auto with real arith;
 Rewrite Rmult_ne_r; Auto.
Apply Rle_trans with 1 := H'6; Auto.
Apply Rledouble; Auto.
Apply Rle_trans with 2 := H'6; Apply Rlt_le; Auto.
Rewrite (Fminus_correct radix); Auto with arith.
Apply ExactMinusIntervalAux with P := P q := q; Auto.
Exists r; Split; Auto.
Apply (FcanonicBound radix b precision); Auto with arith.
Rewrite <- H'0; Ring.
Qed.

Theorem ExactMinusInterval:
  (P:?) (RoundedModeP b radix P) ->
  (p, q:float)
  (Rle R0 p) ->
  (Rle p q) ->
  (Fbounded b p) ->
  (Fbounded b q) -> (Ex [r:float] (Fbounded b r) /\ r==(Rminus q p)) ->
  (r:float) (Fbounded b r) -> (Rle p r) -> (Rle r q) ->
  (Ex [r':float] (Fbounded b r') /\ r'==(Rminus r p)).
Intros P H' p q H'0 H'1 H'2 H'3 H'4 r H'5 H'6 H'7.
Replace (FtoRradix r) with (FtoRradix (Fnormalize radix b precision r));
 [Idtac | Apply (FnormalizeCorrect radix)]; Auto.
Replace (FtoRradix p) with (FtoRradix (Fnormalize radix b precision p));
 [Idtac | Apply (FnormalizeCorrect radix)]; Auto.
Apply ExactMinusIntervalAux1 with P := P q := (Fnormalize radix b precision q);
 Auto; Try Repeat Rewrite (FnormalizeCorrect radix); Auto;
 Apply FnormalizeCanonic; Auto with arith.
Qed.



(* Properties concerning LSB MSB *)

Theorem MSBroundLSB:
  (P:R -> float ->Prop) (RoundedModeP b radix P) ->
  (f1, f2:float) (P f1 f2) -> ~ (is_Fzero (Fminus radix f1 f2)) ->
  (Zlt (MSB radix (Fminus radix f1 f2)) (LSB radix f2)).
Intros P H' f1 f2 H'0 HZ.
Apply (oneExp_lt radix); Auto.
Apply Rlt_le_trans with (Fulp b radix precision f2).
Apply Rle_lt_trans with (FtoRradix (Fabs (Fminus radix f1 f2))).
Unfold FtoRradix; Apply MSB_le_abs; Auto.
Unfold FtoRradix; Rewrite Fabs_correct; Auto with arith; Rewrite Fminus_correct;
 Auto with arith.
Apply RoundedModeUlp with 4 := H'; Auto.
Apply FUlp_Le_LSigB; Auto.
Apply RoundedModeBounded with 1 := H' 2 := H'0; Auto.
Qed.

Theorem LSBMinus:
  (p, q:float) ~ (is_Fzero (Fminus radix p q)) ->
  (Zle (Zmin (LSB radix p) (LSB radix q)) (LSB radix (Fminus radix p q))).
Intros p q H'1.
Elim (LSB_rep_min radix) with p := p; Auto; Intros z E.
Elim (LSB_rep_min radix) with p := q; Auto; Intros z0 E0.
Replace (LSB radix (Fminus radix p q))
     with (LSB
             radix
             (Fminus radix (Float z (LSB radix p)) (Float z0 (LSB radix q)))).
Replace (Zmin (LSB radix p) (LSB radix q))
     with (Fexp (Fminus radix (Float z (LSB radix p)) (Float z0 (LSB radix q))));
 [Idtac | Simpl; Auto].
Apply Fexp_le_LSB; Auto.
Apply sym_equal; Apply LSB_comp; Auto.
Repeat Rewrite Fminus_correct; Auto with arith.
Unfold FtoRradix in E; Unfold FtoRradix in E0; Rewrite E; Rewrite E0; Auto.
Qed.

Theorem LSBPlus:
  (p, q:float) ~ (is_Fzero (Fplus radix p q)) ->
  (Zle (Zmin (LSB radix p) (LSB radix q)) (LSB radix (Fplus radix p q))).
Intros p q H'.
Elim (LSB_rep_min ? radixMoreThanOne p); Intros z E.
Elim (LSB_rep_min ? radixMoreThanOne q); Intros z0 E0.
Replace (LSB radix (Fplus radix p q))
     with (LSB
             radix
             (Fplus radix (Float z (LSB radix p)) (Float z0 (LSB radix q)))).
Replace (Zmin (LSB radix p) (LSB radix q))
     with (Fexp (Fplus radix (Float z (LSB radix p)) (Float z0 (LSB radix q))));
 [Idtac | Simpl; Auto].
Apply Fexp_le_LSB; Auto.
Apply sym_equal; Apply LSB_comp; Auto.
Repeat Rewrite Fplus_correct; Auto with arith.
Unfold FtoRradix in E; Unfold FtoRradix in E0; Rewrite E; Rewrite E0; Auto.
Qed.
End FRoundP.
