
(****************************************************************************
                                                                             
          IEEE754  :  FroundProp                                                     
                                                                             
          Laurent Thery  &  Sylvie Boldo                                                      
                                                                             
*****************************************************************************
*)
Require Export Fround.
Require Export MSB.
Section FRoundP.
Variable b:Fbound.
Variable radix:nat.
Variable precision:nat.

Local FtoRradix := (FtoR radix).
Coercion FtoRradix : float >-> R.
Hypothesis radixMoreThanOne:(lt (S O) radix).
Hypothesis precisionGreaterThanOne:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp radix precision) (1)).

Definition Fulp :=
   [p:float](powerRZ radix (Fexp (Fnormalize radix b precision p))).

Theorem FulpComp:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) -> <R> p==q ->(Fulp p)==(Fulp q).
Intros p q H' H'0 H'1; Unfold Fulp.
Rewrite FcanonicUnique with p := (Fnormalize radix b precision p)
                            q := (Fnormalize radix b precision q)
                            3 := pGivesBound; Auto with float arith.
Apply trans_eqT with (FtoR radix p).
Apply FnormalizeCorrect; Auto.
Apply trans_eqT with (FtoR radix q); Auto.
Apply sym_eqT; Apply FnormalizeCorrect; Auto.
Qed.

Theorem FulpLe: (p:float) (Fbounded b p) ->(Rle (Fulp p) (Float (1) (Fexp p))).
Intros p H'; Unfold Fulp FtoRradix FtoR Fnormalize; Simpl; AutoRewrite [Rsimpl].
Case (Z_zerop (Fnum p)); Simpl; Auto.
Intros H'0; Apply Rle_powerRZ; Auto with real arith.
Apply Rlt_le; Replace R1 with (INR (1)); Auto with real arith.
Intuition.
Intros H'0; Apply Rle_powerRZ; Auto with real arith.
Apply Rlt_le; Replace R1 with (INR (1)); Auto with real arith.
Auto with zarith.
Qed.

Theorem Fulp_zero:
  (x:float) (is_Fzero x) ->(Fulp x)==(powerRZ radix (Zopp (dExp b))).
Intros x; Unfold is_Fzero Fulp Fnormalize; Case (Z_zerop (Fnum x)); Simpl; Auto.
Intros H' H'0; Contradict H'; Auto.
Qed.

Theorem FulpSucCan:
  (p:float) (Fcanonic radix b precision p) ->
  (Rle (Rminus (FSucc b radix precision p) p) (Fulp p)).
Intros p H'.
Replace (Fulp p) with (powerRZ radix (Fexp p)).
2:Unfold Fulp; Replace (Fnormalize radix b precision p) with p; Auto;
   Apply sym_equal; Apply FcanonicUnique with 3 := pGivesBound; Auto with arith;
   Apply FnormalizeCanonic Orelse Apply FnormalizeCorrect; Auto with arith;
   Intuition.
Unfold FtoRradix; Rewrite <- Fminus_correct; Auto with arith.
Case (Z_eq_dec (Fnum p) (Zopp (nNormMin radix precision))); Intros H1'.
Case (Z_eq_dec (Fexp p) (Zopp (dExp b))); Intros H2'.
Rewrite FSuccDiff2; Auto with arith.
Unfold FtoR; Simpl; AutoRewrite [Rsimpl]; Auto with real.
Rewrite FSuccDiff3; Auto with arith.
Unfold FtoR; Simpl; AutoRewrite [Rsimpl]; Auto with real.
Apply Rlt_le; Apply Rlt_powerRZ; Auto with real zarith.
Replace R1 with (INR (1)); Auto with real arith.
Unfold Zpred; Auto with zarith.
Rewrite FSuccDiff1; Auto with arith.
Unfold FtoR; Simpl; AutoRewrite [Rsimpl]; Auto with real.
Qed.

Theorem FulpSuc:
  (p:float) (Fbounded b p) ->
  (Rle (Rminus (FNSucc b radix precision p) p) (Fulp p)).
Intros p H'.
Replace (Fulp p) with (Fulp (Fnormalize radix b precision p)).
Replace (FtoRradix p) with (FtoRradix (Fnormalize radix b precision p)).
Unfold FNSucc; Apply FulpSucCan; Auto with float arith.
Unfold FtoRradix; Apply FnormalizeCorrect; Auto.
Apply FulpComp; Auto with float arith.
Unfold FtoRradix; Apply FnormalizeCorrect; Auto.
Qed.

Theorem FulpPredCan:
  (p:float) (Fcanonic radix b precision p) ->
  (Rle (Rminus p (FPred b radix precision p)) (Fulp p)).
Intros p H'.
Replace (Fulp p) with (powerRZ radix (Fexp p)).
2:Unfold Fulp; Replace (Fnormalize radix b precision p) with p; Auto;
   Apply sym_equal; Apply FcanonicUnique with 3 := pGivesBound; Auto with arith;
   Apply FnormalizeCanonic Orelse Apply FnormalizeCorrect; Auto with arith;
   Intuition.
Unfold FtoRradix; Rewrite <- Fminus_correct; Auto with arith.
Case (Z_eq_dec (Fnum p) (nNormMin radix precision)); Intros H1'.
Case (Z_eq_dec (Fexp p) (Zopp (dExp b))); Intros H2'.
Rewrite FPredDiff2; Auto with arith.
Unfold FtoR; Simpl; AutoRewrite [Rsimpl]; Auto with real.
Rewrite FPredDiff3; Auto with arith.
Unfold FtoR; Simpl; AutoRewrite [Rsimpl]; Auto with real.
Apply Rlt_le; Apply Rlt_powerRZ; Auto with real zarith.
Replace R1 with (INR (1)); Auto with real arith.
Unfold Zpred; Auto with zarith.
Rewrite FPredDiff1; Auto with arith.
Unfold FtoR; Simpl; AutoRewrite [Rsimpl]; Auto with real.
Qed.

Theorem FulpPred:
  (p:float) (Fbounded b p) ->
  (Rle (Rminus p (FNPred b radix precision p)) (Fulp p)).
Intros p H'.
Replace (Fulp p) with (Fulp (Fnormalize radix b precision p)).
Replace (FtoRradix p) with (FtoRradix (Fnormalize radix b precision p)).
Unfold FNPred; Apply FulpPredCan; Auto with float arith.
Unfold FtoRradix; Apply FnormalizeCorrect; Auto.
Apply FulpComp; Auto with float arith.
Unfold FtoRradix; Apply FnormalizeCorrect; Auto.
Qed.

Theorem RoundedModeUlp:
  (P:?) (RoundedModeP b radix P) ->
  (p, q:float) (P p q) ->(Rlt (Rabsolu (Rminus p q)) (Fulp q)).
Intros P H' p q H'0.
Case (Req_EM p q); Intros Eq1.
Rewrite <- Eq1.
Replace (Rminus p p) with R0; [Idtac | Ring].
Rewrite Rabsolu_R0; Auto.
Unfold Fulp FtoRradix FtoR; Simpl; Auto with real arith.
Case H'.
Intros H'1 H'2; Elim H'2; Intros H'3 H'4; Elim H'4; Intros H'5 H'6;
 Case H'5 with 1 := H'0; Clear H'5 H'4 H'2; Intros H'5.
Rewrite Rabsolu_simpl1; Auto.
Cut (Fbounded b q); [Intros B0 | Intuition].
Apply Rlt_le_trans with 2 := (FulpSuc q B0).
Apply Rlt_anti_compatibility with r := (FtoR radix q).
Repeat Rewrite Rplus_Rminus; Auto.
Case (Rle_or_lt (FNSucc b radix precision q) p); Auto.
Intros H'2; Absurd (Rle (FNSucc b radix precision q) q); Auto.
Unfold FtoRradix; Auto with real float arith.
Case H'5; Auto.
Intros H'4 H'7; Elim H'7; Intros H'8 H'9; Apply H'9; Clear H'7; Auto.
Apply (FcanonicBound radix b precision); Auto with float arith.
Apply Rle_anti_compatibility with r := (FtoR radix q).
Repeat Rewrite Rplus_Rminus; Auto.
AutoRewrite [Rsimpl]; Intuition.
Rewrite Rabsolu_simpl2; Auto.
Rewrite Rminus_Ropp; Auto.
Cut (Fbounded b q); [Intros B0 | Intuition].
Apply Rlt_le_trans with 2 := (FulpPred q B0).
Apply Ropp_Rlt; Repeat Rewrite Rminus_Ropp.
Apply Rlt_anti_compatibility with r := (FtoR radix q).
Repeat Rewrite Rplus_Rminus; Auto.
Case (Rle_or_lt p (FNPred b radix precision q)); Auto.
Intros H'2; Absurd (Rle q (FNPred b radix precision q)); Auto.
Unfold FtoRradix; Auto with real float arith.
Case H'5; Auto.
Intros H'4 H'7; Elim H'7; Intros H'8 H'9; Apply H'9; Clear H'7; Auto.
Apply (FcanonicBound radix b precision); Auto with float arith.
Apply Rle_anti_compatibility with r := (FtoR radix q).
Repeat Rewrite Rplus_Rminus; Auto.
AutoRewrite [Rsimpl]; Intuition.
Qed.

Theorem RoundedModeUlp2:
  (P:?) (RoundedModeP b radix P) ->
  (p:R) (q:float) (P p q) ->(Rlt (Rabsolu (Rminus p q)) (Fulp q)).
Intros P H' p q H'0.
Case (Req_EM p q); Intros Eq1.
Rewrite <- Eq1.
Replace (Rminus p p) with R0; [Idtac | Ring].
Rewrite Rabsolu_R0; Auto.
Unfold Fulp FtoRradix FtoR; Simpl; Auto with real arith.
Case H'.
Intros H'1 H'2; Elim H'2; Intros H'3 H'4; Elim H'4; Intros H'5 H'6;
 Case H'5 with 1 := H'0; Clear H'5 H'4 H'2; Intros H'5.
Rewrite Rabsolu_simpl1; Auto.
Cut (Fbounded b q); [Intros B0 | Intuition].
Apply Rlt_le_trans with 2 := (FulpSuc q B0).
Apply Rlt_anti_compatibility with r := (FtoR radix q).
Repeat Rewrite Rplus_Rminus; Auto.
Case (Rle_or_lt (FNSucc b radix precision q) p); Auto.
Intros H'2; Absurd (Rle (FNSucc b radix precision q) q); Auto.
Unfold FtoRradix; Auto with real float arith.
Case H'5; Auto.
Intros H'4 H'7; Elim H'7; Intros H'8 H'9; Apply H'9; Clear H'7; Auto.
Apply (FcanonicBound radix b precision); Auto with float arith.
Apply Rle_anti_compatibility with r := (FtoR radix q).
Repeat Rewrite Rplus_Rminus; Auto.
AutoRewrite [Rsimpl]; Intuition.
Rewrite Rabsolu_simpl2; Auto.
Rewrite Rminus_Ropp; Auto.
Cut (Fbounded b q); [Intros B0 | Intuition].
Apply Rlt_le_trans with 2 := (FulpPred q B0).
Apply Ropp_Rlt; Repeat Rewrite Rminus_Ropp.
Apply Rlt_anti_compatibility with r := (FtoR radix q).
Repeat Rewrite Rplus_Rminus; Auto.
Case (Rle_or_lt p (FNPred b radix precision q)); Auto.
Intros H'2; Absurd (Rle q (FNPred b radix precision q)); Auto.
Unfold FtoRradix; Auto with real float arith.
Case H'5; Auto.
Intros H'4 H'7; Elim H'7; Intros H'8 H'9; Apply H'9; Clear H'7; Auto.
Apply (FcanonicBound radix b precision); Auto with float arith.
Apply Rle_anti_compatibility with r := (FtoR radix q).
Repeat Rewrite Rplus_Rminus; Auto.
AutoRewrite [Rsimpl]; Intuition.
Qed.

Theorem RoundedModeErrorExpStrict:
  (P:?) (RoundedModeP b radix P) ->
  (p, q:float)
  (x:R)
  (Fbounded b p) ->
  (Fbounded b q) -> (P x p) -> <R> q==(Rminus x p) -> ~ <R> q==R0 ->
  (Zlt (Fexp q) (Fexp p)).
Intros.
Apply Zlt_powerRZ with e := (INR radix).
Replace R1 with (INR (1)); Auto with real arith.
Apply Rle_lt_trans with (FtoRradix (Fabs q)).
Replace (powerRZ radix (Fexp q)) with (FtoRradix (Float (1) (Fexp q)));
 Unfold FtoRradix;
 [Apply RleFexpFabs; Auto with arith | Unfold FtoR; Simpl; Ring].
Rewrite (Fabs_correct radix); Auto with arith.
Fold FtoRradix; Rewrite H3.
Apply Rlt_le_trans with (Fulp p);
 [Apply RoundedModeUlp2 with P; Auto | Unfold Fulp].
Apply Rle_powerRZ.
Replace R1 with (INR (1)); Auto with real arith.
Apply FcanonicLeastExp with radix b precision; Auto with real arith.
Apply sym_eqT; Apply FnormalizeCorrect; Auto.
Apply FnormalizeCanonic; Auto with arith.
Qed.

Theorem RoundedModeProjectorIdem:
  (P:?) (p:float) (RoundedModeP b radix P) -> (Fbounded b p) ->(P p p).
Intros P p H' H.
Elim H'; Intros H'0 H'1; Elim H'1; Intros H'2 H'3; Elim H'3; Intros H'4 H'5;
 Clear H'3 H'1.
Case (H'0 p).
Intros x H'6.
Apply (H'2 p p x); Auto.
Apply sym_eqT; Apply (RoundedProjector ? ? P H'); Auto.
Qed.

Theorem RoundedModeBounded:
  (P:?) (r:R) (q:float) (RoundedModeP b radix P) -> (P r q) ->(Fbounded b q).
Intros P r q H' H'0.
Case H'.
Intros H'1 H'2; Elim H'2; Intros H'3 H'4; Elim H'4; Intros H'5 H'6;
 Case H'5 with 1 := H'0; Clear H'4 H'2; Intuition.
Qed.

Theorem RoundedModeProjectorIdemEq:
  (P:R -> float ->Prop)
  (p, q:float)
  (RoundedModeP b radix P) -> (Fbounded b p) -> (P (FtoR radix p) q) -><R> p==q.
Intros P p q H' H'0 H'1.
Cut (MinOrMaxP b radix P);
 [Intros Mn; Case (Mn p q); Auto; Intros Mn1 | Intuition].
Apply sym_eqT; Apply MinEq with 1 := Mn1; Auto.
Apply (RoundedModeProjectorIdem (isMin b radix)); Auto.
Apply MinRoundedModeP with precision := precision; Auto.
Apply sym_eqT; Apply MaxEq with 1 := Mn1; Auto.
Apply (RoundedModeProjectorIdem (isMax b radix)); Auto.
Apply MaxRoundedModeP with precision := precision; Auto.
Qed.

Theorem RoundedModeMult:
  (P:?) (RoundedModeP b radix P) ->
  (r:R) (q, q':float) (P r q) -> (Fbounded b q') -> (Rle r (Rmult radix q')) ->
  (Rle q (Rmult radix q')).
Intros P H' r q q' H'0 H'1.
Replace (Rmult radix q')
     with (FtoRradix (Float (Fnum q') (Zplus (Fexp q') (1)))).
Intros H'2; Case H'2.
Intros H'3; Case H'; Intros H'4 H'5; Elim H'5; Intros H'6 H'7; Elim H'7;
 Intros H'8 H'9; Apply H'9 with 1 := H'3; Clear H'7 H'5; Auto.
Apply RoundedModeProjectorIdem; Auto.
Apply FBoundedScale; Auto.
Intros H'3.
Replace (FtoRradix q) with (FtoRradix (Float (Fnum q') (Zplus (Fexp q') (1))));
 Auto with real.
Apply (RoundedProjector ? ? P H'); Auto.
Apply FBoundedScale; Auto.
Case H'.
Intros H'4 H'5; Elim H'5; Intros H'6 H'7; Clear H'5.
Apply (H'6 r (Float (Fnum q') (Zplus (Fexp q') (1))) q); Auto.
Apply RoundedModeBounded with P := P r := r; Auto.
Rewrite (FvalScale ? radixMoreThanOne b).
Rewrite powerRZ_1; Auto.
Qed.

Theorem RoundedModeMultLess:
  (P:?) (RoundedModeP b radix P) ->
  (r:R) (q, q':float) (P r q) -> (Fbounded b q') -> (Rle (Rmult radix q') r) ->
  (Rle (Rmult radix q') q).
Intros P H' r q q' H'0 H'1.
Replace (Rmult radix q')
     with (FtoRradix (Float (Fnum q') (Zplus (Fexp q') (1)))).
Intros H'2; Case H'2.
Intros H'3; Case H'; Intros H'4 H'5; Elim H'5; Intros H'6 H'7; Elim H'7;
 Intros H'8 H'9; Apply H'9 with 1 := H'3; Clear H'7 H'5; Auto.
Apply RoundedModeProjectorIdem; Auto.
Apply FBoundedScale; Auto.
Intros H'3.
Replace (FtoRradix q) with (FtoRradix (Float (Fnum q') (Zplus (Fexp q') (1))));
 Auto with real.
Apply (RoundedProjector ? ? P H'); Auto.
Apply FBoundedScale; Auto.
Unfold FtoRradix in H'3; Rewrite H'3; Auto.
Case H'.
Intros H'4 H'5; Elim H'5; Intros H'6 H'7; Clear H'5.
Unfold FtoRradix; Rewrite FvalScale; Auto.
Rewrite powerRZ_1; Auto.
Qed.

Theorem RleMinR0:
  (r:R) (min:float) (Rle R0 r) -> (isMin b radix r min) ->(Rle R0 min).
Intros r min H' H'0.
Rewrite <- (FzeroisZero radix b).
Case H'; Intros H'1.
Apply (MonotoneMin b radix) with p := (FtoRradix (Fzero (Zopp (dExp b)))) q := r;
 Auto.
Unfold FtoRradix; Rewrite (FzeroisZero radix b); Auto.
Apply (RoundedModeProjectorIdem (isMin b radix)); Auto.
Apply MinRoundedModeP with precision := precision; Auto with float.
Apply FboundedFzero; Auto.
Replace (FtoR radix (Fzero (Zopp (dExp b)))) with (FtoRradix min);
 Auto with real.
Apply sym_eqT; Apply (ProjectMin b radix).
Apply FboundedFzero; Auto.
Rewrite <- H'1 in H'0; Rewrite <- (FzeroisZero radix b) in H'0; Auto.
Qed.

Theorem RleRoundedR0:
  (P:R -> float ->Prop)
  (p:float) (r:R) (RoundedModeP b radix P) -> (P r p) -> (Rle R0 r) ->(Rle R0 p).
Intros P p r H' H'0 H'1.
Case H'.
Intros H'2 H'3; Elimc H'3; Intros H'3 H'4; Elimc H'4; Intros H'4 H'5;
 Case (H'4 r p); Auto; Intros H'6.
Apply RleMinR0 with r := r; Auto.
Apply Rle_trans with r; Auto; Intuition.
Qed.

Theorem RleMaxR0:
  (r:R) (max:float) (Rle r R0) -> (isMax b radix r max) ->(Rle max R0).
Intros r max H' H'0.
Rewrite <- (FzeroisZero radix b).
Case H'; Intros H'1.
Apply (MonotoneMax b radix) with q := (FtoRradix (Fzero (Zopp (dExp b)))) p := r;
 Auto.
Unfold FtoRradix; Rewrite FzeroisZero; Auto.
Apply (RoundedModeProjectorIdem (isMax b radix)); Auto.
Apply MaxRoundedModeP with precision := precision; Auto.
Apply FboundedFzero; Auto.
Replace (FtoR radix (Fzero (Zopp (dExp b)))) with (FtoRradix max);
 Auto with real.
Apply sym_eqT; Apply (ProjectMax b radix).
Apply FboundedFzero; Auto.
Rewrite H'1 in H'0; Rewrite <- (FzeroisZero radix b) in H'0; Auto.
Qed.

Theorem RleRoundedLessR0:
  (P:R -> float ->Prop)
  (p:float) (r:R) (RoundedModeP b radix P) -> (P r p) -> (Rle r R0) ->(Rle p R0).
Intros P p r H' H'0 H'1.
Case H'.
Intros H'2 H'3; Elimc H'3; Intros H'3 H'4; Elimc H'4; Intros H'4 H'5;
 Case (H'4 r p); Auto; Intros H'6.
Apply Rle_trans with r; Auto; Intuition.
Apply RleMaxR0 with r := r; Auto.
Qed.

Theorem PminPos:
  (p:float)
  (min:float)
  (Rle R0 p) -> (Fbounded b p) -> (isMin b radix (Rmult (Rinv (2)) p) min) ->
  (Ex [c:float] (Fbounded b c) /\ <R> c==(Rminus p min)).
Intros p min H' H'0 H'1.
Cut (Rle min (Rmult (Rinv (2)) p));
 [Intros Rl1; Casec Rl1; Intros Rl1 | Intuition].
Case (eqExpMax ? radixMoreThanOne b min p); Auto.
Intuition.
Rewrite Fabs_correct; Auto with arith.
Rewrite Rabsolu_simpl1; Auto.
Apply Rle_trans with (Rmult (Rinv (2)) p); Auto.
Apply Rlt_le; Auto.
Apply Rle_mult_anticompatibility with z := (INR (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_r; Auto with real; Rewrite Rmult_ne_r;
 Auto with real.
Apply RleMinR0 with r := (Rmult (Rinv (2)) p); Auto.
Apply Rle_mult_anticompatibility with z := (INR (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_r; Auto with real; Rewrite Rmult_ne_r;
 Rewrite Rmult_Or; Auto with real.
Intros min' H'2; Elim H'2; Intros H'3 H'4; Elim H'4; Intros H'5 H'6;
 Clear H'4 H'2.
Case (FboundNext ? radixMoreThanOne b precision) with p := min';
 Auto with arith; Fold FtoRradix.
Intros Smin H'2; Elim H'2; Intros H'4 H'7; Clear H'2.
Exists Smin; Split; Auto.
Rewrite H'7; Auto.
Unfold FtoRradix.
Rewrite <- H'5; Auto.
Replace (Float (Zs (Fnum min')) (Fexp min'))
     with (Float
             (Zminus
                (Fnum (Fshift radix (absolu (Zminus (Fexp p) (Fexp min'))) p))
                (Fnum min')) (Fexp min')); Auto.
Unfold FtoRradix FtoR; Simpl.
Rewrite Rminus_IZR.
Rewrite [x, y, z:R](Rmult_sym (Rminus x y) z); Rewrite Rmult_Rminus_distr;
 Repeat Rewrite [x:Z](Rmult_sym (powerRZ radix x)).
Rewrite Rmult_IZR.
Rewrite Zpower_nat_powerRZ_absolu; Auto with zarith.
Rewrite Rmult_assoc.
Rewrite <- INR_IZR_INZ; Rewrite <- (powerRZ_add radix); Auto with real arith.
Replace (Zplus (Zminus (Fexp p) (Fexp min')) (Fexp min')) with (Fexp p);
 [Auto | Ring].
Apply floatEq; Auto; Simpl.
Apply Zle_antisym.
Apply Zlt_n_Sm_le.
Apply Zsimpl_lt_plus_l with p := (Fnum min'); Auto.
Cut (x, y:Z)(Zplus x (Zminus y x))=y;
 [Intros tmp; Rewrite tmp; Clear tmp | Intros; Ring].
Replace (Zplus (Fnum min') (Zs (Zs (Fnum min'))))
     with (Zmult (2) (Zs (Fnum min'))).
Apply (Rlt_Float_Zlt radix) with r := (Fexp min'); Auto; Fold FtoRradix.
Replace (FtoRradix (Float (Zmult (2) (Zs (Fnum min'))) (Fexp min')))
     with (Rmult (2) (Float (Zs (Fnum min')) (Fexp min'))).
Rewrite <- H'7.
Replace (Float
           (Zmult
              (Fnum p) (Zpower_nat radix (absolu (Zminus (Fexp p) (Fexp min')))))
           (Fexp min'))
     with (Fshift radix (absolu (Zminus (Fexp p) (Fexp min'))) p).
(Unfold FtoRradix; Rewrite FshiftCorrect); Auto.
Apply Rlt_mult_anticompatibility with z := (Rinv (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_l; Auto with real; Rewrite Rmult_ne_r;
 Auto with real.
Case (Rle_or_lt Smin (Rmult (Rinv (2)) (FtoR radix p))); Auto.
Intros H'2; Absurd (Rlt min Smin).
Apply Rlt_not.
Case H'1; Auto.
Intros H'8 H'9; Elim H'9; Intros H'10 H'11; Apply H'11; Clear H'9; Auto.
Rewrite H'7; Unfold FtoRradix; Rewrite <- H'5; Auto.
Unfold FtoR; Simpl; Apply Rlt_monotony_exp; Auto with real zarith.
Unfold Fshift; Simpl.
Replace (Zminus (Fexp p) (absolu (Zminus (Fexp p) (Fexp min'))))
     with (Fexp min'); Auto.
Rewrite inj_abs; Auto.
Ring.
Auto with zarith.
Replace (FtoRradix (Float (Zmult (2) (Zs (Fnum min'))) (Fexp min')))
     with (Rmult (Zmult (2) (Zs (Fnum min'))) (powerRZ radix (Fexp min'))).
Rewrite Rmult_IZR; Auto.
Unfold FtoRradix FtoR; Simpl; Ring.
Simpl; Auto.
Replace (inject_nat (2)) with (Zs (Zs ZERO)).
Repeat Rewrite <- Zmult_Sm_n; Unfold Zs; Ring.
Simpl; Auto.
Apply Zlt_le_S; Auto.
Apply Zsimpl_lt_plus_l with p := (Fnum min'); Auto.
Cut (x, y:Z)(Zplus x (Zminus y x))=y;
 [Intros tmp; Rewrite tmp; Clear tmp | Intros; Ring].
Replace (Zplus (Fnum min') (Fnum min')) with (Zmult (2) (Fnum min')).
Apply (Rlt_Float_Zlt radix) with r := (Fexp min'); Auto; Fold FtoRradix.
Replace (FtoRradix (Float (Zmult (2) (Fnum min')) (Fexp min')))
     with (Rmult (2) (Float (Fnum min') (Fexp min'))).
Replace (Float
           (Zmult
              (Fnum p) (Zpower_nat radix (absolu (Zminus (Fexp p) (Fexp min')))))
           (Fexp min'))
     with (Fshift radix (absolu (Zminus (Fexp p) (Fexp min'))) p).
(Unfold FtoRradix; Rewrite FshiftCorrect); Auto.
Apply Rlt_mult_anticompatibility with z := (Rinv (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_l; Auto with real; Rewrite Rmult_ne_r;
 Auto with real.
Replace (FtoR radix (Float (Fnum min') (Fexp min'))) with (FtoR radix min); Auto.
Unfold Fshift; Simpl.
Replace (Zminus (Fexp p) (absolu (Zminus (Fexp p) (Fexp min'))))
     with (Fexp min'); Auto.
Rewrite inj_abs; Auto.
Ring.
Auto with zarith.
Replace (FtoRradix (Float (Zmult (2) (Fnum min')) (Fexp min')))
     with (Rmult (Zmult (2) (Fnum min')) (powerRZ radix (Fexp min'))).
Rewrite Rmult_IZR; Auto.
Unfold FtoRradix FtoR; Simpl; Ring.
Simpl; Auto.
Replace (inject_nat (2)) with (Zs (Zs ZERO)).
Repeat Rewrite <- Zmult_Sm_n; Unfold Zs; Ring.
Simpl; Auto.
Exists min; Split; Auto.
Intuition.
Rewrite Rl1.
Pattern 2 (FtoRradix p);
 Replace (FtoRradix p) with (Rmult (2) (Rmult (Rinv (2)) p)).
Simpl; Ring.
Rewrite <- Rmult_assoc; Rewrite Rinv_r; Auto with real; Rewrite Rmult_ne_r;
 Auto with real.
Qed.

Theorem div2IsBetweenPos:
  (p:float)
  (min, max:float)
  (Rle R0 p) ->
  (Fbounded b p) ->
  (isMin b radix (Rmult (Rinv (2)) p) min) ->
  (isMax b radix (Rmult (Rinv (2)) p) max) -><R> p==(Rplus min max).
Intros p min max P H' H'0 H'1; Apply Rle_antisym.
Apply Rle_anti_compatibility with r := (Ropp max).
Replace (Rplus (Ropp max) p) with (Rminus p max); [Idtac | Ring].
Replace (Rplus (Ropp max) (Rplus min max)) with (FtoRradix min); [Idtac | Ring].
Rewrite <- (Fminus_correct radix); Auto with arith.
Case H'0.
Intros H'2 H'3; Elim H'3; Intros H'4 H'5; Apply H'5; Clear H'3; Auto.
Apply Sterbenz; Auto.
Intuition.
Apply Rle_trans with (FtoRradix max); Auto.
Apply Rle_mult_anticompatibility with z := (INR (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_r; Auto with real; Rewrite Rmult_ne_r;
 Auto with real.
Apply Rledouble; Auto.
Apply Rle_trans with (FtoRradix min); Auto.
Apply RleMinR0 with r := (Rmult (Rinv (2)) p); Auto.
Apply Rle_mult_anticompatibility with z := (INR (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_r; Auto with real; Rewrite Rmult_ne_r;
 Rewrite Rmult_Or; Auto with real.
Apply Rle_trans with (Rmult (Rinv (2)) p); Intuition.
Case H'1.
Intros H'3 H'6; Elim H'6; Intros H'7 H'8; Apply H'8; Clear H'6; Auto.
Apply Rle_mult_anticompatibility with z := (INR (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_r; Auto with real; Rewrite Rmult_ne_r;
 Auto with real.
Apply Rle_mult_anticompatibility with z := (Rinv (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_l; Auto with real; Rewrite Rmult_ne_r;
 Auto with real.
Intuition.
Rewrite Fminus_correct; Auto with arith.
Apply Rle_anti_compatibility with r := (FtoR radix max).
Replace (Rplus (FtoR radix max) (Rminus (FtoR radix p) (FtoR radix max)))
     with (FtoR radix p); [Idtac | Ring].
Apply Rle_anti_compatibility with r := (Ropp (Rmult (Rinv (2)) p)).
Replace (Rplus (Ropp (Rmult (Rinv (2)) p)) (FtoR radix p))
     with (Rmult (Rinv (2)) p).
Replace (Rplus
           (Ropp (Rmult (Rinv (2)) p))
           (Rplus (FtoR radix max) (Rmult (Rinv (2)) p))) with (FtoR radix max);
 [Intuition | Ring].
Replace (FtoR radix p) with (Rmult (2) (Rmult (Rinv (2)) p)).
Simpl; Ring.
Rewrite <- Rmult_assoc; Rewrite Rinv_r; Auto with real.
Apply Rle_anti_compatibility with r := (Ropp min).
Replace (Rplus (Ropp min) p) with (Rminus p min); [Idtac | Ring].
Replace (Rplus (Ropp min) (Rplus min max)) with (FtoRradix max); [Idtac | Ring].
Case (PminPos p min); Auto.
Intros x H'2; Elim H'2; Intros H'3 H'4; Elim H'4; Clear H'2.
Case H'1.
Intros H'2 H'5; Elim H'5; Intros H'6 H'7; Apply H'7; Clear H'5; Auto.
Unfold FtoRradix in H'4; Rewrite H'4; Auto.
Apply Rle_anti_compatibility with r := (FtoR radix min).
Replace (Rplus (FtoR radix min) (Rminus (FtoR radix p) (FtoR radix min)))
     with (FtoR radix p); [Idtac | Ring].
Apply Rle_anti_compatibility with r := (Ropp (Rmult (Rinv (2)) p)).
Replace (Rplus (Ropp (Rmult (Rinv (2)) p)) (FtoR radix p))
     with (Rmult (Rinv (2)) p).
Replace (Rplus
           (Ropp (Rmult (Rinv (2)) p))
           (Rplus (FtoR radix min) (Rmult (Rinv (2)) p))) with (FtoR radix min);
 [Intuition | Ring].
Replace (FtoR radix p) with (Rmult (2) (Rmult (Rinv (2)) p)).
Simpl; Ring.
Rewrite <- Rmult_assoc; Rewrite Rinv_r; Auto with real.
Qed.

Theorem div2IsBetween:
  (p:float)
  (min, max:float)
  (Fbounded b p) ->
  (isMin b radix (Rmult (Rinv (2)) p) min) ->
  (isMax b radix (Rmult (Rinv (2)) p) max) -><R> p==(Rplus min max).
Intros p min max H' H'0 H'1; Case (Rle_or_lt R0 p); Intros H'2.
Apply div2IsBetweenPos; Auto.
Cut (x, y:R) (Ropp x)==(Ropp y) ->x==y;
 [Intros H'3; Apply H'3; Clear H'3 | Idtac].
Replace (Ropp (Rplus min max)) with (Rplus (Ropp max) (Ropp min));
 [Idtac | Ring].
Repeat Rewrite <- (Fopp_correct radix); Auto with float.
Apply div2IsBetweenPos; Auto with float.
Rewrite (Fopp_correct radix); Auto.
Replace R0 with (Ropp R0); Try Apply Rlt_le; Auto with real.
Replace (Rmult (Rinv (2)) (Fopp p)) with (Ropp (Rmult (Rinv (2)) p));
 Auto with float.
Rewrite (Fopp_correct radix); Auto; Ring; Ring.
Replace (Rmult (Rinv (2)) (Fopp p)) with (Ropp (Rmult (Rinv (2)) p));
 Auto with float.
Rewrite (Fopp_correct radix); Auto; Ring; Ring.
Intros x y H'3; Rewrite <- (Ropp_Ropp x); Rewrite <- (Ropp_Ropp y); Rewrite H'3;
 Auto.
Qed.

Theorem RoundedModeMultAbs:
  (P:?) (RoundedModeP b radix P) ->
  (r:R)
  (q, q':float)
  (P r q) -> (Fbounded b q') -> (Rle (Rabsolu r) (Rmult radix q')) ->
  (Rle (Rabsolu q) (Rmult radix q')).
Intros P H' r q q' H'0 H'1 H'2.
Case (Rle_or_lt R0 r); Intros Rl0.
Rewrite Rabsolu_simpl1; Auto.
Apply RoundedModeMult with P := P r := r; Auto.
Rewrite <- (Rabsolu_simpl1 r); Auto.
Apply RleRoundedR0 with P := P r := r; Auto.
Rewrite Rabsolu_simpl2; Auto.
Replace (Rmult radix q') with (Ropp (Rmult radix (Ropp q')));
 [Apply Rle_Ropp | Ring].
Rewrite <- (Fopp_correct radix).
Apply RoundedModeMultLess with P := P r := r; Auto.
Apply oppBounded; Auto.
Unfold FtoRradix; Rewrite Fopp_correct.
Rewrite <- (Ropp_Ropp r).
Replace (Rmult radix (Ropp (FtoR radix q'))) with (Ropp (Rmult radix q'));
 [Apply Rle_Ropp | Ring]; Auto.
Rewrite <- (Rabsolu_simpl2 r); Auto.
Apply Rlt_le; Auto.
Ring.
Apply RleRoundedLessR0 with P := P r := r; Auto.
Apply Rlt_le; Auto.
Qed.

Theorem isMinComp:
  (r1, r2:R)
  (min, max:float)
  (isMin b radix r1 min) ->
  (isMax b radix r1 max) -> (Rlt min r2) -> (Rlt r2 max) ->
  (isMin b radix r2 min).
Intros r1 r2 min max H' H'0 H'1 H'2; Split.
Intuition.
Split.
Apply Rlt_le; Auto.
Intros f H'3 H'4.
Case H'; Auto.
Intros H'5 H'6; Elim H'6; Intros H'7 H'8; Apply H'8; Clear H'6; Auto.
Case (Rle_or_lt (FtoR radix f) r1); Auto; Intros C1.
Absurd (Rlt (FtoR radix f) max).
Apply Rlt_not.
Case H'0.
Intros H'6 H'9; Elim H'9; Intros H'10 H'11; Apply H'11; Clear H'9; Auto.
Apply Rlt_le; Auto.
Apply Rle_lt_trans with 2 := H'2; Auto.
Qed.

Theorem isMaxComp:
  (r1, r2:R)
  (min, max:float)
  (isMin b radix r1 min) ->
  (isMax b radix r1 max) -> (Rlt min r2) -> (Rlt r2 max) ->
  (isMax b radix r2 max).
Intros r1 r2 min max H' H'0 H'1 H'2; Split.
Intuition.
Split.
Apply Rlt_le; Auto.
Intros f H'3 H'4.
Case H'0; Auto.
Intros H'5 H'6; Elim H'6; Intros H'7 H'8; Apply H'8; Clear H'6; Auto.
Case (Rle_or_lt r1 (FtoR radix f)); Auto; Intros C1.
Absurd (Rlt min (FtoR radix f)).
Apply Rlt_not.
Case H'.
Intros H'6 H'9; Elim H'9; Intros H'10 H'11; Apply H'11; Clear H'9; Auto.
Apply Rlt_le; Auto.
Apply Rlt_le_trans with 1 := H'1; Auto.
Qed.

Theorem roundedModeLessMult:
  (P:R -> float ->Prop)
  (p:float)
  (r:R)
  (RoundedModeP b radix P) -> (P r p) -> (Rle (Float (1) (Zopp (dExp b))) r) ->
  (Rle p (Rmult radix r)).
Intros P p r H' H'0 H'1.
Cut (Rlt R0 (Float (1) (Zopp (dExp b))));
 [Intros Rl0 |
   Unfold FtoRradix FtoR; Simpl; AutoRewrite [Rsimpl]; Auto with real arith].
Cut (Rlt R0 r); [Intros Rl1 | Apply Rlt_le_trans with 1 := Rl0]; Auto.
Cut (Rle R0 r); [Intros Rl2 | Apply Rlt_le; Auto].
Case H'.
Intros H'2 H'3; Elimc H'3; Intros H'3 H'4; Elimc H'4; Intros H'4 H'5;
 Case (H'4 r p); Auto; Intros H'6.
Apply Rle_trans with r; Auto with real.
Intuition.
Rewrite Rmult_sym; Pattern 1 r; Replace r with (Rmult r (1));
 [Apply Rle_monotony | Ring]; Auto with real arith.
Case (MinEx b radix precision) with r := r; Auto with arith; Intros min Hmin.
Cut (Fbounded b (Float (Fnum min) (Zs (Fexp min)))); [Intros F2 | Idtac].
Cut <R> (FtoRradix (Float (Fnum min) (Zs (Fexp min))))==(Rmult radix min);
 [Intros F2Eq | Idtac].
Apply Rle_trans with (FtoRradix (Float (Fnum min) (Zs (Fexp min)))).
Case H'6.
Intros H'7 H'8; Elim H'8; Intros H'9 H'10; Apply H'10; Clear H'8; Auto.
Case (Rle_or_lt r (Float (Fnum min) (Zs (Fexp min)))); Auto; Intros Rlt0.
Absurd (Rle (Float (Fnum min) (Zs (Fexp min))) min).
Apply Rle_not.
Rewrite F2Eq; Auto with real.
Rewrite Rmult_sym.
Pattern 1 (FtoRradix min); Replace (FtoRradix min) with (Rmult min (1));
 Auto with real.
Apply Rlt_monotony; Auto with real arith.
Case (RleMinR0 r min); Auto.
Intros H'8; Case H'1.
Intros H'11; Absurd (Rle (Float (1) (Zopp (dExp b))) min).
Apply Rle_not; Auto.
Rewrite <- H'8; Auto.
Apply (MonotoneMin b radix)
     with p := (FtoRradix (Float (1) (Zopp (dExp b)))) q := r; Auto.
Apply (RoundedModeProjectorIdem (isMin b radix)); Auto.
Apply MinRoundedModeP with precision := precision; Auto.
Repeat Split.
Simpl; Auto with zarith.
Replace (Fnum (Float (1) (Zopp (dExp b)))) with (inject_nat (1));
 [Apply inj_le; Auto with arith | Simpl; Auto].
Apply lt_le_S; Auto.
Apply (lt_vNum ? radixMoreThanOne b precision); Auto with arith.
Simpl; Auto with zarith.
Intros H'11; Absurd <R> min==(Float (1) (Zopp (dExp b))).
Rewrite <- H'8.
Apply imp_not_Req; Left; Auto.
Apply (MinUniqueP b radix r); Auto.
Rewrite <- H'11.
Apply (RoundedModeProjectorIdem (isMin b radix)); Auto.
Apply MinRoundedModeP with precision := precision; Auto.
Repeat Split.
Simpl; Auto with zarith.
Replace (Fnum (Float (1) (Zopp (dExp b)))) with (inject_nat (1));
 [Apply inj_le; Auto with arith | Simpl; Auto].
Apply lt_le_S; Auto.
Apply (lt_vNum ? radixMoreThanOne b precision); Auto with arith.
Simpl; Auto with zarith.
Case Hmin.
Intros H'8 H'11; Elim H'11; Intros H'12 H'13; Apply H'13; Clear H'11; Auto.
Apply Rlt_le; Auto.
Rewrite F2Eq.
Apply Rle_monotony; Auto with real arith.
Replace R0 with (INR O); Auto with real arith.
Intuition.
Unfold FtoRradix FtoR; Simpl.
Rewrite powerRZ_Zs; Auto with real arith; Ring.
Repeat (Split; Simpl).
Intuition.
Intuition.
Apply Zle_trans with (Fexp min); [Intuition | Auto with zarith].
Qed.

Theorem roundedModeMoreMult:
  (P:R -> float ->Prop)
  (p:float)
  (r:R)
  (RoundedModeP b radix P) ->
  (P r p) -> (Rle r (Float (Zopp (1)) (Zopp (dExp b)))) ->
  (Rle (Rmult radix r) p).
Intros P p r H' H'0 H'1.
Cut (Rlt (Float (Zopp (1)) (Zopp (dExp b))) R0);
 [Intros Rl0 |
   Unfold FtoRradix FtoR; Simpl; AutoRewrite [Rsimpl]; Auto with real arith].
2:Replace R0 with (Ropp R0); Auto with real arith; Ring.
Cut (Rlt r R0); [Intros Rl1 | Apply Rle_lt_trans with 2 := Rl0]; Auto.
Cut (Rle r R0); [Intros Rl2 | Apply Rlt_le; Auto].
Case H'.
Intros H'2 H'3; Elimc H'3; Intros H'3 H'4; Elimc H'4; Intros H'4 H'5;
 Case (H'4 r p); Auto; Intros H'6.
Case (MaxEx b radix precision) with r := r; Auto with arith; Intros max Hmax.
Cut (Fbounded b (Float (Fnum max) (Zs (Fexp max)))); [Intros F2 | Idtac].
Cut <R> (FtoRradix (Float (Fnum max) (Zs (Fexp max))))==(Rmult radix max);
 [Intros F2Eq | Idtac].
Apply Rle_trans with (FtoRradix (Float (Fnum max) (Zs (Fexp max)))).
Rewrite F2Eq; Auto with real.
Apply Rle_monotony; Auto with real arith.
Replace R0 with (INR O); Auto with real arith.
Intuition.
Case H'6.
Intros H'7 H'8; Elim H'8; Intros H'9 H'10; Apply H'10; Clear H'8; Auto.
Case (Rle_or_lt (Float (Fnum max) (Zs (Fexp max))) r); Auto; Intros Rlt0.
Absurd (Rle max (Float (Fnum max) (Zs (Fexp max)))).
Apply Rle_not.
Rewrite F2Eq.
Replace (Rmult radix max) with (Ropp (Rmult (Ropp max) radix)); [Idtac | Ring].
Pattern 2 (FtoRradix max);
 Replace (FtoRradix max) with (Ropp (Rmult (Ropp max) (1)));
 [Idtac | Simpl; Ring].
Apply Rlt_Ropp; Apply Rlt_monotony; Auto with real.
Replace R0 with (Ropp R0); [Apply Rlt_Ropp | Ring].
Case (RleMaxR0 r max); Auto.
Intros H'8; Case H'1.
Intros H'11; Absurd (Rle max (Float (Zopp (1)) (Zopp (dExp b)))).
Apply Rle_not; Auto.
Rewrite H'8; Auto.
Apply (MonotoneMax b radix)
     with q := (FtoRradix (Float (Zopp (1)) (Zopp (dExp b)))) p := r; Auto.
Apply (RoundedModeProjectorIdem (isMax b radix)); Auto.
Apply MaxRoundedModeP with precision := precision; Auto.
Repeat Split.
Replace (Fnum (Float (Zopp (1)) (Zopp (dExp b)))) with (Zopp (1));
 [Apply Zle_Zopp; Apply inj_le; Auto with arith | Simpl; Auto].
Apply lt_le_S; Auto.
Apply (lt_vNum ? radixMoreThanOne b precision); Auto with arith.
Simpl; Auto with zarith.
Simpl; Auto with zarith.
Intros H'11; Absurd <R> (Float (Zopp (1)) (Zopp (dExp b)))==max.
Rewrite H'8; Auto.
Apply imp_not_Req; Left; Auto.
Apply (MaxUniqueP b radix r); Auto.
Rewrite H'11.
Apply (RoundedModeProjectorIdem (isMax b radix)); Auto.
Apply MaxRoundedModeP with precision := precision; Auto.
Repeat Split.
Replace (Fnum (Float (Zopp (1)) (Zopp (dExp b)))) with (Zopp (1));
 [Apply Zle_Zopp; Apply inj_le; Auto with arith | Simpl; Auto].
Apply lt_le_S; Auto.
Apply (lt_vNum ? radixMoreThanOne b precision); Auto with arith.
Simpl; Auto with zarith.
Simpl; Auto with zarith.
Case Hmax.
Intros H'8 H'11; Elim H'11; Intros H'12 H'13; Apply H'13; Clear H'11; Auto.
Apply Rlt_le; Auto.
Unfold FtoRradix FtoR; Simpl.
Rewrite powerRZ_Zs; Auto with real arith; Ring.
Repeat (Split; Simpl).
Intuition.
Intuition.
Apply Zle_trans with (Fexp max); [Intuition | Auto with zarith].
Apply Rle_trans with r; Auto with real.
Pattern 2 r; Replace r with (Ropp (Rmult (Ropp r) (1))); [Idtac | Ring].
Replace (Rmult radix r) with (Ropp (Rmult (Ropp r) radix)); [Idtac | Ring].
Apply Rle_Ropp; Apply Rle_monotony; Auto with real arith.
Replace R0 with (Ropp R0); Auto with real arith.
Intuition.
Qed.

Theorem roundedModeAbsMult:
  (P:R -> float ->Prop)
  (p:float)
  (r:R)
  (RoundedModeP b radix P) ->
  (P r p) -> (Rle (Float (1) (Zopp (dExp b))) (Rabsolu r)) ->
  (Rle (Rabsolu p) (Rmult radix (Rabsolu r))).
Intros P p r H' H'0 H'1; Case (Rle_or_lt R0 r); Intros H'2.
Repeat Rewrite Rabsolu_simpl1; Auto.
Apply roundedModeLessMult with P := P; Auto.
Rewrite <- (Rabsolu_simpl1 r); Auto.
Apply (RleRoundedR0 P) with r := r; Auto.
Repeat Rewrite Rabsolu_simpl2; Auto.
Replace (Rmult radix (Ropp r)) with (Ropp (Rmult radix r));
 [Apply Rle_Ropp | Ring].
Apply roundedModeMoreMult with P := P; Auto.
Rewrite <- (Ropp_Ropp r); Rewrite <- (Rabsolu_simpl2 r); Auto.
Replace (Float (Zopp (1)) (Zopp (dExp b)))
     with (Fopp (Float (1) (Zopp (dExp b)))).
Unfold FtoRradix; Rewrite Fopp_correct; Auto with real.
Simpl; Auto.
Apply Rlt_le; Auto.
Apply Rlt_le; Auto.
Apply (RleRoundedLessR0 P) with r := r; Auto.
Apply Rlt_le; Auto.
Qed.

Theorem RleBoundRoundl:
  (P:?) (RoundedModeP b radix P) ->
  (p, q:float) (r:R) (Fbounded b p) -> (Rle p r) -> (P r q) ->(Rle p q).
Intros P H' p q r H'0 H'1 H'2; Case H'1; Intros H'3.
Cut (MonotoneP radix P); [Intros Mn | Intuition].
Apply (Mn p r); Auto.
Apply RoundedModeProjectorIdem with P := P; Auto.
Rewrite RoundedModeProjectorIdemEq with P := P p := p q := q; Auto with real.
Cut (CompatibleP b radix P); [Intros Cp | Intuition].
Apply (Cp r p q); Auto.
Apply RoundedModeBounded with P := P r := r; Auto.
Qed.

Theorem RleBoundRoundr:
  (P:?) (RoundedModeP b radix P) ->
  (p, q:float) (r:R) (Fbounded b p) -> (Rle r p) -> (P r q) ->(Rle q p).
Intros P H' p q r H'0 H'1 H'2; Case H'1; Intros H'3.
Cut (MonotoneP radix P); [Intros Mn | Intuition].
Apply (Mn r p); Auto.
Apply RoundedModeProjectorIdem with P := P; Auto.
Rewrite RoundedModeProjectorIdemEq with P := P p := p q := q; Auto with real.
Cut (CompatibleP b radix P); [Intros Cp | Intuition].
Apply (Cp r p q); Auto.
Apply RoundedModeBounded with P := P r := r; Auto.
Qed.

Theorem RoundAbsMonotoner:
  (P:R -> float ->Prop)
  (p:R)
  (q, r:float)
  (RoundedModeP b radix P) ->
  (Fbounded b r) -> (P p q) -> (Rle (Rabsolu p) r) ->(Rle (Rabsolu q) r).
Intros P p q r H' H'0 H'1 H'2.
Case (Rle_or_lt R0 p); Intros Rl1.
Rewrite Rabsolu_simpl1; Auto.
Apply RleBoundRoundr with P := P r := p; Auto.
Rewrite <- (Rabsolu_simpl1 p); Auto.
Apply RleRoundedR0 with P := P r := p; Auto.
Rewrite Rabsolu_simpl2; Auto.
Rewrite <- (Ropp_Ropp r); Apply Rle_Ropp.
Rewrite <- (Fopp_correct radix); Auto.
Apply RleBoundRoundl with P := P r := p; Auto with float.
Rewrite (Fopp_correct radix); Rewrite <- (Ropp_Ropp p);
 Rewrite <- (Rabsolu_simpl2 p); Auto with real; Apply Rlt_le; Auto.
Apply RleRoundedLessR0 with P := P r := p; Auto; Apply Rlt_le; Auto.
Qed.

Theorem RoundAbsMonotonel:
  (P:R -> float ->Prop)
  (p:R)
  (q, r:float)
  (RoundedModeP b radix P) ->
  (Fbounded b r) -> (P p q) -> (Rle r (Rabsolu p)) ->(Rle r (Rabsolu q)).
Intros P p q r H' H'0 H'1 H'2.
Case (Rle_or_lt R0 p); Intros Rl1.
Rewrite Rabsolu_simpl1; Auto.
Apply RleBoundRoundl with P := P r := p; Auto.
Rewrite <- (Rabsolu_simpl1 p); Auto.
Apply RleRoundedR0 with P := P r := p; Auto.
Rewrite Rabsolu_simpl2; Auto.
Rewrite <- (Ropp_Ropp r); Apply Rle_Ropp.
Rewrite <- (Fopp_correct radix); Auto.
Apply RleBoundRoundr with P := P r := p; Auto with float.
Rewrite (Fopp_correct radix); Rewrite <- (Ropp_Ropp p);
 Rewrite <- (Rabsolu_simpl2 p); Auto with real; Apply Rlt_le; Auto.
Apply RleRoundedLessR0 with P := P r := p; Auto; Apply Rlt_le; Auto.
Qed.
(* Rounded of natural numbers are natural *)

Theorem ZroundZ:
  (P:R -> float ->Prop) (z:Z) (p:float) (RoundedModeP b radix P) -> (P z p) ->
  (Ex [z':Z] <R> p==z').
Intros P z p HP H'.
Case (RoundedModeRep b radix precision) with P := P p := (Float z O) q := p;
 Auto.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Apply Cp with 1 := H'; Auto.
Unfold FtoRradix FtoR; Simpl.
Rewrite Rmult_ne_l; Auto.
Apply RoundedModeBounded with P := P r := (IZR z); Auto.
Intros x H'0; Exists x; Auto.
Unfold FtoRradix; Rewrite H'0.
Unfold FtoRradix FtoR; Simpl.
Rewrite Rmult_ne_l; Auto.
Qed.

Theorem NroundN:
  (P:R -> float ->Prop) (n:nat) (p:float) (RoundedModeP b radix P) -> (P n p) ->
  (Ex [n':nat] <R> p==n').
Intros P n p HP H'.
Case (ZroundZ P (inject_nat n) p); Auto.
Repeat Rewrite <- INR_IZR_INZ; Auto.
Intros x H'0; Exists (absolu x).
Rewrite <- (inj_abs x) in H'0.
Rewrite H'0.
Repeat Rewrite <- INR_IZR_INZ; Auto.
Apply le_IZR; Simpl.
Rewrite <- H'0; Auto.
Apply RleRoundedR0 with P := P r := (INR n); Auto.
Replace R0 with (INR O); Auto with real arith.
Qed.
(* Properties of LSB and MSB *)

Theorem FUlp_Le_LSigB:
  (x:float) (Fbounded b x) ->(Rle (Fulp x) (Float (1) (LSB radix x))).
Intros x H; Unfold is_Fzero Fulp Fnormalize; Case (Z_zerop (Fnum x)); Intros ZH.
Unfold FtoRradix FtoR; Simpl.
Rewrite Rmult_ne_r.
Apply Rle_powerRZ.
Replace R1 with (INR (1)); Auto with real arith.
Apply Zle_trans with (Fexp x); Auto.
Intuition.
Apply Fexp_le_LSB; Auto.
Rewrite LSB_shift with n :=
                       (min
                          (minus precision (Fdigit radix x))
                          (absolu (Zplus (dExp b) (Fexp x)))); Auto.
Unfold FtoRradix FtoR; Simpl.
Rewrite Rmult_ne_r.
Apply Rle_powerRZ; Auto with arith.
Replace R1 with (INR (1)); Auto with real arith.
Exact (Fexp_le_LSB
         radix
         (Fshift
            radix
            (min
               (minus precision (Fdigit radix x))
               (absolu (Zplus (dExp b) (Fexp x)))) x)).
Qed.

Theorem MSBisMin:
  (f1, f2:float)
  (Rle R0 f1) -> (isMin b radix f1 f2) -> ~ (is_Fzero f1) -> ~ (is_Fzero f2) ->
  (MSB radix f1)=(MSB radix f2).
Intros f1 f2 H' H'0 H'1 H'2.
Apply Zle_antisym.
2:Apply MSB_monotone; Auto.
2:Repeat Rewrite Fabs_correct1; Auto with arith.
2:Intuition.
2:Apply RleRoundedR0 with P := (isMin b radix) r := (FtoRradix f1); Auto.
2:Apply MinRoundedModeP with precision := precision; Auto.
Case (Zle_or_lt (MSB radix f1) (MSB radix f2)); Auto.
Intros H'3; Absurd (Rle (Float (1) (Zs (MSB radix f2))) f2).
Apply Rle_not.
Unfold FtoRradix; Rewrite <- Fabs_correct1; Auto with float arith.
Apply abs_lt_MSB; Auto.
Apply RleRoundedR0 with P := (isMin b radix) r := (FtoRradix f1);
 Auto with float.
Apply MinRoundedModeP with precision := precision; Auto.
Case H'0.
Intros H'4 H'5; Elim H'5; Intros H'6 H'7; Apply H'7; Clear H'5; Auto.
Apply (FboundedOne ? radixMoreThanOne b precision); Auto with arith.
Apply Zle_trans with (Fexp f2).
Cut (Fbounded b f2); Intuition.
Apply Zle_trans with (MSB radix f2); Auto with zarith.
Apply Fexp_le_MSB; Auto.
Apply Rle_trans with (FtoR radix (Float (1) (MSB radix f1))); Auto.
Apply oneExp_le; Auto with zarith.
Unfold FtoRradix; Rewrite <- Fabs_correct1 with x := f1; Auto with float arith.
Apply MSB_le_abs; Auto.
Qed.

Theorem MSBtoZero:
  (f1, f2:float)
  (ToZeroP b radix f1 f2) -> ~ (is_Fzero f1) -> ~ (is_Fzero f2) ->
  (MSB radix f1)=(MSB radix f2).
Intros f1 f2 H' H'0 H'1; Casec H'; Intros tmp; Elimc tmp; Intros H1 H2.
Apply MSBisMin; Auto.
Rewrite (MSB_opp radix f1).
Rewrite (MSB_opp radix f2).
Apply MSBisMin; Auto with float.
Unfold FtoRradix; Rewrite Fopp_correct.
Replace R0 with (Ropp R0); Auto with real.
Unfold FtoRradix; Rewrite Fopp_correct; Auto with float.
Qed.

Theorem MSBBoundNotZero:
  (P:R -> float ->Prop) (RoundedModeP b radix P) ->
  (f1, f2:float)
  (P f1 f2) -> ~ <R> f1==R0 -> (Zle (Zopp (dExp b)) (MSB radix f1)) ->
  ~ <R> f2==R0.
Intros P H' f1 f2 H'0 H'1 H'2.
Case (Rle_or_lt R0 f1); Intros Rl1.
Apply imp_not_Req; Right; Red.
Apply Rlt_le_trans with r2 := (FtoRradix (Float (1) (MSB radix f1))); Auto.
Unfold FtoRradix FtoR; Simpl; Rewrite Rmult_ne_r; Auto with real arith.
Cut (Rle (Float (1) (MSB radix f1)) (Fabs f1)); Unfold FtoRradix;
 [Rewrite Fabs_correct; Auto with arith; Rewrite Rabsolu_simpl1; Auto;
   Intros Rl2; Casec Rl2; Intros Rl2 | Apply MSB_le_abs]; Auto.
Cut (MonotoneP radix P); [Intros Mn | Intuition].
Apply (Mn (Float (1) (MSB radix f1)) f1); Auto.
Apply RoundedModeProjectorIdem; Auto.
Apply (FboundedOne radix) with precision := precision;
 Auto with real zarith arith.
Replace (FtoR radix f2) with (FtoR radix (Float (1) (MSB radix f1)));
 Auto with float real.
Apply RoundedModeProjectorIdemEq with P := P; Auto.
Apply (FboundedOne radix) with precision := precision;
 Auto with real zarith arith.
Cut (CompatibleP b radix P); [Intros Cp | Intuition].
Apply (Cp f1) with p := f2; Auto.
Apply RoundedModeBounded with P := P r := (FtoRradix f1); Auto.
Contradict H'1; Unfold FtoRradix; Apply is_Fzero_rep1; Auto.
Apply imp_not_Req; Left.
Apply Rle_lt_trans with r2 := (FtoRradix (Float (Zopp (1)) (MSB radix f1)));
 Auto.
Cut (MonotoneP radix P); [Intros Mn | Intuition].
Cut (Rle f1 (Float (Zopp (1)) (MSB radix f1)));
 [Intros Rle1; Casec Rle1; Intros Rle1 | Idtac].
Apply (Mn f1 (Float (Zopp (1)) (MSB radix f1))); Auto.
Apply RoundedModeProjectorIdem; Auto.
Apply oppBoundedInv; Unfold Fopp; Simpl.
Apply (FboundedOne radix) with precision := precision;
 Auto with real zarith arith.
Replace (FtoRradix f2) with (FtoRradix (Float (Zopp (1)) (MSB radix f1)));
 Auto with real.
Apply RoundedModeProjectorIdemEq with P := P; Auto.
Apply oppBoundedInv; Unfold Fopp; Simpl.
Apply (FboundedOne ? radixMoreThanOne b precision); Auto with real zarith arith.
Cut (CompatibleP b radix P); [Intros Cp | Intuition].
Apply (Cp f1) with p := f2; Auto.
Apply RoundedModeBounded with P := P r := (FtoRradix f1); Auto.
Replace (FtoRradix f1) with (Ropp (FtoRradix (Fabs f1))).
Replace (Float (Zopp (1)) (MSB radix f1)) with (Fopp (Float (1) (MSB radix f1))).
Unfold FtoRradix; Rewrite Fopp_correct; Auto.
Apply Rle_Ropp; Apply MSB_le_abs; Auto.
Contradict H'1; Unfold FtoRradix; Apply is_Fzero_rep1; Auto.
Unfold Fopp; Simpl; Auto.
Unfold FtoRradix; Rewrite Fabs_correct; Auto with arith; Rewrite Rabsolu_simpl2;
 Try Apply Rlt_le; Auto; Ring.
Replace (Float (Zopp (1)) (MSB radix f1)) with (Fopp (Float (1) (MSB radix f1)));
 [Idtac | Unfold Fopp; Simpl; Auto].
Replace R0 with (Ropp R0); [Idtac | Ring].
Unfold FtoRradix; Repeat Rewrite Fopp_correct; Apply Rlt_Ropp.
Unfold FtoR; Simpl; Rewrite Rmult_ne_r; Auto with real arith.
Qed.

Theorem RoundMSBmax:
  (P:R -> float ->Prop)
  (p, q:float)
  (RoundedModeP b radix P) ->
  (P p q) -> ~ <R> p==R0 -> (Zle (Zopp (dExp b)) (MSB radix p)) ->
  (Zle (MSB radix q) (Zs (MSB radix p))).
Intros P p q H' H'0 H'1 H'2.
Apply (oneExp_Zle radix); Auto.
Apply Rle_trans with (FtoRradix (Fabs q)).
Unfold FtoRradix; Apply MSB_le_abs; Auto.
Red; Intros H'3; Absurd <R> q==R0.
Apply MSBBoundNotZero with P := P f1 := p; Auto.
Apply (is_Fzero_rep1 radix); Auto.
Unfold FtoRradix; Rewrite Fabs_correct; Auto with arith; Fold FtoRradix.
Apply RoundAbsMonotoner with P := P p := (FtoRradix p); Auto.
Apply (FboundedOne ? radixMoreThanOne b precision); Auto with zarith.
Unfold FtoRradix; Rewrite <- (Fabs_correct radix); Auto with arith.
Apply Rlt_le; Apply abs_lt_MSB; Auto.
Qed.

Theorem RoundMSBmin:
  (P:R -> float ->Prop)
  (p, q:float)
  (RoundedModeP b radix P) ->
  (P p q) -> ~ <R> p==R0 -> (Zle (Zopp (dExp b)) (MSB radix p)) ->
  (Zle (MSB radix p) (MSB radix q)).
Intros P p q H' H'0 H'1 H'2.
Replace (MSB radix p) with (MSB radix (Float (1) (MSB radix p))).
Apply MSB_monotone; Auto.
Unfold is_Fzero; Simpl; Red; Intros; Discriminate.
Red; Intros H'3; Absurd <R> q==R0.
Apply MSBBoundNotZero with P := P f1 := p; Auto.
Unfold FtoRradix; Apply is_Fzero_rep1; Auto.
Replace (Fabs (Float (1) (MSB radix p))) with (Float (1) (MSB radix p));
 [Idtac | Unfold Fabs; Simpl; Auto].
Rewrite Fabs_correct; Auto with arith; Fold FtoRradix.
Apply RoundAbsMonotonel with P := P p := (FtoRradix p); Auto.
Apply (FboundedOne ? radixMoreThanOne b precision); Auto with zarith.
Unfold FtoRradix; Rewrite <- (Fabs_correct radix); Auto with arith;
 Apply MSB_le_abs; Auto.
Contradict H'1; Unfold FtoRradix; Apply is_Fzero_rep1; Auto.
Unfold MSB Fdigit; Simpl.
Unfold convert; Simpl; Rewrite digit1; Auto.
Replace (inject_nat (1)) with (POS xH); Auto.
Unfold Zpred; Ring.
Qed.

Theorem RoundLSBMax:
  (P:R -> float ->Prop)
  (p, q:float) (RoundedModeP b radix P) -> (P p q) -> ~ (is_Fzero q) ->
  (Zle (LSB radix p) (LSB radix q)).
Intros P p q H' H'0 H'2.
Elim (LSB_rep_min radix) with p := p; Auto; Intros z E.
Case (RoundedModeRep b radix precision) with P := P p := (Float z (LSB radix p))
                                             q := q; Auto.
Cut (CompatibleP b radix P); [Intros Cp | Intuition].
Apply (Cp p (Float z (LSB radix p)) q); Auto.
Apply RoundedModeBounded with P := P r := (FtoRradix p); Auto.
Intros x H'3.
Replace (LSB radix p) with (Fexp (Float x (LSB radix p))); [Idtac | Simpl; Auto].
Replace (LSB radix q) with (LSB radix (Float x (LSB radix p))).
Apply Fexp_le_LSB.
Apply LSB_comp; Auto.
Apply NisFzeroComp with radix := radix x := q; Auto.
Qed.
End FRoundP.
Hints Resolve FulpSucCan FulpSuc FulpPredCan FulpPred :float.
