Require Import Float.
Require Import Fcomp.
Require Import Fprop.
Require Import Fop.
Require Import Fnorm.
Section mf.
Variable radix:nat.
Hypothesis radixMoreThanOne:(lt (1) radix).

Local FtoRradix := (FtoR radix).
Coercion FtoRradix : float >-> R.

Fixpoint maxDiv[v, p:nat]: nat :=
   Cases p of
       O => O
      | (S p') => Cases (dividesP v (exp radix p)) of
                      (left _) => p
                     | (right _) => (maxDiv v p')
                  end
   end.

Theorem maxDivLess: (v, p:nat)(le (maxDiv v p) p).
Intros v p; Elim p; Simpl; Auto.
Intros n H'; Case (dividesP v (mult radix (exp radix n))); Auto.
Qed.

Theorem maxDivLt: (v, p:nat) ~ (Ndivides v (exp radix p)) ->(lt (maxDiv v p) p).
Intros v p; Case p; Simpl; Auto.
Intros H'; Case H'.
Apply Ndivides1.
Intros n H'; Case (dividesP v (mult radix (exp radix n))); Auto.
Intros H'0; Case H'; Auto.
Intros H'0; Generalize (maxDivLess v n); Auto with arith.
Qed.

Theorem maxDivCorrect: (v, p:nat)(Ndivides v (exp radix (maxDiv v p))).
Intros v p; Elim p.
Unfold maxDiv; Replace (exp radix O) with (1); Auto.
Apply Ndivides1.
Simpl.
Intros n H'; Case (dividesP v (mult radix (exp radix n))); Simpl;
 Auto with zarith.
Qed.

Theorem maxDivSimplAux:
  (v, p, q:nat) p=(maxDiv v (S (plus q p))) ->p=(maxDiv v (S p)).
Intros v p q; Elim q.
Simpl; Case (dividesP v (mult radix (exp radix p))); Auto.
Intros n H' H'0.
Apply H'; Auto; Clear H'.
Simpl in H'0; Generalize H'0; Clear H'0.
Case (dividesP v (mult radix (mult radix (exp radix (plus n p))))).
2:Simpl; Auto.
Intros H' H'0; Contradict H'0; Auto with zarith.
Qed.

Theorem maxDivSimpl:
  (v, p, q:nat) (lt p q) -> p=(maxDiv v q) ->p=(maxDiv v (S p)).
Intros v p q H' H'0.
Apply maxDivSimplAux with q := (minus q (S p)); Auto.
Replace (S (plus (minus q (S p)) p)) with q; Auto with zarith.
Qed.

Theorem maxDivSimplInvAux:
  (v, p, q:nat) p=(maxDiv v (S p)) ->p=(maxDiv v (S (plus q p))).
Intros v p q H'; Elim q.
Simpl; Auto.
Intros n; Simpl.
Case (dividesP v (mult radix (mult radix (exp radix (plus n p))))); Auto.
Case (dividesP v (mult radix (exp radix (plus n p)))); Auto.
Intros H'0 H'1 H'2; Contradict H'2; Auto with zarith.
Intros H'0 H'1 H'2; Case H'1.
Intros x H'3; Case H'0; Red; Exists (mult radix x).
Rewrite H'3; Auto with arith.
Repeat Rewrite mult_assoc_r.
Apply f_equal2 with f := mult; Auto.
Rewrite (mult_sym radix); Repeat Rewrite mult_assoc_r.
Apply f_equal2 with f := mult; Auto with arith.
Qed.

Theorem maxDivSimplInv:
  (v, p, q:nat) (lt p q) -> p=(maxDiv v (S p)) ->p=(maxDiv v q).
Intros v p q H' H'0.
Replace q with (S (plus (minus q (S p)) p)); Auto with zarith.
Apply maxDivSimplInvAux; Auto.
Qed.

Theorem maxDivUnique:
  (v, p:nat) p=(maxDiv v (S p)) ->
  (Ndivides v (exp radix p)) /\ ~ (Ndivides v (exp radix (S p))).
Intros v p H'; Split.
Rewrite H'.
Apply maxDivCorrect; Auto.
Red; Intros H'0; Generalize H'; Clear H'.
Simpl.
Case (dividesP v (mult radix (exp radix p))); Simpl; Auto.
Intros H' H'1; Contradict H'1; Auto with zarith.
Qed.

Theorem maxDivUniqueDigit:
  (v:nat) ~ v=O ->
  (Ndivides v (exp radix (maxDiv v (digit radix v)))) /\
  ~ (Ndivides v (exp radix (S (maxDiv v (digit radix v))))).
Intros v H'.
Apply maxDivUnique; Auto.
Apply maxDivSimpl with q := (digit radix v); Auto.
Apply maxDivLt; Auto.
Apply NotDividesDigit; Auto.
Qed.

Theorem maxDivUniqueInverse:
  (v, p:nat) (Ndivides v (exp radix p)) -> ~ (Ndivides v (exp radix (S p))) ->
  p=(maxDiv v (S p)).
Intros v p H' H'0; Simpl.
Case (dividesP v (mult radix (exp radix p))); Auto.
Intros H'1; Case H'0; Simpl; Auto.
Intros H'1.
Generalize H'; Case p; Simpl; Auto.
Intros n H'2; Case (dividesP v (mult radix (exp radix n))); Auto.
Intros H'3; Case H'3; Auto.
Qed.

Theorem maxDivUniqueInverseDigit:
  (v, p:nat)
  ~ v=O -> (Ndivides v (exp radix p)) -> ~ (Ndivides v (exp radix (S p))) ->
  p=(maxDiv v (digit radix v)).
Intros v p H' H'0 H'1.
Apply maxDivSimplInv; Auto.
2:Apply maxDivUniqueInverse; Auto.
Apply exp_anti_monotone_lt with n := radix; Auto.
Apply le_lt_trans with m := v; Auto.
Apply NDividesLt; Auto.
Apply digitMore; Auto.
Qed.

Theorem maxDivPlus:
  (v, n:nat) ~ v=O ->
  (maxDiv (mult v (exp radix n)) (plus (digit radix v) n))=
  (plus (maxDiv v (digit radix v)) n).
Intros v n H.
Replace (plus (digit radix v) n) with (digit radix (mult v (exp radix n))); Auto.
Apply sym_equal.
Apply maxDivUniqueInverseDigit; Auto.
Case (mult_O_le (exp radix n) v); Intros H1.
Case H; Auto.
Red; Intros H'; Rewrite H' in H1.
Contradict H1; Auto with arith.
Rewrite expPlus.
Repeat Rewrite [x:nat](mult_sym x (exp radix n)).
Apply NdividesMult; Auto.
Elim (maxDivUniqueDigit v); Auto.
Replace (S (plus (maxDiv v (digit radix v)) n))
     with (plus (S (maxDiv v (digit radix v))) n); Auto.
Rewrite expPlus.
Repeat Rewrite [x:nat](mult_sym x (exp radix n)).
Red; Intros H'.
Absurd (Ndivides v (exp radix (S (maxDiv v (digit radix v))))).
Elim (maxDivUniqueDigit v); Auto.
Apply NdividesDiv with p := (exp radix n); Auto with arith.
Apply digitAdd; Auto with zarith.
Qed.

Definition LSB :=
   [x:float]
   (Zplus (inject_nat (maxDiv (absolu (Fnum x)) (Fdigit radix x))) (Fexp x)).

Theorem LSB_shift:
  (x:float) (n:nat) ~ (is_Fzero x) ->(LSB x)=(LSB (Fshift radix n x)).
Intros x n H'; Unfold LSB Fdigit; Simpl.
Repeat Rewrite absolu_comp_mult; Repeat Rewrite absolu_comp_Rpower.
Rewrite (absolu_INR radix); Auto with arith.
Rewrite digitAdd; Auto with arith.
2:Apply absolu_lt_nz; Auto.
Rewrite maxDivPlus; Auto.
Rewrite inj_plus; Unfold Zminus.
Rewrite (Zplus_sym (Fexp x) (Zopp n)); Rewrite Zplus_assoc.
Rewrite <- Zplus_assoc with z := (Zopp n); Auto.
Rewrite Zplus_inverse_r; Rewrite Zero_right; Auto.
Apply sym_not_eq; Apply lt_O_neq; Apply absolu_lt_nz; Auto.
Qed.

Theorem LSB_comp:
  (x, y:float) (n:nat) ~ (is_Fzero x) -> <R> x==y ->(LSB x)=(LSB y).
Intros x y H' H'0 H'1.
Case (FshiftCorrectSym radix) with 2 := H'1; Auto.
Intros m1 H'2; Elim H'2; Intros m2 E; Clear H'2.
Rewrite (LSB_shift x m1); Auto.
Rewrite E; Auto.
Apply sym_equal; Apply LSB_shift; Auto.
Apply (NisFzeroComp radix) with x := x; Auto.
Qed.

Theorem LSB_opp: (x:float)(LSB x)=(LSB (Fopp x)).
Intros x; Unfold LSB; Simpl.
Rewrite absolu_Zopp; Auto.
Rewrite Fdigit_opp; Auto.
Qed.

Theorem LSB_abs: (x:float)(LSB x)=(LSB (Fabs x)).
Intros x; Unfold LSB; Simpl.
Rewrite Fdigit_abs; Auto.
Case (Fnum x); Auto.
Qed.

Definition MSB :=
   [x:float](Zpred (Zplus (inject_nat (Fdigit radix x)) (Fexp x))).

Theorem MSB_shift:
  (x:float) (n:nat) ~ (is_Fzero x) ->(MSB x)=(MSB (Fshift radix n x)).
Intros; Unfold MSB Fshift Fdigit; Simpl.
Repeat Rewrite absolu_comp_mult; Repeat Rewrite absolu_comp_Rpower; Auto.
Rewrite (absolu_INR radix); Auto with arith.
Rewrite digitAdd; Auto with arith.
2:Apply absolu_lt_nz; Auto.
Repeat Rewrite inj_plus; Unfold Zpred; Ring; Auto.
Qed.

Theorem MSB_comp:
  (x, y:float) (n:nat) ~ (is_Fzero x) -> <R> x==y ->(MSB x)=(MSB y).
Intros x y H' H'0 H'1.
Case (FshiftCorrectSym radix) with 2 := H'1; Auto.
Intros m1 H'2; Elim H'2; Intros m2 E; Clear H'2.
Rewrite (MSB_shift x m1); Auto.
Rewrite E; Auto.
Apply sym_equal; Apply MSB_shift; Auto.
Apply (NisFzeroComp radix) with x := x; Auto.
Qed.

Theorem MSB_opp: (x:float)(MSB x)=(MSB (Fopp x)).
Intros x; Unfold MSB; Simpl.
Rewrite Fdigit_opp; Auto.
Qed.

Theorem MSB_abs: (x:float)(MSB x)=(MSB (Fabs x)).
Intros x; Unfold MSB; Simpl.
Rewrite Fdigit_abs; Auto.
Qed.

Theorem LSB_le_MSB: (x:float) ~ (is_Fzero x) ->(Zle (LSB x) (MSB x)).
Intros x H'; Unfold LSB MSB.
Cut (Zlt (maxDiv (absolu (Fnum x)) (Fdigit radix x)) (Fdigit radix x));
 Unfold Zpred; Auto with zarith.
Apply inj_lt.
Apply maxDivLt; Auto.
Unfold Fdigit; Apply NotDividesDigit; Auto.
Apply sym_not_eq; Apply lt_O_neq; Apply absolu_lt_nz; Auto.
Qed.

Theorem Fexp_le_LSB: (x:float)(Zle (Fexp x) (LSB x)).
Intros x; Unfold LSB.
Auto with zarith.
Qed.

Theorem Ulp_Le_LSigB: (x:float)(Rle (Float (1) (Fexp x)) (Float (1) (LSB x))).
Intros x; Apply (oneExp_le radix); Auto.
Apply Fexp_le_LSB; Auto.
Qed.

Theorem Fexp_le_MSB: (x:float) ~ (is_Fzero x) ->(Zle (Fexp x) (MSB x)).
Intros x H'; Unfold MSB.
Cut ~ <Z> (Fdigit radix x)=ZERO; Unfold Zpred; Auto with zarith.
Unfold Fdigit.
Red; Intros H'0; Absurd (digit radix (absolu (Fnum x)))=O; Auto with zarith.
Apply sym_not_eq; Apply lt_O_neq; Apply digitNotZero; Auto.
Apply absolu_lt_nz; Auto.
Qed.

Theorem MSB_le_abs:
  (x:float) ~ (is_Fzero x) ->(Rle (Float (1) (MSB x)) (Fabs x)).
Intros x H'; Unfold MSB FtoRradix FtoR; Simpl.
Replace (Zpred (Zplus (Fdigit radix x) (Fexp x)))
     with (Zplus (Zpred (Fdigit radix x)) (Fexp x));
 [Idtac | Unfold Zpred; Ring].
Rewrite powerRZ_add; Auto with real arith.
AutoRewrite [Rsimpl].
Repeat Rewrite [r:R](Rmult_sym r (powerRZ radix (Fexp x))).
Apply Rle_monotony; Auto with real arith.
Rewrite <- inj_pred; Auto.
Rewrite powerRZ_is_exp; Auto with arith.
Rewrite Zabs_absolu.
Rewrite <- INR_IZR_INZ; Apply Rle_INR; Auto.
Unfold Fdigit; Auto with arith.
Apply digitLess; Auto.
Apply absolu_lt_nz; Auto.
Unfold Fdigit.
Red; Intros H'0; Absurd (digit radix (absolu (Fnum x)))=O; Auto with zarith.
Apply sym_not_eq; Apply lt_O_neq; Apply digitNotZero; Auto.
Apply absolu_lt_nz; Auto.
Qed.

Theorem abs_lt_MSB: (x:float)(Rlt (Fabs x) (Float (1) (Zs (MSB x)))).
Intros x.
Rewrite (MSB_abs x).
Unfold MSB FtoRradix FtoR.
Rewrite <- Zs_pred; Simpl.
Rewrite powerRZ_add; Auto with real arith.
AutoRewrite [Rsimpl].
Repeat Rewrite [r:R](Rmult_sym r (powerRZ radix (Fexp x))).
Apply Rlt_monotony; Auto with real arith.
Rewrite powerRZ_is_exp; Auto with arith.
Rewrite Zabs_absolu.
Rewrite <- INR_IZR_INZ; Apply Rlt_INR; Auto.
Unfold Fdigit; Auto with arith.
Unfold Fabs; Simpl; Rewrite Zabs_absolu; Rewrite absolu_INR.
Apply digitMore; Auto with real arith.
Qed.

Theorem LSB_le_abs:
  (x:float) ~ (is_Fzero x) ->(Rle (Float (1) (LSB x)) (Fabs x)).
Intros x H'; Apply Rle_trans with (FtoRradix (Float (1) (MSB x))).
Apply (oneExp_le radix); Auto.
Apply LSB_le_MSB; Auto.
Apply MSB_le_abs; Auto.
Qed.

Theorem MSB_monotoneAux:
  (x, y:float) (Rle (Fabs x) (Fabs y)) -> (Fexp x)=(Fexp y) ->
  (Zle (MSB x) (MSB y)).
Intros x y H' H'0; Unfold MSB.
Rewrite <- H'0.
Cut (Zle (Fdigit radix x) (Fdigit radix y));
 [Unfold Zpred; Auto with zarith | Idtac].
Unfold Fdigit; Apply inj_le.
Apply digit_monotone; Auto; Apply ZleLe.
Repeat Rewrite <- Zabs_absolu.
Apply le_IZR.
Apply Rle_mult_anticompatibility with z := (powerRZ radix (Fexp x));
 Auto with real arith.
Repeat Rewrite (Rmult_sym (powerRZ radix (Fexp x))); Auto.
Pattern 2 (Fexp x); Rewrite H'0; Auto.
Qed.

Theorem MSB_monotone:
  (x, y:float) ~ (is_Fzero x) -> ~ (is_Fzero y) -> (Rle (Fabs x) (Fabs y)) ->
  (Zle (MSB x) (MSB y)).
Intros x y H' H'0 H'1; Rewrite (MSB_abs x); Rewrite (MSB_abs y).
Case (Zle_or_lt (Fexp (Fabs x)) (Fexp (Fabs y))); Simpl; Intros Zle1.
Rewrite MSB_shift with x := (Fabs y)
                       n := (absolu (Zminus (Fexp (Fabs y)) (Fexp (Fabs x)))).
Apply MSB_monotoneAux; Auto.
Unfold FtoRradix; Repeat Rewrite Fabs_correct; Auto with real arith.
Rewrite FshiftCorrect; Auto with real arith.
Repeat Rewrite Fabs_correct; Auto with real arith.
Repeat Rewrite Rabsolu_Rabsolu; Repeat Rewrite <- Fabs_correct;
 Auto with real arith.
Unfold Fshift; Simpl.
Rewrite inj_abs; [Ring | Auto with zarith].
Apply Fabs_Fzero; Auto.
Rewrite MSB_shift with x := (Fabs x)
                       n := (absolu (Zminus (Fexp (Fabs x)) (Fexp (Fabs y)))).
Apply MSB_monotoneAux; Auto.
Unfold FtoRradix; Repeat Rewrite Fabs_correct; Auto with real arith.
Rewrite FshiftCorrect; Auto with real arith.
Repeat Rewrite Fabs_correct; Auto with real arith.
Repeat Rewrite Rabsolu_Rabsolu; Repeat Rewrite <- Fabs_correct;
 Auto with real arith.
Unfold Fshift; Simpl.
Rewrite inj_abs; [Ring | Auto with zarith].
Apply Fabs_Fzero; Auto.
Qed.

Theorem MSB_le_multAux:
  (x, y:float) ~ (is_Fzero x) -> ~ (is_Fzero y) ->
  (Zle (Zplus (MSB x) (MSB y)) (MSB (Fmult x y))).
Intros x y H' H'0; Unfold MSB Fmult Fdigit; Simpl.
Replace (Zplus
           (Zpred (Zplus (digit radix (absolu (Fnum x))) (Fexp x)))
           (Zpred (Zplus (digit radix (absolu (Fnum y))) (Fexp y))))
     with (Zpred
             (Zplus
                (Zplus
                   (digit radix (absolu (Fnum x)))
                   (Zpred (digit radix (absolu (Fnum y)))))
                (Zplus (Fexp x) (Fexp y)))); [Idtac | Unfold Zpred; Ring].
Cut (Zle
       (Zplus
          (digit radix (absolu (Fnum x)))
          (Zpred (digit radix (absolu (Fnum y)))))
       (digit radix (absolu (Zmult (Fnum x) (Fnum y)))));
 [Unfold Zpred; Auto with zarith | Idtac].
Rewrite <- inj_pred; Auto.
Rewrite <- inj_plus.
Apply inj_le; Auto.
Rewrite absolu_comp_mult.
Rewrite <- digitAdd; Auto with arith.
Apply digit_monotone; Auto with arith.
Apply lte_comp_mult; Auto with arith.
Apply digitLess; Auto with arith.
Apply absolu_lt_nz; Auto.
Apply absolu_lt_nz; Auto.
Apply sym_not_eq; Apply lt_O_neq; Apply digitNotZero; Auto.
Apply absolu_lt_nz; Auto.
Qed.

Theorem MSB_le_mult:
  (x, y:float) ~ (is_Fzero x) -> ~ (is_Fzero y) ->
  (Rle
     (Fmult (Float (1) (MSB x)) (Float (1) (MSB y)))
     (Float (1) (MSB (Fmult x y)))).
Intros x y H' H'0.
Rewrite <- oneZplus.
Apply (oneExp_le radix); Auto.
Apply MSB_le_multAux; Auto.
Qed.

Theorem mult_le_MSBAux:
  (x, y:float) ~ (is_Fzero x) -> ~ (is_Fzero y) ->
  (Zle (MSB (Fmult x y)) (Zs (Zplus (MSB x) (MSB y)))).
Intros x y H' H'0; Unfold MSB Fmult Fdigit; Simpl.
Replace (Zs
           (Zplus
              (Zpred (Zplus (digit radix (absolu (Fnum x))) (Fexp x)))
              (Zpred (Zplus (digit radix (absolu (Fnum y))) (Fexp y)))))
     with (Zpred
             (Zplus
                (Zplus
                   (digit radix (absolu (Fnum x)))
                   (digit radix (absolu (Fnum y)))) (Zplus (Fexp x) (Fexp y))));
 [Idtac | Unfold Zpred Zs; Ring].
Cut (Zle
       (digit radix (absolu (Zmult (Fnum x) (Fnum y))))
       (Zplus (digit radix (absolu (Fnum x))) (digit radix (absolu (Fnum y)))));
 [Unfold Zpred; Auto with zarith | Idtac].
Rewrite <- inj_plus.
Apply inj_le; Auto.
Rewrite absolu_comp_mult.
Rewrite <- digitAdd; Auto with arith.
Apply digit_monotone; Auto with arith.
Apply lte_comp_mult; Auto with arith.
Apply lt_le_weak; Apply digitMore; Auto with arith.
Apply absolu_lt_nz; Auto.
Qed.

Theorem mult_le_MSB:
  (x, y:float) ~ (is_Fzero x) -> ~ (is_Fzero y) ->
  (Rle
     (Float (1) (MSB (Fmult x y)))
     (Rmult radix (Fmult (Float (1) (MSB x)) (Float (1) (MSB y))))).
Intros x y H' H'0; Rewrite <- oneZplus.
Replace (Rmult radix (Float (1) (Zplus (MSB x) (MSB y))))
     with (FtoRradix (Float (1) (Zs (Zplus (MSB x) (MSB y))))).
Apply (oneExp_le radix); Auto.
Apply mult_le_MSBAux; Auto.
Unfold FtoRradix FtoR; Simpl.
Rewrite powerRZ_Zs; Auto with real arith; Ring.
Qed.

Theorem MSB_mix:
  (x, y:float) ~ (is_Fzero x) -> ~ (is_Fzero y) ->
  (Rlt
     (Rmult (Fabs x) (Float (1) (MSB y)))
     (Rmult radix (Rmult (Fabs y) (Float (1) (MSB x))))).
Intros x y H' H'0; Rewrite (MSB_abs x); Rewrite (MSB_abs y).
Apply Rle_lt_trans with (Rmult (Fabs x) (Fabs y)); Auto with real.
Apply Rle_monotony; Auto with real.
Unfold FtoRradix; Rewrite Fabs_correct; Auto with real arith.
Rewrite <- MSB_abs; Apply MSB_le_abs; Auto.
Rewrite (Rmult_sym (Fabs x)).
Replace (Rmult radix (Rmult (Fabs y) (Float (1) (MSB (Fabs x)))))
     with (Rmult (Fabs y) (Rmult radix (Float (1) (MSB (Fabs x)))));
 [Idtac | Ring].
Apply Rlt_monotony; Auto with real.
Unfold FtoRradix FtoR; Simpl; Auto with real arith.
Rewrite Rmult_sym; Replace R0 with (Rmult (powerRZ radix (Fexp y)) R0);
 [Idtac | Ring].
Apply Rlt_monotony; Auto with real arith.
Rewrite Zabs_absolu.
Replace R0 with (INR O); [Idtac | Simpl; Auto]; Rewrite <- INR_IZR_INZ;
 Apply Rlt_INR.
Apply absolu_lt_nz; Auto.
Replace (Rmult radix (Float (1) (MSB (Fabs x))))
     with (FtoRradix (Float (1) (Zs (MSB (Fabs x))))).
Rewrite <- MSB_abs; Apply abs_lt_MSB; Auto.
Unfold FtoRradix FtoR; Simpl.
Rewrite powerRZ_Zs; Auto with real arith; Ring.
Qed.

Theorem LSB_rep:
  (x, y:float) ~ (is_Fzero y) -> (Zle (LSB x) (LSB y)) ->
  (Ex [z:Z] <R> y==(Float z (Fexp x))).
Intros x y H' H'0.
Case (Zle_or_lt (Fexp x) (Fexp y)); Intros Zl1.
Exists (Zmult (Fnum y) (Zpower_nat radix (absolu (Zminus (Fexp y) (Fexp x))))).
Pattern 2 (Fexp x);
 Replace (Fexp x) with (Zminus (Fexp y) (absolu (Zminus (Fexp y) (Fexp x)))).
Unfold FtoRradix;
 Rewrite <- (FshiftCorrect radix) with n := (absolu (Zminus (Fexp y) (Fexp x)))
                                       x := y; Auto.
Rewrite inj_abs; Try Ring; Auto with zarith.
Exists (ZNquotient (Fnum y) (exp radix (absolu (Zminus (Fexp x) (Fexp y))))).
Unfold FtoRradix;
 Rewrite <- (FshiftCorrect radix) with n := (absolu (Zminus (Fexp x) (Fexp y)))
                                       x :=
                                       (Float
                                          (ZNquotient
                                             (Fnum y)
                                             (exp
                                                radix
                                                (absolu
                                                   (Zminus (Fexp x) (Fexp y)))))
                                          (Fexp x)); Auto.
Unfold Fshift; Simpl.
Cut (Zle ZERO (Zminus (Fexp x) (Fexp y)));
 [Intros Le1; Repeat Rewrite inj_abs | Auto with zarith]; Auto.
Unfold FtoR; Simpl; Auto.
Replace (Zminus (Fexp x) (Zminus (Fexp x) (Fexp y))) with (Fexp y);
 [Idtac | Ring].
Replace (Zmult
           (ZNquotient (Fnum y) (exp radix (absolu (Zminus (Fexp x) (Fexp y)))))
           (Zpower_nat radix (absolu (Zminus (Fexp x) (Fexp y))))) with (Fnum y);
 Auto.
Rewrite <- (inj_abs (Zpower_nat radix (absolu (Zminus (Fexp x) (Fexp y))))).
Rewrite absolu_comp_Rpower.
Rewrite absolu_INR.
Apply ZNquotientProp; Auto with arith.
Apply NdividesNquotient; Auto with arith.
Apply NdividesTrans
     with m := (exp radix (maxDiv (absolu (Fnum y)) (Fdigit radix y))).
Apply maxDivCorrect.
Apply NdividesLessExp; Auto.
Apply ZleLe.
Rewrite inj_abs; Auto with zarith.
Apply Zsimpl_le_plus_l with p := (Fexp y).
Apply Zle_trans with (LSB x).
Replace (Zplus (Fexp y) (Zminus (Fexp x) (Fexp y))) with (Fexp x);
 [Idtac | Ring].
Apply Fexp_le_LSB.
Rewrite Zplus_sym; Auto.
Apply Zpower_NR0; Auto with zarith arith.
Qed.

Theorem LSB_rep_min: (p:float)(Ex [z:Z] <R> p==(Float z (LSB p))).
Intros p;
 Exists (ZNquotient (Fnum p) (exp radix (absolu (Zminus (LSB p) (Fexp p))))).
Unfold FtoRradix FtoR LSB; Simpl.
Rewrite powerRZ_add; Auto with real arith.
Rewrite <- Rmult_assoc.
Replace (Zminus
           (Zplus (maxDiv (absolu (Fnum p)) (Fdigit radix p)) (Fexp p)) (Fexp p))
     with (inject_nat (maxDiv (absolu (Fnum p)) (Fdigit radix p))); Auto.
Rewrite absolu_INR.
Rewrite powerRZ_is_exp; Auto with arith.
Repeat Rewrite INR_IZR_INZ; Auto.
Rewrite <- Rmult_IZR.
Rewrite <- ZNquotientProp; Auto with arith.
Rewrite <- NdividesNquotient; Auto with arith.
Apply maxDivCorrect.
Ring.
Qed.
End mf.

