Require Import Arith.
Require Export Float.
Require Export Fop.
Require Export Fnorm.
Require Export MSB.
Section MSBProp.
Variable b:Fbound.
Variable precision:nat.
Variable radix:nat.

Local FtoRradix := (FtoR radix).
Coercion FtoRradix : float >-> R.
Hypothesis radixMoreThanOne:(lt (S O) radix).
Hypothesis precisionGreaterThanOne:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp radix precision) (1)).

Theorem normal1:
  (a:float) (Fnormal radix b precision a) ->
  (Rle (Float (1) (MSB radix a)) (Rabsolu a)).
Unfold MSB; Intros.
Elim H; Intros H0 H1; Rewrite H1.
Unfold FtoRradix; Unfold FtoR; Simpl.
Repeat Rewrite <- INR_IZR_INZ.
Apply Rle_trans with (powerRZ radix (Zpred (Zplus precision (Fexp a))));
 [Right; Ring | Idtac].
Apply Rle_trans with (powerRZ radix (Zplus (pred precision) (Fexp a))).
Apply Rle_powerRZ; Auto with real arith.
Apply Rlt_le; Auto with real arith.
Replace (Zpred (Zplus precision (Fexp a)))
     with (Zplus (Zpred precision) (Fexp a));
 [Apply Zle_reg_r | Unfold Zpred; Ring].
Case precision.
Simpl; Auto with zarith.
Intros n0.
Rewrite inj_pred; Auto with zarith.
Apply Rle_trans
     with (Rmult (Rabsolu (Fnum a)) (Rabsolu (powerRZ radix (Fexp a))));
 [Idtac | Right; Apply sym_eqT; Apply Rabsolu_mult].
Apply Rle_trans
     with (Rmult (powerRZ radix (pred precision)) (powerRZ radix (Fexp a)));
 [Right; Apply powerRZ_add | Idtac].
Replace R0 with (INR (0)); Auto with real; Apply NEq_INR; Apply notEqLt;
 Apply lt_trans with (1); Auto with arith; Apply radixMoreThanOne.
Replace (Rabsolu (powerRZ radix (Fexp a))) with (powerRZ radix (Fexp a));
 [Apply Rle_monotony_r | Apply sym_eqT; Apply Rabsolu_simpl1].
Apply powerRZ_le; Auto with arith.
Apply Rlt_trans with (INR (1)); Auto with real.
Rewrite powerRZ_is_exp; Auto with arith.
Generalize H1; Unfold Fdigit; Intros H2.
Rewrite <- H2.
Rewrite Rabsolu_Zabs; Rewrite Zabs_absolu.
Repeat Rewrite <- INR_IZR_INZ; Apply Rle_INR.
Apply digitLess.
Apply absolu_lt_nz.
Contradict H2; Rewrite H2; Simpl.
Rewrite digitO.
Apply sym_not_eq; Apply notEqLt.
Apply lt_trans with (1); Auto with arith; Apply precisionGreaterThanOne.
Apply powerRZ_le.
Replace R0 with (INR (0)); Auto with real; Apply Rlt_INR;
 Repeat Rewrite <- INR_IZR_INZ; Apply lt_trans with (1); Auto with arith;
 Apply radixMoreThanOne.
Qed.

Theorem bounded1:
  (a:float) (Fbounded b a) ->
  (Rlt (Rabsolu a) (Float (1) (Zplus precision (Fexp a)))).
Intros.
Apply Rle_lt_trans with (FtoRradix (Float (vNum b) (Fexp a))).
Unfold FtoRradix; Apply MaxFloat with precision := precision; Auto.
Unfold FtoRradix; Unfold FtoR; Simpl.
Replace (Rmult R1 (powerRZ radix (Zplus precision (Fexp a))))
     with (powerRZ radix (Zplus precision (Fexp a))); [Idtac | Ring].
Rewrite powerRZ_add.
Apply Rlt_monotony_r.
Apply powerRZ_lt.
Replace R0 with (INR (0)); Auto with real; Apply Rlt_INR;
 Repeat Rewrite <- INR_IZR_INZ; Apply lt_trans with (1); Auto with arith;
 Apply radixMoreThanOne.
Rewrite pGivesBound.
Repeat Rewrite <- INR_IZR_INZ.
Replace (INR (minus (exp radix precision) (1)))
     with (Rminus (exp radix precision) (1)).
Apply Rlt_le_trans with (INR (exp radix precision)).
Replace (Rminus (exp radix precision) (1))
     with (Rplus (exp radix precision) (Ropp (1))); [Idtac | Ring].
Pattern 1 (exp radix precision).
Replace (INR (exp radix precision)) with (Rplus (exp radix precision) R0);
 [Idtac | Ring].
Apply Rlt_compatibility.
Replace R0 with (Ropp R0); Auto with real.
Apply eq_Rle.
Apply sym_eqT; Apply powerRZ_is_exp.
Apply lt_trans with (1); Auto with arith; Apply radixMoreThanOne.
Apply sym_eqT; Apply minus_INR.
Replace (1) with (exp radix (0)).
Apply exp_monotone_le.
Apply radixMoreThanOne.
Apply lt_le_weak.
Apply lt_trans with (1); Auto with arith.
Unfold exp; Auto with arith.
Replace R0 with (INR (0)); Auto with real; Apply NEq_INR; Apply notEqLt;
 Apply lt_trans with (1); Auto with arith; Apply radixMoreThanOne.
Qed.

Theorem normal2:
  (x, y:float) (Fnormal radix b precision x) -> (Fbounded b y) ->
  (Rlt
     (Rmult (Rabsolu y) (Float (1) (Fexp x)))
     (Rmult radix (Rmult (Rabsolu x) (Float (1) (Fexp y))))).
Intros.
Apply Rle_lt_trans with (Rmult (Float (vNum b) (Fexp y)) (Float (1) (Fexp x))).
Apply Rle_monotony_r.
Unfold FtoRradix; Unfold FtoR; Simpl.
Replace (Rmult R1 (powerRZ radix (Fexp x))) with (powerRZ radix (Fexp x));
 [Idtac | Ring].
Apply powerRZ_le; Auto with real arith.
Unfold FtoRradix; Apply MaxFloat with precision := precision; Auto.
Replace (Rmult (Float (vNum b) (Fexp y)) (Float (1) (Fexp x)))
     with (Rmult (Rmult (vNum b) (Float (1) (Fexp x))) (Float (1) (Fexp y))).
Replace (Rmult radix (Rmult (Rabsolu x) (Float (1) (Fexp y))))
     with (Rmult (Rmult radix (Rabsolu x)) (Float (1) (Fexp y))); [Idtac | Ring].
Apply Rlt_monotony_r.
Unfold FtoRradix; Unfold FtoR; Simpl.
Replace (Rmult R1 (powerRZ radix (Fexp y))) with (powerRZ radix (Fexp y));
 [Idtac | Ring].
Apply powerRZ_lt; Auto with real arith.
Replace (Rmult (vNum b) (Float (1) (Fexp x)))
     with (FtoRradix (Float (vNum b) (Fexp x))).
Replace (FtoRradix (Float (vNum b) (Fexp x)))
     with (Rmult radix (Rmult (Rinv radix) (Float (vNum b) (Fexp x)))).
Apply Rlt_monotony; Auto with real arith.
Replace (Rmult (Rinv radix) (Float (vNum b) (Fexp x)))
     with (FtoRradix (Float (vNum b) (Zpred (Fexp x)))).
Unfold FtoRradix.
Replace (Rabsolu (FtoR radix x)) with (FtoR radix (Fabs x)).
Apply FnormalBoundAbs with precision := precision; Auto with real arith.
Apply Fabs_correct; Auto with real arith.
Unfold FtoRradix; Unfold FtoR; Simpl; Ring.
Pattern 2 (Fexp x).
Replace (Fexp x) with (Zs (Zpred (Fexp x))).
Rewrite powerRZ_Zs; Auto with real arith.
Repeat Rewrite <- INR_IZR_INZ.
Replace (Rmult
           (Rmult radix (powerRZ radix (Zpred (Fexp x))))
           (Rmult (vNum b) (Rinv radix)))
     with (Rmult
             (Rmult (vNum b) (powerRZ radix (Zpred (Fexp x))))
             (Rmult (Rinv radix) radix)); [Idtac | Ring].
Replace (Rmult (Rinv radix) radix) with R1; Ring.
Apply sym_eqT; Apply Rinv_r; Auto with real arith.
Apply sym_eq; Apply Zs_pred.
Ring.
Replace (Rmult (Rinv radix) radix) with R1; Ring.
Apply sym_eqT; Apply Rinv_r; Auto with real arith.
Unfold FtoRradix; Unfold FtoR; Simpl; Ring.
Repeat Rewrite <- INR_IZR_INZ.
Auto with real.
Ring.
Replace (Rmult (vNum b) (Float (1) (Fexp y)))
     with (FtoRradix (Float (vNum b) (Fexp y))); Auto with real.
Unfold FtoRradix; Unfold FtoR; Simpl; Ring.
Repeat Rewrite <- INR_IZR_INZ.
Auto with real.
Qed.
End MSBProp.
