
(****************************************************************************
                                                                             
          IEEE754  :  PradixE.v                                               
                                                                             
          Laurent Thery  & Laurence Rideau                                    
                                                                             
*****************************************************************************
Proof of Malcom's test program
*)

Require Programs.
Require Export FroundPlus.
Section prog.

(* Usual hypothesis *)
Variable b:Fbound.
Variable radix:nat.
Variable precision:nat.

Local FtoRradix := (FtoR radix).
Coercion FtoRradix : float >-> R.
Hypothesis radixMoreThanOne:(lt (S O) radix).
Hypothesis precisionGreaterThanOne:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp radix precision) (1)).

(* An arbitrary rounding mode *)
Variable P:R -> float ->Prop.
Hypothesis ProundedMode:(RoundedModeP b radix P).

(* The operations + - *  and their property *)
Variable fplus:float -> float ->float.
Variable fminus:float -> float ->float.
Variable fmult:float -> float ->float.
Hypothesis fplusCorrect:
           (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
           (P (Rplus p q) (fplus p q)).
Hypothesis fminusCorrect:
           (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
           (P (Rminus p q) (fminus p q)).
Hypothesis fmultCorrect:
           (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
           (P (Rmult p q) (fmult p q)).

(* 0,1 et 2 *)
Variables f0, f1, f2:float.
Variable f0Bounded:(Fbounded b f0).
Variable f0Correct:<R> f0==O.
Variable f1Bounded:(Fbounded b f1).
Variable f1Correct:<R> f1==(1).
Variable f2Bounded:(Fbounded b f2).
Variable f2Correct:<R> f2==(2).

Theorem FcanonicLePos:
  (p, q:float)
  (Fcanonic radix b precision p) ->
  (Fcanonic radix b precision q) -> (Rle R0 p) -> (Rle p q) ->
  (Zlt (Fexp p) (Fexp q)) \/ (Fexp p)=(Fexp q) /\ (Zle (Fnum p) (Fnum q)).
Intros p q H' H'0 H'1 H'2.
Case H'2; Intros H'3.
Case (FcanonicLtPos ? radixMoreThanOne b precision) with p := p q := q;
 Auto with zarith arith.
Rewrite (FcanonicUnique ? radixMoreThanOne b precision) with p := p q := q;
 Auto with zarith arith.
Qed.

Theorem boundedNorMinGivesExp:
  (x:Z)
  (p:float)
  (Fbounded b p) ->
  (Zle (Zopp (dExp b)) x) ->
  (Rle (Float (nNormMin radix precision) x) p) -> (Rle p (Float (vNum b) x)) ->
  (Fexp (Fnormalize radix b precision p))=x.
Intros x p H' H'0 H'1 H'2.
Cut (Rle R0 p); [Intros Rle1 | Idtac].
Cut (Fcanonic radix b precision (Float (vNum b) x)); [Intros Fc1 | Idtac].
Cut (Fcanonic radix b precision (Float (nNormMin radix precision) x));
 [Intros Fc2 | Idtac].
Case (FcanonicLePos (Fnormalize radix b precision p) (Float (vNum b) x));
 Try Rewrite (FnormalizeCorrect radix); Simpl; Auto.
Apply FnormalizeCanonic; Auto with arith.
Intros H'3.
Case (FcanonicLePos
        (Float (nNormMin radix precision) x) (Fnormalize radix b precision p));
 Try Rewrite (FnormalizeCorrect radix); Simpl; Auto with float.
Apply FnormalizeCanonic; Auto with arith.
Apply (LeFnumZERO radix); Simpl; Auto with zarith.
Intros H'4.
Contradict H'4; Auto with zarith.
Intros H'4; Elim H'4; Auto.
Intros H'3; Elim H'3; Auto.
Left; Repeat Split; Simpl; Auto with zarith float.
Unfold Fdigit; Simpl.
Rewrite absolu_INR; Apply (digitnNormMin ? radixMoreThanOne b precision);
 Auto with arith.
Left; Repeat Split; Simpl; Auto with zarith float.
Unfold Fdigit; Simpl.
Rewrite absolu_INR; Apply (digitVNumiSPrecision ? radixMoreThanOne b precision);
 Auto with arith.
Apply Rle_trans with 2 := H'1.
Apply (LeFnumZERO radix); Simpl; Auto with zarith.
Qed.

Theorem LtFnumZERO: (x:float) (Zlt ZERO (Fnum x)) ->(Rlt R0 x).
Intros x; Case x; Unfold FtoRradix FtoR; Simpl.
Intros Fnum Fexp H'; Replace R0 with (Rmult Fnum R0);
 [Apply Rlt_monotony | Ring]; Auto with real arith.
Replace R0 with (IZR ZERO); Auto with real.
Qed.

Theorem InBinade:
  (p, q, r:float)
  (e:Z)
  (Fbounded b p) ->
  (Fbounded b q) ->
  (P (Rplus p q) r) ->
  (Zle (Zopp (dExp b)) e) ->
  (Rle (Float (exp radix (pred precision)) e) p) ->
  (Rle p (Float (vNum b) e)) -> (Rlt O q) -> (Rlt q (powerRZ radix e)) ->
  <R> r==p \/ <R> r==(Rplus p (powerRZ radix e)).
Intros p q r e H' H'0 H'1 H'2 H'3 H'4 H'5 H'6.
Cut (Rlt p (Rplus p q)); [Intros Rlt1 | Idtac].
Cut (Rlt (Rplus p q) (FNSucc b radix precision p)); [Intros Rlt2 | Idtac].
Cut (isMin b radix (Rplus p q) p); [Intros Min1 | Idtac].
Cut (isMax b radix (Rplus p q) (FNSucc b radix precision p));
 [Intros Max1 | Idtac].
Cut (MinOrMaxP b radix P); [Intros MinOrMax | Intuition].
Case (MinOrMax (Rplus p q) r); Auto; Intros H1.
Left.
Apply (MinUniqueP b radix (Rplus p q)); Auto.
Right.
Cut (Rplus p (powerRZ radix e))==(FNSucc b radix precision p);
 [Intros Eq1; Rewrite Eq1 | Idtac].
Apply (MaxUniqueP b radix (Rplus p q)); Auto.
Replace (FtoRradix (FNSucc b radix precision p))
     with (Rplus
             (Fnormalize radix b precision p)
             (Rminus
                (FNSucc b radix precision p) (Fnormalize radix b precision p)));
 [Idtac | Ring].
Unfold FNSucc; Rewrite <- (Fminus_correct radix); Auto with arith;
 Rewrite (FSuccDiff1 b radix precision); Auto with arith.
Rewrite (boundedNorMinGivesExp e); Auto with zarith.
Rewrite (FnormalizeCorrect radix); Auto; Unfold FtoRradix FtoR; Simpl; Ring.
Apply sym_not_equal; Apply Zlt_not_eq.
Apply Zle_lt_trans with ZERO; Auto with zarith.
Apply (LtR0Fnum radix); Auto.
Rewrite FnormalizeCorrect; Fold FtoRradix; Auto.
Apply Rlt_le_trans with 2 := H'3.
Apply LtFnumZERO; Simpl; (Replace ZERO with (INZ O); Auto with zarith arith).
Apply MinMax; Auto with arith.
Contradict Rlt1.
Rewrite Rlt1; Auto with real.
Apply MinBinade with precision := precision; Auto with arith.
Apply Rlt_le; Auto.
Replace (FtoRradix (FNSucc b radix precision p))
     with (Rplus
             (Fnormalize radix b precision p)
             (Rminus
                (FNSucc b radix precision p) (Fnormalize radix b precision p)));
 [Idtac | Ring].
Unfold FNSucc; Rewrite <- (Fminus_correct radix); Auto with arith;
 Rewrite (FSuccDiff1 b radix precision); Auto with arith.
Rewrite (boundedNorMinGivesExp e); Auto with zarith.
Rewrite (FnormalizeCorrect radix); Auto; Fold FtoRradix.
Replace (FtoRradix (Float (1) e)) with (powerRZ radix e); Auto with real.
Apply Rlt_compatibility; Auto.
Unfold FtoRradix FtoR nNormMin; Simpl; Ring.
Apply sym_not_equal; Apply Zlt_not_eq.
Apply Zle_lt_trans with ZERO; Auto with zarith.
Apply (LtR0Fnum radix); Auto.
Rewrite FnormalizeCorrect; Fold FtoRradix; Auto.
Apply Rlt_le_trans with 2 := H'3.
Apply LtFnumZERO; Simpl; (Replace ZERO with (INZ O); Auto with zarith arith).
Pattern 1 (FtoRradix p); Replace (FtoRradix p) with (Rplus p R0); [Idtac | Ring].
Apply Rlt_compatibility; Auto.
Qed.
(* Properties to conclude for the first loop *)

Theorem Loop0:
  (p:float) (Fbounded b p) ->
  (n:nat) (le (1) n) -> (le n (vNum b)) -> <R> p==n ->
  <R> (fplus p f1)==(Rplus p f1).
Intros p H' n H'0 H'1 H'2.
Case (FboundNext ? radixMoreThanOne b precision) with p := (Float n O);
 Auto with arith; Fold FtoRradix.
Repeat Split; Simpl; Auto with zarith arith.
Intros x H'3; Elim H'3; Intros H'4 H'5; Clear H'3.
Cut (Rplus p f1)==(FtoRradix x); [Intros Eq1; Rewrite Eq1 | Idtac].
Apply sym_eqT; Apply (RoundedModeProjectorIdemEq b radix precision) with P := P;
 Auto; Fold FtoRradix.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Apply Cp with 1 := (fplusCorrect ? ? H' f1Bounded); Auto.
Apply RoundedModeBounded with radix := radix P := P r := (Rplus p f1); Auto.
Rewrite H'5; Rewrite H'2; Rewrite f1Correct; Unfold FtoRradix FtoR Zs; Simpl;
 Rewrite plus_IZR; Simpl; Repeat Rewrite <- INR_IZR_INZ; Ring.
Qed.

Theorem Loop1:
  (p:float) (Fbounded b p) ->
  (n:nat) (lt (vNum b) n) -> (le n (mult radix (vNum b))) -> <R> p==n ->
  <R> (fplus p f1)==p \/ <R> (fplus p f1)==(Rplus p radix).
Intros p H' n H'0 H'1 H'2.
Replace (INR radix) with (powerRZ radix (1)); [Idtac | Simpl; Ring].
Apply InBinade with q := f1; Auto with zarith.
Rewrite H'2;
 Replace (FtoRradix (Float (exp radix (pred precision)) (1)))
      with (INR (exp radix (S (pred precision))));
 [Apply Rle_INR |
   Unfold FtoRradix FtoR; Simpl; Rewrite Rmult_INR;
    Repeat Rewrite <- INR_IZR_INZ; Ring].
Replace (S (pred precision)) with precision;
 [Generalize H'0; Rewrite pGivesBound; Case (exp radix precision); Simpl;
   Auto with arith; Intros n1; Rewrite <- minus_n_O; Auto with arith |
   Generalize precisionGreaterThanOne; Case precision; Simpl; Auto with arith;
    Intros tmp; Inversion tmp].
Rewrite H'2;
 Replace (FtoRradix (Float (vNum b) (1))) with (INR (mult radix (vNum b)));
 Auto with real arith; Unfold FtoRradix FtoR; Simpl; Rewrite Rmult_INR;
 Repeat Rewrite <- INR_IZR_INZ; Ring.
Rewrite f1Correct; Auto with real arith.
Rewrite f1Correct; Replace (powerRZ radix (1)) with (INR radix);
 Auto with real arith.
Qed.

Theorem Loop2:
  (p:float) (Fbounded b p) ->
  (n:nat) (le (1) n) -> (le n (vNum b)) -> <R> p==n ->
  <R> (fminus (fplus p f1) p)==f1.
Intros p H' n H'0 H'1 H'2.
Apply sym_eqT; Apply (RoundedModeProjectorIdemEq b radix precision) with P := P;
 Auto.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Cut (Fbounded b (fplus p f1)); [Intros Fbp | Auto].
Apply Cp with 1 := (fminusCorrect ? ? Fbp H'); Auto.
Rewrite Loop0 with n := n; Auto.
Fold FtoRradix; Ring.
Apply RoundedModeBounded with radix := radix P := P r := (Rminus (fplus p f1) p);
 Auto.
Apply RoundedModeBounded with radix := radix P := P r := (Rplus p f1); Auto.
Qed.

Theorem Loop3:
  (p:float) (Fbounded b p) ->
  (n:nat) (lt (vNum b) n) -> (le n (mult radix (vNum b))) -> <R> p==n ->
  ~ <R> (fminus (fplus p f1) p)==f1.
Intros p H' n H'0 H'1 H'2.
Case (Loop1 p) with n := n; Auto; Intros Eq1; Red; Intros Eq2.
Absurd <R> f1==f0.
Rewrite f1Correct; Rewrite f0Correct; Auto with real.
Rewrite <- Eq2.
Apply sym_eqT; Apply (RoundedModeProjectorIdemEq b radix precision) with P := P;
 Auto.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Cut (Fbounded b (fplus p f1)); [Intros Fbp | Auto].
Apply Cp with 1 := (fminusCorrect ? ? Fbp H'); Auto.
Rewrite Eq1; Fold FtoRradix; Rewrite f0Correct; Ring.
Apply RoundedModeBounded with radix := radix P := P r := (Rminus (fplus p f1) p);
 Auto.
Apply RoundedModeBounded with radix := radix P := P r := (Rplus p f1); Auto.
Absurd <R> f1==radix.
Rewrite f1Correct; Apply NEq_INR; Contradict radixMoreThanOne;
 Rewrite radixMoreThanOne; Auto with arith.
Rewrite <- Eq2.
Replace (INR radix) with (FtoRradix (Float (1) (1))).
Apply sym_eqT; Apply (RoundedModeProjectorIdemEq b radix precision) with P := P;
 Auto.
Repeat Split; Simpl; Auto with zarith.
Replace (POS xH) with (inject_nat (1)); Auto; Apply inj_le.
Apply lt_le_S; Apply (lt_vNum radix) with precision := precision;
 Auto with arith.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Cut (Fbounded b (fplus p f1)); [Intros Fbp | Auto].
Apply Cp with 1 := (fminusCorrect ? ? Fbp H'); Auto.
Rewrite Eq1; Unfold FtoR; Simpl; Ring.
Apply RoundedModeBounded with radix := radix P := P r := (Rminus (fplus p f1) p);
 Auto.
Apply RoundedModeBounded with radix := radix P := P r := (Rplus p f1); Auto.
Unfold FtoRradix FtoR; Simpl; Ring.
Qed.

Theorem Loop4:
  (p:float) (Fbounded b p) ->
  (n:nat) (le (1) n) -> (le n (vNum b)) -> <R> p==n ->(Rlt p (fmult f2 p)).
Intros p H' n H'0 H'1 H'2.
Case (le_or_lt (mult (2) n) (vNum b)); Intros le1.
Replace (FtoRradix (fmult f2 p)) with (FtoRradix (Float (mult (2) n) O)).
Rewrite H'2; Unfold FtoRradix FtoR; Simpl.
Apply Rlt_anti_compatibility with r := (Ropp n); Auto.
Replace (Rplus (Ropp n) n) with (INR O); [Idtac | Simpl; Ring].
Repeat Rewrite <- INR_IZR_INZ.
Replace (Rplus (Ropp n) (Rmult (plus n (plus n O)) R1)) with (INR n);
 [Auto with real arith | Repeat Rewrite plus_INR; Simpl; Ring].
Apply (RoundedModeProjectorIdemEq b radix precision) with P := P; Auto.
Repeat Split; Simpl; Auto with zarith.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Apply Cp with 1 := (fmultCorrect ? ? f2Bounded H'); Auto.
Rewrite H'2; Rewrite f2Correct; Unfold FtoRradix FtoR; Simpl;
 Repeat Rewrite <- INR_IZR_INZ; Repeat Rewrite plus_INR; Simpl; Ring.
Apply RoundedModeBounded with radix := radix P := P r := (Rmult f2 p); Auto.
Cut (FtoRradix (Float (nNormMin radix precision) (1)))==(exp radix precision);
 [Intros Eq1 | Unfold FtoRradix FtoR nNormMin; Simpl].
Apply Rlt_le_trans with (FtoRradix (Float (nNormMin radix precision) (1))).
Rewrite H'2; Rewrite Eq1; Simpl.
Apply Rlt_INR.
Apply le_lt_trans with 1 := H'1; Rewrite pGivesBound.
Cut (lt O (exp radix precision)); Auto with arith.
Apply (RleBoundRoundl b radix precision) with P := P r := (Rmult f2 p); Auto.
Repeat Split; Simpl; Auto with zarith arith.
Apply inj_le; Apply nNrMMimLevNum; Auto with arith.
Fold FtoRradix; Rewrite Eq1.
Rewrite f2Correct; Rewrite H'2; Rewrite <- Rmult_INR.
Apply Rle_INR.
Generalize le1; Rewrite pGivesBound; Cut (lt O (exp radix precision));
 Auto with arith.
Case (exp radix precision); Simpl; Auto with arith; Intros n0;
 Rewrite <- minus_n_O; Auto with arith.
Replace (exp radix precision) with (exp radix (S (pred precision))).
Simpl; Rewrite Rmult_INR; Repeat Rewrite <- INR_IZR_INZ; Ring.
Generalize precisionGreaterThanOne; Case precision; Simpl; Auto with arith.
Intros H'3; Inversion H'3; Auto.
Qed.
(* Properties to conclude for the second loop *)

Theorem BLoop1:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
  (m, n:nat)
  (le (exp radix precision) m) ->
  (le m (mult radix (vNum b))) ->
  <R> p==m -> (le (1) n) -> (lt n radix) -> <R> q==n ->
  <R> (fplus p q)==p \/ <R> (fplus p q)==(Rplus p radix).
Intros p q H' H'0 m n H'1 H'2 H'3 H'4 H'5 H'6.
Replace (INR radix) with (powerRZ radix (1)); [Idtac | Simpl; Ring].
Apply InBinade with q := q; Auto with zarith.
Rewrite H'3;
 Replace (FtoRradix (Float (exp radix (pred precision)) (1)))
      with (INR (exp radix (S (pred precision))));
 [Apply Rle_INR |
   Unfold FtoRradix FtoR; Simpl; Rewrite Rmult_INR;
    Repeat Rewrite <- INR_IZR_INZ; Ring].
Replace (S (pred precision)) with precision;
 [Auto with arith |
   Generalize precisionGreaterThanOne; Case precision; Simpl; Auto with arith;
    Intros tmp; Inversion tmp].
Rewrite H'3;
 Replace (FtoRradix (Float (vNum b) (1))) with (INR (mult radix (vNum b)));
 Auto with real arith; Unfold FtoRradix FtoR; Simpl; Rewrite Rmult_INR;
 Repeat Rewrite <- INR_IZR_INZ; Ring.
Rewrite H'6; Auto with real.
Rewrite H'6; Replace (powerRZ radix (1)) with (INR radix); Auto with real arith.
Qed.

Theorem BLoop2:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
  (m, n:nat)
  (le (exp radix precision) m) ->
  (le m (mult radix (vNum b))) ->
  <R> p==m -> (le (1) n) -> (lt n radix) -> <R> q==n ->
  ~ <R> (fminus (fplus p q) p)==q.
Intros p q H' H'0 m n H'1 H'2 H'3 H'4 H'5 H'6.
Case (BLoop1 p q) with n := n m := m; Auto; Intros Eq1; Red; Intros Eq2.
Absurd <R> q==f0.
Rewrite H'6; Rewrite f0Correct.
Apply imp_not_Req.
Right; Red; Apply Rlt_INR.
Apply lt_le_trans with 2 := H'4; Auto.
Rewrite <- Eq2.
Apply sym_eqT; Apply (RoundedModeProjectorIdemEq b radix precision) with P := P;
 Auto.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Cut (Fbounded b (fplus p q)); [Intros Fbp | Auto].
Apply Cp with 1 := (fminusCorrect ? ? Fbp H'); Auto.
Rewrite Eq1; Fold FtoRradix; Rewrite f0Correct; Ring.
Apply RoundedModeBounded with radix := radix P := P r := (Rminus (fplus p q) p);
 Auto.
Apply RoundedModeBounded with radix := radix P := P r := (Rplus p q); Auto.
Absurd <R> q==radix.
Rewrite H'6; Apply NEq_INR; Auto with arith.
Contradict H'5; Rewrite H'5; Auto with arith.
Rewrite <- Eq2.
Replace (INR radix) with (FtoRradix (Float (1) (1))).
Apply sym_eqT; Apply (RoundedModeProjectorIdemEq b radix precision) with P := P;
 Auto.
Repeat Split; Simpl; Auto with zarith.
Replace (POS xH) with (inject_nat (1)); Auto; Apply inj_le.
Apply lt_le_S; Apply (lt_vNum radix) with precision := precision;
 Auto with arith.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Cut (Fbounded b (fplus p q)); [Intros Fbp | Auto].
Apply Cp with 1 := (fminusCorrect ? ? Fbp H'); Auto.
Rewrite Eq1; Unfold FtoR; Simpl; Ring.
Apply RoundedModeBounded with radix := radix P := P r := (Rminus (fplus p q) p);
 Auto.
Apply RoundedModeBounded with radix := radix P := P r := (Rplus p q); Auto.
Unfold FtoRradix FtoR; Simpl; Ring.
Qed.

Theorem BLoop3:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
  (m:nat)
  (le (exp radix precision) m) ->
  (le m (mult radix (vNum b))) -> <R> p==m -> <R> q==radix ->
  <R> (fminus (fplus p q) p)==q.
Intros p q H' H'0 m H'1 H'2 H'3 H'4.
Cut <R> (fplus p q)==(Rplus p q); [Intros Eq1 | Idtac].
Apply sym_eqT; Apply (RoundedModeProjectorIdemEq b radix precision) with P := P;
 Auto.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Cut (Fbounded b (fplus p q)); [Intros Fbp | Auto].
Apply Cp with 1 := (fminusCorrect ? ? Fbp H'); Auto; Fold FtoRradix.
Rewrite Eq1; Ring.
Apply RoundedModeBounded with radix := radix P := P r := (Rminus (fplus p q) p);
 Auto.
Apply RoundedModeBounded with radix := radix P := P r := (Rplus p q); Auto.
Case (FboundNext ? radixMoreThanOne b precision) with p :=
                                                      (Fnormalize
                                                         radix b precision p);
 Auto with arith; Fold FtoRradix.
Apply FcanonicBound with radix := radix precision := precision; Auto.
Apply FnormalizeCanonic with radix := radix precision := precision;
 Auto with arith.
Intros p' H'5; Elim H'5; Intros H'6 H'7; Clear H'5.
Cut (Fexp (Fnormalize radix b precision p))=(1); [Intros Eq1 | Idtac].
Cut <R> (Rplus p q)==(FtoRradix p'); [Intros Eq2 | Idtac].
Rewrite Eq2.
Apply sym_eqT; Apply (RoundedModeProjectorIdemEq b radix precision) with P := P;
 Auto.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Apply Cp with 1 := (fplusCorrect ? ? H' H'0); Auto; Fold FtoRradix.
Apply RoundedModeBounded with radix := radix P := P r := (Rplus p q); Auto.
Rewrite H'7.
Rewrite H'4.
Unfold FtoRradix FtoR; Simpl.
Replace (Rmult
           (Zs (Fnum (Fnormalize radix b precision p)))
           (powerRZ radix (Fexp (Fnormalize radix b precision p))))
     with (Rplus
             (Rmult
                (Fnum (Fnormalize radix b precision p))
                (powerRZ radix (Fexp (Fnormalize radix b precision p))))
             (powerRZ radix (Fexp (Fnormalize radix b precision p)))).
Replace (Rmult
           (Fnum (Fnormalize radix b precision p))
           (powerRZ radix (Fexp (Fnormalize radix b precision p))))
     with (FtoRradix (Fnormalize radix b precision p)); Auto.
Rewrite (FnormalizeCorrect radix); Auto.
Rewrite Eq1; Unfold FtoR; Simpl; Ring.
Unfold Zs; Simpl; Rewrite plus_IZR; Ring.
Apply boundedNorMinGivesExp; Auto with zarith.
Rewrite H'3; Unfold FtoRradix FtoR nNormMin; Simpl.
Repeat Rewrite <- INR_IZR_INZ;
 Replace (Rmult (exp radix (pred precision)) (Rmult radix R1))
      with (Rmult radix (exp radix (pred precision))); [Idtac | Ring].
Rewrite <- Rmult_INR; Apply Rle_INR.
Generalize H'1 precisionGreaterThanOne; Case precision; Simpl; Auto with arith.
Intros H'5 H'8; Inversion H'8; Auto.
Rewrite H'3; Unfold FtoRradix FtoR nNormMin; Simpl.
Repeat Rewrite <- INR_IZR_INZ;
 Replace (Rmult (vNum b) (Rmult radix R1)) with (Rmult radix (vNum b));
 [Idtac | Ring].
Rewrite <- Rmult_INR; Apply Rle_INR; Auto.
Qed.

Theorem BLoop4:
  (p:float) (Fbounded b p) ->
  (n:nat) (le (1) n) -> (lt n radix) -> <R> p==n ->(Rlt p (fplus p f1)).
Intros p H' n H'0 H'1 H'2.
Rewrite Loop0 with n := n; Auto.
Rewrite H'2; Rewrite f1Correct; Rewrite <- plus_INR; Apply Rlt_INR;
 Auto with zarith.
Apply le_trans with radix; Auto with arith.
Rewrite pGivesBound.
Pattern 1 radix; Replace radix with (minus (S (exp radix (1))) (1));
 [Idtac | Simpl; Rewrite <- minus_n_O; Ring].
Apply minus_le; Auto with arith.
Qed.

Lemma Loop6c:
  (p, b0:float) (Fbounded b p) ->
  (n:nat)
  (le (1) n) ->
  (le n (vNum b)) -> <R> p==n -> (Fbounded b b0) /\ <R> b0==radix ->
  (Rlt p (fmult p b0)).
Intros p b0 H' n H'0 H'1 H'2 H'3.
Casec H'3; Intros Fbb0 H'3.
Case (le_or_lt (mult n radix) (vNum b)); Intros le1.
Replace (FtoRradix (fmult p b0)) with (FtoRradix (Float (mult n radix) O)).
Rewrite H'2; Unfold FtoRradix FtoR; Simpl.
Pattern 1 n; Replace n with (mult n (1)); Auto with real arith.
Repeat Rewrite <- INR_IZR_INZ.
Replace (Rmult (mult n radix) R1) with (INR (mult n radix));
 Auto with real arith.
Apply (RoundedModeProjectorIdemEq b radix precision) with P := P; Auto.
Repeat Split; Simpl; Auto with zarith.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Apply Cp with 1 := (fmultCorrect ? ? H' Fbb0); Auto.
Rewrite H'3; Rewrite H'2; Unfold FtoRradix FtoR; Simpl;
 Repeat Rewrite <- INR_IZR_INZ; Simpl; Ring.
Rewrite (Rmult_sym radix n); Auto with arith real.
Apply RoundedModeBounded with radix := radix P := P r := (Rmult p b0); Auto.
Cut (FtoRradix (Float (nNormMin radix precision) (1)))==(exp radix precision);
 [Intros Eq1 | Unfold FtoRradix FtoR nNormMin; Simpl].
Apply Rlt_le_trans with (FtoRradix (Float (nNormMin radix precision) (1))).
Rewrite H'2; Rewrite Eq1; Simpl.
Apply Rlt_INR.
Apply le_lt_trans with 1 := H'1; Rewrite pGivesBound.
Cut (lt O (exp radix precision)); Auto with arith.
Apply (RleBoundRoundl b radix precision) with P := P r := (Rmult p b0); Auto.
Repeat Split; Simpl; Auto with zarith arith.
Apply inj_le; Apply nNrMMimLevNum; Auto with arith.
Fold FtoRradix; Rewrite Eq1.
Rewrite H'3; Rewrite H'2; Rewrite <- Rmult_INR.
Apply Rle_INR.
Generalize le1; Rewrite pGivesBound; Cut (lt O (exp radix precision));
 Auto with arith.
Case (exp radix precision); Simpl; Auto with arith; Intros n0;
 Rewrite <- minus_n_O; Auto with arith.
Replace (exp radix precision) with (exp radix (S (pred precision))).
Simpl; Rewrite Rmult_INR; Repeat Rewrite <- INR_IZR_INZ; Ring.
Generalize precisionGreaterThanOne; Case precision; Simpl; Auto with arith.
Intros H'4; Inversion H'4; Auto.
Qed.

Local feq := (Feq_bool radix).

Correctness i_exp
  let x = ref f1 in
  let y = ref f1 in
  let n = ref O in
  begin
    while (feq (fminus (fplus !x f1) !x) f1) do
        { invariant (Ex [m:nat] ((le (1) m)/\ (le m (mult radix (vNum b)))/\
                                <R> x == m)) /\ (Fbounded b x)
          variant (minus (mult radix (vNum b)) (absolu (Int_part x))) for lt }
        x := (fmult f2 !x)
      done
     {(Ex [m:nat] ((le (exp radix precision) m)/\ (le m (mult radix (vNum b)))/\
                                <R> x == m)) /\ (Fbounded b x)
     }
    ;
    while (not (feq (fminus (fplus !x !y) !x) !y)) do
        { invariant  (Ex [m:nat] (le (1) m)/\ (le m radix) /\
                                <R> y == m) /\ (Fbounded b y)
          variant (minus radix (absolu (Int_part y))) for lt }
        y := (fplus !y f1)
      done {(Fbounded b y)/\<R> y == radix}; 
    x:=f1;
    while (feq (fminus (fplus !x f1) !x) f1) do
        { invariant (Ex [m:nat] ((le (1) m)/\ (le m (mult radix (vNum b)))/\
                                <R> x == m)/\ m= (exp radix n)) /\ (Fbounded b x)
          variant (minus (mult radix (vNum b)) (absolu (Int_part x))) for lt }
        x := (fmult !x !y); n := (S !n)
      done
    end
    { <R> y == radix /\ n = precision }
.
(* subgoal 1*)
Casec Pre2; Intros Pre2; Casec Pre2; Intros vx0 Pre2; Casec Pre2;
 Intros Hv1 Pre2; Casec Pre2; Intros Hv2 Hv3 Fb1.
Cut (ex
       nat
       [m:nat]
       (le (1) m) /\ ((le m (mult radix (vNum b))) /\ <R> (fmult f2 x0)==m)).
Intros Pre2; Casec Pre2; Intros va1 Pre2; Casec Pre2; Intros Hv1' Pre2;
 Casec Pre2; Intros Hv2' Hv3'.
Split.
Rewrite Hv3; Rewrite Hv3'; Repeat Rewrite Int_part_INR;
 Repeat Rewrite absolu_INR.
Apply lt_minus_inv; Auto with arith.
Apply lt_Rlt.
Rewrite <- Hv3'; Rewrite <- Hv3.
Apply (Loop4 x0 Fb1 vx0); Auto.
Case (le_or_lt (exp radix precision) vx0); Auto.
Intros lte1; Absurd <R> (fminus (fplus x0 f1) x0)==f1.
Apply Loop3 with n := vx0; Auto.
Rewrite pGivesBound.
Cut (lt O (exp radix precision)); Auto with arith zarith.
Generalize Feq_bool_correct_t; Unfold Feq.
Intros H'; Apply (H' radix); Auto.
Rewrite pGivesBound.
Cut (lt O (exp radix precision)); Auto with arith zarith.
Split; Auto.
Exists va1; Split; Auto.
Apply RoundedModeBounded with radix := radix P := P r := (Rmult f2 x0); Auto.
Case (NroundN b radix precision) with P := P n := (mult (2) vx0)
                                      p := (fmult f2 x0); Auto.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Apply Cp with 1 := (fmultCorrect ? ? f2Bounded Fb1); Auto.
Rewrite Hv3; Rewrite f2Correct; Auto with real arith.
Apply RoundedModeBounded with radix := radix P := P r := (Rmult f2 x0); Auto.
Intros x H'; Exists x; Repeat Split; Auto.
Apply le_Rle.
Rewrite <- f1Correct; Auto.
Rewrite <- H'; Auto.
Apply (RleBoundRoundl b radix precision) with P := P r := (Rmult f2 x0); Auto.
Rewrite Hv3; Rewrite f2Correct.
Fold FtoRradix; Rewrite f1Correct.
Rewrite <- Rmult_INR.
Apply Rle_INR; Generalize Hv1; Case vx0; Simpl; Auto with arith.
Apply le_Rle.
Cut <R> (Float (vNum b) (1))==(mult radix (vNum b)); [Intros Eq1 | Idtac].
Rewrite <- Eq1; Rewrite <- H'.
Apply (RleBoundRoundr b radix precision) with P := P r := (Rmult f2 x0); Auto.
Repeat Split; Simpl; Auto with zarith.
Fold FtoRradix; Rewrite Eq1.
Repeat Rewrite <- INR_IZR_INZ; Rewrite Rmult_INR; Auto.
Apply Rle_Rmult_comp; Auto.
Rewrite f2Correct; Replace R0 with (INR O); Auto with real arith.
Rewrite Hv3; Replace R0 with (INR O); Auto with real arith.
Rewrite f2Correct; Auto with real arith.
Rewrite Hv3.
Case (le_or_lt vx0 (vNum b)); Auto with real arith.
Intros H'0; Absurd <R> (fminus (fplus x0 f1) x0)==f1.
Apply Loop3 with n := vx0; Auto.
Generalize (Feq_bool_correct_t radix); Unfold Feq; Intros Eq2; Apply Eq2; Auto.
Unfold FtoRradix FtoR; Simpl; Repeat Rewrite <- INR_IZR_INZ; Rewrite Rmult_INR;
 Ring.

(* subgoal 2*)
Casec Pre2; Intros Pre2; Casec Pre2; Intros vx0 Pre2; Casec Pre2;
 Intros Hv1 Pre2; Casec Pre2; Intros Hv2 Hv3 Fb1.
Split; Auto; Auto.
Exists vx0; Split; Auto.
Case (le_or_lt (exp radix precision) vx0); Auto with real arith.
Intros H'0; Absurd <R> (fminus (fplus x0 f1) x0)==f1.
Generalize (Feq_bool_correct_f radix); Unfold Feq; Intros Eq2; Apply Eq2; Auto.
Apply Loop2 with n := vx0; Auto.
Rewrite pGivesBound.
Cut (lt O (exp radix precision)); Auto with arith zarith.

(* subgoal 3*)
Split; Auto.
Exists (1); Repeat Split; Auto.
Replace (1) with (mult (1) (1)); [Apply lte_comp_mult | Simpl]; Auto with arith.
Apply lt_le_S; Auto.
Apply lt_vNum with radix := radix precision := precision; Auto with arith.

(* subgoal 4*)
Elim Post3; Intros H'0 H'1; Elim H'0; Intros vx0 E; Elim E; Intros Hv1 H'3;
 Elim H'3; Intros Hv2 Hv3; Clear H'3 E H'0 Post3.
Elim Pre4; Intros H' Hv1'; Elim H'; Intros vy0 E; Elim E; Intros Hv2' H'3;
 Elim H'3; Intros Hv3' Hv4'; Clear H'3 E H' Pre4.
Cut (ex nat [m:nat](le (1) m) /\ ((le m radix) /\ <R> (fplus y0 f1)==m));
 [Intros Ex1 | Idtac].
Split; [Idtac | Split]; Auto.
Elim Ex1; Intros vb1 E; Elim E; Intros Hv1'' H'0; Elim H'0; Intros Hv2'' Hv3'';
 Clear H'0 E Ex1.
Rewrite Hv4'; Rewrite Hv3''; Repeat Rewrite Int_part_INR;
 Repeat Rewrite absolu_INR.
Apply lt_minus_inv; Auto with arith.
Apply lt_Rlt.
Rewrite <- Hv4'; Rewrite <- Hv3''.
Apply (BLoop4 y0 Hv1' vy0); Auto.
Case (le_or_lt radix vy0); Auto; Intros le1.
Absurd <R> (fminus (fplus x0 y0) x0)==y0; Auto.
Generalize (Feq_bool_correct_f radix); Unfold Feq; Intros Eq2; Apply Eq2; Auto.
Apply BLoop3 with m := vx0; Auto with real arith.
(Rewrite Hv4'; Apply Rle_antisym); Auto with real arith.
Apply RoundedModeBounded with radix := radix P := P r := (Rplus y0 f1); Auto.
Case (le_lt_or_eq ? ? Hv3'); Intros le1.
Case (NroundN b radix precision) with P := P n := (S vy0) p := (fplus y0 f1);
 Auto.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Apply Cp with 1 := (fplusCorrect ? ? Hv1' f1Bounded); Auto.
Rewrite Hv4'.
Rewrite f1Correct; Simpl; Case vy0; Auto; Ring.
Apply RoundedModeBounded with radix := radix P := P r := (Rplus y0 f1); Auto.
Intros x H'; Exists x; Repeat Split; Auto.
Apply le_Rle.
Rewrite <- H'; Rewrite <- f1Correct; Auto.
Apply (RleBoundRoundl b radix precision) with P := P r := (Rplus y0 f1); Auto.
Fold FtoRradix; Rewrite Hv4'; Rewrite f1Correct; Auto.
Rewrite <- plus_INR; Auto with real arith.
Apply le_Rle.
Rewrite <- H'.
Replace (INR radix) with (FtoRradix (Float (1) (1))).
Apply (RleBoundRoundr b radix precision) with P := P r := (Rplus y0 f1); Auto.
Repeat Split; Simpl; Auto with zarith.
Change (Zle (1) (vNum b)); Apply inj_le.
Apply lt_le_S; Auto.
Apply lt_vNum with radix := radix precision := precision; Auto with arith.
Rewrite Hv4'; Rewrite f1Correct; Unfold FtoR; Simpl; Rewrite Rmult_ne_r;
 Rewrite Rmult_ne_l.
Replace R1 with (INR (1)); Auto with real arith; Rewrite <- plus_INR;
 Apply Rle_INR; Rewrite plus_sym; Simpl; Auto with arith.
Unfold FtoRradix FtoR; Simpl; Rewrite Rmult_ne_r; Rewrite Rmult_ne_l; Auto.
Absurd <R> (fminus (fplus x0 y0) x0)==y0.
Generalize (Feq_bool_correct_f radix); Unfold Feq; Intros Eq2; Apply Eq2; Auto.
Apply BLoop3 with m := vx0; Auto with real arith.
Rewrite Hv4'; Rewrite le1; Auto with real.

(* subgoal 5*)
Elim Post3; Intros H' Hv1; Elim H';
Intros va0 E;Elim E;Intros Hv3 H'4; Elim H'4;Intros Hv4 Hv5;Clear H'4 E H' Post3.  

Elim Pre4; Intros H' Hv1'; Elim H'; Intros vb0 E; Elim E; Intros Hv2' H'2;
 Elim H'2; Intros Hv3' Hv4'; Clear H'2 E H' Pre4.
Rewrite Hv4'.
Case (le_lt_or_eq ? ? Hv3'); Intros le1.
Absurd <R> (fminus (fplus x0 y0) x0)==y0.
Apply BLoop2 with n:= vb0 m:= va0;Auto.
Generalize (Feq_bool_correct_t radix); Unfold Feq; Intros Eq2; Apply Eq2; Auto.
Rewrite le1; Auto.

(* subgoal 6*)
Split; [Exists (1); Split; [Idtac | Split] | Idtac]; Auto with arith.

(* subgoal 7*)
Casec Pre6; Intros Pre6; Casec Pre6; Intros va0 Pre6; Casec Pre6;
 Intros Pre6 Hv1'; Casec Pre6; Intros Hv1 Pre6; Casec Pre6; Intros Hv2 Hv3 Fb1.
Casec Post2; Intros Fbb0 Post2.
Cut (ex
       nat
       [m:nat]
       ((le (1) m) /\ ((le m (mult radix (vNum b))) /\ <R> (fmult x2 y0)==m)) /\
       <nat> m=(exp radix (S n0))).
Intros H'; Split; [Idtac | Split; [Try Assumption | Idtac]].
Casec H'; Intros va1 Pre6; Casec Pre6; Intros Pre6 H'1; Casec Pre6;
 Intros Hv1' Pre6; Casec Pre6; Intros Hv2' Hv3'.
Rewrite Hv3; Rewrite Hv3'; Repeat Rewrite Int_part_INR;
 Repeat Rewrite absolu_INR.
Apply lt_minus_inv; Auto.
Rewrite H'1; Rewrite Hv1'; Auto with arith.
(* Fbounded (fmult x2 y0)).*)
Apply RoundedModeBounded with radix := radix P := P r := (Rmult x2 y0); Auto.
Case (NroundN b radix precision) with P := P n := (mult va0 radix)
                                      p := (fmult x2 y0); Auto.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Apply Cp with 1 := (fmultCorrect ? ? Fb1 Fbb0); Auto.
Rewrite Hv3; Rewrite Post2; Auto with real arith.
Apply RoundedModeBounded with radix := radix P := P r := (Rmult x2 y0); Auto.
Intros x H'; Exists x; Repeat Split; Auto.
Apply le_Rle; Rewrite <- f1Correct; Rewrite <- H'.
Apply (RleBoundRoundl b radix precision) with P := P r := (Rmult x2 y0); Auto.
Rewrite Hv3; Rewrite Post2; Fold FtoRradix; Rewrite f1Correct.
Rewrite <- Rmult_INR.
Apply Rle_INR; Generalize Hv1; Case va0; Simpl; Auto with arith.
Apply le_Rle.
Cut <R> (Float (vNum b) (1))==(mult radix (vNum b)); [Intros Eq1 | Idtac].
Rewrite <- Eq1; Rewrite <- H'.
Apply (RleBoundRoundr b radix precision) with P := P r := (Rmult x2 y0); Auto.
Repeat Split; Simpl; Auto with zarith.
Fold FtoRradix; Rewrite Eq1.
Repeat Rewrite <- INR_IZR_INZ; Rewrite Rmult_INR; Auto with arith real.
Replace (Rmult radix (vNum b)) with (Rmult (vNum b) radix); Auto with real arith.
Apply Rle_Rmult_comp; Auto with real arith.
Rewrite Hv3; Replace R0 with (INR O); Auto with real arith.
Rewrite Post2; Replace R0 with (INR O); Auto with real arith.
Rewrite Hv3.
Case (le_or_lt va0 (vNum b)); Auto with real arith.
Intros H'0; Absurd <R> (fminus (fplus x2 f1) x2)==f1.
Apply Loop3 with n := va0; Auto.
Generalize (Feq_bool_correct_t radix); Unfold Feq; Intros Eq2; Apply Eq2; Auto.
Rewrite Post2; Auto with arith real.
Unfold FtoRradix FtoR; Simpl; Repeat Rewrite <- INR_IZR_INZ; Rewrite Rmult_INR;
 Ring.
Apply INR_inv.
Rewrite <- H'.
Cut <R> (exp radix (S n0))==(Float (Fnum x2) (Zs (Fexp x2)));
 [Intros Eq1; Rewrite Eq1 | Idtac].
Apply sym_eqT; Apply (RoundedModeProjectorIdemEq b radix precision) with P := P;
 Auto.
Repeat Split; Simpl; Auto with arith zarith.
Elim Fb1; Intros H'0; Elim H'0; Intros H'1 H'2; Auto.
Elim Fb1; Intros H'0; Elim H'0; Intros H'1 H'2; Auto.
Apply Zle_trans with (Fexp x2); Auto with zarith.
Elim Fb1; Intros H'0; Elim H'0; Intros H'1 H'2; Auto.
Cut (CompatibleP b radix P); [Intros Cp | Intuition]; Auto.
Apply Cp with 1 := (fmultCorrect ? ? Fb1 Fbb0); Auto.
Unfold FtoR; Simpl.
Rewrite powerRZ_Zs; Auto with real zarith.
Replace (Rmult (Fnum x2) (Rmult radix (powerRZ radix (Fexp x2))))
     with (Rmult (Rmult (Fnum x2) (powerRZ radix (Fexp x2))) radix);
 Auto with real zarith.
Ring.
(* Fbounded (fmult a2 b0)).*)
Apply RoundedModeBounded with radix := radix P := P r := (Rmult x2 y0); Auto.
Unfold FtoRradix FtoR; Simpl.
Rewrite <- Hv1'.
Cut <R>
      (INR (mult radix va0))==
      (Rmult radix (Rmult (Fnum x2) (powerRZ radix (Fexp x2))));
 [Intros Eq2; Rewrite Eq2 | Idtac].
Rewrite powerRZ_Zs; Auto with real arith; Ring.
Fold FtoRradix.
Rewrite (Rmult_INR radix va0).
Rewrite <- Hv3; Auto.

(* subgoal 8*)
Split; Auto.
Exists (1); Repeat Split; Auto.
Replace (1) with (mult (1) (1)); [Apply lte_comp_mult | Simpl]; Auto with arith.
Apply lt_le_S; Auto.
Apply lt_vNum with radix := radix precision := precision; Auto with arith.

(* subgoal 9*)
Casec Post2; Intros Fbb0 Post2.
Split; [Try Assumption | Idtac].
Apply exp_anti_eq with n := radix; Auto.
Casec Post1; Intros Post1 H'0; Casec Post1; Intros Post1 H'2; Casec Post1;
 Intros m E; Casec E; Intros H'3 H'4; Rewrite <- H'4; Casec H'3; Intros H' H'1;
 Casec H'1; Intros H'3 H'5.
Case (le_lt_dec m (vNum b)).
Intros H'1.
Absurd <bool> (feq (fminus (fplus x2 f1) x2) f1)=true.
Rewrite H'0; Auto with v62.
Unfold feq; Apply Feq_bool_correct_r; Auto.
LApply (Loop2 x2);
 [Intros H'7; LApply (H'7 m);
   [Intros H'9; LApply H'9;
     [Intros H'10; LApply H'10;
       [Intros H'11; Try Exact H'11; Clear H'10 H'9 | Clear H'10 H'9] |
       Clear H'9] | Idtac] | Idtac]; Auto.
Rewrite pGivesBound; Intros H'1.
Cut (le (exp radix precision) m); Intros; Auto with arith zarith.
Cut (le m (exp radix precision)); Intros; Auto with arith.
Cut (lt (exp radix precision) (exp radix (S precision))); Intros;
 Auto with arith.
Cut (lt m (exp radix (S precision))); Rewrite H'4; Intros; Auto with arith.
Cut (lt n0 (S precision)); Intros; Auto with arith.
Apply (exp_anti_monotone_lt radix); Auto.
Rewrite <- H'4; Auto.
Apply le_lt_trans with m := (minus (exp radix (S precision)) (1));
 Auto with arith zarith.
Replace (exp radix (S precision)) with (mult radix (exp radix precision));
 Auto with arith zarith.
Apply le_trans with m := (minus (mult radix (exp radix precision)) radix);
 Auto with arith zarith.
Replace (minus (mult radix (exp radix precision)) radix)
     with (mult radix (minus (exp radix precision) (1))); Auto with arith zarith.
Rewrite <- pGivesBound; Auto.
Pattern 5 radix; Replace radix with (mult (1) radix); Ring;
 Rewrite (mult_minus_distr (exp radix precision) (1) radix);
 Auto with arith zarith.
Qed.
End prog.

