
(****************************************************************************
                                                                             
          IEEE754  :  ThreeSum2                                               
                                                                             
          Laurent Thery & Sylvie Boldo                                        
                                                                             
*****************************************************************************
*)
Require Export Closest2Plus.
Section F2.
Variable b:Fbound.
Variable precision:nat.

Local FtoRradix := (FtoR (2)).
Coercion FtoRradix : float >-> R.

Theorem TwoMoreThanOne: (lt (S O) (2)).
Auto.
Qed.
Hints Resolve TwoMoreThanOne.
Hypothesis precisionNotZero:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp (2) precision) (1)).
Variables p, q, r, u, v, w, p', q', r':float.
Hypothesis Fp:(Fbounded b p).
Hypothesis Fq:(Fbounded b q).
Hypothesis Fr:(Fbounded b r).
Hypothesis Fu:(Fbounded b u).
Hypothesis Fv:(Fbounded b v).
Hypothesis Fw:(Fbounded b w).
Hypothesis Fp':(Fbounded b p').
Hypothesis Fq':(Fbounded b q').
Hypothesis Fr':(Fbounded b r').
Hypothesis epq:(Zle (Fexp q) (Fexp p)).
Hypothesis eqr:(Zle (Fexp r) (Fexp q)).
Hypothesis uDef:(Closest b (2) (Rplus q r) u).
Hypothesis vDef:<R> v==(Rminus (Rplus q r) u).
Hypothesis p'Def:(Closest b (2) (Rplus p u) p').
Hypothesis wDef:<R> w==(Rminus (Rplus p u) p').
Hypothesis q'Def:(Closest b (2) (Rplus w v) q').
Hypothesis r'Def:<R> r'==(Rminus (Rplus w v) q').

Theorem vNumPos: (Rle R0 (vNum b)).
Apply Rlt_le.
Replace R0 with (INR (0)).
Apply Rlt_INR.
Apply lt_vNum with precision := precision radix := (2); Auto with real zarith.
Auto with real.
Qed.

Theorem TwoSumNul:
  (f, g, x:float)
  (Closest b (2) (Rplus f g) x) ->
  <R> x==R0 -> (Fbounded b f) -> (Fbounded b g) ->(Rminus (Rplus f g) x)==R0.
Intros.
Replace (FtoRradix x) with R0; Ring.
Apply sym_eqT.
Cut <R> (FtoR (2) (Fzero (Zopp (dExp b))))==R0.
Intro.
Replace R0 with (FtoR (2) (Fzero (Zopp (dExp b)))); Auto.
Apply plusExact1
     with b := b radix := (2) p := g q := f r := (Fzero (Zopp (dExp b)))
          precision := precision; Auto with arith.
Cut (CompatibleP b (2) (Closest b (2))).
Unfold CompatibleP.
Intro.
Apply H4 with r1 := (Rplus f g) p := x; Auto with real.
Apply Rplus_sym; Auto with real.
Rewrite H3; Auto.
Repeat Split; Simpl; Auto with zarith.
Apply ClosestCompatible.
Simpl.
Case (Zmin_or (Fexp g) (Fexp f)); Intro; Rewrite H4; [Elim H2 | Elim H1]; Auto.
Unfold FtoR; Simpl; Auto.
Ring.
Qed.

Theorem bound3Sum:
  ~ <R> r'==R0 ->
  (Rlt
     (Rabsolu (Rplus q' r'))
     (Rmult (Rmult (Rmult (3) (2)) (Rinv (vNum b))) (Rabsolu p'))).
Intros H'; Case (Req_EM w R0); Intros Hw.
Contradict H'.
Rewrite r'Def; Auto.
Rewrite Hw; AutoRewrite [Rsimpl].
Apply eq_Rminus.
Unfold FtoRradix; Apply ClosestIdem with b := b; Auto.
Replace (FtoR (2) v) with (Rplus w v); Auto.
Rewrite Hw; AutoRewrite [Rsimpl]; Auto.
Rewrite r'Def; Rewrite Rplus_Rminus.
Apply Rle_lt_trans with 1 := (Rabsolu_triang w v).
Apply Rlt_le_trans
     with (Rmult
             (Rmult (Rmult (3) (2)) (Rinv (vNum b)))
             (Rmult (Rinv (2)) (Rmax (Rabsolu p) (Rabsolu u)))).
Replace (Rmult (Rmult (3) (2)) (Rinv (vNum b)))
     with (Rplus
             (Rmult (Rmult (2) (2)) (Rinv (vNum b))) (Rmult (2) (Rinv (vNum b))));
 [Idtac | Simpl; Ring].
Rewrite [x, y:R](Rmult_sym x (Rmult (Rinv (2)) y)); Rewrite Rmult_Rplus_distr;
 Repeat Rewrite [y:R](Rmult_sym (Rmult (Rinv (2)) y)).
Apply Rplus_lt.
Replace (Rmult
           (Rmult (Rmult (2) (2)) (Rinv (vNum b)))
           (Rmult (Rinv (2)) (Rmax (Rabsolu p) (Rabsolu u))))
     with (Rmult
             (Rmult (2) (Rinv (2)))
             (Rmult (Rmult (2) (Rinv (vNum b))) (Rmax (Rabsolu p) (Rabsolu u))));
 [Rewrite Rinv_r; Auto with real arith; Rewrite Rmult_ne_r | Ring].
Rewrite wDef.
Rewrite <- Rminus_Ropp; Rewrite Rabsolu_Ropp.
Apply (plusErrorBound2 b precision); Auto.
Contradict Hw.
Rewrite wDef.
Apply eq_Rminus.
Unfold FtoRradix; Rewrite (is_Fzero_rep1 (2) ? Hw); Auto.
Rewrite <- (FzeroisZero (2) b).
Apply sym_eqT; Apply (plusExact1 b (2) precision); Auto.
Apply (ClosestCompatible b (2) (Rplus p u) (Rplus (FtoR (2) p) (FtoR (2) u)) p');
 Auto.
Rewrite FzeroisZero; Apply is_Fzero_rep1; Auto.
Apply FboundedFzero; Auto.
Simpl; Unfold Zmin; Case (Zcompare (Fexp p) (Fexp u)); Intuition.
Apply Rlt_le_trans
     with (Rmult (Rmult (2) (Rinv (vNum b))) (Rmult (Rinv (2)) (Rabsolu u))).
Rewrite vDef.
Rewrite <- Rminus_Ropp; Rewrite Rabsolu_Ropp.
Replace (Rmult (Rmult (2) (Rinv (vNum b))) (Rmult (Rinv (2)) (Rabsolu u)))
     with (Rmult (Rmult (Rabsolu u) (Rinv (2))) (Rmult (2) (Rinv (vNum b))));
 [Idtac | Ring].
Unfold FtoRradix; Apply (plusErrorBound1 b (2) precision); Auto.
Contradict Hw.
Rewrite wDef.
Unfold FtoRradix; Rewrite (is_Fzero_rep1 (2) ? Hw).
AutoRewrite [Rsimpl].
Apply eq_Rminus.
Unfold FtoRradix; Apply ClosestIdem with b := b; Auto.
Replace (FtoR (2) p) with (Rplus p u); Auto.
Unfold FtoRradix; Rewrite (is_Fzero_rep1 (2) ? Hw).
AutoRewrite [Rsimpl]; Auto.
Apply Rle_monotony; Auto with real.
Replace R0 with (Rmult (2) R0); Auto with real; Apply Rle_monotony;
 Auto with real arith.
Replace R0 with (INR O); Auto with real arith.
Apply Rlt_le; Apply Rlt_Rinv; Auto with real arith.
Replace R0 with (INR O); Auto with real arith.
Apply Rlt_INR.
Apply lt_vNum with radix := (2) precision := precision; Auto with arith.
Apply Rle_monotony; Auto with real arith.
Apply Rlt_le; Apply Rlt_Rinv; Auto with real arith.
Apply RmaxLess2; Auto.
Apply Rle_monotony; Auto with real arith.
Replace R0 with (Rmult (Rmult (3) (2)) R0); Auto with real; Apply Rle_monotony;
 Auto with real arith.
Replace (Rmult (3) (2)) with (INR (6)); [Idtac | Simpl; Ring].
Replace R0 with (INR O); Auto with real arith.
Apply Rlt_le; Apply Rlt_Rinv; Auto with real arith.
Replace R0 with (INR O); Auto with real arith.
Apply Rlt_INR.
Apply lt_vNum with radix := (2) precision := precision; Auto with arith.
Apply (plusClosestLowerBound b precision); Auto with real.
Contradict Hw.
Rewrite wDef.
Unfold FtoRradix; Rewrite <- Hw; Ring.
Ring.
Qed.

Theorem exp3Sum:
  (Ex [p'':float]
  (Ex [q'':float]
  (Ex [r'':float]
  ((Fbounded b p'') /\ ((Fbounded b q'') /\ (Fbounded b r''))) /\
  ((<R> p''==p' /\ (<R> q''==q' /\ <R> r''==r')) /\
   ((Zle (Fexp r) (Fexp r'')) /\
    (((Zle (Fexp r'') (Fexp q'')) /\ (Zle (Fexp q'') (Fexp p''))) /\
     (Fexp r'')=(Fexp r))))))).
Cut (Fbounded b (Fzero (Fexp r))); [Intros Fbr | Apply FboundedZeroSameExp];
 Auto.
Case (Req_EM v R0); Intros Hv.
Cut <R> (Fzero (Fexp r))==r';
 [Intros EqZr |
   Rewrite (FzeroisReallyZero (2)); Rewrite r'Def; Rewrite Hv;
    Rewrite Rplus_ne_l; Apply sym_eqT; Apply eq_Rminus; Unfold FtoRradix;
    Apply (ClosestIdem b (2)); Auto;
    Apply (ClosestCompatible b (2) (Rplus w v) w q'); Auto; Rewrite Hv; Ring].
Case (plusExpMin b (2) precision) with P := (Closest b (2)) 5 := uDef; Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Intros u' H'; Elim H'; Intros H'0 H'1; Elim H'1; Intros H'2 H'3; Clear H'1 H'.
Case (Req_EM w R0); Intros Hw.
Case (plusExpMin b (2) precision) with P := (Closest b (2)) p := p q := u'
                                       pq := p'; Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply (ClosestCompatible
         b (2) (Rplus p u) (Rplus (FtoR (2) p) (FtoR (2) u')) p'); Auto.
Rewrite H'2; Auto.
Intros p'' H'; Elim H'; Intros H'1 H'4; Elim H'4; Intros H'5 H'6; Clear H'4 H'.
Exists p''; Exists (Fzero (Fexp r)); Exists (Fzero (Fexp r)); Split;
 [Idtac | Split]; (Split; [Idtac | Split]); Auto.
Rewrite (FzeroisReallyZero (2)); Auto.
Rewrite <- (FzeroisZero (2) b).
Unfold FtoRradix; Apply (ClosestIdem b (2)); Auto.
Apply FboundedFzero; Auto.
Apply (ClosestCompatible b (2) (Rplus w v) (Fzero (Zopp (dExp b))) q'); Auto.
Rewrite Hw; Rewrite Hv; Unfold FtoRradix; Rewrite (FzeroisZero (2) b); Ring.
Simpl; Auto with zarith.
Split; Simpl; Auto with zarith.
Simpl.
Apply Zle_trans with 2 := H'6.
Apply Zmin_Zle; Auto.
Apply Zle_trans with (Fexp q); Auto.
Rewrite <- (Zmin_le2 (Fexp q) (Fexp r)); Auto.
Case (errorBoundedPlus b (2) precision) with p := p q := u' pq := p'; Auto.
Apply (ClosestCompatible
         b (2) (Rplus p u) (Rplus (FtoR (2) p) (FtoR (2) u')) p'); Auto.
Rewrite H'2; Auto.
Intros w' H'; Elim H'; Intros H'1 H'4; Elim H'4; Intros H'5 H'6; Clear H'4 H'.
Exists p'; Exists w'; Exists (Fzero (Fexp r)); Split; [Idtac | Split];
 (Split; [Idtac | Split]); Auto.
Unfold FtoRradix; Apply (ClosestIdem b (2)); Auto.
Apply (ClosestCompatible b (2) (Rplus w v) (FtoR (2) w') q'); Auto.
Rewrite Hv; AutoRewrite [Rsimpl]; Auto.
Rewrite wDef; Unfold FtoRradix; Rewrite <- H'2; Auto.
Simpl; Auto with zarith.
Split; Simpl; Auto with zarith.
Simpl; Rewrite H'6.
Apply Zmin_Zle; Auto.
Apply Zle_trans with (Fexp q); Auto.
Apply Zle_trans with 2 := H'3; Auto with zarith.
Rewrite Zmin_le2; Auto with zarith.
Rewrite H'6.
Apply Zlt_le_weak.
Apply plusExact1bis with b := b radix := (2) precision := precision; Auto.
Apply (ClosestCompatible
         b (2) (Rplus p u) (Rplus (FtoR (2) p) (FtoR (2) u')) p'); Auto.
Rewrite H'2; Auto.
Contradict Hw.
Rewrite wDef; Auto.
Unfold FtoRradix; Rewrite <- H'2.
Rewrite <- Hw; Ring; Ring.
Case (Req_EM w R0); Intros Hw.
Cut <R> (Fzero (Fexp r))==r';
 [Intros EqZr |
   Rewrite (FzeroisReallyZero (2)); Rewrite r'Def; Rewrite Hw;
    Rewrite Rplus_ne_r; Apply sym_eqT; Apply eq_Rminus; Unfold FtoRradix;
    Apply (ClosestIdem b (2)); Auto;
    Apply (ClosestCompatible b (2) (Rplus w v) (FtoR (2) v) q'); Auto;
    Rewrite Hw; Unfold FtoRradix; Ring].
Case (plusExpMin b (2) precision) with P := (Closest b (2)) 5 := uDef; Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Intros u' H'; Elim H'; Intros H'0 H'1; Elim H'1; Intros H'2 H'3; Try Exact H'3;
 Clear H'1 H'.
Case (errorBoundedPlus b (2) precision) with p := q q := r pq := u'; Auto.
Apply (ClosestCompatible
         b (2) (Rplus (FtoR (2) q) (FtoR (2) r))
         (Rplus (FtoR (2) q) (FtoR (2) r)) u); Auto.
Intros v' H'; Elim H'; Intros H'1 H'4; Elim H'4; Intros H'5 H'6; Clear H'4 H'.
Case (plusExpMin b (2) precision) with P := (Closest b (2)) p := p q := u'
                                       pq := p'; Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply (ClosestCompatible
         b (2) (Rplus p u) (Rplus (FtoR (2) p) (FtoR (2) u')) p'); Auto.
Rewrite H'2; Auto.
Intros p'' H'; Elim H'; Intros H'4 H'7; Elim H'7; Intros H'8 H'9; Clear H'7 H'.
Exists p''; Exists v'; Exists (Fzero (Fexp r)); Split; [Idtac | Split];
 (Split; [Idtac | Split]); Auto.
Replace (FtoRradix v') with (FtoRradix v); [Idtac | Rewrite vDef; Auto].
Unfold FtoRradix; Apply (ClosestIdem b (2)); Auto;
 Apply (ClosestCompatible b (2) (Rplus w v) (FtoR (2) v) q'); Auto; Rewrite Hw;
 Unfold FtoRradix; Ring.
Unfold FtoRradix; Rewrite H'1; Auto.
Rewrite H'2; Auto.
Simpl; Auto with zarith.
Split; Simpl; Auto with zarith.
Rewrite H'6; Rewrite Zmin_le2; Auto with zarith.
Apply Zle_trans with 2 := H'9.
Apply Zmin_Zle; Auto.
Rewrite H'6; Auto.
Apply Zle_trans with (Fexp q); Auto with zarith.
Apply Zle_min_l; Auto.
Rewrite H'6; Auto.
Case (plusExpMin b (2) precision) with P := (Closest b (2)) 5 := uDef; Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Intros u' H'; Elim H'; Intros H'0 H'1; Elim H'1; Intros H'2 H'3; Try Exact H'3;
 Clear H'1 H'.
Case (errorBoundedPlus b (2) precision) with p := q q := r pq := u'; Auto.
Apply (ClosestCompatible
         b (2) (Rplus (FtoR (2) q) (FtoR (2) r))
         (Rplus (FtoR (2) q) (FtoR (2) r)) u); Auto.
Intros v' H'; Elim H'; Intros H'1 H'4; Elim H'4; Intros H'5 H'6; Clear H'4 H'.
Case (errorBoundedPlus b (2) precision) with p := p q := u' pq := p'; Auto.
Apply (ClosestCompatible
         b (2) (Rplus p u) (Rplus (FtoR (2) p) (FtoR (2) u')) p'); Auto.
Rewrite H'2; Auto.
Intros w' H'; Elim H'; Intros H'4 H'7; Elim H'7; Intros H'8 H'9; Clear H'7 H'.
Case (plusExpBound b (2) precision) with P := (Closest b (2)) p := w' q := v'
                                         pq := q'; Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply (ClosestCompatible
         b (2) (Rplus w v) (Rplus (FtoR (2) w') (FtoR (2) v')) q'); Auto.
Rewrite wDef; Unfold FtoRradix; Rewrite H'4; Auto.
Unfold FtoRradix in vDef; Rewrite vDef; Unfold FtoRradix; Rewrite <- H'2;
 Rewrite H'1; Auto.
Intros q'' H'; Elim H'; Intros H'7 H'10; Elim H'10; Intros H'11 H'12; Elim H'12;
 Intros H'13 H'14; Try Exact H'14; Clear H'12 H'10 H'.
Case (errorBoundedPlus b (2) precision) with p := v' q := w' pq := q''; Auto.
Apply (ClosestCompatible
         b (2) (Rplus w v) (Rplus (FtoR (2) v') (FtoR (2) w')) q'); Auto.
Rewrite wDef; Rewrite vDef; Rewrite H'4; Rewrite H'1; Unfold FtoRradix;
 Repeat Rewrite H'2; Ring; Ring.
Intros r'' H'; Elim H'; Intros H'10 H'12; Elim H'12; Intros H'15 H'16;
 Clear H'12 H'.
Exists p'; Exists q''; Exists r''; Split; [Idtac | Split];
 (Split; [Idtac | Split]); Auto.
Rewrite r'Def; Rewrite wDef; Rewrite vDef; Unfold FtoRradix; Rewrite H'10;
 Rewrite H'4; Rewrite H'1; Repeat Rewrite H'2; Rewrite H'11; Ring; Ring.
Rewrite H'16; Apply Zmin_Zle; Auto with zarith.
Rewrite H'6; Apply Zmin_Zle; Auto with zarith.
Rewrite H'9; Apply Zmin_Zle; Auto with zarith.
Apply Zle_trans with 2 := H'3; Apply Zmin_Zle; Auto with zarith.
Split.
Rewrite H'16; Rewrite Zmin_sym; Auto.
Apply Zle_trans with 1 := H'14.
Rewrite Zmax_le1; Auto.
Apply Zlt_le_S; Auto.
Rewrite H'9.
Apply plusExact1bis with b := b radix := (2) precision := precision; Auto.
Apply (ClosestCompatible
         b (2) (Rplus p u) (Rplus (FtoR (2) p) (FtoR (2) u')) p'); Auto.
Rewrite H'2; Auto.
Contradict Hw; Auto.
Rewrite wDef; Auto.
Unfold FtoRradix; Rewrite <- H'2.
Rewrite <- Hw; Ring; Ring.
Rewrite H'9.
Apply Zmin_Zle; Auto.
Rewrite H'6; Rewrite Zmin_le2; Auto with zarith.
Rewrite H'6; Auto.
Rewrite H'16.
Rewrite H'6.
Rewrite H'9.
Replace (Zmin (Fexp q) (Fexp r)) with (Fexp r).
Apply Zmin_le1.
Apply Zmin_Zle; Auto.
Apply Zle_trans with (Fexp q); Auto.
Rewrite <- (Zmin_le2 (Fexp q) (Fexp r)); Auto.
Apply sym_eq; Apply Zmin_le2; Auto.
Qed.

Theorem OutSum3:
  ~ <R> r'==R0 ->
  (Rlt
     (Float (1) (Fexp r))
     (Rmult
        (3)
        (Rmult
           (2)
           (Rmult
              (2)
              (Rmult
                 (Rabsolu p')
                 (Rinv (Rmult (vNum b) (Rminus (Rmult (2) (vNum b)) (2))))))))).
Intros.
Generalize exp3Sum.
Intros H'; Elim H'; Intros p'' E; Elim E; Intros q'' E0; Elim E0; Intros r'' E1;
 Elim E1; Intros H'0 H'1; Elim H'1; Intros H'2 H'3; Elim H'3; Intros H'4 H'5;
 Elim H'5; Intros H'6 H'7; Clear H'5 H'3 H'1 E1 E0 E H'.
Elim H'6; Intros H' H'1; Clear H'6.
Elim H'2; Intros H'3 H'5; Elim H'5; Intros H'6 H'8; Clear H'5 H'2.
Elim H'0; Intros H'2 H'5; Elim H'5; Intros H'9 H'10; Clear H'5 H'0.
Cut ~ (is_Fzero q'); [Intros Z1 | Idtac].
2:Contradict H; Auto with real float.
2:Replace (FtoRradix r') with (Rminus (Rplus w v) q'); Auto.
2:Apply TwoSumNul; Auto.
2:Unfold is_Fzero in H.
2:Unfold FtoRradix FtoR.
2:Replace (Fnum q') with ZERO; Ring.
Apply Rle_lt_trans with (Rabsolu r'').
Apply Rle_trans with (FtoR (2) (Float (1) (Fexp r''))).
Unfold FtoRradix.
Replace (Fexp r) with (Fexp r''); Auto with real.
Rewrite <- (Fabs_correct (2)); Auto with arith.
Apply RleFexpFabs; Auto with arith.
Fold FtoRradix.
Rewrite H'8; Auto.
Apply Rlt_trans
     with (Rmult (Rmult (Rabsolu q') (Rinv (2))) (Rmult (2) (Rinv (vNum b)))).
Replace (FtoRradix r'') with (Rminus (Rplus w v) q').
Replace (Rabsolu (Rminus (Rplus w v) q')) with (Rabsolu (Rminus q' (Rplus w v))).
Apply plusErrorBound1
     with b := b radix := (2) precision := precision p := w q := v;
 Auto with arith.
Apply Rabsolu_minus_sym.
Rewrite H'8; Auto.
Cut (Rlt
       (Rabsolu q')
       (Rplus
          (Rmult (3) (Rmult (2) (Rmult (Rabsolu p') (Rinv (vNum b)))))
          (Rmult (Rabsolu q') (Rinv (vNum b))))) ->
    (Rlt
       (Rmult (Rmult (Rabsolu q') (Rinv (2))) (Rmult (2) (Rinv (vNum b))))
       (Rmult
          (3)
          (Rmult
             (2)
             (Rmult
                (2)
                (Rmult
                   (Rabsolu p')
                   (Rinv (Rmult (vNum b) (Rminus (Rmult (2) (vNum b)) (2))))))))).
Intros C; Apply C.
Apply Rlt_trans
     with (Rplus
             (Rmult (3) (Rmult (2) (Rmult (Rabsolu p') (Rinv (vNum b)))))
             (Rabsolu r')).
Apply Rle_lt_trans with (Rplus (Rabsolu (Rplus q' r')) (Rabsolu r')).
Replace (Rabsolu q') with (Rabsolu (Rplus (Rplus q' r') (Ropp r'))).
Replace (Rabsolu r') with (Rabsolu (Ropp r')).
Apply Rabsolu_triang.
Apply Rabsolu_Ropp.
Replace (Rplus (Rplus q' r') (Ropp r')) with (FtoRradix q'); Auto.
Ring.
Replace (Rplus (Rabsolu (Rplus q' r')) (Rabsolu r'))
     with (Rplus (Rabsolu r') (Rabsolu (Rplus q' r'))).
Replace (Rplus
           (Rmult (3) (Rmult (2) (Rmult (Rabsolu p') (Rinv (vNum b)))))
           (Rabsolu r'))
     with (Rplus
             (Rabsolu r')
             (Rmult (3) (Rmult (2) (Rmult (Rabsolu p') (Rinv (vNum b)))))).
Apply Rlt_compatibility.
Replace (Rmult (3) (Rmult (2) (Rmult (Rabsolu p') (Rinv (vNum b)))))
     with (Rmult (Rmult (Rmult (3) (2)) (Rinv (vNum b))) (Rabsolu p')).
Apply bound3Sum; Auto.
Ring.
Ring.
Ring.
Apply Rlt_compatibility.
Replace (FtoRradix r') with (Rminus (Rplus w v) q'); Auto.
Replace (Rabsolu (Rminus (Rplus w v) q')) with (Rabsolu (Rminus q' (Rplus w v))).
Replace (Rmult (Rabsolu q') (Rinv (vNum b)))
     with (Rmult (Rmult (Rabsolu q') (Rinv (2))) (Rmult (2) (Rinv (vNum b)))).
Apply plusErrorBound1
     with b := b radix := (2) precision := precision r := q' p := w q := v; Auto.
Ring.
Replace (Rmult (Rinv (2)) (2)) with R1; Ring.
Apply sym_eqT.
Apply Rinv_r; Auto with real.
Replace (Rminus q' (Rplus w v)) with (Ropp (Rminus (Rplus w v) q')).
Apply Rabsolu_Ropp.
Ring.
Intros H'0.
Replace (Rmult (Rmult (Rabsolu q') (Rinv (2))) (Rmult (2) (Rinv (vNum b))))
     with (Rmult (Rmult (2) (Rinv (2))) (Rmult (Rinv (vNum b)) (Rabsolu q')));
 Try Ring.
Rewrite Rinv_r; Auto with real; Rewrite Rmult_ne_r.
Replace (Rminus (Rmult (2) (vNum b)) (2)) with (Rmult (2) (Rminus (vNum b) R1));
 [Idtac | Simpl; Ring; Ring].
Cut (lt (1) (vNum b));
 [Intros lt1 | Apply le_lt_trans with (nNormMin (2) precision)];
 Auto with float arith.
Cut (Rlt R1 (vNum b)); [Intros Rlt1 | Idtac]; Auto with real float arith.
Repeat Rewrite Rinv_Rmult; Auto with real.
Replace (Rmult
           (3)
           (Rmult
              (2)
              (Rmult
                 (2)
                 (Rmult
                    (Rabsolu p')
                    (Rmult
                       (Rinv (vNum b))
                       (Rmult (Rinv (2)) (Rinv (Rminus (vNum b) R1))))))))
     with (Rmult
             (Rmult (2) (Rinv (2)))
             (Rmult
                (Rinv (vNum b))
                (Rmult
                   (3)
                   (Rmult (2) (Rmult (Rabsolu p') (Rinv (Rminus (vNum b) R1)))))));
 [Idtac | Ring].
Rewrite Rinv_r; Auto with real; Rewrite Rmult_ne_r.
Apply Rlt_monotony; Auto with real.
Apply Rlt_Rinv; Auto with real.
Apply Rlt_trans with R1; Auto with real.
Apply Rlt_mult_anticompatibility with z := (Rminus (vNum b) R1).
Replace (Rmult
           (Rminus (vNum b) R1)
           (Rmult
              (3) (Rmult (2) (Rmult (Rabsolu p') (Rinv (Rminus (vNum b) R1))))))
     with (Rmult
             (Rmult (Rinv (Rminus (vNum b) R1)) (Rminus (vNum b) R1))
             (Rmult (3) (Rmult (2) (Rabsolu p')))); [Idtac | Ring].
Rewrite Rinv_l; Auto with real.
Rewrite Rmult_ne_r.
Replace (Rmult (Rminus (vNum b) R1) (Rabsolu q'))
     with (Rplus (Rmult (vNum b) (Rabsolu q')) (Ropp (Rabsolu q'))); Try Ring.
Replace (Rmult (3) (Rmult (2) (Rabsolu p')))
     with (Rplus
             (Rplus (Rmult (3) (Rmult (2) (Rabsolu p'))) (Rabsolu q'))
             (Ropp (Rabsolu q'))); Try Ring.
Apply Rlt_compatibility_r.
Replace (Rplus (Rmult (3) (Rmult (2) (Rabsolu p'))) (Rabsolu q'))
     with (Rmult
             (Rmult (vNum b) (Rinv (vNum b)))
             (Rplus (Rmult (3) (Rmult (2) (Rabsolu p'))) (Rabsolu q')));
 Auto with real float arith.
Replace (Rmult
           (Rmult (vNum b) (Rinv (vNum b)))
           (Rplus (Rmult (3) (Rmult (2) (Rabsolu p'))) (Rabsolu q')))
     with (Rmult
             (vNum b)
             (Rmult
                (Rinv (vNum b))
                (Rplus (Rmult (3) (Rmult (2) (Rabsolu p'))) (Rabsolu q'))));
 Try Ring.
Apply Rlt_monotony; Auto with real.
Apply Rlt_trans with R1; Auto with real.
Replace (Rmult
           (Rinv (vNum b))
           (Rplus (Rmult (3) (Rmult (2) (Rabsolu p'))) (Rabsolu q')))
     with (Rplus
             (Rmult (3) (Rmult (2) (Rmult (Rabsolu p') (Rinv (vNum b)))))
             (Rmult (Rabsolu q') (Rinv (vNum b)))); Try Ring; Auto.
Rewrite Rinv_r; Auto with real float arith.
Contradict Rlt1.
Replace (INR (vNum b)) with (Rplus (Rminus (vNum b) R1) R1); Try Ring.
Rewrite Rlt1; Auto with real float arith.
Replace (Rplus R0 R1) with R1; Try Ring; Auto with real.
Replace R0 with (Rminus R1 R1); Try Ring; Unfold Rminus; Auto with real.
Apply Rlt_compatibility_r; Auto with real.
Contradict Rlt1.
Replace (INR (vNum b)) with (Rplus (Rminus (vNum b) R1) R1); Try Ring.
Rewrite Rlt1; Auto with real float arith.
Replace (Rplus R0 R1) with R1; Try Ring; Auto with real.
Auto with real float arith.
Contradict Rlt1.
Case (without_div_Od ? ? Rlt1).
Intros H'5; Absurd (Rlt O (2)); Auto with real arith.
Rewrite H'5; Simpl; Auto with real.
Intros H'5.
Replace (INR (vNum b)) with (Rplus (Rminus (vNum b) R1) R1); Try Ring.
Rewrite H'5; Auto with real float arith.
Replace (Rplus R0 R1) with R1; Try Ring; Auto with real.
Qed.

Theorem TwoSumNonNul:
  (p, q, r, pq:float)
  (Fbounded b p) ->
  (Fbounded b q) ->
  (Closest b (2) (Rplus p q) pq) ->
  <R> r==(Rminus (Rplus p q) pq) -> ~ <R> r==R0 ->~ <R> pq==R0.
Intros.
Contradict H3.
Rewrite H2.
Apply TwoSumNul; Auto with real.
Qed.

Theorem TwoSumOneNul:
  (p, q, pq:float)
  (Fbounded b p) ->
  (Fbounded b q) ->
  (Fbounded b pq) -> (Closest b (2) (Rplus p q) pq) -> <R> p==R0 -><R> pq==q.
Intros.
Generalize H2.
Rewrite H3.
Replace (Rplus R0 q0) with (FtoRradix q0); [Idtac | Ring].
Intro.
Apply sym_eqT;
 Apply (RoundedModeProjectorIdemEq b (2) precision) with P := (Closest b (2));
 Auto with real float arith zarith.
Apply (ClosestRoundedModeP b (2) precision); Auto with real float arith zarith.
Qed.

Theorem TwoSumOneNul2:
  (p, q, pq, r:float)
  (Fbounded b p) ->
  (Fbounded b q) ->
  (Fbounded b pq) ->
  (Closest b (2) (Rplus p q) pq) ->
  <R> p==R0 -> <R> r==(Rminus (Rplus p q) pq) -><R> r==R0.
Intros.
Rewrite H4; Rewrite H3.
Apply eq_Rminus; Ring.
Apply sym_eqT; Apply TwoSumOneNul with p := p0; Auto.
Qed.
End F2.
