(****************************************************************************
                                                                             
          IEEE754  :  ThreeSumProps                                           
                                                                             
          Sylvie Boldo                                          
                                                                             
*****************************************************************************
*)

Require Import Arith.
Require Import Omega.
Require Export Float.
Require Export Fop.
Require Export Fnorm.
Require Export MSB.
Require Export ThreeSum2.
Section F2.
Variable b:Fbound.
Variable precision:nat.

Local FtoRradix := (FtoR (2)).
Coercion FtoRradix : float >-> R.
Hypothesis precisionNotZero:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp (2) precision) (1)).
Variables p, q, r, u, v, w, p', q', r':float.
Hypothesis Fp:(Fbounded b p).
Hypothesis Fq:(Fbounded b q).
Hypothesis Fr:(Fbounded b r).
Hypothesis Fu:(Fbounded b u).
Hypothesis Fv:(Fbounded b v).
Hypothesis Fw:(Fbounded b w).
Hypothesis Fp':(Fbounded b p').
Hypothesis Fq':(Fbounded b q').
Hypothesis Fr':(Fbounded b r').
Hypothesis epq:(Zle (Fexp q) (Fexp p)).
Hypothesis eqr:(Zle (Fexp r) (Fexp q)).
Hypothesis uDef:(Closest b (2) (Rplus q r) u).
Hypothesis vDef:<R> v==(Rminus (Rplus q r) u).
Hypothesis p'Def:(Closest b (2) (Rplus p u) p').
Hypothesis wDef:<R> w==(Rminus (Rplus p u) p').
Hypothesis q'Def:(Closest b (2) (Rplus w v) q').
Hypothesis r'Def:<R> r'==(Rminus (Rplus w v) q').
Hypothesis Ngd:(Rle R1 (Rmult (vNum b) (Rminus R1 (Rinv (2))))).
Hypothesis Hplus:
           (<R> p==R0 \/ (<R> q==R0 \/ <R> r==R0)) \/
           (Rle
              (Rabsolu q)
              (Rmult
                 (vNum b) (Rminus (Float (1) (Fexp p)) (Float (1) (Fexp r))))).
Hypothesis Ngd2:
           (Rle (6) (Rmult (vNum b) (Rminus R1 (Rmult (Rinv (2)) (Rinv (2)))))).

Theorem FTS:
  (Rle
     (Rabsolu q)
     (Rmult (vNum b) (Rminus (Float (1) (Fexp p)) (Float (1) (Fexp r))))) ->
  (Ex [u'':float] <R> u''==u /\ ((Fbounded b u'') /\ (Zle (Fexp u'') (Fexp p)))).
Intros H.
Case (Req_EM u R0); Intros Hu.
Exists (Fzero (Fexp p)).
Split.
Replace (FtoRradix u) with R0; Auto with real; Unfold FtoRradix;
 Apply FzeroisReallyZero.
Split; Auto with zarith.
Apply FboundedZeroSameExp; Auto.
Exists (Fnormalize (2) b precision u).
Cut <R> (Fnormalize (2) b precision u)==u; Intros.
Split; Auto with float.
2:Unfold FtoRradix; Apply FnormalizeCorrect; Auto with arith.
Split.
Apply FcanonicBound with precision := precision radix := (2).
Apply FnormalizeCanonic; Auto with arith.
Cut (Fcanonic (2) b precision (Fnormalize (2) b precision u)).
2:Apply FnormalizeCanonic; Auto with arith.
Intros H1; Case H1.
Intros H2.
Cut (Zle (Zpred (Fexp (Fnormalize (2) b precision u))) (Zpred (Fexp p))) ->
    (Zle (Fexp (Fnormalize (2) b precision u)) (Fexp p)).
Intros H'.
Apply H'.
Apply Zlt_n_Sm_le.
Rewrite <- (Zs_pred (Fexp p)).
2:Unfold Zpred; Intros.
2:Apply Zsimpl_le_plus_l with p := (NEG xH).
2:Replace (Zplus (NEG xH) (Fexp (Fnormalize (2) b precision u)))
       with (Zplus (Fexp (Fnormalize (2) b precision u)) (NEG xH));
   Replace (Zplus (NEG xH) (Fexp p)) with (Zplus (Fexp p) (NEG xH));
   Auto with zarith.
Apply Zlt_powerRZ with e := (INR (2)); Auto with real arith zarith.
Replace R1 with (INR (1)); Auto with real arith zarith.
Apply Rlt_mult_anticompatibility with z := (IZR (vNum b)).
Apply Rlt_le_trans
     with (Rmult
             (absolu (Fnum (Fnormalize (2) b precision u)))
             (powerRZ (2) (Fexp (Fnormalize (2) b precision u)))).
Unfold Zpred; Rewrite powerRZ_add; Auto with real arith zarith.
Rewrite <- INR_IZR_INZ.
Replace (Rmult
           (vNum b)
           (Rmult
              (powerRZ (2) (Fexp (Fnormalize (2) b precision u)))
              (powerRZ (2) (NEG xH))))
     with (Rmult
             (Rmult (vNum b) (powerRZ (2) (NEG xH)))
             (powerRZ (2) (Fexp (Fnormalize (2) b precision u))));
 [Idtac | Ring].
Apply Rlt_monotony_r; Auto with real arith zarith.
Case H2.
Intros.
Replace (powerRZ (2) (NEG xH)) with (Rinv (2)).
2:Unfold powerRZ; Simpl; Auto with real.
Generalize H4; Unfold Fdigit; Intros.
Rewrite pGivesBound.
Apply Rlt_le_trans with (Rmult (exp (2) precision) (Rinv (2))).
Apply Rlt_monotony_r; Auto with real.
Apply Rlt_INR; Auto with arith.
Apply Rle_trans with (INR (exp (2) (pred precision))).
2:Apply Rle_INR; Auto with arith.
2:Pattern 1 precision.
2:Rewrite <- H5.
2:Apply digitLess.
2:Apply absolu_lt_nz.
2:Contradict Hu.
2:Replace (FtoRradix u) with (FtoRradix (Fnormalize (2) b precision u)); Auto.
2:Unfold FtoRradix; Unfold FtoR; Rewrite Hu; Simpl; Ring.
Rewrite <- (powerRZ_is_exp (2) (pred precision)); Auto.
Rewrite <- (powerRZ_is_exp (2) precision); Auto.
Replace (Rmult (powerRZ (2) (INZ precision)) (Rinv (2)))
     with (powerRZ (2) (Zpred (INZ precision))).
Apply Rle_powerRZ; Auto with arith zarith real.
Replace R1 with (INR (1)); Auto with arith zarith real.
Unfold INZ.
Case precision.
Simpl; Auto with zarith.
Intros n0.
Rewrite inj_pred; Auto with zarith.
Unfold Zpred.
Rewrite (powerRZ_add (2) (INZ precision) (NEG xH)); Auto with real.
Replace (powerRZ (2) (NEG xH)) with (Rinv (2)); Auto with real.
Unfold powerRZ; Auto with arith zarith real.
Replace (Rmult
           (Fnum (Fnormalize (2) b precision u))
           (powerRZ (2) (Fexp (Fnormalize (2) b precision u))))
     with (FtoRradix (Fnormalize (2) b precision u)).
2:Unfold FtoRradix; Unfold FtoR; Simpl; Auto with real.
Replace (Rmult
           (absolu (Fnum (Fnormalize (2) b precision u)))
           (powerRZ (2) (Fexp (Fnormalize (2) b precision u)))) with (Rabsolu u).
Generalize MaxEx; Intros Max.
LApply (Max b (2) precision);
 [Intros H'3; LApply H'3;
   [Intros H'4; LApply H'4;
     [Intros H'5; Specialize 1 H'5 with r := (Rabsolu (Rplus q r)); Intros H'6;
       Clear H'4 H'3 | Clear H'4 H'3] | Clear H'3] | Idtac]; Auto.
Elim H'6.
Intros max H'1; Clear H'6; Clear Max.
2:Auto with arith zarith.
Case (ClosestMinOrMax b (2) (Rabsolu (Rplus q r)) (Fabs u)); Auto; Intros.
Apply ClosestFabs with precision := precision; Auto.
Apply Rle_trans with (Rabsolu (Rplus q r)).
Replace (Rabsolu u) with (FtoRradix (Fabs u)).
Case H3; Intros; Auto.
Elim H5; Intros; Auto.
Unfold FtoRradix; Apply Fabs_correct; Auto with arith.
Apply Rle_trans with (Rplus (Rabsolu q) (Rabsolu r)).
Apply Rabsolu_triang.
Apply Rle_trans with (Rplus (Rabsolu q) (Rmult (vNum b) (powerRZ (2) (Fexp r)))).
Apply Rle_compatibility.
Replace (Rmult (vNum b) (powerRZ (2) (Fexp r)))
     with (FtoR (2) (Float (vNum b) (Fexp r))).
Unfold FtoRradix; Apply MaxFloat with precision := precision; Auto.
Unfold FtoR; Simpl; Ring.
Rewrite <- INR_IZR_INZ; Auto with real.
Apply Rle_anti_compatibility with (Ropp (Rmult (vNum b) (powerRZ (2) (Fexp r)))).
Replace (Rplus
           (Ropp (Rmult (vNum b) (powerRZ (2) (Fexp r))))
           (Rplus (Rabsolu q) (Rmult (vNum b) (powerRZ (2) (Fexp r)))))
     with (Rabsolu q); [Idtac | Ring].
Rewrite <- INR_IZR_INZ; Auto with real.
Replace (Rplus
           (Ropp (Rmult (vNum b) (powerRZ (2) (Fexp r))))
           (Rmult (vNum b) (powerRZ (2) (Fexp p))))
     with (Rmult (vNum b) (Rminus (Float (1) (Fexp p)) (Float (1) (Fexp r))));
 Auto with real float.
Unfold FtoRradix; Unfold FtoR; Simpl; Ring.
Specialize 1 H'5 with r := (Rmult (vNum b) (powerRZ (2) (Fexp p))); Intros H'0.
Elim H'0.
Intros max2 H'3; Clear H'0; Clear H'5.
Apply Rle_trans with (FtoRradix max2).
Case (Req_EM (Rabsolu (Rplus q r)) (Rmult (vNum b) (powerRZ (2) (Fexp p))));
 Intros.
Apply eq_Rle.
Generalize H'3; Rewrite <- H4; Intros.
Unfold FtoRradix.
Replace (Rabsolu (FtoR (2) u)) with (FtoR (2) (Fabs u));
 [Idtac | Apply Fabs_correct; Auto with arith].
Unfold FtoRradix;
 Apply MaxEq with r := (Rabsolu (Rplus q r)) b := b radix := (2); Auto.
Unfold FtoRradix.
Replace (Rabsolu (FtoR (2) u)) with (FtoR (2) (Fabs u));
 [Idtac | Apply Fabs_correct; Auto with arith].
Apply (MonotoneMax
         b (2) (Rabsolu (Rplus q r)) (Rmult (vNum b) (powerRZ (2) (Fexp p)))
         (Fabs u) max2); Auto.
Apply Rle_lt; Auto.
Apply Rle_trans with (Rabsolu (Rplus q r)).
Auto with real.
Apply Rle_trans with (Rplus (Rabsolu q) (Rabsolu r)).
Apply Rabsolu_triang.
Apply Rle_trans with (Rplus (Rabsolu q) (Rmult (vNum b) (powerRZ (2) (Fexp r)))).
Apply Rle_compatibility.
Replace (Rmult (vNum b) (powerRZ (2) (Fexp r)))
     with (FtoR (2) (Float (vNum b) (Fexp r))).
Unfold FtoRradix; Apply MaxFloat with precision := precision; Auto.
Unfold FtoR; Simpl; Ring.
Rewrite <- INR_IZR_INZ; Auto with real.
Apply Rle_anti_compatibility with (Ropp (Rmult (vNum b) (powerRZ (2) (Fexp r)))).
Replace (Rplus
           (Ropp (Rmult (vNum b) (powerRZ (2) (Fexp r))))
           (Rplus (Rabsolu q) (Rmult (vNum b) (powerRZ (2) (Fexp r)))))
     with (Rabsolu q); [Idtac | Ring].
Replace (Rplus
           (Ropp (Rmult (vNum b) (powerRZ (2) (Fexp r))))
           (Rmult (vNum b) (powerRZ (2) (Fexp p))))
     with (Rmult (vNum b) (Rminus (Float (1) (Fexp p)) (Float (1) (Fexp r))));
 Auto with real float.
Unfold FtoRradix; Unfold FtoR; Simpl; Ring.
Apply eq_Rle.
Rewrite <- INR_IZR_INZ; Auto with real.
Generalize H'3;
 Replace (Rmult (vNum b) (powerRZ (2) (Fexp p)))
      with (FtoRradix (Float (vNum b) (Fexp p))); Intros.
Apply sym_eqT; Apply (ProjectMax b (2) (Float (vNum b) (Fexp p)) max2); Auto.
Unfold Fbounded; Simpl.
Split; Auto with zarith.
Elim Fp; Auto.
Unfold FtoRradix; Unfold FtoR; Simpl; Ring; Auto with real.
Rewrite <- INR_IZR_INZ; Auto with real.
Replace (Rabsolu u) with (Rabsolu (Fnormalize (2) b precision u));
 Auto with real.
Unfold FtoRradix; Unfold FtoR; Simpl.
Rewrite (Rabsolu_mult
           (Fnum (Fnormalize (2) b precision u))
           (powerRZ (Rplus R1 R1) (Fexp (Fnormalize (2) b precision u)))).
Replace (Rabsolu (Fnum (Fnormalize (2) b precision u)))
     with (IZR (absolu (Fnum (Fnormalize (2) b precision u)))).
Rewrite <- INR_IZR_INZ; Auto with real.
Replace (Rabsolu (powerRZ (Rplus R1 R1) (Fexp (Fnormalize (2) b precision u))))
     with (powerRZ (Rplus R1 R1) (Fexp (Fnormalize (2) b precision u)));
 Auto with real.
Apply sym_eqT; Apply Rabsolu_simpl1.
Apply powerRZ_le; Auto with real arith.
Rewrite <- (Zabs_absolu (Fnum (Fnormalize (2) b precision u))).
Apply sym_eqT; Apply Rabsolu_Zabs.
Replace R0 with (INR (0)).
Rewrite <- INR_IZR_INZ.
Apply Rlt_INR.
Apply lt_vNum with precision := precision radix := (2); Auto with real zarith.
Auto with arith real.
Replace R0 with (INR (0)); Auto with arith real.
Intros H'; Case H'; Intros.
Elim H3; Intros.
Rewrite H4.
Elim Fp; Auto.
Qed.

Theorem FTSinSum3_allcases:
  ~ <R> p==R0 ->
  (Ex [u'':float] <R> u''==u /\ ((Fbounded b u'') /\ (Zle (Fexp u'') (Fexp p)))).
Intros.
Elim Hplus; [Intros H' | Intros H'].
Elim H'; [Clear H' | Intros H'0; Clear H'].
Intros.
Contradict H; Auto.
Elim H'0; Intros; Clear H'0.
Exists r.
Split.
Generalize uDef.
Rewrite H0.
Replace (Rplus R0 r) with (FtoRradix r); [Idtac | Ring].
Intro; Unfold FtoRradix; Apply ClosestIdem with b := b; Auto.
Split; Auto.
Apply Zle_trans with (Fexp q); Auto.
Exists q.
Split.
Generalize uDef.
Rewrite H0.
Replace (Rplus q R0) with (FtoRradix q); [Idtac | Ring].
Intro; Unfold FtoRradix; Apply ClosestIdem with b := b; Auto.
Split; Auto.
Apply FTS; Auto.
Qed.

Theorem PuissLessThanHalf:
  (z:Z)(Rle (Rmult (Float (1) z) (Rinv (2))) (Float (1) z)).
Intros.
Pattern 2 (FtoRradix (Float (1) z));
 Replace (FtoRradix (Float (1) z)) with (Rmult (Float (1) z) R1);
 [Apply Rle_monotony | Ring].
Unfold FtoRradix; Unfold FtoR; Simpl.
Replace (Rmult R1 (powerRZ (Rplus R1 R1) z)) with (powerRZ (Rplus R1 R1) z);
 [Apply powerRZ_le; Auto with real | Ring].
Apply Rlt_le.
Replace R1 with (Rinv R1); [Apply Rlt_Rinv_R1 | Idtac]; Auto with real.
Qed.

Theorem ThreeSumLoop:
  (Ex [p'':float]
  (Ex [q'':float]
  (Ex [r'':float]
  ((Fbounded b p'') /\ ((Fbounded b q'') /\ (Fbounded b r''))) /\
  (((<R> p''==p' /\ (<R> q''==q' /\ <R> r''==r')) /\
    ((Zle (Fexp r) (Fexp r'')) /\
     (((Zle (Fexp r'') (Fexp q'')) /\ (Zle (Fexp q'') (Fexp p''))) /\
      (Fexp r'')=(Fexp r)))) /\
   (<R> r'==R0 /\
    (<R> p''==R0 \/
     (Rle
        (Rabsolu q'')
        (Rmult (vNum b) (Rminus (Float (1) (Fexp p'')) (Float (1) (Fexp r))))))
    \/
    ~ <R> r'==R0 /\
    (Rle
       (Rabsolu r'')
       (Rmult (vNum b) (Rminus (Float (1) (Fexp q'')) (Float (1) (Fexp r)))))))))).
Generalize plusExactExp; Intros PEE.
LApply (PEE b (2) precision);
 [Intros H'2; LApply H'2;
   [Intros H'3; LApply H'3;
     [Intros H'4; LApply (H'4 q r u);
       [Intros H'8; LApply H'8;
         [Intros H'9; LApply H'9;
           [Intros H'10; Clear H'9 H'8 H'3 H'2 | Clear H'9 H'8 H'3 H'2] |
           Clear H'8 H'3 H'2] | Clear H'3 H'2] | Clear H'3 H'2] | Clear H'2] |
   Idtac]; Auto.
Elim H'10; Intros v' E; Elim E; Intros u' E0; Clear E H'10.
Elim E0; Intros H' H'0; Elim H'0; Intros H'1 H'2; Elim H'2; Intros H'3 H'5;
 Elim H'5; Intros H'6 H'7; Clear H'5 H'2 H'0 E0.
Elim H'7; Intros H'0 H'2; Clear H'7.
Elim H'2; Intros H'5 H'4; Clear H'2.
Cut (Fbounded b (Fzero (Fexp r)));
 [Intros Fbr | Apply FboundedZeroSameExp; Auto].
Generalize H'0.
Generalize H'0; Replace (Zmin (Fexp q) (Fexp r)) with (Fexp r);
 [Intros | Apply sym_eq; Apply Zmin_le2; Auto].
Case (Req_EM p R0); Intros Hp.
Exists u'; Exists v'; Exists (Fzero (Fexp r)).
Split; Auto.
Split; Auto.
Split; Auto.
Split; Auto.
Apply sym_eqT; Apply TwoSumOneNul with precision := precision p := p b := b;
 Auto.
Replace (FtoR (2) u') with (FtoR (2) u); Auto with real.
Cut <R> w==R0; [Intros v'Def | Idtac].
Split; Auto.
Apply sym_eqT; Apply TwoSumOneNul with precision := precision p := w b := b;
 Auto.
Replace (FtoR (2) v') with (FtoR (2) v); Auto with real.
Replace (FtoR (2) v')
     with (Rminus (Rplus (FtoR (2) q) (FtoR (2) r)) (FtoR (2) u'));
 Auto with real.
Replace (FtoR (2) u') with (FtoR (2) u); Auto with real.
Replace (FtoRradix (Fzero (Fexp r))) with R0; Apply sym_eqT;
 [Idtac | Unfold FtoRradix; Apply FzeroisReallyZero].
Apply TwoSumOneNul2 with b := b precision := precision p := w q := v pq := q';
 Auto.
Apply TwoSumOneNul2 with b := b precision := precision p := p q := u pq := p';
 Auto.
Split; Auto with float zarith.
Split; Auto with float zarith.
Split; Auto with float zarith.
Simpl; Rewrite H'0.
Replace (Zmin (Fexp q) (Fexp r)) with (Fexp r);
 [Auto with zarith | Apply sym_eq; Apply Zmin_le2; Auto with zarith].
Left.
Split; Auto with float zarith.
Apply TwoSumOneNul2 with b := b precision := precision p := w q := v pq := q';
 Auto.
Apply TwoSumOneNul2 with b := b precision := precision p := p q := u pq := p';
 Auto.
Case (Req_EM u' R0); Intros Hu'; Auto.
Right.
Case (Req_EM v' R0); Intros Hv'; Auto.
Rewrite Hv'.
Replace (Rabsolu R0) with (Rmult (vNum b) R0);
 [Apply Rle_monotony | Auto with real]; Auto.
Apply vNumPos with precision := precision; Auto.
Apply Rle_anti_compatibility with r := (FtoRradix (Float (1) (Fexp r))).
Replace (Rplus (Float (1) (Fexp r)) R0) with (FtoRradix (Float (1) (Fexp r)));
 [Replace (Rplus
             (Float (1) (Fexp r))
             (Rminus (Float (1) (Fexp u')) (Float (1) (Fexp r))))
       with (FtoRradix (Float (1) (Fexp u')));
   [Unfold FtoRradix; Unfold FtoR; Simpl | Ring] | Ring].
Replace (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp r)))
     with (powerRZ (Rplus R1 R1) (Fexp r));
 [Replace (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp u')))
       with (powerRZ (Rplus R1 R1) (Fexp u')); [Apply Rle_powerRZ | Ring] | Ring].
Replace (Rplus R1 R1) with (INR (2)); [Replace R1 with (INR (1)) | Idtac];
 Auto with real arith.
Apply Zle_trans with (Fexp v'); Auto with zarith.
Ring; Apply sym_eqT; Auto with real.
Apply Rabsolu_R0.
Apply Rle_trans with (Rmult (FtoR (2) (Float (1) (Fexp u'))) (Rinv (2)));
 [Unfold FtoRradix;
   Apply ClosestErrorBound
        with b := b x := (Rplus q r) q := v' precision := precision |
   Apply Rle_trans with (FtoR (2) (Float (1) (Fexp u')))]; Auto with real arith.
Apply (ClosestCompatible b (2) (Rplus q r) (Rplus q r) u); Auto.
Apply PuissLessThanHalf.
Unfold FtoRradix; Unfold FtoR; Simpl.
Apply Rle_trans
     with (Rmult
             (Rmult (vNum b) (Rminus R1 (Rinv (2))))
             (powerRZ (Rplus R1 R1) (Fexp u'))).
Apply Rle_monotony_r; Auto.
Apply powerRZ_le; Auto with real.
Replace (Rmult
           (Rmult (vNum b) (Rminus R1 (Rinv (2))))
           (powerRZ (Rplus R1 R1) (Fexp u')))
     with (Rmult
             (vNum b)
             (Rminus
                (powerRZ (Rplus R1 R1) (Fexp u'))
                (powerRZ (Rplus R1 R1) (Zpred (Fexp u'))))).
Apply Rle_monotony.
Apply vNumPos with precision := precision; Auto.
Replace (Rminus
           (powerRZ (Rplus R1 R1) (Fexp u'))
           (powerRZ (Rplus R1 R1) (Zpred (Fexp u'))))
     with (Rplus
             (powerRZ (Rplus R1 R1) (Fexp u'))
             (Ropp (powerRZ (Rplus R1 R1) (Zpred (Fexp u'))))); [Idtac | Ring].
Replace (Rminus
           (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp u')))
           (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp r))))
     with (Rplus
             (powerRZ (Rplus R1 R1) (Fexp u'))
             (Ropp (powerRZ (Rplus R1 R1) (Fexp r)))); [Idtac | Ring].
Apply Rle_compatibility.
Apply Rle_Ropp.
Apply Rle_powerRZ.
Replace (Rplus R1 R1) with (INR (2)); [Replace R1 with (INR (1)) | Idtac];
 Auto with real arith.
Apply Zlt_n_Sm_le.
Rewrite <- (Zs_pred (Fexp u')).
Replace (Fexp r) with (Fexp v'); Auto.
Apply ClosestErrorExpStrict
     with x := (Rplus q r) b := b radix := (2) precision := precision; Auto.
Apply (ClosestCompatible b (2) (Rplus q r) (Rplus q r) u); Auto.
Unfold Zpred.
Rewrite (powerRZ_add (Rplus R1 R1) (Fexp u') (NEG xH)); Auto.
Replace (powerRZ (Rplus R1 R1) (NEG xH)) with (Rinv (2)).
Ring; Ring.
Unfold powerRZ; Simpl.
Replace (Rmult (Rplus R1 R1) R1) with (Rplus R1 R1); Auto with real.
Replace (Rplus R1 R1) with (INR (2)); Auto with real.
Case (Req_EM v R0); Intros Hv.
LApply (PEE b (2) precision);
 [Intros H'10; LApply H'10;
   [Intros H'11; LApply H'11;
     [Intros H'12; LApply (H'12 p u' p');
       [Intros H'16; LApply H'16;
         [Intros H'17; LApply H'17;
           [Intros H'18; Try Exact H'18; Clear H'17 H'16 H'11 H'10 |
             Clear H'17 H'16 H'11 H'10] | Clear H'16 H'11 H'10] |
         Clear H'11 H'10] | Clear H'11 H'10] | Clear H'10] | Idtac]; Auto.
2:Replace (FtoR (2) u') with (FtoR (2) u); Auto with real.
Clear H'12; Clear H'4.
Elim H'18; Intros w' E; Elim E; Intros p'' E0; Clear E H'18.
Elim E0; Intros H'4 H'10; Elim H'10; Intros H'11 H'12; Elim H'12;
 Intros H'13 H'14; Clear H'12 H'10 E0.
Elim H'14; Intros H'10 H'12; Elim H'12; Intros H'15 H'16; Clear H'12 H'14.
Elim H'16; Intros; Clear H'16.
Exists p''; Exists w'; Exists (Fzero (Fexp r)).
Cut <R> w'==w;
 [Intros w'Def |
   Unfold FtoRradix; Replace (FtoR (2) w) with (Rminus (Rplus p u) p');
    Unfold FtoRradix; Replace (FtoR (2) u) with (FtoR (2) u');
    Replace (FtoR (2) p') with (FtoR (2) p''); Auto with real].
Split; Auto.
Split; Auto.
Split; Auto.
Split; Auto.
Split; Auto.
Apply sym_eqT; Apply TwoSumOneNul with precision := precision p := v b := b;
 Auto.
Replace (FtoR (2) w') with (FtoR (2) w); Auto with real.
Replace (Rplus (FtoR (2) v) (FtoR (2) w)) with (Rplus (FtoR (2) w) (FtoR (2) v));
 Auto with real.
Ring.
Replace (FtoRradix (Fzero (Fexp r))) with R0; Apply sym_eqT;
 [Idtac | Unfold FtoRradix; Apply FzeroisReallyZero].
Apply TwoSumOneNul2 with b := b precision := precision p := v q := w pq := q';
 Auto.
Replace (Rplus (FtoR (2) v) (FtoR (2) w)) with (Rplus (FtoR (2) w) (FtoR (2) v));
 Auto with real.
Ring.
Replace (Rplus (FtoR (2) v) (FtoR (2) w)) with (Rplus (FtoR (2) w) (FtoR (2) v));
 Auto with real.
Ring.
Split; Auto with float zarith.
Split; Auto with float zarith.
Split; Auto with float zarith.
Simpl; Rewrite H'15.
Case (Zmin_or (Fexp p) (Fexp u')); Intros E; Rewrite E; Auto.
Apply Zle_trans with (Fexp q); Auto.
Apply Zle_trans with (Fexp v'); Auto with zarith.
Left.
Split; Auto with float zarith.
Apply TwoSumOneNul2 with b := b precision := precision p := v q := w pq := q';
 Auto.
Replace (Rplus (FtoR (2) v) (FtoR (2) w)) with (Rplus (FtoR (2) w) (FtoR (2) v));
 [Auto with real | Ring].
Replace (Rplus (FtoR (2) v) (FtoR (2) w)) with (Rplus (FtoR (2) w) (FtoR (2) v));
 [Auto with real | Ring].
Case (Req_EM p'' R0); Intros Hp''; Auto.
Right.
Case (Req_EM w' R0); Intros Hw'; Auto.
Rewrite Hw'.
Replace (Rabsolu R0) with (Rmult (vNum b) R0);
 [Apply Rle_monotony | Auto with real]; Auto.
Apply vNumPos with precision := precision; Auto.
Apply Rle_anti_compatibility with r := (FtoRradix (Float (1) (Fexp r))).
Replace (Rplus (Float (1) (Fexp r)) R0) with (FtoRradix (Float (1) (Fexp r)));
 [Replace (Rplus
             (Float (1) (Fexp r))
             (Rminus (Float (1) (Fexp p'')) (Float (1) (Fexp r))))
       with (FtoRradix (Float (1) (Fexp p'')));
   [Unfold FtoRradix; Unfold FtoR; Simpl | Ring] | Ring].
Replace (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp r)))
     with (powerRZ (Rplus R1 R1) (Fexp r));
 [Replace (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp p'')))
       with (powerRZ (Rplus R1 R1) (Fexp p'')); [Apply Rle_powerRZ | Ring] |
   Ring].
Replace (Rplus R1 R1) with (INR (2)); [Replace R1 with (INR (1)) | Idtac];
 Auto with real arith.
Apply Zle_trans with (Fexp w'); Auto with zarith.
Simpl; Rewrite H'15.
Case (Zmin_or (Fexp p) (Fexp u')); Intros E; Rewrite E; Auto.
Apply Zle_trans with (Fexp q); Auto.
Apply Zle_trans with (Fexp v'); Auto with zarith.
Ring; Apply sym_eqT; Auto with real.
Apply Rabsolu_R0.
Apply Rle_trans with (Rmult (FtoR (2) (Float (1) (Fexp p''))) (Rinv (2)));
 [Unfold FtoRradix;
   Apply ClosestErrorBound with b := b x := (Rplus p u') precision := precision |
   Apply Rle_trans with (FtoR (2) (Float (1) (Fexp p'')))]; Auto with real arith.
Apply (ClosestCompatible b (2) (Rplus p u') (Rplus p u') p'); Auto.
Unfold FtoRradix; Replace (FtoR (2) u') with (FtoR (2) u); Auto with real.
Apply PuissLessThanHalf.
Unfold FtoRradix; Unfold FtoR; Simpl.
Apply Rle_trans
     with (Rmult
             (Rmult (vNum b) (Rminus R1 (Rinv (2))))
             (powerRZ (Rplus R1 R1) (Fexp p''))).
Apply Rle_monotony_r; Auto.
Apply powerRZ_le; Auto with real.
Replace (Rmult
           (Rmult (vNum b) (Rminus R1 (Rinv (2))))
           (powerRZ (Rplus R1 R1) (Fexp p'')))
     with (Rmult
             (vNum b)
             (Rminus
                (powerRZ (Rplus R1 R1) (Fexp p''))
                (powerRZ (Rplus R1 R1) (Zpred (Fexp p''))))).
Apply Rle_monotony.
Apply vNumPos with precision := precision; Auto.
Replace (Rminus
           (powerRZ (Rplus R1 R1) (Fexp p''))
           (powerRZ (Rplus R1 R1) (Zpred (Fexp p''))))
     with (Rplus
             (powerRZ (Rplus R1 R1) (Fexp p''))
             (Ropp (powerRZ (Rplus R1 R1) (Zpred (Fexp p''))))); [Idtac | Ring].
Replace (Rminus
           (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp p'')))
           (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp r))))
     with (Rplus
             (powerRZ (Rplus R1 R1) (Fexp p''))
             (Ropp (powerRZ (Rplus R1 R1) (Fexp r)))); [Idtac | Ring].
Apply Rle_compatibility.
Apply Rle_Ropp.
Apply Rle_powerRZ.
Replace (Rplus R1 R1) with (INR (2)); [Replace R1 with (INR (1)) | Idtac];
 Auto with real arith.
Apply Zlt_n_Sm_le.
Rewrite <- (Zs_pred (Fexp p'')).
Apply Zle_lt_trans with (Fexp w'); Auto.
Simpl; Rewrite H'15.
Case (Zmin_or (Fexp p) (Fexp u')); Intros E; Rewrite E; Auto.
Apply Zle_trans with (Fexp q); Auto.
Apply Zle_trans with (Fexp v'); Auto with zarith.
Apply ClosestErrorExpStrict
     with x := (Rplus p u') b := b radix := (2) precision := precision; Auto.
Apply (ClosestCompatible b (2) (Rplus p u') (Rplus p u') p'); Auto.
Replace (FtoRradix u') with (FtoRradix u); Auto with real.
Unfold Zpred.
Rewrite (powerRZ_add (Rplus R1 R1) (Fexp p'') (NEG xH)); Auto.
Replace (powerRZ (Rplus R1 R1) (NEG xH)) with (Rinv (2)).
Ring; Ring.
Unfold powerRZ; Simpl.
Replace (Rmult (Rplus R1 R1) R1) with (Rplus R1 R1); Auto with real.
Replace (Rplus R1 R1) with (INR (2)); Auto with real.
Generalize FTSinSum3_allcases; Intros.
LApply H; [Intros H'9; Clear H | Clear H]; Auto.
Elim H'9; Intros u'' E; Clear H'9.
Elim E; Intros H'9 H'10; Clear E.
Elim H'10; Intros; Clear H'10.
LApply (PEE b (2) precision);
 [Intros H'10; LApply H'10;
   [Intros H'11; LApply H'11;
     [Intros H'12; LApply (H'12 p u'' p');
       [Intros H'16; LApply H'16;
         [Intros H'17; LApply H'17;
           [Intros H'18; Try Exact H'18; Clear H'17 H'16 H'11 H'10 |
             Clear H'17 H'16 H'11 H'10] | Clear H'16 H'11 H'10] |
         Clear H'11 H'10] | Clear H'11 H'10] | Clear H'10] | Idtac]; Auto.
2:Replace (FtoR (2) u'') with (FtoR (2) u); Auto with real.
Clear H'12; Clear H'4.
Elim H'18; Intros w' E; Elim E; Intros p'' E0; Clear E H'18.
Elim E0; Intros H'4 H'10; Elim H'10; Intros H'11 H'12; Elim H'12;
 Intros H'13 H'14; Clear H'12 H'10 E0.
Elim H'14; Intros H'10 H'12; Elim H'12; Intros H'15 H'16; Clear H'12 H'14.
Elim H'16; Intros; Clear H'16.
Cut <R> v==v';
 [Intros v'Def |
   Rewrite vDef; Auto with real; Unfold FtoRradix;
    Replace (FtoR (2) u) with (FtoR (2) u'); Auto with real].
LApply (PEE b (2) precision);
 [Intros H'16; LApply H'16;
   [Intros H'17; LApply H'17;
     [Intros H'18; LApply (H'18 w' v' q');
       [Intros H'22; LApply H'22;
         [Intros H'23; LApply H'23;
           [Intros H'24; Try Exact H'24; Clear H'23 H'22 H'17 H'16 |
             Clear H'23 H'22 H'17 H'16] | Clear H'22 H'17 H'16] |
         Clear H'17 H'16] | Clear H'17 H'16] | Clear H'16] | Idtac]; Auto.
Clear H'18.
Elim H'24; Intros r'' E; Elim E; Intros q'' E0; Elim E0; Intros H'12 H'14;
 Elim H'14; Intros H'16 H'17; Elim H'17; Intros H'18 H'19;
 Clear H'17 H'14 E0 E H'24.
Elim H'19; Intros H'14 H'17; Elim H'17; Intros H'20 H'21; Clear H'17 H'19.
Elim H'21; Intros; Clear H'21.
2:Replace (FtoR (2) w') with (FtoR (2) w);
   Replace (FtoR (2) v') with (FtoR (2) v); Auto with real.
2:Replace (FtoR (2) w) with (Rminus (Rplus p u) p'); Unfold FtoRradix;
   Replace (FtoR (2) u) with (FtoR (2) u'');
   Replace (FtoR (2) p') with (FtoR (2) p''); Auto with real.
Cut <R> w'==w;
 [Intros w'Def |
   Unfold FtoRradix; Replace (FtoR (2) w) with (Rminus (Rplus p u) p');
    Unfold FtoRradix; Replace (FtoR (2) u) with (FtoR (2) u'');
    Replace (FtoR (2) p') with (FtoR (2) p''); Auto with real].
Cut <R> r'==r'';
 [Intros r'Def |
   Rewrite r'Def; Unfold FtoRradix; Replace (FtoR (2) w) with (FtoR (2) w');
    Replace (FtoR (2) v) with (FtoR (2) v');
    Replace (FtoR (2) q') with (FtoR (2) q''); Auto with real].
Clear PEE.
Case (Req_EM w' R0); Intros Hw'.
Exists p''; Exists v'; Exists (Fzero (Fexp r)).
Split; Auto.
Split; Auto.
Split; Auto.
Split; Auto.
Split; Auto.
Apply sym_eqT; Apply TwoSumOneNul with b := b precision := precision p := w;
 Auto.
Replace (FtoR (2) v') with (FtoR (2) v); Auto with real.
Replace (FtoR (2) w) with (FtoR (2) w'); Auto with real.
Replace (FtoRradix (Fzero (Fexp r))) with R0;
 [Apply sym_eqT | Apply sym_eqT; Unfold FtoRradix; Apply FzeroisReallyZero].
Rewrite r'Def; Unfold FtoRradix; Replace (FtoR (2) w) with (FtoR (2) w');
 Replace (FtoR (2) w') with R0; Auto.
Apply eq_Rminus; Ring.
Apply sym_eqT; Apply TwoSumOneNul with b := b precision := precision p := w;
 Auto.
Replace (FtoR (2) w) with (FtoR (2) w'); Auto with real.
Split; Auto.
Simpl; Auto with zarith.
Split; Auto.
Split; Auto.
Simpl; Auto.
Rewrite H'8; Auto with zarith.
Apply Zle_trans with (Fexp w'); [Rewrite H'15 | Auto with zarith].
Case (Zmin_or (Fexp p) (Fexp u'')); Intros E; Rewrite E; Auto with zarith.
Apply Zlt_le_weak.
Apply ClosestErrorExpStrict
     with b := b radix := (2) precision := precision x := (Rplus q r); Auto.
Apply (ClosestCompatible b (2) (Rplus q r) (Rplus q r) u); Auto.
Replace (FtoR (2) u'') with (FtoR (2) u'); Auto with real.
Replace (FtoR (2) u'') with (FtoR (2) u); Auto with real.
Replace (FtoR (2) v') with (FtoR (2) v); Auto with real.
Left.
Split; Auto.
Rewrite r'Def; Unfold FtoRradix; Replace (FtoR (2) w) with (FtoR (2) w');
 Replace (FtoR (2) w') with R0; Auto.
Apply eq_Rminus; Ring.
Apply sym_eqT; Apply TwoSumOneNul with b := b precision := precision p := w;
 Auto.
Replace (FtoR (2) w) with (FtoR (2) w'); Auto with real.
Right.
Apply Rle_trans with (Rmult (Float (1) (Fexp u'')) (Rinv (2))).
Unfold FtoRradix;
 Apply ClosestErrorBound with b := b precision := precision x := (Rplus q r);
 Auto with real.
Apply (ClosestCompatible b (2) (Rplus q r) (Rplus q r) u); Auto.
Replace (FtoR (2) u'') with (FtoR (2) u);
 Replace (FtoR (2) v') with (FtoR (2) v); Auto with real.
Apply Rle_trans with (FtoRradix (Float (1) (Fexp u''))); Auto with real.
Apply PuissLessThanHalf.
Apply Rle_trans with (FtoRradix (Float (1) (Fexp p''))); Auto with real.
Unfold FtoRradix; Unfold FtoR; Simpl.
Replace (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp u'')))
     with (powerRZ (Rplus R1 R1) (Fexp u''));
 [Replace (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp p'')))
       with (powerRZ (Rplus R1 R1) (Fexp p'')); [Idtac | Ring] | Ring].
Apply Rle_powerRZ; Auto with real.
Replace (Rplus R1 R1) with (INR (2)); Auto with real arith zarith.
Replace R1 with (INR (1)); Auto with real arith zarith.
Apply Zle_trans with (Fexp w'); Auto.
Rewrite H'15.
Replace (Zmin (Fexp p) (Fexp u'')) with (Fexp u'');
 [Auto with zarith | Apply sym_eq; Apply Zmin_le2; Auto].
Unfold FtoRradix; Unfold FtoR; Simpl.
Apply Rle_trans
     with (Rmult
             (Rmult (vNum b) (Rminus R1 (Rinv (2))))
             (powerRZ (Rplus R1 R1) (Fexp p''))).
Apply Rle_monotony_r; Auto.
Apply powerRZ_le; Auto with real.
Replace (Rmult
           (Rmult (vNum b) (Rminus R1 (Rinv (2))))
           (powerRZ (Rplus R1 R1) (Fexp p'')))
     with (Rmult
             (vNum b)
             (Rminus
                (powerRZ (Rplus R1 R1) (Fexp p''))
                (powerRZ (Rplus R1 R1) (Zpred (Fexp p''))))).
Apply Rle_monotony.
Apply vNumPos with precision := precision; Auto.
Replace (Rminus
           (powerRZ (Rplus R1 R1) (Fexp p''))
           (powerRZ (Rplus R1 R1) (Zpred (Fexp p''))))
     with (Rplus
             (powerRZ (Rplus R1 R1) (Fexp p''))
             (Ropp (powerRZ (Rplus R1 R1) (Zpred (Fexp p''))))); [Idtac | Ring].
Replace (Rminus
           (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp p'')))
           (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp r))))
     with (Rplus
             (powerRZ (Rplus R1 R1) (Fexp p''))
             (Ropp (powerRZ (Rplus R1 R1) (Fexp r)))); [Idtac | Ring].
Apply Rle_compatibility.
Apply Rle_Ropp.
Apply Rle_powerRZ.
Replace (Rplus R1 R1) with (INR (2)); [Replace R1 with (INR (1)) | Idtac];
 Auto with real arith.
Apply Zlt_n_Sm_le.
Rewrite <- (Zs_pred (Fexp p'')).
Apply Zlt_le_trans with (Fexp w'); Auto.
Rewrite <- H'8.
Rewrite H'15.
Replace (Zmin (Fexp p) (Fexp u'')) with (Fexp u'').
Apply ClosestErrorExpStrict
     with b := b radix := (2) precision := precision x := (Rplus q r); Auto.
Apply (ClosestCompatible b (2) (Rplus q r) (Rplus q r) u); Auto.
Replace (FtoR (2) u'') with (FtoR (2) u'); Auto with real.
Replace (FtoR (2) u'') with (FtoR (2) u); Auto with real.
Replace (FtoR (2) v') with (FtoR (2) v); Auto with real.
Apply sym_eq; Apply Zmin_le2; Auto.
Unfold Zpred.
Rewrite (powerRZ_add (Rplus R1 R1) (Fexp p'') (NEG xH)); Auto.
Replace (powerRZ (Rplus R1 R1) (NEG xH)) with (Rinv (2)).
Ring; Ring.
Unfold powerRZ; Simpl.
Replace (Rmult (Rplus R1 R1) R1) with (Rplus R1 R1); Auto with real.
Replace (Rplus R1 R1) with (INR (2)); Auto with real.
Exists p''; Exists q''; Exists r''.
Split; Auto.
Split; Auto.
Split; Auto.
Split; Auto.
Apply Zle_trans with (Fexp v'); Auto with zarith.
Rewrite H'20.
Replace (Zmin (Fexp w') (Fexp v')) with (Fexp v'); Auto with zarith.
Apply sym_eq; Apply Zmin_le2.
Rewrite H'15.
Case (Zmin_or (Fexp p) (Fexp u'')); Intros L; Rewrite L; Auto with zarith.
Apply Zlt_le_weak.
Apply ClosestErrorExpStrict
     with b := b radix := (2) precision := precision x := (Rplus q r); Auto.
Apply (ClosestCompatible b (2) (Rplus q r) (Rplus q r) u); Auto.
Replace (FtoR (2) u'') with (FtoR (2) u'); Auto with real.
Replace (FtoR (2) u'') with (FtoR (2) u); Auto with real.
Replace (FtoR (2) v') with (FtoR (2) v); Auto with real.
Split; Auto.
Split; Auto.
Apply Zle_trans with (Zs (Zmax (Fexp w') (Fexp v'))); Auto.
Replace (Fexp p'') with (Zs (Zpred (Fexp p''))).
Apply Zle_n_S.
Apply Zgt_S_le.
Replace (Zs (Zpred (Fexp p''))) with (Fexp p'').
Apply Zlt_gt.
Replace (Zmax (Fexp w') (Fexp v')) with (Fexp w').
Apply ClosestErrorExpStrict
     with b := b radix := (2) precision := precision x := (Rplus p u''); Auto.
Apply (ClosestCompatible b (2) (Rplus p u) (Rplus p u'') p'); Auto.
Replace (FtoR (2) u'') with (FtoR (2) u); Auto with real.
Apply sym_eq; Apply Zmax_le1; Auto.
Apply Zlt_le_weak.
Replace (Fexp w') with (Fexp u'').
Apply ClosestErrorExpStrict
     with b := b radix := (2) precision := precision x := (Rplus q r); Auto.
Apply (ClosestCompatible b (2) (Rplus q r) (Rplus q r) u); Auto.
Replace (FtoR (2) u'') with (FtoR (2) u'); Auto with real.
Replace (FtoR (2) u'') with (FtoR (2) u); Auto with real.
Replace (FtoR (2) v') with (FtoR (2) v); Auto with real.
Rewrite H'15; Apply sym_eq; Apply Zmin_le2; Auto.
Apply Zs_pred.
Apply sym_eq; Apply Zs_pred.
Replace (Fexp r) with (Fexp v'); Auto.
Rewrite H'20; Apply Zmin_le2.
Replace (Fexp w') with (Fexp u'').
Apply Zlt_le_weak.
Apply ClosestErrorExpStrict
     with b := b radix := (2) precision := precision x := (Rplus q r); Auto.
Apply (ClosestCompatible b (2) (Rplus q r) (Rplus q r) u); Auto.
Replace (FtoR (2) u'') with (FtoR (2) u'); Auto with real.
Replace (FtoR (2) u'') with (FtoR (2) u); Auto with real.
Replace (FtoR (2) v') with (FtoR (2) v); Auto with real.
Rewrite H'15; Apply sym_eq; Apply Zmin_le2; Auto.
Case (Req_EM r' R0); Intros Hr'.
Left; Split; Auto.
Right.
Replace (FtoRradix q'') with (Rplus w' v').
Apply Rle_trans with (Rplus (Rabsolu w') (Rabsolu v')).
Apply Rabsolu_triang.
Apply Rle_trans
     with (Rplus
             (Rabsolu w')
             (Rmult
                (Rmult (Rabsolu (FtoR (2) u)) (Rinv (2)))
                (Rmult (2) (Rinv (vNum b))))).
Apply Rle_compatibility.
Rewrite <- v'Def; Rewrite vDef.
Apply Rlt_le.
Apply Rle_lt_trans with (Rabsolu (Rminus u (Rplus q r))).
Apply eq_Rle; Replace (Rminus (Rplus q r) u) with (Ropp (Rminus u (Rplus q r)));
 [Apply Rabsolu_Ropp | Ring].
Unfold FtoRradix; Apply plusErrorBound1 with precision := precision; Auto.
Unfold is_Fzero.
Contradict Hw'.
Apply TwoSumOneNul2 with precision := precision p := u q := p pq := p' b := b;
 Auto.
Replace (Rplus (FtoR (2) u) (FtoR (2) p)) with (Rplus (FtoR (2) p) (FtoR (2) u));
 Auto; Ring.
Unfold FtoR; Simpl; Rewrite Hw'; Ring.
Replace (FtoR (2) w') with (FtoR (2) w); Auto with real.
Replace (Rplus (FtoR (2) u) (FtoR (2) p)) with (Rplus (FtoR (2) p) (FtoR (2) u));
 Auto; Ring.
Replace (Rmult
           (Rmult (Rabsolu (FtoR (2) u)) (Rinv (2))) (Rmult (2) (Rinv (vNum b))))
     with (Rmult (Rabsolu (FtoR (2) u)) (Rinv (vNum b))).
2:Replace (Rmult
             (Rmult (Rabsolu (FtoR (2) u)) (Rinv (2)))
             (Rmult (2) (Rinv (vNum b))))
       with (Rmult
               (Rmult (Rabsolu (FtoR (2) u)) (Rinv (vNum b)))
               (Rmult (Rinv (2)) (2))); [Idtac | Ring].
2:Replace (Rmult (Rinv (2)) (2)) with R1; Ring.
2:Apply sym_eqT; Apply Rinv_r; Auto with real arith.
Apply Rle_trans
     with (Rplus
             (Rmult
                (Rmult (2) (Rinv (vNum b)))
                (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u''))))
             (Rmult (Rabsolu (FtoR (2) u)) (Rinv (vNum b)))).
Replace (Rplus (Rabsolu w') (Rmult (Rabsolu (FtoR (2) u)) (Rinv (vNum b))))
     with (Rplus (Rmult (Rabsolu (FtoR (2) u)) (Rinv (vNum b))) (Rabsolu w'));
 [Idtac | Ring].
Replace (Rplus
           (Rmult
              (Rmult (2) (Rinv (vNum b)))
              (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u''))))
           (Rmult (Rabsolu (FtoR (2) u)) (Rinv (vNum b))))
     with (Rplus
             (Rmult (Rabsolu (FtoR (2) u)) (Rinv (vNum b)))
             (Rmult
                (Rmult (2) (Rinv (vNum b)))
                (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u'')))));
 [Idtac | Ring].
Apply Rle_compatibility.
Replace (FtoRradix w') with (Rminus (Rplus p u'') p''); Auto with real.
Apply Rlt_le.
Unfold FtoRradix.
Apply Rle_lt_trans
     with (Rabsolu (Rminus (FtoR (2) p'') (Rplus (FtoR (2) p) (FtoR (2) u'')))).
Apply eq_Rle;
 Replace (Rminus (Rplus (FtoR (2) p) (FtoR (2) u'')) (FtoR (2) p''))
      with (Ropp (Rminus (FtoR (2) p'') (Rplus (FtoR (2) p) (FtoR (2) u''))));
 [Apply Rabsolu_Ropp | Ring].
Apply plusErrorBound2 with b := b precision := precision; Auto.
Apply (ClosestCompatible b (2) (Rplus p u) (Rplus p u'') p'); Auto.
Replace (FtoR (2) u'') with (FtoR (2) u); Auto with real.
Cut ~ (FtoRradix p'')==R0 ->~ (is_Fzero p'').
Intros Hi; Apply Hi.
Apply TwoSumNonNul with b := b precision := precision r := w' p := p q := u'';
 Auto.
Apply (ClosestCompatible b (2) (Rplus p u) (Rplus p u'') p'); Auto.
Replace (FtoR (2) u'') with (FtoR (2) u); Auto with real.
Unfold FtoRradix; Unfold FtoR; Simpl.
Unfold is_Fzero.
Intros Hi.
Contradict Hi.
Rewrite Hi; Ring.
Apply Rle_trans
     with (Rplus
             (Rmult
                (Rmult (2) (Rinv (vNum b)))
                (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u''))))
             (Rmult
                (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u'')))
                (Rinv (vNum b)))).
Apply Rle_compatibility.
Apply Rle_monotony_r.
Apply Rlt_le.
Apply Rlt_Rinv.
Replace R0 with (INR (0)); Auto with real.
Apply Rlt_INR; Apply lt_vNum with precision := precision radix := (2);
 Auto with real arith.
Replace (FtoR (2) u) with (FtoR (2) u''); Auto with real.
Apply RmaxLess2.
Replace (Rplus
           (Rmult
              (Rmult (2) (Rinv (vNum b)))
              (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u''))))
           (Rmult
              (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u'')))
              (Rinv (vNum b))))
     with (Rmult
             (Rmult (3) (Rinv (vNum b)))
             (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u''))));
 [Idtac | Ring; Ring].
Apply Rle_trans with (Rmult (Rabsolu p'') (Rmult (6) (Rinv (vNum b)))).
Replace (Rmult
           (Rmult (3) (Rinv (vNum b)))
           (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u''))))
     with (Rmult
             (Rmult (3) (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u''))))
             (Rinv (vNum b))); [Idtac | Ring].
Replace (Rmult (Rabsolu p'') (Rmult (6) (Rinv (vNum b))))
     with (Rmult (Rmult (Rabsolu p'') (6)) (Rinv (vNum b))); [Idtac | Ring].
Apply Rle_monotony_r.
Apply Rlt_le.
Apply Rlt_Rinv.
Replace R0 with (INR (0)); Auto with real.
Apply Rlt_INR; Apply lt_vNum with precision := precision radix := (2);
 Auto with real arith.
Replace (Rmult (3) (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u''))))
     with (Rmult
             (Rmult
                (Rinv (2))
                (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u'')))) (6));
 [Idtac | Ring; Ring; Auto with real].
Apply Rle_monotony_r; Auto with real.
Apply Rlt_le.
Replace R0 with (INR (0)); Auto with real.
Apply Rlt_INR; Auto with real arith.
Apply plusClosestLowerBound with b := b precision := precision; Auto.
Apply (ClosestCompatible b (2) (Rplus p u) (Rplus p u'') p'); Auto.
Replace (FtoR (2) u'') with (FtoR (2) u); Auto with real.
Contradict Hw'.
Unfold FtoRradix; Rewrite H'10.
Rewrite Hw'; Ring.
Replace (Rmult (Rinv (2)) (6)) with (INR (3)); Ring.
Apply r_Rmult_mult with r := (INR (2)).
Replace (Rmult (2) (Rmult (6) (Rinv (2)))) with (INR (6)).
Replace (6) with (mult (2) (3)); Auto with arith.
Apply sym_eqT; Apply Rmult_INR; Auto with real arith.
Replace (Rmult (2) (Rmult (6) (Rinv (2))))
     with (Rmult (6) (Rmult (2) (Rinv (2)))); [Idtac | Ring].
Replace (Rmult (2) (Rinv (2))) with R1; Auto with real.
Apply sym_eqT; Apply Rinv_r; Auto with real arith.
Replace R0 with (INR (0)); Auto with real arith.
Apply Rle_trans with (Rmult (6) (powerRZ (Rplus R1 R1) (Fexp p''))).
Replace (Rmult (Rabsolu p'') (Rmult (6) (Rinv (vNum b))))
     with (Rmult (6) (Rmult (Rabsolu p'') (Rinv (vNum b))));
 [Apply Rle_monotony | Ring].
Replace R0 with (INR (0)); Auto with real arith.
Replace (powerRZ (Rplus R1 R1) (Fexp p''))
     with (Rmult (Float (vNum b) (Fexp p'')) (Rinv (vNum b)));
 [Apply Rle_monotony_r | Unfold FtoRradix; Unfold FtoR; Simpl; Ring].
Apply Rlt_le.
Apply Rlt_Rinv.
Replace R0 with (INR (0)); Auto with real.
Apply Rlt_INR; Apply lt_vNum with precision := precision radix := (2);
 Auto with real arith.
Unfold FtoRradix; Apply MaxFloat with precision := precision; Auto.
Rewrite <- INR_IZR_INZ.
Replace (Rmult
           (Rinv (vNum b)) (Rmult (powerRZ (Rplus R1 R1) (Fexp p'')) (vNum b)))
     with (Rmult
             (powerRZ (Rplus R1 R1) (Fexp p'')) (Rmult (vNum b) (Rinv (vNum b))));
 [Idtac | Ring].
Replace (Rmult (vNum b) (Rinv (vNum b))) with R1; Try Ring.
Replace (Rmult (Rinv (vNum b)) (vNum b)) with (Rmult (vNum b) (Rinv (vNum b)));
 [Idtac | Ring].
Apply sym_eqT; Apply Rinv_r.
Apply sym_not_eqT; Apply imp_not_Req.
Left.
Replace R0 with (INR (0)); Auto with real.
Apply Rlt_INR; Apply lt_vNum with precision := precision radix := (2);
 Auto with real arith.
Apply Rle_trans
     with (Rmult
             (Rmult (vNum b) (Rminus R1 (Rmult (Rinv (2)) (Rinv (2)))))
             (powerRZ (Rplus R1 R1) (Fexp p''))).
Apply Rle_monotony_r.
Apply powerRZ_le; Auto with real arith.
Auto.
Replace (Rmult
           (Rmult (vNum b) (Rminus R1 (Rmult (Rinv (2)) (Rinv (2)))))
           (powerRZ (Rplus R1 R1) (Fexp p'')))
     with (Rmult
             (vNum b)
             (Rminus
                (powerRZ (Rplus R1 R1) (Fexp p''))
                (powerRZ (Rplus R1 R1) (Zpred (Zpred (Fexp p''))))));
 [Apply Rle_monotony | Ring].
Apply vNumPos with precision := precision; Auto.
Replace (Rminus
           (powerRZ (Rplus R1 R1) (Fexp p''))
           (powerRZ (Rplus R1 R1) (Zpred (Zpred (Fexp p'')))))
     with (Rplus
             (powerRZ (Rplus R1 R1) (Fexp p''))
             (Ropp (powerRZ (Rplus R1 R1) (Zpred (Zpred (Fexp p''))))));
 [Idtac | Ring].
Unfold FtoRradix; Unfold FtoR; Simpl.
Replace (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp p'')))
     with (powerRZ (Rplus R1 R1) (Fexp p'')); [Idtac | Ring].
Replace (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp r)))
     with (powerRZ (Rplus R1 R1) (Fexp r)); [Idtac | Ring].
Replace (Rminus
           (powerRZ (Rplus R1 R1) (Fexp p'')) (powerRZ (Rplus R1 R1) (Fexp r)))
     with (Rplus
             (powerRZ (Rplus R1 R1) (Fexp p''))
             (Ropp (powerRZ (Rplus R1 R1) (Fexp r)))); [Idtac | Ring].
Apply Rle_compatibility.
Apply Rle_Ropp.
Apply Rle_powerRZ.
Replace (Rplus R1 R1) with (INR (2)); Auto with real arith.
Replace R1 with (INR (1)); Auto with real arith.
Apply Zgt_S_le.
Replace (Zs (Zpred (Zpred (Fexp p'')))) with (Zpred (Fexp p'')).
2:Apply Zs_pred.
Apply Zlt_gt.
Apply Zlt_le_trans with (Fexp u'').
Replace (Fexp r) with (Fexp v'); Auto.
Apply ClosestErrorExpStrict
     with b := b radix := (2) precision := precision x := (Rplus q r); Auto.
Apply (ClosestCompatible b (2) (Rplus q r) (Rplus q r) u); Auto.
Replace (FtoR (2) u'') with (FtoR (2) u'); Auto with real.
Replace (FtoR (2) u'') with (FtoR (2) u); Auto with real.
Replace (FtoR (2) v') with (FtoR (2) v); Auto with real.
Apply Zgt_S_le.
Replace (Zs (Zpred (Fexp p''))) with (Fexp p'').
2:Apply Zs_pred.
Replace (Fexp u'') with (Fexp w'); Auto.
Apply Zlt_gt.
Apply ClosestErrorExpStrict
     with b := b radix := (2) precision := precision x := (Rplus p u''); Auto.
Apply (ClosestCompatible b (2) (Rplus p u'') (Rplus p u'') p'); Auto.
Replace (FtoRradix u'') with (FtoRradix u); Auto with real.
Rewrite H'15; Apply Zmin_le2; Auto.
Replace (Rplus
           (Ropp
              (Rmult
                 (powerRZ (Rplus R1 R1) (Fexp p''))
                 (Rmult (Rinv (2)) (Rmult (Rinv (2)) (vNum b)))))
           (Rmult (powerRZ (Rplus R1 R1) (Fexp p'')) (vNum b)))
     with (Rplus
             (Rmult (powerRZ (Rplus R1 R1) (Fexp p'')) (vNum b))
             (Ropp
                (Rmult
                   (powerRZ (Rplus R1 R1) (Fexp p''))
                   (Rmult (Rinv (2)) (Rmult (Rinv (2)) (vNum b))))));
 [Idtac | Ring].
Apply Rplus_plus_r.
Apply eq_Ropp.
Replace (powerRZ (Rplus R1 R1) (Zpred (Zpred (Fexp p''))))
     with (Rmult
             (powerRZ (Rplus R1 R1) (Fexp p'')) (Rmult (Rinv (2)) (Rinv (2))));
 [Ring | Idtac].
Unfold Zpred.
Rewrite (powerRZ_add (Rplus R1 R1) (Zplus (Fexp p'') (NEG xH)) (NEG xH)).
Rewrite (powerRZ_add (Rplus R1 R1) (Fexp p'') (NEG xH)).
Replace (Rinv (2)) with (powerRZ (Rplus R1 R1) (NEG xH)); [Ring | Idtac].
Unfold powerRZ; Simpl; Ring.
Replace (Rmult (Rplus R1 R1) R1) with (Rplus R1 R1); [Auto with real | Ring].
Replace (Rplus R1 R1) with (INR (2)); Replace R0 with (INR (0));
 Auto with real arith.
Replace (Rplus R1 R1) with (INR (2)); Replace R0 with (INR (0));
 Auto with real arith.
Replace (Rplus
           (Rmult
              (2)
              (Rmult
                 (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u'')))
                 (Rinv (vNum b))))
           (Rmult
              (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u'')))
              (Rinv (vNum b))))
     with (Rmult
             (Rplus (2) R1)
             (Rmult
                (Rmax (Rabsolu (FtoR (2) p)) (Rabsolu (FtoR (2) u'')))
                (Rinv (vNum b)))).
Replace (Rplus (2) R1) with (INR (3)).
Ring.
Replace R1 with (INR (1)); Auto with real arith.
Ring.
Replace (Rplus w' v') with (Rplus r'' q'').
Replace (FtoRradix r'') with R0; Ring; Auto with real.
Replace (FtoRradix r'') with (FtoRradix r'); Auto.
Replace (FtoRradix r'')
     with (Rminus (Rplus (FtoR (2) w') (FtoR (2) v')) (FtoR (2) q''));
 Auto with real.
Unfold FtoRradix; Ring; Ring.
Right; Split; Auto.
Unfold FtoRradix;
 Apply Rle_trans with (Rmult (FtoR (2) (Float (1) (Fexp q''))) (Rinv (2))).
Apply ClosestErrorBound with b := b precision := precision x := (Rplus w' v');
 Auto.
Apply (ClosestCompatible b (2) (Rplus w' v') (Rplus w' v') q'); Auto.
Replace (FtoRradix w') with (FtoRradix w);
 Replace (FtoRradix v') with (FtoRradix v); Auto with real.
Apply Rle_trans with (FtoR (2) (Float (1) (Fexp q''))).
Apply PuissLessThanHalf.
Unfold FtoR; Simpl.
Apply Rle_trans
     with (Rmult
             (Rmult (vNum b) (Rminus R1 (Rinv (2))))
             (powerRZ (Rplus R1 R1) (Fexp q''))).
Apply Rle_monotony_r; Auto.
Apply powerRZ_le; Auto with real.
Replace (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp q'')))
     with (powerRZ (Rplus R1 R1) (Fexp q''));
 Replace (Rmult R1 (powerRZ (Rplus R1 R1) (Fexp r)))
      with (powerRZ (Rplus R1 R1) (Fexp r)); Try Ring.
Replace (Rmult
           (Rmult (vNum b) (Rminus R1 (Rinv (2))))
           (powerRZ (Rplus R1 R1) (Fexp q'')))
     with (Rmult
             (vNum b)
             (Rmult (Rminus R1 (Rinv (2))) (powerRZ (Rplus R1 R1) (Fexp q''))));
 [Apply Rle_monotony | Ring; Ring].
Apply vNumPos with precision := precision; Auto.
Replace (Rmult (Rminus R1 (Rinv (2))) (powerRZ (Rplus R1 R1) (Fexp q'')))
     with (Rminus
             (powerRZ (Rplus R1 R1) (Fexp q''))
             (powerRZ (Rplus R1 R1) (Zpred (Fexp q'')))).
Replace (Rminus
           (powerRZ (Rplus R1 R1) (Fexp q''))
           (powerRZ (Rplus R1 R1) (Zpred (Fexp q''))))
     with (Rplus
             (powerRZ (Rplus R1 R1) (Fexp q''))
             (Ropp (powerRZ (Rplus R1 R1) (Zpred (Fexp q''))))); [Idtac | Ring].
Replace (Rminus
           (powerRZ (Rplus R1 R1) (Fexp q'')) (powerRZ (Rplus R1 R1) (Fexp r)))
     with (Rplus
             (powerRZ (Rplus R1 R1) (Fexp q''))
             (Ropp (powerRZ (Rplus R1 R1) (Fexp r)))); [Idtac | Ring].
Apply Rle_compatibility.
Apply Rle_Ropp.
Apply Rle_powerRZ.
Replace (Rplus R1 R1) with (INR (2)); Replace R1 with (INR (1));
 Auto with real arith.
Apply Zgt_S_le.
Replace (Zs (Zpred (Fexp q''))) with (Fexp q'').
2:Apply Zs_pred.
Apply Zlt_gt.
Replace (Fexp r) with (Fexp r'').
Apply ClosestErrorExpStrict
     with b := b radix := (2) precision := precision x := (Rplus w' v'); Auto.
Apply (ClosestCompatible b (2) (Rplus w' v') (Rplus w' v') q'); Auto.
Replace (FtoRradix w') with (FtoRradix w);
 Replace (FtoRradix v') with (FtoRradix v); Auto with real.
Replace (FtoR (2) r'') with (FtoR (2) r'); Auto with real.
Replace (Fexp r) with (Fexp v'); Auto.
Rewrite H'20; Apply Zmin_le2.
Replace (Fexp w') with (Fexp u'').
Apply Zlt_le_weak.
Apply ClosestErrorExpStrict
     with b := b radix := (2) precision := precision x := (Rplus q r); Auto.
Apply (ClosestCompatible b (2) (Rplus q r) (Rplus q r) u); Auto.
Replace (FtoR (2) u'') with (FtoR (2) u'); Auto with real.
Replace (FtoR (2) u'') with (FtoR (2) u); Auto with real.
Replace (FtoR (2) v') with (FtoR (2) v); Auto with real.
Rewrite H'15; Apply sym_eq; Apply Zmin_le2; Auto.
Ring; Ring.
Replace (powerRZ (Rplus R1 R1) (Zpred (Fexp q'')))
     with (Rmult (Rinv (2)) (powerRZ (Rplus R1 R1) (Fexp q''))); Auto with real.
Ring.
Unfold Zpred.
Rewrite (powerRZ_add (Rplus R1 R1) (Fexp q'') (NEG xH)); Auto with real.
Replace (powerRZ (Rplus R1 R1) (NEG xH)) with (Rinv (2));
 [Ring | Unfold powerRZ; Simpl].
Replace (Rmult (Rplus R1 R1) R1) with (Rplus R1 R1); Ring.
Replace (Rplus R1 R1) with (INR (2)); Auto with real arith.
Qed.
End F2.
