
(****************************************************************************
                                                                             
          IEEE754  :  TwoSum                                                  
                                                                             
          Laurent Thery & Sylvie Boldo                                        
                                                                             
*****************************************************************************
*)
Require Export Fast2Sum.
Section EFast.
Variable b:Fbound.
Variable precision:nat.

Local FtoRradix := (FtoR (2)).
Coercion FtoRradix : float >-> R.
Hypothesis precisionGreaterThanOne:(lt (1) precision).
Hypothesis pGivesBound:(vNum b)=(minus (exp (2) precision) (1)).
Variable Iplus:float -> float ->float.
Hypothesis IplusCorrect:
           (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
           (Closest b (2) (Rplus p q) (Iplus p q)).
Hypothesis IplusComp:
           (p, q, r, s:float)
           (Fbounded b p) ->
           (Fbounded b q) ->
           (Fbounded b r) -> (Fbounded b s) -> <R> p==r -> <R> q==s ->
           <R> (Iplus p q)==(Iplus r s).
Hypothesis IplusSym:(p, q:float)(Iplus p q)=(Iplus q p).
Hypothesis IplusOp:(p, q:float)(Fopp (Iplus p q))=(Iplus (Fopp p) (Fopp q)).
Variable Iminus:float -> float ->float.
Hypothesis IminusPlus:(p, q:float)(Iminus p q)=(Iplus p (Fopp q)).

Theorem IplusOl:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) -> <R> p==R0 ->
  <R> (Iplus p q)==q.
Intros p q H' H'0 H'1.
Rewrite IplusSym.
Apply (IplusOr b); Auto.
Qed.

Local IminusCorrect := (IminusCorrect b Iplus IplusCorrect Iminus IminusPlus).

Local IplusBounded := (IplusBounded b Iplus IplusCorrect).

Local IminusBounded := (IminusBounded b Iplus IplusCorrect Iminus IminusPlus).

Local IminusId := (IminusId b Iplus IplusCorrect Iminus IminusPlus).

Theorem MKnuth:
  (p, q:float)
  (Fbounded b p) ->
  (Fbounded b q) -> <R> (Iminus (Iplus p q) p)==(Rminus (Iplus p q) p) ->
  <R>
    (Iplus
       (Iminus p (Iminus (Iplus p q) (Iminus (Iplus p q) p)))
       (Iminus q (Iminus (Iplus p q) p)))==(Rminus (Rplus p q) (Iplus p q)).
Intros p q H' H'0 H'1.
Rewrite IplusOl; Auto.
Case (errorBoundedPlus b (2) precision) with p := p q := q pq := (Iplus p q);
 Auto.
Intros x H'2; Elim H'2; Intros H'3 H'4; Elim H'4; Intros H'5 H'6; Clear H'4 H'2.
Unfold FtoRradix; Rewrite <- H'3.
Apply sym_eqT; Apply (ClosestIdem b); Auto.
Apply (ClosestCompatible
         b (2) (Rminus q (Iminus (Iplus p q) p)) (FtoR (2) x)
         (Iminus q (Iminus (Iplus p q) p))); Auto.
Apply IminusCorrect; Auto.
Repeat (Try Apply IplusBounded; Try Apply IminusBounded; Auto); Auto.
Rewrite H'3.
Rewrite H'1.
Unfold FtoRradix; Ring; Ring.
Repeat (Try Apply IplusBounded; Try Apply IminusBounded; Auto); Auto.
Repeat (Try Apply IplusBounded; Try Apply IminusBounded; Auto); Auto.
Repeat (Try Apply IplusBounded; Try Apply IminusBounded; Auto); Auto.
Replace R0 with (FtoRradix (Iminus p p)).
Repeat Rewrite IminusPlus; Apply IplusComp; Repeat Rewrite <- IminusPlus; Auto.
Apply oppBounded.
Repeat (Try Apply IplusBounded; Try Apply IminusBounded; Auto); Auto.
Apply oppBounded; Auto.
Unfold FtoRradix; Repeat Rewrite Fopp_correct.
Replace (FtoR (2) (Iminus (Iplus p q) (Iminus (Iplus p q) p))) with (FtoR (2) p);
 Auto.
Apply (ClosestIdem b); Auto.
Apply (ClosestCompatible
         b (2) (Rminus (Iplus p q) (Iminus (Iplus p q) p)) (FtoR (2) p)
         (Iminus (Iplus p q) (Iminus (Iplus p q) p))); Auto.
Apply IminusCorrect; Auto.
Repeat (Try Apply IplusBounded; Try Apply IminusBounded; Auto); Auto.
Repeat (Try Apply IplusBounded; Try Apply IminusBounded; Auto); Auto.
Rewrite H'1; Auto.
Unfold FtoRradix; Ring; Ring.
Repeat (Try Apply IplusBounded; Try Apply IminusBounded; Auto); Auto.
Apply IminusId; Auto.
Qed.

Theorem IplusCorrectEq:
  (p, q, pq:float)
  (r:R)
  (Fbounded b p) ->
  (Fbounded b q) -> (Fbounded b pq) -> <R> r==pq -> <R> (Rplus p q)==pq ->
  <R> (Iplus p q)==r.
Intros p q pq r H' H'0 H'1 H'2 H'3; Rewrite H'2.
Unfold FtoRradix; Apply sym_eqT; Apply (ClosestIdem b (2)); Auto.
Apply (ClosestCompatible b (2) (Rplus p q) pq (Iplus p q)); Auto.
Apply IplusBounded; Auto.
Qed.

Theorem IminusCorrectEq:
  (p, q, pq:float)
  (r:R)
  (Fbounded b p) ->
  (Fbounded b q) -> (Fbounded b pq) -> <R> r==pq -> <R> (Rminus p q)==pq ->
  <R> (Iminus p q)==r.
Intros p q pq r H' H'0 H'1 H'2 H'3; Rewrite H'2.
Unfold FtoRradix; Apply sym_eqT; Apply (ClosestIdem b (2)); Auto.
Apply (ClosestCompatible b (2) (Rminus p q) pq (Iminus p q)); Auto.
Apply IminusCorrect; Auto.
Apply IminusBounded; Auto.
Qed.

Theorem Iminus2Exact:
  (p, q:float) (Rle R0 p) -> (Rle p q) -> (Fbounded b p) -> (Fbounded b q) ->
  <R> (Iminus q (Iminus q p))==(Rminus q (Iminus q p)).
Intros p q H' H'0 H'1 H'2.
Case (minusRoundRep b (2) precision) with P := (Closest b (2)) p := p q := q
                                          qmp := (Iminus q p); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply IminusCorrect; Auto.
Intros pq H'3; Elim H'3; Intros H'4 H'5; Clear H'3.
Apply IminusCorrectEq with pq := pq; Auto.
Apply IminusBounded; Auto.
Qed.

Theorem MKnuth1:
  (p, q:float)
  (Fbounded b p) ->
  (Fbounded b q) ->
  <R> (Iminus q (Iminus (Iplus p q) p))==(Rminus q (Iminus (Iplus p q) p)) ->
  <R>
    (Iminus (Iplus p q) (Iminus (Iplus p q) p))==
    (Rminus (Iplus p q) (Iminus (Iplus p q) p)) ->
  <R>
    (Iplus
       (Iminus p (Iminus (Iplus p q) (Iminus (Iplus p q) p)))
       (Iminus q (Iminus (Iplus p q) p)))==(Rminus (Rplus p q) (Iplus p q)).
Intros p q H' H'0 H'1 H'2.
Cut <R>
      (Iminus p (Iminus (Iplus p q) (Iminus (Iplus p q) p)))==
      (Rminus p (Iminus (Iplus p q) (Iminus (Iplus p q) p)));
 [Intros H'2 | Idtac].
Case (errorBoundedPlus b (2) precision) with p := p q := q pq := (Iplus p q);
 Auto.
Intros pq H'4; Elim H'4; Intros H'5 H'6; Elim H'6; Intros H'7 H'8; Clear H'6 H'4.
Apply IplusCorrectEq with pq := pq; Auto.
Repeat (Try Apply IplusBounded; Try Apply IminusBounded; Auto); Auto.
Repeat (Try Apply IplusBounded; Try Apply IminusBounded; Auto); Auto.
Repeat Rewrite H'3; Repeat Rewrite H'2; Repeat Rewrite H'1.
Unfold FtoRradix; Rewrite H'5.
Ring.
Ring.
Rewrite H'2.
Case (errorBoundedPlus b (2) precision) with p := (Iplus p q) q := (Fopp p)
                                             pq := (Iminus (Iplus p q) p); Auto.
Repeat (Try Apply IplusBounded; Try Apply IminusBounded; Auto); Auto.
Apply oppBounded; Auto.
Apply (ClosestCompatible
         b (2) (Rminus (Iplus p q) p) (Rplus (Iplus p q) (Fopp p))
         (Iminus (Iplus p q) p)); Auto.
Apply IminusCorrect; Auto; Apply IplusBounded; Auto.
Rewrite (Fopp_correct (2)); Auto.
Repeat (Try Apply IplusBounded; Try Apply IminusBounded; Auto); Auto.
Intros pq H'3; Elim H'3; Intros H'4 H'5; Elim H'5; Intros H'6 H'7; Clear H'5 H'3.
Rewrite (IminusPlus p); Apply IplusCorrectEq with pq := (Fopp pq); Auto.
Apply oppBounded;
 Repeat (Try Apply IplusBounded; Try Apply IminusBounded; Auto); Auto.
Apply oppBounded; Auto.
Unfold FtoRradix; Rewrite (Fopp_correct (2)); Auto; Rewrite H'4;
 Rewrite (Fopp_correct (2)); Auto; Ring; Ring.
Unfold FtoRradix; Repeat Rewrite (Fopp_correct (2)); Auto; Rewrite H'4;
 Rewrite (Fopp_correct (2)); Auto.
Unfold FtoRradix in H'2; Rewrite H'2; Ring.
Qed.

Theorem MKnuth2:
  (p, q:float)
  (Rle (Rabsolu q) (Rabsolu p)) -> (Fbounded b p) -> (Fbounded b q) ->
  <R>
    (Iplus
       (Iminus p (Iminus (Iplus p q) (Iminus (Iplus p q) p)))
       (Iminus q (Iminus (Iplus p q) p)))==(Rminus (Rplus p q) (Iplus p q)).
Intros p q H' H'0 H'1.
Apply MKnuth; Auto.
Apply (MDekker b precision); Auto.
Qed.

Theorem IminusOp: (p, q:float)(Fopp (Iminus p q))=(Iminus (Fopp p) (Fopp q)).
Intros p q; Repeat Rewrite IminusPlus; Repeat Rewrite <- IplusOp; Auto.
Qed.

Theorem MKnuthOpp:
  (p, q:float)
  <R>
    (Iplus
       (Iminus p (Iminus (Iplus p q) (Iminus (Iplus p q) p)))
       (Iminus q (Iminus (Iplus p q) p)))==
    (Fopp
       (Iplus
          (Iminus
             (Fopp p)
             (Iminus
                (Iplus (Fopp p) (Fopp q))
                (Iminus (Iplus (Fopp p) (Fopp q)) (Fopp p))))
          (Iminus (Fopp q) (Iminus (Iplus (Fopp p) (Fopp q)) (Fopp p))))).
Intros p q; Repeat (Rewrite <- IplusOp Orelse Rewrite <- IminusOp);
 Rewrite Fopp_Fopp; Auto.
Qed.

Theorem MKnuth3:
  (p, q:float)
  (Rle R0 q) ->
  (Rle q (Rmult (2) (Ropp p))) ->
  (Rle (Ropp p) q) -> (Fbounded b p) -> (Fbounded b q) ->
  <R>
    (Iplus
       (Iminus p (Iminus (Iplus p q) (Iminus (Iplus p q) p)))
       (Iminus q (Iminus (Iplus p q) p)))==(Rminus (Rplus p q) (Iplus p q)).
Intros p q H' H'0 H'1 H'2 H'3.
Apply MKnuth; Auto.
Cut <R> (FtoRradix (Iplus p q))==(Rplus p q); [Intros Eq1 | Idtac].
Apply IminusCorrectEq with pq := q; Auto.
Apply IplusBounded; Auto.
Rewrite Eq1; Ring.
Rewrite Eq1; Ring.
Apply IplusCorrectEq with pq := (Fminus (2) q (Fopp p)); Auto.
Apply Sterbenz; Auto.
Apply oppBounded; Auto.
2:Repeat Rewrite (Fopp_correct (2)); Auto.
2:Rewrite (Fminus_correct (2)); Repeat Rewrite (Fopp_correct (2));
   Unfold FtoRradix; Auto; Ring; Ring.
2:Rewrite (Fminus_correct (2)); Repeat Rewrite (Fopp_correct (2));
   Unfold FtoRradix; Auto; Ring; Ring.
Apply Rle_trans with 2 := H'1.
Apply Rle_mult_anticompatibility with z := (INR (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_r; Auto with real arith; Rewrite Rmult_ne_r.
Rewrite (Fopp_correct (2)).
Apply Rle_trans with 1 := H'1; Auto.
Qed.

Theorem MKnuth4:
  (p, q:float)
  (Rlt R0 (Ropp p)) ->
  (Rlt R0 q) ->
  (Rlt (Rmult (2) (Ropp p)) q) -> (Fbounded b p) -> (Fbounded b q) ->
  <R>
    (Iplus
       (Iminus p (Iminus (Iplus p q) (Iminus (Iplus p q) p)))
       (Iminus q (Iminus (Iplus p q) p)))==(Rminus (Rplus p q) (Iplus p q)).
Intros p q H' H'0 H'1 H'2 H'3.
Cut (Rle (Rmult (Rinv (2)) q) (Iplus p q)); [Intros Rle1 | Idtac].
Cut (Rle (Iplus p q) (Iminus (Iplus p q) p)); [Intros Rle2 | Idtac].
Cut (Rle (Iminus (Iplus p q) p) (Rmult (2) (Iplus p q))); [Intros Rle3 | Idtac].
Cut (Rle (Rmult (2) (Iplus p q)) (Rmult (2) q)); [Intros Rle4 | Idtac].
Apply MKnuth1; Auto.
Apply IminusCorrectEq with pq := (Fminus (2) q (Iminus (Iplus p q) p)); Auto.
Apply IminusBounded; Auto.
Apply IplusBounded; Auto.
Rewrite <- (Fopp_Fopp (Fminus (2) q (Iminus (Iplus p q) p))).
Apply oppBounded.
Rewrite Fopp_Fminus.
Apply Sterbenz; Auto.
Apply IminusBounded; Auto.
Apply IplusBounded; Auto.
Apply Rle_trans with 1 := Rle1; Auto.
Apply Rle_trans with 1 := Rle3; Auto.
Apply sym_eqT; Apply (Fminus_correct (2)); Auto.
Apply sym_eqT; Apply (Fminus_correct (2)); Auto.
Apply IminusCorrectEq with pq := (Fminus (2) (Iplus p q) (Iminus (Iplus p q) p));
 Auto.
Apply IplusBounded; Auto.
Apply IminusBounded; Auto.
Apply IplusBounded; Auto.
Rewrite <- (Fopp_Fopp (Fminus (2) (Iplus p q) (Iminus (Iplus p q) p))).
Apply oppBounded.
Rewrite Fopp_Fminus.
Apply Sterbenz; Auto.
Apply IminusBounded; Auto.
Apply IplusBounded; Auto.
Apply IplusBounded; Auto.
Apply Rle_trans with 2 := Rle2; Auto.
Apply Rle_mult_anticompatibility with z := (INR (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_r; Auto with real arith; Rewrite Rmult_ne_r.
Apply Rledouble.
Apply Rle_trans with 2 := Rle1.
Apply Rle_mult_anticompatibility with z := (INR (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_r; Auto with real arith;
 Rewrite Rmult_ne_r; Rewrite Rmult_Or; Apply Rlt_le; Auto.
Apply sym_eqT; Apply (Fminus_correct (2)); Auto.
Apply sym_eqT; Apply (Fminus_correct (2)); Auto.
Apply Rle_mult_anticompatibility with z := (Rinv (2)); Auto with real.
Repeat Rewrite <- Rmult_assoc; Repeat Rewrite Rinv_l; Auto with real arith;
 Repeat Rewrite Rmult_ne_r; Auto.
Cut (MonotoneP (2) (Closest b (2))); [Intros Cp | Idtac].
Apply (Cp (Rplus p q) q); Auto.
Apply Rlt_anti_compatibility with r := (Ropp q); Auto.
Replace (Rplus (Ropp q) (Rplus p q)) with (Ropp (Ropp p)); [Idtac | Ring].
Replace (Rplus (Ropp q) q) with (Ropp R0); [Auto with real | Ring].
Apply (RoundedModeProjectorIdem b (2) (Closest b (2))); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply ClosestMonotone; Auto.
Apply (RoundedModeMult b (2))
     with P := (Closest b (2)) r := (Rminus (Iplus p q) p); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply IminusCorrect; Auto.
Apply IplusBounded; Auto.
Apply IplusBounded; Auto.
Apply Rle_anti_compatibility with r := (Ropp (Iplus p q)); Auto.
Replace (Rplus (Ropp (Iplus p q)) (Rminus (Iplus p q) p)) with (Ropp p);
 [Idtac | Ring].
Replace (Rplus (Ropp (Iplus p q)) (Rmult (2) (FtoR (2) (Iplus p q))))
     with (FtoR (2) (Iplus p q)); [Idtac | Simpl; Unfold FtoRradix; Ring; Ring].
Rewrite <- (Fopp_correct (2)); Auto.
Cut (MonotoneP (2) (Closest b (2))); [Intros Cp | Idtac].
Apply (Cp (Fopp p) (Rplus p q)); Auto.
Rewrite (Fopp_correct (2)); Auto.
Apply Rlt_anti_compatibility with r := (Ropp p); Auto.
Replace (Rplus (Ropp p) (Ropp (FtoR (2) p))) with (Rmult (2) (Ropp p));
 [Idtac | Simpl; Ring].
Replace (Rplus (Ropp p) (Rplus p q)) with (FtoRradix q); [Auto | Simpl; Ring].
Apply (RoundedModeProjectorIdem b (2) (Closest b (2))); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply oppBounded; Auto.
Apply ClosestMonotone; Auto.
Cut (MonotoneP (2) (Closest b (2))); [Intros Cp | Idtac].
Apply (Cp (Iplus p q) (Rminus (Iplus p q) p)); Auto.
Apply Rlt_anti_compatibility with r := (Ropp (Iplus p q)); Auto.
Replace (Rplus (Ropp (Iplus p q)) (Iplus p q)) with R0; [Auto | Simpl; Ring].
Replace (Rplus (Ropp (Iplus p q)) (Rminus (Iplus p q) p)) with (Ropp p);
 [Auto | Simpl; Ring].
Apply (RoundedModeProjectorIdem b (2) (Closest b (2))); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply IplusBounded; Auto.
Apply IminusCorrect; Auto.
Apply IplusBounded; Auto.
Apply ClosestMonotone; Auto.
Apply (FmultRadixInv b (2) precision) with y := (Rplus p q); Auto.
Apply Rlt_mult_anticompatibility with z := (INR (2)); Auto with real.
Repeat Rewrite <- Rmult_assoc; Repeat Rewrite Rinv_r; Auto with real arith;
 Repeat Rewrite Rmult_ne_r; Auto.
Apply Rlt_anti_compatibility with r := (Ropp q); Auto.
Apply Rlt_anti_compatibility with r := (Rmult (2) (Ropp p)); Auto.
Replace (Rplus (Rmult (2) (Ropp p)) (Rplus (Ropp q) (FtoR (2) q)))
     with (Rmult (2) (Ropp p)); [Auto | Unfold FtoRradix; Simpl; Ring].
Replace (Rplus (Rmult (2) (Ropp p)) (Rplus (Ropp q) (Rmult (2) (Rplus p q))))
     with (FtoRradix q); [Auto | Simpl; Ring].
Qed.

Theorem MKnuth5:
  (p, q:float) (Rlt R0 p) -> (Rlt p q) -> (Fbounded b p) -> (Fbounded b q) ->
  <R>
    (Iplus
       (Iminus p (Iminus (Iplus p q) (Iminus (Iplus p q) p)))
       (Iminus q (Iminus (Iplus p q) p)))==(Rminus (Rplus p q) (Iplus p q)).
Intros p q H' H'0 H'1 H'2.
Cut <R>
      (Iminus (Iplus p q) (Iminus (Iplus p q) p))==
      (Rminus (Iplus p q) (Iminus (Iplus p q) p)); [Intros Eq1 | Idtac].
Apply MKnuth1; Auto.
Cut (Rle (Iminus (Iplus p q) p) (Iplus p q)); [Intros Rle1 | Idtac].
Cut (Rle (Iplus p q) (Rmult (2) q)); [Intros Rle2 | Idtac].
Case (Rle_or_lt q (Iminus (Iplus p q) p)); Intros Rle3.
Apply IminusCorrectEq with pq := (Fminus (2) q (Iminus (Iplus p q) p)); Auto.
Apply IminusBounded; Auto.
Apply IplusBounded; Auto.
Rewrite <- (Fopp_Fopp (Fminus (2) q (Iminus (Iplus p q) p))).
Apply oppBounded.
Rewrite Fopp_Fminus.
Apply Sterbenz; Auto.
Apply IminusBounded; Auto.
Apply IplusBounded; Auto.
Apply Rle_trans with 2 := Rle3; Auto.
Apply Rle_mult_anticompatibility with z := (INR (2)); Auto with real.
Rewrite <- Rmult_assoc; Rewrite Rinv_r; Auto with real arith; Rewrite Rmult_ne_r.
Apply Rledouble.
Apply Rlt_le; Apply Rlt_trans with 1 := H'; Auto.
Apply Rle_trans with 1 := Rle1; Auto.
Apply sym_eqT; Apply (Fminus_correct (2)); Auto.
Apply sym_eqT; Apply (Fminus_correct (2)); Auto.
Case (ExactMinusInterval b (2) precision) with P := (Closest b (2))
                                               p := (Iminus (Iplus p q) p)
                                               q := (Iplus p q) r := q; Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply (RleRoundedR0 b (2) precision)
     with P := (Closest b (2)) r := (Rminus (Iplus p q) p); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply IminusCorrect; Auto.
Apply IplusBounded; Auto.
Apply Rle_anti_compatibility with r := (FtoRradix p); Auto.
Replace (Rplus p R0) with (FtoRradix p); [Idtac | Ring].
Replace (Rplus p (Rminus (Iplus p q) p)) with (FtoRradix (Iplus p q));
 [Idtac | Ring].
Cut (MonotoneP (2) (Closest b (2))); [Intros Cp | Idtac].
Apply (Cp p (Rplus p q)); Auto.
Apply Rlt_anti_compatibility with r := (Ropp p); Auto.
Replace (Rplus (Ropp p) p) with R0; [Idtac | Ring].
Replace (Rplus (Ropp p) (Rplus p q)) with (FtoRradix q);
 [Apply Rlt_trans with 1 := H'; Auto | Ring].
Apply (RoundedModeProjectorIdem b (2) (Closest b (2))); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply ClosestMonotone; Auto.
Apply IminusBounded; Auto.
Apply IplusBounded; Auto.
Apply IplusBounded; Auto.
Exists (Iminus (Iplus p q) (Iminus (Iplus p q) p)); Split; Auto.
Repeat (Apply IplusBounded Orelse Apply IminusBounded); Auto.
Apply Rlt_le; Auto.
Cut (MonotoneP (2) (Closest b (2))); [Intros Cp | Idtac].
Apply (Cp q (Rplus p q)); Auto.
Apply Rlt_anti_compatibility with r := (Ropp q); Auto.
Replace (Rplus (Ropp q) q) with R0; [Idtac | Ring].
Replace (Rplus (Ropp q) (Rplus p q)) with (FtoRradix p); [Auto | Ring].
Apply (RoundedModeProjectorIdem b (2) (Closest b (2))); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply ClosestMonotone; Auto.
Intros x H'4; Elim H'4; Intros H'5 H'6; Clear H'4.
Apply IminusCorrectEq with pq := x; Auto.
Repeat (Apply IplusBounded Orelse Apply IminusBounded); Auto.
Apply (RoundedModeMult b (2)) with P := (Closest b (2)) r := (Rplus p q); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply Rle_anti_compatibility with r := (Ropp q); Auto.
Replace (Rplus (Ropp q) (Rplus p q)) with (FtoRradix p); [Idtac | Ring].
Replace (Rplus (Ropp q) (Rmult (2) (FtoR (2) q))) with (FtoRradix q);
 [Apply Rlt_le; Auto | Unfold FtoRradix; Simpl; Ring].
Cut (MonotoneP (2) (Closest b (2))); [Intros Cp | Idtac].
Apply (Cp (Rminus (Iplus p q) p) (Iplus p q)); Auto.
Apply Rlt_anti_compatibility with r := (Ropp (Iplus p q)); Auto.
Replace (Rplus (Ropp (Iplus p q)) (Rminus (Iplus p q) p)) with (Ropp p);
 [Idtac | Ring].
Replace (Rplus (Ropp (Iplus p q)) (Iplus p q)) with (Ropp R0);
 [Auto with real | Ring].
Apply IminusCorrect; Auto.
Apply IplusBounded; Auto.
Apply (RoundedModeProjectorIdem b (2) (Closest b (2))); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply IplusBounded; Auto.
Apply ClosestMonotone; Auto.
Case (minusRoundRep b (2) precision) with P := (Closest b (2)) p := p
                                          q := (Iplus p q)
                                          qmp := (Iminus (Iplus p q) p); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply Rlt_le; Auto.
Cut (MonotoneP (2) (Closest b (2))); [Intros Cp | Idtac].
Apply (Cp (FtoRradix p) (Rplus p q)); Auto.
Apply Rlt_anti_compatibility with r := (Ropp p); Auto.
Replace (Rplus (Ropp p) p) with R0; [Idtac | Ring].
Replace (Rplus (Ropp p) (Rplus p q)) with (FtoRradix q);
 [(Apply Rlt_trans with 1 := H'; Auto) | Ring].
Apply (RoundedModeProjectorIdem b (2) (Closest b (2))); Auto.
Apply ClosestRoundedModeP with precision := precision; Auto.
Apply ClosestMonotone; Auto.
Apply IminusCorrect; Auto.
Apply IplusBounded; Auto.
Apply IplusBounded; Auto.
Intros x H'4; Elim H'4; Intros H'5 H'6; Clear H'4.
Apply IminusCorrectEq with pq := x; Auto.
Apply IplusBounded; Auto.
Repeat (Apply IplusBounded Orelse Apply IminusBounded); Auto.
Qed.

Theorem MKnuth6:
  (p, q:float)
  <R> (Iplus p q)==(Rplus p q) -> (Fbounded b p) -> (Fbounded b q) ->
  <R>
    (Iplus
       (Iminus p (Iminus (Iplus p q) (Iminus (Iplus p q) p)))
       (Iminus q (Iminus (Iplus p q) p)))==(Rminus (Rplus p q) (Iplus p q)).
Intros p q H' H'0 H'1.
Apply MKnuth; Auto.
Apply IminusCorrectEq with pq := q; Auto.
Apply IplusBounded; Auto.
Rewrite H'; Ring.
Rewrite H'; Ring.
Qed.

Theorem MKnuth7:
  (p, q:float) (Rlt (Rabsolu p) q) -> (Fbounded b p) -> (Fbounded b q) ->
  <R>
    (Iplus
       (Iminus p (Iminus (Iplus p q) (Iminus (Iplus p q) p)))
       (Iminus q (Iminus (Iplus p q) p)))==(Rminus (Rplus p q) (Iplus p q)).
Intros p q H' H'0 H'1.
Cut (Rlt R0 q); [Intros Rlt1 | Apply Rle_lt_trans with 2 := H'; Auto with real].
Case (Rle_or_lt R0 p); Intros Rle1; [Casec Rle1; Intros Rle1 | Idtac].
Apply MKnuth5; Auto.
Rewrite <- (Rabsolu_simpl1 p); Auto; Apply Rlt_le; Auto.
Apply MKnuth6; Auto.
Apply IplusCorrectEq with pq := q; Auto; Rewrite <- Rle1; Ring.
Case (Rle_or_lt q (Rmult (2) (Ropp p))); Intros Rle2.
Apply MKnuth3; Auto.
Apply Rlt_le; Auto.
Rewrite <- (Rabsolu_simpl2 p); Auto; Apply Rlt_le; Auto.
Apply MKnuth4; Auto.
Replace R0 with (Ropp R0); [Auto with real | Ring].
Qed.

Theorem Knuth:
  (p, q:float) (Fbounded b p) -> (Fbounded b q) ->
  <R>
    (Iplus
       (Iminus p (Iminus (Iplus p q) (Iminus (Iplus p q) p)))
       (Iminus q (Iminus (Iplus p q) p)))==(Rminus (Rplus p q) (Iplus p q)).
Intros p q H' H'0; Case (Rle_or_lt (Rabsolu q) (Rabsolu p)); Intros Rle1.
Apply MKnuth2; Auto.
Case (Rle_or_lt R0 q); Intros Rle1.
Apply MKnuth7; Auto.
Rewrite <- (Rabsolu_simpl1 q); Auto; Apply Rlt_le; Auto.
Rewrite MKnuthOpp.
Rewrite (Fopp_correct (2)).
Replace (Rminus (Rplus p q) (Iplus p q))
     with (Ropp (Rminus (Rplus (Fopp p) (Fopp q)) (Iplus (Fopp p) (Fopp q)))).
Cut (x, y:R) x==y ->(Ropp x)==(Ropp y);
 [Intros tmp; Apply tmp; Clear tmp | Intros x y tmp; Rewrite tmp; Auto].
Apply MKnuth7; Auto.
Repeat Rewrite (Fopp_correct (2)); Auto.
Rewrite <- (Rabsolu_simpl2 (FtoR (2) q)); Try Apply Rlt_le; Auto.
Rewrite Rabsolu_Ropp; Auto.
Apply oppBounded; Auto.
Apply oppBounded; Auto.
Rewrite <- IplusOp; Repeat Rewrite (Fopp_correct (2)); Unfold FtoRradix; Ring.
Qed.
End EFast.
